Package: bhmbasket
Title: Bayesian Hierarchical Models for Basket Trials
Version: 1.1.0
Authors@R: 
  c(
    person(given = 'Stephan',
           family = 'Wojciekowski',
           role = c('aut', 'cre'),
           email = 'stephan.wojciekowski@boehringer-ingelheim.com'),
    person(given = 'Tathagata',
           family = 'Chattopadhyay',
           role = 'aut',
           email = 'tchattopadhyay6@gmail.com'))
Description: Provides functions for the evaluation of basket
    trial designs with binary endpoints. Operating characteristics of a
    basket trial design are assessed by simulating trial data according to
    scenarios, analyzing the data with Bayesian hierarchical models (BHMs), and
    assessing decision probabilities on stratum and trial-level based on Go / No-go decision making.
    The package is build for high flexibility regarding decision rules,
    number of interim analyses, number of strata, and recruitment.
    The BHMs proposed by
    Berry et al. (2013) <doi:10.1177/1740774513497539>
    and Neuenschwander et al. (2016) <doi:10.1002/pst.1730>,
    as well as a model that combines both approaches are implemented.
    Functions are provided to implement Bayesian decision rules as for example
    proposed by Fisch et al. (2015) <doi:10.1177/2168479014533970>.
    In addition, posterior point estimates (mean/median) and credible intervals
    for response rates and some model parameters can be calculated.
    For simulated trial data, bias and mean squared errors of posterior
    point estimates for response rates can be provided.
License: GPL-3
URL: https://CRAN.R-project.org/package=bhmbasket,
        https://github.com/Boehringer-Ingelheim/bhmbasket
BugReports: https://github.com/Boehringer-Ingelheim/bhmbasket/issues
Depends: R (>= 3.6.0)
Imports: foreach, doRNG, rjags, checkmate
Suggests: doFuture, future, knitr, rmarkdown, RBesT, testthat (>=
        3.0.0)
VignetteBuilder: knitr
Encoding: UTF-8
RoxygenNote: 7.3.2
Language: en-US
SystemRequirements: JAGS 4.x.y (http://mcmc-jags.sourceforge.net)
Config/testthat/edition: 3
NeedsCompilation: no
Packaged: 2026-04-14 11:51:27 UTC; wojcie
Author: Stephan Wojciekowski [aut, cre],
  Tathagata Chattopadhyay [aut]
Maintainer: Stephan Wojciekowski <stephan.wojciekowski@boehringer-ingelheim.com>
Repository: CRAN
Date/Publication: 2026-04-15 07:20:02 UTC
Built: R 4.6.0; ; 2026-04-27 07:06:00 UTC; unix

# Generated by roxygen2: do not edit by hand

S3method(print,analysis_list)
S3method(print,decision_list)
S3method(print,scenario_list)
export(combinePriorParameters)
export(continueRecruitment)
export(createTrial)
export(getAverageNSubjects)
export(getEstimates)
export(getGoDecisions)
export(getGoProbabilities)
export(getMuVar)
export(getPriorParameters)
export(invLogit)
export(loadAnalyses)
export(loadScenarios)
export(logit)
export(negateGoDecisions)
export(performAnalyses)
export(saveAnalyses)
export(saveScenarios)
export(scaleRoundList)
export(setPriorParametersBerry)
export(setPriorParametersExNex)
export(setPriorParametersExNexAdj)
export(setPriorParametersPooled)
export(setPriorParametersStratified)
export(setPriorParametersStratifiedMix)
export(simulateScenarios)

<!-- badges: start -->
[![Linux](https://github.com/Boehringer-Ingelheim/bhmbasket/actions/workflows/linux.yml/badge.svg?branch=test)](https://github.com/Boehringer-Ingelheim/bhmbasket/actions/workflows/linux.yml)
[![Windows](https://github.com/Boehringer-Ingelheim/bhmbasket/actions/workflows/windows.yml/badge.svg?branch=test)](https://github.com/Boehringer-Ingelheim/bhmbasket/actions/workflows/windows.yml)
[![macOS](https://github.com/Boehringer-Ingelheim/bhmbasket/actions/workflows/macos.yml/badge.svg?branch=test)](https://github.com/Boehringer-Ingelheim/bhmbasket/actions/workflows/macos.yml)
[![Codecov test coverage](https://codecov.io/gh/Boehringer-Ingelheim/bhmbasket/branch/test/graph/badge.svg)](https://app.codecov.io/gh/Boehringer-Ingelheim/bhmbasket?branch=test)
<!-- badges: end -->


# The `bhmbasket` package

Simulate and analyze basket trial designs with binary endpoints using Bayesian hierarchical models.

## Installation

Install the package from CRAN using:

```{r}
install.packages("bhmbasket")
```

The development version can be installed from this repository:

```{r}
# install.packages("remotes")
remotes::install_github("https://github.com/Boehringer-Ingelheim/bhmbasket")
```

## ----include = FALSE----------------------------------------------------------
knitr::opts_chunk$set(
  eval     = TRUE,  # en- / disables R code evaluation globally
  cache    = FALSE,  # en- / disables R code caching globally
  collapse = TRUE,
  comment  = "#>"
)

## ----setup--------------------------------------------------------------------
library(bhmbasket)

rng_seed <- 123
set.seed(rng_seed)

## -----------------------------------------------------------------------------
chugh_trial <- createTrial(
  n_subjects   = c(15, 13, 12, 28, 29, 29, 26, 5, 2, 20),
  n_responders = c(2, 0, 1, 6, 7, 3, 5, 1, 0, 3))

## ----chugh--------------------------------------------------------------------
chugh_prior_parameters <- setPriorParametersExNex(
  mu_mean   = c(-1.735, 0.847),
  mu_sd     = c(0.146, 0.266)^-0.5,
  tau_scale = 1,
  mu_j      = rep(-1.734, 10),
  tau_j     = rep(0.128^-0.5, 10),
  w_j       = c(0.5, 0, 0.5))

if (file.exists("chugh_analysis.rds")) {
  
  chugh_analysis <- readRDS("chugh_analysis.rds")
  
} else {
  
  chugh_analysis <- performAnalyses(
  scenario_list         = chugh_trial,
  method_names          = "exnex",
  prior_parameters_list = chugh_prior_parameters,
  seed                  = rng_seed,
  n_mcmc_iterations     = 5e4,
  n_cores               = 2L,
  verbose               = FALSE)
  
  saveRDS(chugh_analysis, "chugh_analysis.rds")
  
}

## -----------------------------------------------------------------------------
getEstimates(analyses_list = chugh_analysis)

## ----simulateScenarios--------------------------------------------------------
scenarios_list <- simulateScenarios(
  n_subjects_list     = list(c(20, 20, 10, 10),
                             c(20, 20, 10, 10),
                             c(20, 20, 10, 10)),
  response_rates_list = list(c(0.1, 0.1, 0.1, 0.1),
                             c(0.1, 0.1, 0.3, 0.3),
                             c(0.1, 0.1, 0.1, 0.5)),
  scenario_numbers    = c(1, 3, 4),
  n_trials            = 1e4)

## ----performAnalyses----------------------------------------------------------
prior_parameters <- setPriorParametersExNex(
  mu_mean   = c(logit(0.1), logit(0.3)),
  mu_sd     = c(3.18, 1.94),
  tau_scale = 1,
  mu_j      = rep(logit(0.2), 4),
  tau_j     = rep(2.5, 4),
  w_j       = c(0.25, 0.25, 0.5))

if (file.exists("analyses_list.rds")) {
  
  analyses_list <- readRDS("analyses_list.rds")
  
} else {
  
  analyses_list <- performAnalyses(
    scenario_list         = scenarios_list,
    method_names          = "exnex",
    prior_parameters_list = prior_parameters,
    seed                  = rng_seed,
    n_mcmc_iterations     = 5e4,
    n_cores               = 2L)
  
  saveRDS(analyses_list, "analyses_list.rds")
  
}

## ----include = FALSE----------------------------------------------------------
## check for unnecessary data
if (any(grepl("mu_",
              colnames(analyses_list$scenario_1$quantiles_list$exnex[[1]])))) {
  
  ## remove unnecessary data
  
  ## ...in quantiles_list
  analyses_list$scenario_1$quantiles_list$exnex <-   lapply(analyses_list$scenario_1$quantiles_list$exnex, function   (x) {
    x[-c(2, 6), -c(1:6, 11, 12)]
  })
  analyses_list$scenario_3$quantiles_list$exnex <-   lapply(analyses_list$scenario_3$quantiles_list$exnex, function   (x) {
    x[-c(2, 6), -c(1:6, 11, 12)]
  })
  analyses_list$scenario_4$quantiles_list$exnex <-   lapply(analyses_list$scenario_4$quantiles_list$exnex, function   (x) {
    x[-c(2, 6), -c(1:6, 11, 12)]
  })
  
  ## ... on quantiles
  analyses_list$scenario_1$analysis_parameters$quantiles <-   c(0.025, 0.100, 0.200, 0.500, 0.975)
  analyses_list$scenario_3$analysis_parameters$quantiles <-   c(0.025, 0.100, 0.200, 0.500, 0.975)
  analyses_list$scenario_4$analysis_parameters$quantiles <-   c(0.025, 0.100, 0.200, 0.500, 0.975)
  
  ## ... on subjects and responders
  analyses_list$scenario_1$scenario_data$n_subjects <- NULL
  analyses_list$scenario_3$scenario_data$n_subjects <- NULL
  analyses_list$scenario_4$scenario_data$n_subjects <- NULL
  analyses_list$scenario_1$scenario_data$n_responders <- NULL
  analyses_list$scenario_3$scenario_data$n_responders <- NULL
  analyses_list$scenario_4$scenario_data$n_responders <- NULL
  
  analyses_list$scenario_1$quantiles_list$exnex[[1]]
    
  saveRDS(analyses_list, "analyses_list.rds", compress = "xz")
}

## ----getBiasMSE---------------------------------------------------------------
estimates <- getEstimates(
    analyses_list   = analyses_list,
    point_estimator = "mean")

scaleRoundList(
  list         = estimates,
  scale_param  = 100,
  round_digits = 2)

## ----eval = FALSE-------------------------------------------------------------
# cohort_names    = "p_1"
# boundary_rules  = quote(c(x[1] > 0.1))
# evidence_levels = 0.9

## ----eval = FALSE-------------------------------------------------------------
# cohort_names    = c("p_1", "p_1")
# boundary_rules  = quote(c(x[1] > 0.1 & x[2] > 0.2))
# evidence_levels = c(0.9, "mean")

## ----getGoDecisions-----------------------------------------------------------
decisions_list <- getGoDecisions(
  analyses_list   = analyses_list,
  cohort_names    = c("p_1", "p_1",
                      "p_2", "p_2",
                      "p_3", "p_3",
                      "p_4", "p_4"),
  evidence_levels = c(0.9, "mean",
                      0.9, "mean",
                      0.8, "mean",
                      0.8, "mean"),
  boundary_rules  = quote(c(x[1] > 0.1 & x[2] > 0.2,
                            x[3] > 0.1 & x[4] > 0.2,
                            x[5] > 0.1 & x[6] > 0.2,
                            x[7] > 0.1 & x[8] > 0.2)))

go_probabilities <- getGoProbabilities(decisions_list)

scaleRoundList(
  list         = go_probabilities,
  scale_param  = 100,
  round_digits = 0)

## ----knitr_options, include = FALSE-------------------------------------------
knitr::opts_chunk$set(
  eval     = FALSE,  # en- / disables R code evaluation globally
  cache    = FALSE,  # en- / disables R code caching globally
  collapse = TRUE,
  comment  = "#>"
)

## ----setup--------------------------------------------------------------------
# library(bhmbasket)
# library(doFuture)
# library(future.batchtools)
# 
# rng_seed <- 5440
# set.seed(rng_seed)

## ----SLURM_Setup--------------------------------------------------------------
# ## Adapt the SLURM template to requirements
# job_time  <- 1   # time for job in hours
# n_workers <- 24  # number of worker nodes
# n_cpus    <- 16  # number of cpus per worker node
# gb_memory <- 2   # memory [GB] per cpu
# 
# slurm <- tweak(batchtools_slurm,
#            template  = system.file('templates/slurm-simple.tmpl',
#                                    package = 'batchtools'),
#            workers   = n_workers,
#            resources = list(
#              walltime  = 60 * 60 * job_time,
#              ncpus     = n_cpus,
#              memory    = 1000 * gb_memory))
# 
# ## Register the parallel backend
# registerDoFuture()
# 
# ## Specify how the futures should be resolved
# plan(list(slurm, multisession))

## -----------------------------------------------------------------------------
# scenarios_list <- simulateScenarios(
#     n_subjects_list     = list(c(10, 20, 30)),
#     response_rates_list = list(c(0.1, 0.2, 3)),
#     n_trials            = 10)
# 
#   analyses_list <- performAnalyses(
#     scenario_list       = scenarios_list,
#     target_rates        = c(0.1, 0.1, 0.1),
#     calc_differences    = matrix(c(3, 2, 2, 1), ncol = 2),
#     n_mcmc_iterations   = 100)
# 
#   getEstimates(analyses_list)

## R CMD check appeasement for forach loop
utils::globalVariables("k")
utils::globalVariables("i")

## Code inspired by https://stackoverflow.com/questions/3318333/split-a-vector-into-chunks
## Answer by mathheadinclouds
chunkVector <- function(x, n_chunks) {
  
  if (n_chunks <= 1) {
    
    chunk_list <- list(x)
    
  } else {
    
    chunk_list <- unname(split(x, cut(seq_along(x), n_chunks, labels = FALSE)))
    
  }
  
  return (chunk_list)
  
}

convertVector2Matrix <- function (vector) {
  
  if (is.null(dim(vector))) {
    vector <- t(as.matrix(vector))
  }
  
  return (vector)
  
}

## Code inspired by https://www.r-bloggers.com/2019/01/hash-me-if-you-can/ 
createHashTable <- function (keys, values, size = length(keys)) {
  
  # initialize environment to store key - value assignments, i.e. hash table
  hash <- new.env(hash = TRUE, parent = emptyenv(), size = length(keys))
  
  # assign values to keys
  assignHashes <- Vectorize(assign, vectorize.args = c("x", "value"))
  assignHashes(x = keys, value = values, envir = hash)
  
  return (hash)
  
}

cummulativeMovingAverage <- function (x) {
  return (cumsum(x) / seq_along(x))
}

firstUpper <- function (string) {
  
  substr(string, 1, 1) <- toupper(substr(string, 1, 1))
  return (string)
  
}

getBlankString <- function (length) {
  
  do.call(paste0, as.list(rep(" ", length)))
  
}

getHashKeys <- function (x) {
  
  x <- convertVector2Matrix(x)
  
  # apply(x, 1, rlang::hash)
  
  if (is.null(colnames(x))) {
    
    colnames_x <- as.character(paste0("x", seq_len(ncol(x))))
    
  } else {
    
    colnames_x <- colnames(x)
    
  }
  
  apply(x, 1, function (y) paste0(colnames_x, ":", as.character(y), collapse = "|"))
  
}

## Code inspired by https://www.r-bloggers.com/2019/01/hash-me-if-you-can/ 
getHashValues <- function(search_keys, hash_table) {
  
  unname(mget(search_keys, hash_table))
  
}


getListLevel <- function (
  
  list
  
) {
  
  if (!is.list(list)) {
    
    return (0)
    
  } else {
    
    list <- list[[1]]
    
    return (1 + getListLevel(list))
    
  }
  
}

getRowIndexOfVectorInMatrix <- function (
  
  vector_to_be_found,
  matrix_to_be_searched
  
) {
  
  n_col <- ncol(matrix_to_be_searched)
  
  if (length(vector_to_be_found) != n_col) {
    stop ("The length of the vector must be equal to the number of columns of the matrix")
  }
  
  index <- apply(convertVector2Matrix(sapply(seq_len(n_col), function (i) {
    
    matrix_to_be_searched[, i] %in% vector_to_be_found[i]
    
  })), 1, all)
  
  return (which(index))
  
}

#' @title invLogit
#' @description This function returns the inverse logit of the input argument.
#' @param theta A numeric
#' @details This function is an alias for `stats::binomial()$linkinv`
#' @return Inverse logit of theta
#' @rdname invlogit
#' @author Stephan Wojciekowski
#' @examples
#' invLogit(logit(0.3))
#' invLogit(c(-Inf, 0, Inf))
#' @seealso
#'  \code{\link[stats]{family}}
#' @export
invLogit <- function (theta) {
  
  error_theta <- simpleError("Please provide a numeric for the argument 'theta'")
  
  if (missing(theta))     stop (error_theta)
  if (!is.numeric(theta)) stop (error_theta)
  
  stats::binomial()$linkinv(theta)
  
}

listPerMethod <- function (
  
  list_per_scenario
  
) {
  
  all_method_names <- unique(as.vector(sapply(list_per_scenario, names)))
  
  ## Create list to hold output
  out_list <- vector(mode = "list", length = length(all_method_names))
  names(out_list) <- all_method_names
  
  for (n in seq_along(out_list)) {
    
    out_list[[n]] <- vector(mode = "list", length = length(list_per_scenario))
    names(out_list[[n]]) <- paste0("scenario_", getScenarioNumbers(list_per_scenario))
    
  }
  
  ## Copy contents to new output list
  for (name in all_method_names) {
    
    for (s in seq_along(list_per_scenario)) {
      
      out_list[[name]][[s]] <- list_per_scenario[[s]][[name]]
      
    }
    
  }
  
  return (out_list)
  
}

#' @title logit
#' @description This function returns the logit of the input argument.
#' @param p A numeric in (0, 1)
#' @details This function is an alias for `stats::binomial()$linkfun`
#' @return logit of p
#' @rdname logit
#' @author Stephan Wojciekowski
#' @examples
#' logit(invLogit(0.3))
#' logit(c(0, 0.5, 1))
#' @seealso
#'  \code{\link[stats]{family}}
#' @export

logit <- function (p) {
  
  error_p <- simpleError("Please provide a numeric in (0, 1) for the argument 'p'")
  
  if (missing(p)) stop(error_p)
  
  # replace is.numeric
  if (!checkmate::test_numeric(p, any.missing = FALSE)) stop(error_p)
  if (any(p < 0 | p > 1)) stop(error_p)
  
  stats::binomial()$linkfun(p)
}


roundList <- function (list, round_digits, list_levels) {
  
  if (is.null(round_digits)) return(list)
  
  round_expression <- quote({
    if (checkmate::test_numeric(a)) {
      round(a, round_digits)
    } else {
      stop(simpleError("The list must contain numerics"))
    }
  })
  
  if (list_levels == 1) {
    return(lapply(list, function(a) eval(round_expression)))
    
  } else if (list_levels == 2) {
    return(lapply(list, function(x)
      lapply(x, function(a) eval(round_expression))))
    
  } else if (list_levels == 3) {
    return(lapply(list, function(x)
      lapply(x, function(y)
        lapply(y, function(a) eval(round_expression)))))
    
  } else {
    stop("lists with a nested depth greater than 3 are not supported")
  }
}

scaleList <- function (list, scale_param, list_levels) {
  
  scale_expression <- quote({
    if (checkmate::test_numeric(a)) {
      a * scale_param
    } else {
      stop(simpleError("The list must contain numerics"))
    }
  })
  
  if (list_levels == 1) {
    return(lapply(list, function(a) eval(scale_expression)))
    
  } else if (list_levels == 2) {
    return(lapply(list, function(x)
      lapply(x, function(a) eval(scale_expression))))
    
  } else if (list_levels == 3) {
    return(lapply(list, function(x)
      lapply(x, function(y)
        lapply(y, function(a) eval(scale_expression)))))
    
  } else {
    stop("lists with a nested depth greater than 3 are not supported")
  }
}

#' @title scaleRoundList
#' @description This function applies scaling and rounding to each item of a list of numerics
#' @param list The list to which the scaling and rounding should be applied to.
#' @param scale_param A numeric for the scaling of each item of the list, Default: `1`
#' @param round_digits An integer for the number of digits.
#' If `NULL`, no rounding will be applied, Default: `NULL`
#' @return A list of scaled and rounded numerics
#' @rdname scaleRoundList
#' @author Stephan Wojciekowski
#' @examples
#' some_list <- as.list(runif(5))
#' scaleRoundList(some_list, scale_param = 100, round_digits = 2)
#'
#' scenarios_list <- simulateScenarios(
#'   n_subjects_list     = list(c(10, 20, 30)),
#'   response_rates_list = list(c(0.1, 0.2, 0.3)),
#'   n_trials            = 10)
#'
#' analyses_list <- performAnalyses(
#'   scenario_list       = scenarios_list,
#'   target_rates        = rep(0.5, 3),
#'   n_mcmc_iterations   = 100)
#'
#' scaleRoundList(
#'   list         = getEstimates(analyses_list),
#'   scale_param  = 100,
#'   round_digits = 2)
#' @export
scaleRoundList <- function(
    
  list,
  scale_param  = 1,
  round_digits = NULL
  
) {
  
  error_list  <- simpleError(
    "Please provide a list (of lists) of numerics for the argument 'list'")
  error_scale_param <- simpleError(
    "Providing a positive numeric for the argument 'scale_param'")
  error_round_digits <- simpleError(
    "Providing a positive integer for the argument 'round_digits'")
  
  if (missing(list)) stop(error_list)
  if (!is.list(list)) stop(error_list)
  
  checkmate::assert_number(scale_param, finite = TRUE,
                           .var.name = error_scale_param)
  checkmate::assertTRUE(scale_param > 0, .var.name = error_scale_param)

  if (!is.null(round_digits)) {
    checkmate::assert_integerish(round_digits, lower = 0, len = 1,
                                 .var.name = error_round_digits)
  }
  
########### ######### ########## ########### #############
  
  list_levels <- getListLevel(list)
  
  out_list <- roundList(
    scaleList(list, scale_param, list_levels),
    round_digits,
    list_levels
  )
  
  return(out_list)
  
}

substrRight <- function (x, n) {
  substr(x, nchar(x) - n + 1, nchar(x))
}

# Tests for chunkVector --------------------------------------------------------

# ------------------------------------------------------------------
# Test: chunkVector returns a single chunk when n_chunks <= 1
# Input:
#   - x = 1:10
#   - n_chunks = 1
# Behaviour:
#   - returns list(x) (one chunk containing all elements)
# Expectations:
#   - output is a list of length 1
#   - the only element equals x exactly
# Why:
#   - ensures the "no chunking" path is stable and predictable.
# ------------------------------------------------------------------
test_that("chunkVector returns single chunk when n_chunks <= 1", {

  x <- 1:10

  out <- chunkVector(x, n_chunks = 1)

  expect_type(out, "list")

  expect_length(out, 1)

  expect_identical(out[[1]], x)

})

# ------------------------------------------------------------------
# Test: chunkVector splits into multiple chunks and preserves all elements
# Input:
#   - x = 1:10
#   - n_chunks = 2, 3
# Behaviour:
#   - splits x into n_chunks chunks
# Expectations:
#   - output is a list with requested length
#   - unlisted output contains exactly the original elements
# Why:
#   - verifies chunking doesn't drop or duplicate elements.
# ------------------------------------------------------------------
test_that("chunkVector splits into multiple chunks and preserves elements", {

  x <- 1:10

  out2 <- chunkVector(x, n_chunks = 2)

  expect_type(out2, "list")
  expect_length(out2, 2)
  expect_identical(sort(unlist(out2)), x)

  out3 <- chunkVector(x, n_chunks = 3)

  expect_type(out3, "list")
  expect_length(out3, 3)
  expect_identical(sort(unlist(out3)), x)

})


# Tests for convertVector2Matrix -----------------------------------------------

# ------------------------------------------------------------------
# Test: convertVector2Matrix converts a vector to a 1-row matrix
# Input:
#   - vector = c(1,2,3,4)
# Behaviour:
#   - wraps into a matrix and transposes to 1 x length(vector)
# Expectations:
#   - result is a matrix with dim (1,4)
#   - content equals the original vector
# Why:
#   - many downstream functions assume matrix input.
# ------------------------------------------------------------------
test_that("convertVector2Matrix converts vector to 1-row matrix", {

  v <- 1:4

  out <- convertVector2Matrix(v)

  expect_true(is.matrix(out))

  expect_equal(dim(out), c(1, 4))

  expect_identical(as.integer(out[1, ]), v)

})

# ------------------------------------------------------------------
# Test: convertVector2Matrix returns matrix unchanged
# Input:
#   - matrix input
# Behaviour:
#   - should not modify an existing matrix
# Expectations:
#   - identical to the original matrix
# Why:
#   - prevents accidental reshaping when caller already provides a matrix.
# ------------------------------------------------------------------
test_that("convertVector2Matrix leaves matrices unchanged", {

  m <- matrix(1:6, nrow = 2)

  out <- convertVector2Matrix(m)

  expect_true(is.matrix(out))

  expect_identical(out, m)

})


# Tests for createHashTable + getHashValues ------------------------------------

# ------------------------------------------------------------------
# Test: createHashTable stores key-value pairs retrievable via getHashValues
# Input:
#   - keys = c("a","b","c")
#   - values = list(10,20,30)
# Behaviour:
#   - creates hash and assigns values to keys
# Expectations:
#   - returned object is an environment
#   - getHashValues returns values in the requested key order
# Why:
#   - mapUniqueTrials relies on retrieval from hashed keys.
# ------------------------------------------------------------------
test_that("createHashTable and getHashValues store and retrieve values", {

  keys <- c("a", "b", "c")
  vals <- list(10, 20, 30)

  h <- createHashTable(keys, vals)

  expect_true(is.environment(h))

  out <- getHashValues(c("b", "a"), h)

  expect_type(out, "list")

  expect_equal(out, list(20, 10))

})


# Tests for getHashKeys ---------------------------------------------------------

# ------------------------------------------------------------------
# Test: getHashKeys uses default x1..xp names if colnames are missing
# Input:
#   - 2x2 matrix without colnames
# Behaviour:
#   - returns character keys "x1:<val>|x2:<val>" per row
# Expectations:
#   - correct key strings for each row
# Why:
#   - hashing must be stable even when matrices lack colnames.
# ------------------------------------------------------------------
test_that("getHashKeys uses default x* names without colnames", {

  x <- matrix(c(1, 2, 3, 4), nrow = 2, byrow = TRUE)

  out <- getHashKeys(x)

  expect_type(out, "character")
  expect_length(out, 2)

  expect_identical(out[1], "x1:1|x2:2")
  expect_identical(out[2], "x1:3|x2:4")

})

# ------------------------------------------------------------------
# Test: getHashKeys uses existing colnames when present
# Input:
#   - 2x2 matrix with colnames c("A","B")
# Behaviour:
#   - returns "A:<val>|B:<val>" per row
# Expectations:
#   - correct key strings for each row
# Why:
#   - ensures keys remain consistent when names exist.
# ------------------------------------------------------------------
test_that("getHashKeys uses colnames when present", {

  x <- matrix(c(1, 2, 3, 4), nrow = 2, byrow = TRUE)
  colnames(x) <- c("A", "B")

  out <- getHashKeys(x)

  expect_identical(out[1], "A:1|B:2")
  expect_identical(out[2], "A:3|B:4")

})


# Tests for cummulativeMovingAverage -------------------------------------------

# ------------------------------------------------------------------
# Test: cummulativeMovingAverage returns running mean
# Input:
#   - x = c(2,4,6,8)
# Behaviour:
#   - returns cumsum(x) / seq_along(x)
# Expectations:
#   - equals c(2,3,4,5)
# Why:
#   - check for helper used in summaries.
# ------------------------------------------------------------------
test_that("cummulativeMovingAverage returns running mean", {

  x <- c(2, 4, 6, 8)

  out <- cummulativeMovingAverage(x)

  expect_equal(out, c(2, 3, 4, 5))

})


# Tests for firstUpper ----------------------------------------------------------

# ------------------------------------------------------------------
# Test: firstUpper capitalizes only the first character
# Input:
#   - "hello", "Hello", "h"
# Behaviour:
#   - replaces first character with uppercase
# Expectations:
#   - correct transformed strings
# Why:
#   - ensures string formatting is deterministic.
# ------------------------------------------------------------------
test_that("firstUpper capitalizes first character", {

  expect_identical(firstUpper("hello"), "Hello")
  expect_identical(firstUpper("Hello"), "Hello")
  expect_identical(firstUpper("h"), "H")

})


# Tests for getBlankString ------------------------------------------------------

# ------------------------------------------------------------------
# Test: getBlankString returns a string of spaces of requested length
# Input:
#   - length = 0, 3, 10
# Behaviour:
#   - concatenates that many " " characters
# Expectations:
#   - correct string and nchar
# Why:
#   - used for formatting and alignment in text outputs.
# ------------------------------------------------------------------
test_that("getBlankString returns requested number of spaces", {
  out0 <- getBlankString(0)
  expect_type(out0, "character")
  expect_length(out0, 0)

  out5 <- getBlankString(5)
  expect_identical(out5, "     ")
  expect_type(out5, "character")
  expect_length(out5, 1)
})


# Tests for getListLevel --------------------------------------------------------

# ------------------------------------------------------------------
# Test: getListLevel returns nesting depth
# Input:
#   - non-list and nested lists
# Behaviour:
#   - returns 0 for non-list; otherwise 1 + depth of first element
# Expectations:
#   - correct depths for 0..3 levels
# Why:
#   - scaleRoundList relies on correct depth to traverse lists.
# ------------------------------------------------------------------
test_that("getListLevel returns correct nesting depth", {

  expect_identical(getListLevel(1), 0)
  expect_identical(getListLevel(list(1)), 1)
  expect_identical(getListLevel(list(list(1))), 2)
  expect_identical(getListLevel(list(list(list(1)))), 3)

})


# Tests for getRowIndexOfVectorInMatrix ----------------------------------------

# ------------------------------------------------------------------
# Test: getRowIndexOfVectorInMatrix finds exact matching row
# Input:
#   - matrix with known rows
#   - vectors equal to those rows
# Behaviour:
#   - returns the indices of rows where all columns match
# Expectations:
#   - correct row index returned; empty integer for no match
# Why:
#   - used to map trial patterns to unique rows.
# ------------------------------------------------------------------
test_that("getRowIndexOfVectorInMatrix finds matching rows", {

  m <- matrix(
    c(1, 2,
      3, 4,
      1, 9),
    ncol  = 2,
    byrow = TRUE
  )

  expect_identical(getRowIndexOfVectorInMatrix(c(1, 2), m), 1L)
  expect_identical(getRowIndexOfVectorInMatrix(c(3, 4), m), 2L)
  expect_identical(getRowIndexOfVectorInMatrix(c(1, 9), m), 3L)

  expect_length(getRowIndexOfVectorInMatrix(c(7, 7), m), 0)

})

# ------------------------------------------------------------------
# Test: getRowIndexOfVectorInMatrix errors if vector length != ncol(matrix)
# Input:
#   - vector of wrong length
# Behaviour:
#   - stops with informative error
# Expectations:
#   - error is thrown
# Why:
#   - prevents silent partial matching / wrong indexing.
# ------------------------------------------------------------------
test_that("getRowIndexOfVectorInMatrix errors on length mismatch", {

  m <- matrix(1:4, ncol = 2)

  expect_error(
    getRowIndexOfVectorInMatrix(c(1, 2, 3), m),
    "length of the vector must be equal"
  )

})


# Tests for invLogit ------------------------------------------------------------

# ------------------------------------------------------------------
# Test: invLogit matches stats::binomial()$linkinv and inverts logit
# Input:
#   - theta = 0 and theta = logit(p)
# Behaviour:
#   - returns inverse logit
# Expectations:
#   - invLogit(0) = 0.5
#   - invLogit(logit(p)) = p
# Why:
#   - core link utilities must be numerically correct.
# ------------------------------------------------------------------
test_that("invLogit returns correct inverse-logit values", {

  expect_equal(invLogit(0), 0.5)

  p <- c(0.1, 0.25, 0.8)

  expect_equal(invLogit(logit(p)), p, tolerance = 1e-12)

})

# ------------------------------------------------------------------
# Test: invLogit validates inputs
# Input:
#   - missing theta, non-numeric theta
# Behaviour:
#   - stops with the documented error
# Expectations:
#   - error thrown
# Why:
#   - input validation prevents silent misuse.
# ------------------------------------------------------------------
test_that("invLogit errors for missing or non-numeric input", {

  expect_error(invLogit(), "Please provide a numeric")
  expect_error(invLogit("x"), "Please provide a numeric")

})


# Tests for logit ---------------------------------------------------------------

# ------------------------------------------------------------------
# Test: logit matches stats::binomial()$linkfun and inverts invLogit
# Input:
#   - p = 0.5 and random p values in (0,1)
# Behaviour:
#   - returns log(p/(1-p))
# Expectations:
#   - logit(0.5) = 0
#   - invLogit(logit(p)) = p
# Why:
#   - correctness of link function is fundamental for Berry/ExNex models.
# ------------------------------------------------------------------
test_that("logit returns correct log-odds and inverts invLogit", {

  expect_equal(logit(0.5), 0)

  p <- c(0.1, 0.25, 0.8)

  expect_equal(invLogit(logit(p)), p, tolerance = 1e-12)

})

# ------------------------------------------------------------------
# Test: logit validates inputs
# Input:
#   - missing p, non-numeric p, p outside [0,1]
# Behaviour:
#   - stops with the documented error
# Expectations:
#   - error thrown
# Why:
#   - prevents invalid probability inputs leaking into models.
# ------------------------------------------------------------------
test_that("logit errors for invalid input", {

  expect_error(logit(), "Please provide a numeric")
  expect_error(logit("x"), "Please provide a numeric")
  expect_error(logit(c(-0.1, 0.2)), "Please provide a numeric")
  expect_error(logit(c(0.2, 1.2)), "Please provide a numeric")

})


# Tests for listPerMethod ------------------------------------------------------

# ------------------------------------------------------------------
# Test: listPerMethod reshapes scenario-first lists into method-first lists
# Input:
#   - analyses_list created via simulateScenarios() + performAnalyses()
#   - estimates_per_method from getEstimates() (method-first)
#   - scenario-first object reconstructed from estimates_per_method
# Behaviour:
#   - listPerMethod() should convert scenario-first -> method-first
# Expectations:
#   - Output equals estimates_per_method (same methods, same scenarios, same matrices)
# Why:
#   - Verifies listPerMethod performs pure reshaping on real package outputs.
# ------------------------------------------------------------------
test_that("listPerMethod reshapes scenario-first lists into method-first lists", {

  skip_if_not_installed("foreach")
  foreach::registerDoSEQ()
  set.seed(123)

  method_names <- c("pooled", "stratified")
  target_rates <- c(0.5, 0.5)

  scen <- simulateScenarios(
    n_subjects_list     = list(c(10, 10), c(20, 20)),
    response_rates_list = list(c(0.4, 0.4), c(0.6, 0.6)),
    n_trials = 2
  )

  analyses <- performAnalyses(
    scenario_list      = scen,
    evidence_levels    = c(0.05, 0.5, 0.95),
    target_rates       = target_rates,
    method_names       = method_names,
    n_mcmc_iterations  = 200,
    verbose            = FALSE
  )

  estimates_per_method <- getEstimates(
    analyses_list   = analyses,
    point_estimator = "median",
    alpha_level     = 0.10
  )

  method_names_from_obj <- names(estimates_per_method)
  scenario_names        <- names(estimates_per_method[[1]])

  expect_true(all(grepl("^scenario_[0-9]+$", scenario_names)))

  scenario_first <- vector("list", length = length(scenario_names))
  names(scenario_first) <- scenario_names

  for (s in scenario_names) {
    scenario_first[[s]] <- lapply(
      method_names_from_obj,
      function(m) estimates_per_method[[m]][[s]]
    )
    names(scenario_first[[s]]) <- method_names_from_obj
  }

  reshaped <- listPerMethod(scenario_first)

  expect_setequal(names(reshaped), method_names_from_obj)
  expect_setequal(names(reshaped[[method_names_from_obj[1]]]), scenario_names)

  for (m in method_names_from_obj) {
    for (s in scenario_names) {
      expect_equal(reshaped[[m]][[s]], estimates_per_method[[m]][[s]])
    }
  }
})


# Tests for roundList -----------------------------------------------------------

# ------------------------------------------------------------------
# Test: roundList rounds numerics for nesting levels 1..3
# Input:
#   - list structures with depths 1, 2, 3
# Behaviour:
#   - rounds each numeric entry to round_digits
# Expectations:
#   - correct rounded outputs
# Why:
#   - ensures list traversal is correct for supported nesting depths.
# ------------------------------------------------------------------
test_that("roundList rounds list entries for levels 1..3", {

  l1 <- list(1.234, 2.345)
  expect_equal(roundList(l1, round_digits = 1, list_levels = 1), list(1.2, 2.3))

  l2 <- list(a = list(1.234), b = list(2.345))
  expect_equal(roundList(l2, round_digits = 2, list_levels = 2), list(a = list(1.23), b = list(2.35)))

  l3 <- list(a = list(b = list(1.234)))
  expect_equal(roundList(l3, round_digits = 0, list_levels = 3), list(a = list(b = list(1))))

})

# ------------------------------------------------------------------
# Test: roundList errors for non-numeric entries and unsupported depth
# Input:
#   - list containing non-numeric
#   - list_levels > 3
# Behaviour:
#   - stops with informative error
# Expectations:
#   - error thrown
# Why:
#   - avoids silently returning wrong types.
# ------------------------------------------------------------------
test_that("roundList errors for non-numerics and unsupported depth", {

  expect_error(roundList(list("x"), round_digits = 1, list_levels = 1), "must contain numerics")

  expect_error(roundList(list(1), round_digits = 1, list_levels = 4), "greater than 3")

})


# Tests for scaleList -----------------------------------------------------------

# ------------------------------------------------------------------
# Test: scaleList scales numerics for nesting levels 1..3
# Input:
#   - list structures with depths 1, 2, 3
# Behaviour:
#   - multiplies each numeric by scale_param
# Expectations:
#   - correct scaled outputs
# Why:
#   - used by scaleRoundList; must behave deterministically.
# ------------------------------------------------------------------
test_that("scaleList scales list entries for levels 1..3", {

  l1 <- list(1, 2)
  expect_equal(scaleList(l1, scale_param = 10, list_levels = 1), list(10, 20))

  l2 <- list(a = list(1), b = list(2))
  expect_equal(scaleList(l2, scale_param = 0.5, list_levels = 2), list(a = list(0.5), b = list(1)))

  l3 <- list(a = list(b = list(2)))
  expect_equal(scaleList(l3, scale_param = 3, list_levels = 3), list(a = list(b = list(6))))

})

# ------------------------------------------------------------------
# Test: scaleList errors for non-numeric entries and unsupported depth
# Input:
#   - list containing non-numeric
#   - list_levels > 3
# Behaviour:
#   - stops with informative error
# Expectations:
#   - error thrown
# Why:
#   - protects against silent type coercion.
# ------------------------------------------------------------------
test_that("scaleList errors for non-numerics and unsupported depth", {

  expect_error(scaleList(list("x"), scale_param = 2, list_levels = 1), "must contain numerics")

  expect_error(scaleList(list(1), scale_param = 1, list_levels = 4), "greater than 3")

})


# Tests for scaleRoundList ------------------------------------------------------

# ------------------------------------------------------------------
# Test: scaleRoundList applies scaling then rounding
# Input:
#   - list of numerics
# Behaviour:
#   - scales by scale_param then rounds to round_digits
# Expectations:
#   - correct scaled+rounded values
# Why:
#   - user-facing helper; expected to behave consistently.
# ------------------------------------------------------------------
test_that("scaleRoundList scales then rounds list entries", {

  skip_if_not_installed("checkmate")

  l <- list(1.234, 2.345)

  out <- scaleRoundList(list = l, scale_param = 10, round_digits = 1)

  expect_equal(out, list(12.3, 23.5))

})

# ------------------------------------------------------------------
# Test: scaleRoundList returns unchanged list if round_digits is NULL
# Input:
#   - list of numerics, round_digits = NULL
# Behaviour:
#   - scales by scale_param, does no rounding when round_digits is NULL
# Expectations:
#   - equals scaled values (or original if scale_param = 1)
# Why:
#   - matches documented behaviour and avoids surprising rounding.
# ------------------------------------------------------------------
test_that("scaleRoundList skips rounding when round_digits is NULL", {

  skip_if_not_installed("checkmate")

  l <- list(1.234, 2.345)

  out <- scaleRoundList(list = l, scale_param = 1, round_digits = NULL)

  expect_equal(out, l)

})

# ------------------------------------------------------------------
# Test: scaleRoundList validates inputs
# Input:
#   - missing list / non-list
#   - invalid scale_param
#   - invalid round_digits
# Behaviour:
#   - stops with informative errors
# Expectations:
#   - error thrown
# Why:
#   - ensures correct user-facing input checks.
# ------------------------------------------------------------------
test_that("scaleRoundList errors for invalid inputs", {

  skip_if_not_installed("checkmate")

  expect_error(scaleRoundList(), "Please provide a list")
  expect_error(scaleRoundList(list = 1), "Please provide a list")

  l <- list(1.234)

  expect_error(scaleRoundList(list = l, scale_param = -1))
  expect_error(scaleRoundList(list = l, round_digits = -1))

})


# Tests for substrRight ---------------------------------------------------------

# ------------------------------------------------------------------
# Test: substrRight returns rightmost n characters
# Input:
#   - x = "abcdef", n = 1,2,6
# Behaviour:
#   - returns substring from the right
# Expectations:
#   - correct suffixes
# Why:
#   - used for parsing parameter/cohort suffix indices.
# ------------------------------------------------------------------
test_that("substrRight returns rightmost n characters", {

  expect_identical(substrRight("abcdef", 1), "f")
  expect_identical(substrRight("abcdef", 2), "ef")
  expect_identical(substrRight("abcdef", 6), "abcdef")

})

#'
#' #' @title PlotCumMovAvgGoProbs
#' #' @description This function plots the cummulative moving average of the go probabilities per
#' #' scenario for a given method.
#' #' @param scenario.decisions.list A list with decisions per scenario created with GetGoDecisions()
#' #' @param method.name A strings for the name of the method to be used.
#' #' @rdname PlotCumMovAvgGoProbs
#' #' @export
#' PlotCumMovAvgGoProbs <- function (
#'
#'   scenario.decisions.list,
#'   method.name
#'
#' ) {
#'
#'   n.scenarios <- length(scenario.decisions.list)
#'
#'   par(mfrow = c(n.scenarios, 1),
#'       mar   = c(2, 4, 0, 0) + 0.1)
#'
#'   for (s in seq_along(scenario.decisions.list)) {
#'
#'     if (!method.name %in%
#'         scenario.decisions.list[[s]]$analysis.data$analysis.parameters$method.names) {
#'
#'       stop ("method.name not in scenario.decisions.list")
#'
#'     }
#'
#'     decisions <- scenario.decisions.list[[s]]$decisions.list[[method.name]]
#'
#'     cumMA <- apply(decisions, 2, CummulativeMovingAverage)
#'
#'     plot(0, type = "n",
#'          xlab = "",
#'          ylab = "Go Probability",
#'          xlim = c(1, nrow(cumMA)),
#'          ylim = c(0, 1))
#'
#'     grid()
#'
#'     for (i in seq_len(ncol(cumMA))) {
#'       lines(cumMA[, i], col = i)
#'     }
#'
#'     legend("topright", bty = "n", lty = 1, cex = 1.25,
#'            col = rgb(0, 0, 0, max = 1, alpha = 0),
#'            legend = paste0(
#'              "Scenario ",
#'              scenario.decisions.list[[s]]$scenario.data$scenario.number))
#'
#'     if (s == 1) {
#'
#'       legend("topleft", legend = FirstUpper(colnames(decisions)),
#'              lty = 1, col = seq_len(ncol(cumMA)))
#'
#'     }
#'
#'   }
#'
#' }

# Setup helpers for tests

setup_applicable_prev_trials <- function() {
  set.seed(123)
  
  scen_initial <- simulateScenarios(
    n_subjects_list     = list(c(10, 20, 30)),
    response_rates_list = list(c(0.4, 0.6, 0.8)),
    n_trials            = 10
  )
  
  analysis_initial <- performAnalyses(
    scenario_list       = scen_initial,
    target_rates        = c(0.5, 0.5, 0.5),
    method_names        = "pooled",
    n_mcmc_iterations   = 50,
    verbose             = FALSE
  )
  
  go_initial <- getGoDecisions(
    analyses_list   = analysis_initial,
    cohort_names    = c("p_1", "p_2", "p_3"),
    evidence_levels = c(0.5, 0.5, 0.5),
    boundary_rules  = quote(c(TRUE, TRUE, TRUE))
  )
  
  scen_next <- continueRecruitment(
    n_subjects_add_list = list(c(5, 5, 5)),
    decisions_list      = go_initial,
    method_name         = "pooled"
  )
  
  analysis_with_diff <- performAnalyses(
    scenario_list       = scen_initial,
    target_rates        = c(0.5, 0.5, 0.5),
    method_names        = "pooled",
    calc_differences    = c(3, 2),
    n_mcmc_iterations   = 50,
    verbose             = FALSE
  )
  
  go_with_diff <- getGoDecisions(
    analyses_list   = analysis_with_diff,
    cohort_names    = c("p_1", "p_2", "p_3"),
    evidence_levels = c(0.5, 0.5, 0.5),
    boundary_rules  = quote(c(TRUE, TRUE, TRUE))
  )
  
  scen_next_diff <- continueRecruitment(
    n_subjects_add_list = list(c(5, 5, 5)),
    decisions_list      = go_with_diff,
    method_name         = "pooled"
  )
  
  method_names_prev <- analysis_initial[[1]]$analysis_parameters$method_names
  quantiles_prev    <- analysis_initial[[1]]$analysis_parameters$quantiles
  n_coh_prev        <- ncol(scen_next[[1]]$n_subjects)
  
  list(
    scen_initial       = scen_initial,
    analysis_initial   = analysis_initial,
    go_initial         = go_initial,
    scen_next          = scen_next,
    analysis_with_diff = analysis_with_diff,
    go_with_diff       = go_with_diff,
    scen_next_diff     = scen_next_diff,
    method_names_prev  = method_names_prev,
    quantiles_prev     = quantiles_prev,
    n_coh_prev         = n_coh_prev
  )
}

make_scenario_list_pa <- function() {
  set.seed(123)
  simulateScenarios(
    n_subjects_list     = list(c(10, 20, 30)),
    response_rates_list = list(c(0.3, 0.5, 0.7)),
    n_trials            = 5
  )
}

setup_two_scenario_pooled <- function(
    n_subj,
    rr1,
    rr2,
    n_trials = 30,
    n_mcmc_iterations = 20,
    seed = NULL
) {
  if (!is.null(seed)) set.seed(seed)
  
  scen <- simulateScenarios(
    n_subjects_list     = list(n_subj, n_subj),
    response_rates_list = list(rr1, rr2),
    n_trials            = n_trials
  )
  
  analyses <- performAnalyses(
    scenario_list      = scen,
    method_names       = "pooled",
    target_rates       = rep(0.5, length(n_subj)),
    n_mcmc_iterations  = n_mcmc_iterations,
    verbose            = FALSE
  )
  
  list(
    scen     = scen,
    analyses = analyses
  )
}

getAllCohortNames <- function (

  analyses_list

) {

  Reduce(
    intersect,
    sapply(analyses_list, function (x) {
      lapply(x$quantiles_list, function (y) {

        post_names <- Reduce(intersect, lapply(y, colnames))
        indices    <- grep("p_", post_names)

        post_names[indices]

      })
    })
  )

}

#' @title getAverageNSubjects
#' @md
#' @description This function calculates the average number of subjects per scenario.
#' @param scenario_list An object of class `scenario_list`,
#' as created with \code{\link[bhmbasket]{simulateScenarios}} or
#' \code{\link[bhmbasket]{continueRecruitment}}
#' @details This function can be useful to assess decision rules with regard to the average number
#' of subjects across scenarios when performing interim analyses.
#' @return A named list of vectors for the average number of subjects in each scenario.
#' @rdname getAverageNSubjects
#' @seealso
#'  \code{\link[bhmbasket]{simulateScenarios}}
#'  \code{\link[bhmbasket]{continueRecruitment}}
#' @examples
#' interim_scenarios <- simulateScenarios(
#'   n_subjects_list     = list(c(10, 20, 30)),
#'   response_rates_list = list(rep(0.9, 3)),
#'   n_trials            = 10)
#'
#' interim_analyses <- performAnalyses(
#'   scenario_list       = interim_scenarios,
#'   target_rates        = rep(0.5, 3),
#'   n_mcmc_iterations   = 100)
#'
#' interim_gos <- getGoDecisions(
#'   analyses_list       = interim_analyses,
#'   cohort_names        = c("p_1", "p_2", "p_3"),
#'   evidence_levels     = c(0.5, 0.8, 0.5),
#'   boundary_rules      = quote(c(x[1] > 0.8, x[2] > 0.6, x[3] > 0.7)))
#'
#' scenarios_list <- continueRecruitment(
#'   n_subjects_add_list = list(c(30, 20, 10)),
#'   decisions_list      = interim_gos,
#'   method_name         = "exnex_adj")
#'
#' getAverageNSubjects(scenarios_list)
#'
#' @author Stephan Wojciekowski
#' @export
getAverageNSubjects <- function (

  scenario_list

) {

  error_scenario_list <-
    "Providing an object of class 'scenario_list' for the argument 'scenario_list'"

  checkmate::assertClass(
    scenario_list,
    classes   = "scenario_list",
    .var.name = error_scenario_list
  )

  lapply(scenario_list, function (x) colMeans(x$n_subjects))

}

#' @title getEstimates
#' @md
#' @description This function calculates the point estimates and credible intervals per cohort,
#' as well as estimates of the biases and the mean squared errors of the point estimates per cohort.
#' @param analyses_list An object of class `analysis_list`,
#' as created with \code{\link[bhmbasket]{performAnalyses}}
#' @param add_parameters A vector of strings naming additional parameters
#' from the Bayesian hierarchical models, e.g. `c('mu', 'tau')`.
#' If `NULL`, no additional parameters will be evaluated,
#' Default: `NULL`
#' @param point_estimator A string indicating the type of estimator used for calculation of
#' bias and MSE. Must be one of `'median'` or `'mean'`
#' @param alpha_level A numeric in (0, 1) for the level of the credible interval.
#' Only values corresponding to quantiles saved in \code{\link[bhmbasket]{performAnalyses}}
#' will work, Default: `0.05`
#' @details Bias and MSE will only be calculated for response rate estimates of simulated trials.
#' For additional parameters, bias and MSE will not be calculated.
#'
#' Possible additional parameters for the Bayesian hierarchical models are
#' `c('mu', 'tau')` for `'berry'`, `'exnex'`, `'exnex_mix'`, `'exnex_adj'`,
#' and `'exnex_adj_mix'`.
#' The ExNex-based models can also access posterior weights.
#' `paste0("w_", seq_len(n_cohorts))`.
#' @return A named list of matrices of estimates of response rates and credible intervals.
#' Estimates of bias and MSE are included for response rate estimates of simulated trials.
#' @rdname getEstimates
#' @seealso
#'  \code{\link[bhmbasket]{createTrial}}
#'  \code{\link[bhmbasket]{simulateScenarios}}
#'  \code{\link[bhmbasket]{performAnalyses}}
#' @examples
#'   scenarios_list <- simulateScenarios(
#'     n_subjects_list     = list(c(10, 20, 30)),
#'     response_rates_list = list(c(0.1, 0.2, 3)),
#'     n_trials            = 10)
#'
#'   analyses_list <- performAnalyses(
#'     scenario_list       = scenarios_list,
#'     target_rates        = c(0.1, 0.1, 0.1),
#'     calc_differences    = matrix(c(3, 2, 2, 1), ncol = 2),
#'     n_mcmc_iterations   = 100)
#'
#'   getEstimates(analyses_list)
#'   getEstimates(analyses_list   = analyses_list,
#'                add_parameters  = c("mu", "tau", "w_1", "w_2", "w_3"),
#'                point_estimator = "mean",
#'                alpha_level     = 0.1)
#'
#'   outcome <- createTrial(
#'     n_subjects          = c(10, 20, 30),
#'     n_responders        = c( 1,  2,  3))
#'
#'   outcome_analysis <- performAnalyses(
#'     scenario_list       = outcome,
#'     target_rates        = c(0.1, 0.1, 0.1),
#'     n_mcmc_iterations   = 100)
#'
#'   getEstimates(outcome_analysis)
#'   getEstimates(analyses_list  = outcome_analysis,
#'                add_parameters = c("mu", "w_1", "w_2", "w_3"))
#' @author Stephan Wojciekowski
#' @export
getEstimates <- function (

  analyses_list,
  add_parameters  = NULL,
  point_estimator = "median",
  alpha_level     = 0.05

) {

  error_analyses_list <-
    "Providing an object of class 'analysis_list' for the argument 'analyses_list'"
  error_add_parameters <- paste(
    "Providing either NULL or a vector of strings for the argument 'add_parameters',",
    "naming additional parameters of the applied model(s)"
  )
  error_point_estimator <-
    "Providing either 'median' or 'mean' for the argument 'point_estimator'"
  error_alpha_level <-
    "Providing a numeric in (0, 1) for the argument 'alpha_level'"

  checkmate::assertClass(
    analyses_list,
    classes   = "analysis_list",
    .var.name = error_analyses_list
  )

  checkmate::assertCharacter(
    add_parameters,
    null.ok   = TRUE,
    .var.name = error_add_parameters
  )

  checkmate::assertChoice(
    point_estimator,
    choices   = c("median", "mean"),
    .var.name = error_point_estimator
  )

  checkmate::assertNumber(
    alpha_level,
    finite    = TRUE,
    .var.name = error_alpha_level
  )
  checkmate::assertTRUE(
    alpha_level > 0 & alpha_level < 1,
    .var.name = error_alpha_level
  )

  available_quantiles <- round(analyses_list[[1]]$analysis_parameters$quantiles, 9)
  asked_quantiles     <- round(c(alpha_level / 2, 1 - alpha_level / 2), 9)
  if (any(!asked_quantiles %in% available_quantiles)) {
    stop(paste(
      "The 'alpha_level' must be among the stored quantiles in 'analyses_list',",
      "e.g. 1 - alpha_level must be among the evidence_levels in performAnalyses()"
    ))
  }

  ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ###

  cohort_names <- getAllCohortNames(analyses_list)
  diff_indices <- grepl("diff", cohort_names)

  ## Filter for cohort names in add_parameters
  add_parameters <- add_parameters[!grepl("p_", add_parameters)]

  ## Lists to hold the results
  results_list <- vector(mode = "list", length = length(analyses_list))
  names(results_list) <- names(analyses_list)

  for (s in seq_along(analyses_list)) {

    true_rr <- analyses_list[[s]]$scenario_data$response_rates

    ## calculate the historic response rates
    hist_index <- true_rr <= 0 | true_rr >= 1
    if (any(hist_index)) {

      hist_rr <- sapply(which(hist_index), function (x) {
        true_rr[x] / analyses_list[[s]]$scenario_data$n_subjects[1, x]
      })

      true_rr[hist_index] <- hist_rr

    }

    ## calculate the differences in true rr if necessary
    if (any(diff_indices)) {

      diff_cohorts <- sapply(
        strsplit(substrRight(cohort_names[diff_indices], 2), ""),
        as.numeric
      )

      true_diff_rr <- t(apply(diff_cohorts, 2, function (x) {
        diff(true_rr[x])
      }))

    } else {

      true_diff_rr <- NULL

    }

    ## must be after historic response rates
    true_rr <- cbind(true_rr, true_diff_rr)

    ## prepare loop
    method_names <- analyses_list[[s]]$analysis_parameters$method_names

    results_per_method_list <- vector(mode = "list", length = length(method_names))
    names(results_per_method_list) <- method_names

    ## prepare check that additional parameters occur in at least one of the methods
    if (!is.null(add_parameters)) {
      occurences <- vector(length = length(method_names))
    }

    for (n in seq_along(method_names)) {

      if (!is.null(add_parameters)) {

        add_index <- add_parameters %in%
          colnames(analyses_list[[s]]$quantiles_list[[n]][[1]])

        if (any(add_index)) {
          occurences[n] <- TRUE
        }

        parameter_names <- c(cohort_names, add_parameters[add_index])

      } else {

        parameter_names <- cohort_names

      }

      ## Setting up gamma levels
      n_parameters <- length(parameter_names)

      gamma_levels <- c(
        rep("mean",              n_parameters),
        rep("sd",                n_parameters),
        rep(1 - alpha_level / 2, n_parameters),
        rep(0.5,                 n_parameters),
        rep(alpha_level / 2,     n_parameters)
      )

      matrix_estimates <- do.call(
        rbind,
        getPosteriorGammaQuantiles(
          method_name              = method_names[n],
          gamma_levels             = gamma_levels,
          quantiles                = analyses_list[[s]]$analysis_parameters$quantiles,
          posterior_quantiles_list = analyses_list[[s]]$quantiles_list,
          cohort_names             = rep(parameter_names, times = 5)
        )
      )

      ## Mean, SD & Quantiles
      post_quantiles <- matrix(colMeans(matrix_estimates), ncol = 5)

      rownames(post_quantiles) <- parameter_names
      colnames(post_quantiles) <- c(
        "Mean", "SD",
        paste0(c(alpha_level / 2, 0.5, 1 - alpha_level / 2) * 100, "%")
      )

      ## Bias and MSE
      if (point_estimator == "median") {
        indices_point_est <- n_parameters * 3 + seq_len(n_parameters)
      } else {
        indices_point_est <- seq_len(n_parameters)
      }

      matrix_estimates <- matrix_estimates[, indices_point_est]

      ## if only a single trial (i.e. a trial outcome) has been evaluated
      if (is.null(dim(matrix_estimates))) {

        estimates <- post_quantiles

      } else {

        ## find all estimates for response rates
        rr_index         <- grepl("p_", colnames(matrix_estimates))
        matrix_estimates <- matrix_estimates[, rr_index]

        point_estimates  <- as.matrix(colMeans(matrix_estimates))
        var_estimates    <- as.matrix(apply(matrix_estimates, 2, stats::var))

        bias_estimates <- point_estimates - t(true_rr)
        mse_estimates  <- bias_estimates^2 + var_estimates

        colnames(bias_estimates) <- "Bias"
        colnames(mse_estimates)  <- "MSE"

        ## Combine results
        ## introduce NAs for all values that are not response rates
        na_matrix      <- matrix(NA, nrow = sum(!rr_index), ncol = 1)

        bias_estimates <- rbind(bias_estimates, na_matrix)
        mse_estimates  <- rbind(mse_estimates,  na_matrix)

        estimates <- cbind(post_quantiles, bias_estimates, mse_estimates)

      }

      ## Save results
      results_per_method_list[[method_names[n]]] <- estimates

    }

    if (!is.null(add_parameters) && all(!occurences)) {
      stop(paste(
        "The additional parameters provided in 'add_parameters' do not occur",
        "in any of the methods stored in 'analyses_list'"
      ))
    }

    results_list[[s]] <- results_per_method_list

  }

  return(listPerMethod(results_list))

}

getGammaIndices <- function (

  gamma_levels,
  quantiles

) {

  gamma_indices <- sapply(gamma_levels, function (g) {

    if (is.character(g)) {

      if (g == "mean") {

        gamma_index <- length(quantiles) + 1L

      } else if (g == "sd") {

        gamma_index <- length(quantiles) + 2L

      } else {

        g_numeric <- tryCatch({as.numeric(g)}, warning = function(w) w)

        if (inherits(g_numeric, "warning")) {
          stop(paste(
            "The only strings allowed for the argument 'evidence_levels' are",
            "'mean' and 'sd'"
          ))
        }

        gamma_index <- getNumericGammaIndex(g_numeric, quantiles)

      }

    } else {

      gamma_index <- getNumericGammaIndex(g, quantiles)

    }

    return(gamma_index)

  })

  return(gamma_indices)

}

getGoBoundaries <- function (

  scenario_analysis_list,
  cohort_names,
  go_rates,
  gamma_levels_list,
  method_names

) {"dummy function"}

#' @title getGoDecisions
#' @description This function applies decision rules to the analyzed trials.
#' The resulting \code{decision_list} can be further processed with
#' \code{\link[bhmbasket]{getGoProbabilities}} or
#' \code{\link[bhmbasket]{continueRecruitment}}.
#' @param analyses_list An object of class \code{analysis_list},
#' as created with \code{\link[bhmbasket]{performAnalyses}}
#' @param cohort_names A vector of strings with the names of the cohorts, e.g.
#' \code{c('p_1', 'p_2')}
#' @param evidence_levels A vector of numerics in \code{(0, 1)} for the
#' posterior probability thresholds for the cohorts.
#' Will be recycled to match the number of methods in the \code{analyses_list}
#' @param boundary_rules A quote of a vector for the boundary rules,
#' \code{quote(c(...))}, see details.
#' The number of decisions to be taken must match the number of cohorts.
#' Will be recycled to match the number of methods in the \code{analyses_list}
#' @param overall_min_gos A positive integer for the minimum number of
#' cohort-wise go decisions required for an overall go decision
#'  Default: \code{1}
#' @return An object of class \code{decision_list}
#' @details This function applies decision rules of the following type to the
#' outcomes of (simulated) basket trials with binary endpoints:
#' \deqn{P(p_j|data > p_{B,j}) > \gamma,}
#' where \eqn{p_j|data} is the posterior response rate of cohort \eqn{j},
#' \eqn{p_{B,j}} is the response rate boundary of cohort \eqn{j},
#' and \eqn{\gamma} is the evidence level.
#' This rule can equivalently be written as \deqn{q_{1-\gamma,j} > p_{B,j},}
#' where \eqn{q_{1-\gamma,j}} is the \eqn{1-\gamma}-quantile of the posterior
#' response rate of cohort \eqn{j}.
#'
#' The arguments \code{cohort_names} and \code{evidence_levels} determine
#' \eqn{q_{1-\gamma,j}}, where the entries of \code{cohort_names} and
#' \code{evidence_levels} are matched corresponding to their order.
#'
#' The argument \code{boundary_rules} provides the rules that describe what
#' should happen with  the posterior quantiles \eqn{q_{1-\gamma,j}}.
#' The first posterior quantile determined by the first items of
#' \code{cohort_names} and \code{evidence_levels} is referred to as \code{x[1]},
#' the second as \code{x[2]}, etc.
#' Using the \code{quote(c(...))}-notation,
#' many different rules can be implemented.
#' A decision rule for only one cohort would be
#' \code{boundary_rules = quote(c(x[1] > 0.1))},
#' \code{cohort_names = 'p_1'}, and \code{evidence_levels = 0.5},
#' which implements the rule \eqn{P(p_1|data > 0.1) > 0.5}.
#' The number of decisions to be taken must match the number of cohorts, i.e.
#' for each cohort there must be a decision rule in the vector separated by a comma.
#' See the example section for a decision rule for more than one cohort and
#' the example of \code{\link[bhmbasket]{negateGoDecisions}}
#' for the implementation of a more complex decision rule.
#' @seealso
#'  \code{\link[bhmbasket]{performAnalyses}}
#'  \code{\link[bhmbasket]{getGoProbabilities}}
#'  \code{\link[bhmbasket]{negateGoDecisions}}
#'  \code{\link[bhmbasket]{continueRecruitment}}
#' @rdname getGoDecisions
#' @examples
#' scenarios_list <- simulateScenarios(
#'   n_subjects_list     = list(c(10, 20, 30)),
#'   response_rates_list = list(c(0.1, 0.1, 0.9)),
#'   n_trials            = 10)
#'
#' analyses_list <- performAnalyses(
#'   scenario_list      = scenarios_list,
#'   target_rates       = rep(0.5, 3),
#'   n_mcmc_iterations  = 100)
#'
#' ## Decision rule for more than one cohort
#' decisions_list <- getGoDecisions(
#'   analyses_list   = analyses_list,
#'   cohort_names    = c("p_1", "p_2", "p_3"),
#'   evidence_levels = c(0.5, 0.5, 0.8),
#'   boundary_rules  = quote(c(x[1] > 0.7, x[2] < 0.3, x[3] < 0.6)))
#'
#' ## Decision rule for only two of the three cohorts
#' decisions_list <- getGoDecisions(
#'   analyses_list   = analyses_list,
#'   cohort_names    = c("p_1", "p_3"),
#'   evidence_levels = c(0.5, 0.8),
#'   boundary_rules  = quote(c(x[1] > 0.7, TRUE, x[3] < 0.6)),
#'   overall_min_gos = 2L)
#'
#' ## Different decision rules for each method
#' ## This works the same way for the different evidence_levels
#' decisions_list <- getGoDecisions(
#'   analyses_list   = analyses_list,
#'   cohort_names    = c("p_1", "p_2", "p_3"),
#'   evidence_levels = c(0.5, 0.5, 0.8),
#'   boundary_rules  = list(
#'     quote(c(x[1] > 0.1,  x[2] < 0.5,  x[3] < 0.1)),   # "berry"
#'     quote(c(x[1] > 0.2,  x[2] < 0.4,  x[3] < 0.2)),   # "exnex"
#'     quote(c(x[1] > 0.25, x[2] < 0.35, x[3] < 0.25)),  # "exnex_adj"
#'     quote(c(x[1] > 0.3,  x[2] < 0.3,  x[3] < 0.3)),   # "exnex_adj_mix"
#'     quote(c(x[1] > 0.35, x[2] < 0.25, x[3] < 0.35)),  # "exnex_mix"
#'     quote(c(x[1] > 0.4,  x[2] < 0.2,  x[3] < 0.4)),   # "pooled"
#'     quote(c(x[1] > 0.45, x[2] < 0.15, x[3] < 0.45)),  # "stratified"
#'     quote(c(x[1] > 0.5,  x[2] < 0.1,  x[3] < 0.5))    # "stratified_mix"
#'   ))
#' @author Stephan Wojciekowski
#' @export
getGoDecisions <- function (

  analyses_list,

  cohort_names,
  evidence_levels,
  boundary_rules,

  overall_min_gos = 1

) {

  error_analyses_list <-
    "Providing an object of class 'analysis_list' for the argument 'analyses_list'"
  error_cohort_names <- paste(
    "Providing a vector of strings for the argument 'cohort_names',",
    "e.g. c('p_1','p_2')"
  )
  error_evidence_levels <-
    "Providing a vector of numerics in (0, 1) for the argument 'evidence_levels'"
  error_boundary_rules <- paste(
    "Providing a quote(c(...)) for the argument 'boundary_rules'.",
    "The vector c(...) inside the quote() must have the same length as the number of cohorts.",
    "See ?getGoDecisions for details"
  )
  error_overall_min_gos <-
    "Providing a positive integer for the argument 'overall_min_gos'"

  checkmate::assertClass(
    analyses_list,
    classes   = "analysis_list",
    .var.name = error_analyses_list
  )

  checkmate::assertCharacter(
    cohort_names,
    min.len   = 1,
    any.missing = FALSE,
    .var.name = error_cohort_names
  )

  checkmate::assertInt(
    overall_min_gos,
    lower     = 1,
    .var.name = error_overall_min_gos
  )

  checkmate::assertSubset(
    cohort_names,
    choices   = colnames(analyses_list[[1]]$quantiles_list[[1]][[1]]),
    .var.name = error_cohort_names
  )

  if (missing(evidence_levels)) {
    stop(error_evidence_levels)
  }
  if (missing(boundary_rules)) {
    stop(error_boundary_rules)
  }

  if (is.list(evidence_levels)) {
    for (i in seq_along(evidence_levels)) {
      check.evidence.levels(
        evidence_levels[[i]],
        cohort_names,
        analyses_list,
        error_evidence_levels
      )
    }
  } else {
    check.evidence.levels(
      evidence_levels,
      cohort_names,
      analyses_list,
      error_evidence_levels
    )
  }

  check_boundary_rules <- tryCatch({
    x <- stats::runif(n = length(cohort_names), min = 0.001, max = 0.999)
    if (is.list(boundary_rules)) {
      for (i in seq_along(boundary_rules)) {
        if (!is.language(boundary_rules[[i]]))                                   stop()
        if (!identical(boundary_rules[[i]][1], quote(c())))                      stop()
        if (!identical(
          length(boundary_rules[[i]]) - 1L,
          ncol(analyses_list$scenario_1$scenario_data$n_subjects)
        )) stop()
        eval(boundary_rules[[i]])
      }
    } else {
      if (!is.language(boundary_rules))                                          stop()
      if (!identical(boundary_rules[1], quote(c())))                             stop()
      ## fix number of decisions to number of cohorts
      if (!identical(
        length(boundary_rules) - 1L,
        ncol(analyses_list$scenario_1$scenario_data$response_rates)
      )) stop()
      eval(boundary_rules)
    }
    rm(x)
  }, error = function (e) e)

  if (inherits(check_boundary_rules, "error")) {
    stop(error_boundary_rules)
  }

  ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ###

  gamma_levels <- evidence_levels

  ## Get method names
  method_names_matrix <- t(sapply(
    analyses_list,
    function (x) x$analysis_parameters$method_names
  ))
  if (!all(sapply(seq_len(nrow(method_names_matrix)), function (x) {
    identical(method_names_matrix[1, ], method_names_matrix[x, ])
  }))) {
    stop("The scenarios where analysed with different methods")
  }
  method_names <- method_names_matrix[1, ]

  ## in case only one method has been used for all scenarios
  if (all(sapply(seq_along(method_names), function (x) {
    method_names[1] == method_names[x]
  }))) {
    method_names <- method_names[1]
  }

  ## check for input consistency
  if (!is.list(boundary_rules)) {
    boundary_rules <- list(boundary_rules)
  }
  if (!is.list(gamma_levels)) {
    gamma_levels <- list(gamma_levels)
  }

  if (length(method_names) < length(boundary_rules)) {
    stop(paste0(
      "The lengths of 'boundary_rules' must be less than or equal to",
      " the length of 'method_names'"
    ))
  } else if (length(method_names) > length(boundary_rules)) {
    boundary_rules <- rep(boundary_rules, length.out = length(method_names))
  }
  if (length(method_names) < length(gamma_levels)) {
    stop(paste0(
      "The lengths of 'evidence_levels' must be less than or equal to",
      " the length of 'method_names'"
    ))
  } else if (length(method_names) > length(gamma_levels)) {
    gamma_levels <- rep(gamma_levels, length.out = length(method_names))
  }

  decisions_list <- vector(mode = "list", length = length(analyses_list))
  names(decisions_list) <- names(analyses_list)

  for (s in seq_along(decisions_list)) {

    analysis_data <- analyses_list[[s]]

    methods_decisions_list <- vector(mode = "list", length = length(method_names))
    names(methods_decisions_list) <- method_names

    for (n in seq_along(method_names)) {

      go_decisions <- getGoDecisionsByCohort(
        gamma_quantiles = getPosteriorGammaQuantiles(
          method_name              = method_names[n],
          gamma_levels             = gamma_levels[[n]],
          quantiles                = analysis_data$analysis_parameters$quantiles,
          posterior_quantiles_list = analysis_data$quantiles_list,
          cohort_names             = cohort_names
        ),
        decision_rule = boundary_rules[[n]]
      )

      ## combine new decision outcomes with previous decisions (and convert to logical)
      previous_gos <- analysis_data$scenario_data$previous_analyses$go_decisions[, -1]
      go_decisions <- go_decisions * previous_gos > 0

      ## Overall go:
      overall_go   <- apply(go_decisions, 1, function (x) sum(x) >= overall_min_gos)
      go_decisions <- cbind(overall = overall_go, go_decisions)

      ## store
      methods_decisions_list[[n]] <- go_decisions

    }

    decisions_list[[s]] <-
      list(
        decisions_list = methods_decisions_list,
        analysis_data  = list(
          quantiles_list      = analyses_list[[s]]$quantiles_list,
          analysis_parameters = analyses_list[[s]]$analysis_parameters
        ),
        scenario_data  = analyses_list[[s]]$scenario_data,
        decision_rules = list(
          cohort_names   = cohort_names,
          gamma_levels   = gamma_levels,
          boundary_rules = boundary_rules
        )
      )
  }

  names(decisions_list) <- names(analyses_list)

  class(decisions_list) <- "decision_list"

  return(decisions_list)

}

getGoDecisionsByCohort <- function (

  gamma_quantiles,

  boundary_rates = NULL,
  decision_rule  = quote(x > boundary_rates)

) {

  go_decisions_list <- lapply(
    gamma_quantiles,
    function (x) {
      eval(decision_rule)
    }
  )

  go_decisions <- matrix(
    unlist(go_decisions_list),
    nrow = length(go_decisions_list),
    byrow = TRUE
  )

  colnames(go_decisions) <- paste0("decision_", seq_len(ncol(go_decisions)))

  return(go_decisions)

}

#' @title getGoProbabilities
#' @md
#' @description Calculates the Go probabilities for given decisions
#' @param go_decisions_list An object of class `decision_list`,
#' as returned by \code{\link[bhmbasket]{getGoDecisions}}
#' @param nogo_decisions_list An object of class `decision_list`,
#' as returned by \code{\link[bhmbasket]{getGoDecisions}}, Default: `NULL`
#' @return A list of matrices of Go (and Consider and NoGo) probabilities
#' @details If only `go_decisions_list` is provided
#' (i.e. `nogo_decisions_list` is `NULL`),
#' only Go probabilities will be calculated.
#' If both `go_decisions_list` and `nogo_decisions_list` are provided,
#' Go, Consider, and NoGo probabilities will be calculated.
#' @seealso
#'  \code{\link[bhmbasket]{getGoDecisions}}
#' @rdname getGoProbabilities
#' @examples
#' scenarios_list <- simulateScenarios(
#'   n_subjects_list     = list(c(10, 20)),
#'   response_rates_list = list(rep(0.9, 2)),
#'   n_trials            = 10)
#'
#' analyses_list <- performAnalyses(
#'   scenario_list       = scenarios_list,
#'   target_rates        = rep(0.5, 2),
#'   n_mcmc_iterations   = 100)
#'
#' go_decisions_list <- getGoDecisions(
#'   analyses_list       = analyses_list,
#'   cohort_names        = c("p_1", "p_2"),
#'   evidence_levels     = c(0.5, 0.8),
#'   boundary_rules      = quote(c(x[1] > 0.8, x[2] > 0.6)))
#'
#' nogo_decisions_list <- getGoDecisions(
#'   analyses_list       = analyses_list,
#'   cohort_names        = c("p_1", "p_2"),
#'   evidence_levels     = c(0.5, 0.8),
#'   boundary_rules      = quote(c(x[1] < 0.5, x[2] < 0.3)))
#'
#' getGoProbabilities(go_decisions_list)
#' getGoProbabilities(go_decisions_list, nogo_decisions_list)
#' @author Stephan Wojciekowski
#' @export
getGoProbabilities <- function (

  go_decisions_list,
  nogo_decisions_list = NULL

) {

  error_go_decisions_list <-
    paste("Providing an object of class 'decision_list'",
          "for the argument 'go_decisions_list'")
  error_nogo_decisions_list <- paste(
    "Providing either NULL or an object of class 'decision_list'",
    "for the argument 'nogo_decisions_list'"
  )
  error_format_mismatch <- paste(
    "The decision_lists 'go_decisions_list' and 'go_decisions_list'",
    "do not follow the same format"
  )

  checkmate::assertClass(
    go_decisions_list,
    classes   = "decision_list",
    .var.name = error_go_decisions_list
  )

  checkmate::assertClass(
    nogo_decisions_list,
    classes   = "decision_list",
    null.ok   = TRUE,
    .var.name = error_nogo_decisions_list
  )

  if (!is.null(nogo_decisions_list)) {
    checkmate::assertTRUE(
      identical(
        dim(go_decisions_list[[1]]$decisions_list[[1]]),
        dim(nogo_decisions_list[[1]]$decisions_list[[1]])
      ),
      .var.name = error_format_mismatch
    )
  }

  ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##

  go_probs_per_scenario <- vector(
    mode   = "list",
    length = length(go_decisions_list)
  )
  names(go_probs_per_scenario) <- names(go_decisions_list)

  for (s in seq_along(go_probs_per_scenario)) {

    method_names <- names(go_decisions_list[[s]]$decisions_list)

    go_probs_per_method_list <- vector(
      mode   = "list",
      length = length(method_names)
    )
    names(go_probs_per_method_list) <- method_names

    for (n in seq_along(method_names)) {

      go_decisions     <- go_decisions_list[[s]]$decisions_list[[n]]
      decisions_matrix <- t(as.matrix(colMeans(go_decisions)))
      row.names(decisions_matrix) <- "Go"

      if (!is.null(nogo_decisions_list)) {

        nogo_decisions <- nogo_decisions_list[[s]]$decisions_list[[n]]
        nogo_probs     <- t(as.matrix(colMeans(nogo_decisions)))
        consider_probs <- round(1 - nogo_probs - decisions_matrix, 9)

        if (!isTRUE(all.equal(sum(go_decisions * nogo_decisions), 0))) {
          stop(paste(
            "There are cohorts for which both go and nogo decisions",
            "are TRUE. Please revise your decision rules."
          ))
        }

        decisions_matrix <- rbind(decisions_matrix, consider_probs, nogo_probs)
        row.names(decisions_matrix) <- c("Go", "Consider", "NoGo")

      }

      go_probs_per_method_list[[n]] <- decisions_matrix

    }

    go_probs_per_scenario[[s]] <- go_probs_per_method_list

  }

  go_probs_per_method <- listPerMethod(go_probs_per_scenario)

  return(go_probs_per_method)

}

getNumericGammaIndex <- function (

  g_numeric,
  quantiles

) {

  error_gamma_range <-
    "Providing numerics in (0, 1) for the argument 'gamma_levels'"

  if (checkmate::testNumeric(g_numeric) && all(g_numeric > 0 & g_numeric < 1)) {

    gamma_index <- which(round(1 - quantiles, 5) == round(g_numeric, 5))

    error_gamma_values <- paste0(
      "gamma must be one of ",
      paste(round(1 - quantiles, 5), collapse = ", ")
    )

    checkmate::assertTRUE(
      is.numeric(gamma_index) && length(gamma_index) > 0,
      .var.name = error_gamma_values
    )

  } else {

    stop(
      "gamma_levels must consist of posterior quantiles, 'mean' or 'sd'"
    )

  }

  return(gamma_index)

}

getPosteriorGammaQuantiles <- function (

  method_name,
  posterior_quantiles_list,
  gamma_levels,
  quantiles,
  cohort_names

) {

  gamma_indices <- getGammaIndices(
    gamma_levels = gamma_levels,
    quantiles    = quantiles
  )

  posterior_quantiles <-
    posterior_quantiles_list[[gsub("_mu", "", method_name)]]

  cohort_indices <- sapply(cohort_names, function (n) {
    grep(n, colnames(posterior_quantiles[[1]]), fixed = TRUE)[1]
  })

  if (length(gamma_indices) != length(cohort_indices)) {
    stop(paste(
      "The number of gamma indices must be equal",
      "to the number of cohorts to be analyzed."
    ))
  }

  u_gamma_indices  <- unique(gamma_indices)
  u_cohort_indices <- unique(cohort_indices)

  posterior_gamma_quantiles <- lapply(posterior_quantiles, function (x) {
    as.vector(t(x[u_gamma_indices, u_cohort_indices]))
  })

  for (i in seq_along(posterior_gamma_quantiles)) {
    names(posterior_gamma_quantiles[[i]]) <-
      colnames(posterior_quantiles[[1]])[cohort_indices]
  }

  return(posterior_gamma_quantiles)

}

getScenarioNumbers <- function (analyses_list) {

  as.numeric(sub("scenario_", "", names(analyses_list)))

}

#' @export
print.decision_list <- function (x, digits = 2, ...) {

  n_scenarios    <- length(x)
  scenario_names <- names(x)

  n_methods      <- length(x[[1]]$decisions_list)
  method_names   <- names(x[[1]]$decisions_list)

  go_probs       <- getGoProbabilities(x)

  if (!is.null(x[[1]]$decision_rules)) {

    decision_rules <- x[[1]]$decision_rules

    out_rules <- lapply(decision_rules$boundary_rules, as.character)

    for (n in seq_len(n_methods)) {

      out_rules <- lapply(
        out_rules,
        gsub,
        pattern     = paste0("x\\[", n, "\\]"),
        replacement = decision_rules$cohort_names[n]
      )

    }

    out_rules <- lapply(out_rules, gsub, pattern = "p_", replacement = "P(p_")

    for (n in seq_len(n_methods)) {

      out_rules_n <- lapply(strsplit(out_rules[[n]][-1], "&&"), trimws)

      out_rules_n <- unlist(utils::as.relistable(out_rules_n))

      out_rules_n[grepl("P\\(", out_rules_n)] <-
        paste0(
          out_rules_n[grepl("P\\(", out_rules_n)],
          ") > ", decision_rules$gamma_levels[[n]]
        )

      out_rules_n <- utils::relist(out_rules_n)

      indexAND <- sapply(out_rules_n, length) == 2

      if (any(indexAND)) {

        out_rules_n[[which(indexAND)]] <- sapply(
          out_rules_n[indexAND],
          function (y) paste0(y[1], " && ", y[2])
        )

      }

      out_rules[[n]] <- unlist(unclass(out_rules_n))

    }

  }

  cat(
    "decision_list of ", n_scenarios, " scenario", ifelse(n_scenarios == 1, "", "s"),
    " with ", n_methods, " method", ifelse(n_methods == 1, "", "s"), "\n\n",
    sep = ""
  )

  for (n in seq_along(scenario_names)) {

    mat_out <- do.call(
      rbind,
      lapply(go_probs, function (y) y[[n]])
    )

    rownames(mat_out) <- paste0(
      "    - ",
      paste0(
        firstUpper(method_names),
        sapply(method_names, function (y) {
          getBlankString(max(nchar(method_names)) - nchar(y) + 1)
        })
      )
    )

    cat("  -", scenario_names[n], "\n")
    print(round(mat_out, digits = digits))

    cat("\n")

  }

}

is.decision_list <- function (x) {

  if (missing(x)) stop("Please provide an object for the argument 'x'")

  inherits(x, "decision_list")

}

#' @title negateGoDecisions
#' @md
#' @description Negates the go decisions derived with
#' \code{\link[bhmbasket]{getGoDecisions}}.
#' @param go_decisions_list An object of class `decision_list`,
#' as returned by \code{\link[bhmbasket]{getGoDecisions}}
#' @param overall_min_nogos Either a non-negative integer or the string `all`
#' for the minimum number of cohort-level NoGo decisions required for
#' an overall NoGo decision, Default: `all`
#' @return A list of NoGo decisions of class `decision_list`
#' @details This function is intended for implementing decision rules with a
#' consider zone as
#' e.g. proposed in "Bayesian design of proof-of-concept trials" by
#' Fisch et al. (2015).
#' This approach involves two criteria, Significance and Relevance.
#' \itemize{
#'   \item Significance: high evidence that the treatment effect is greater
#'   than some smaller value (e.g. treatment effect under H0)
#'   \item Relevance: moderate evidence that the treatment effect is greater
#'   than some larger value (e.g. treatment effect under a certain alternative)
#' }
#' The decision for a cohort is then taken as follows:
#' \itemize{
#'   \item Go decision: Significance and Relevance
#'   \item Consider decision: either Significance, or Relevance, but not both
#'   \item NoGo decision: no Significance and no Relevance
#' }
#' In the example below, the following criteria for are implemented for each of
#' the three cohorts:
#' \itemize{
#'   \item Significance: \eqn{P(p_j > 0.4) > 0.95}
#'   \item Relevance: \eqn{P(p_j > 0.8) > 0.5}
#' }
#' @seealso
#'  \code{\link[bhmbasket]{getGoDecisions}}
#' @rdname negateGoDecisions
#' @examples
#' scenarios_list <- simulateScenarios(
#'   n_subjects_list     = list(c(10, 20, 30)),
#'   response_rates_list = list(rep(0.9, 3)),
#'   n_trials            = 10)
#'
#' analysis_list <- performAnalyses(
#'   scenario_list      = scenarios_list,
#'   target_rates       = rep(0.5, 3),
#'   n_mcmc_iterations  = 100)
#'
#' go_decisions_list <- getGoDecisions(
#'   analyses_list   = analysis_list,
#'   cohort_names    = c("p_1", "p_2", "p_3",
#'                       "p_1", "p_2", "p_3"),
#'   evidence_levels = c(0.5,  0.5,  0.5,
#'                       0.95, 0.95, 0.95),
#'   boundary_rules  = quote(c(x[1] > 0.8 & x[4] > 0.4,
#'                             x[2] > 0.8 & x[5] > 0.4,
#'                             x[3] > 0.8 & x[6] > 0.4)))
#'
#' nogo_decisions <- negateGoDecisions(getGoDecisions(
#'   analyses_list   = analysis_list,
#'   cohort_names    = c("p_1", "p_2", "p_3",
#'                       "p_1", "p_2", "p_3"),
#'   evidence_levels = c(0.5,  0.5,  0.5,
#'                       0.95, 0.95, 0.95),
#'   boundary_rules  = quote(c(x[1] > 0.8 | x[4] > 0.4,
#'                             x[2] > 0.8 | x[5] > 0.4,
#'                             x[3] > 0.8 | x[6] > 0.4))))
#'
#' getGoProbabilities(go_decisions_list, nogo_decisions)
#' @author Stephan Wojciekowski
#' @references Fisch, Roland, et al.
#' "Bayesian design of proof-of-concept trials."
#' \emph{Therapeutic innovation & regulatory science} 49.1 (2015): 155-162.
#' @export
negateGoDecisions <- function (

  go_decisions_list,
  overall_min_nogos = "all"

) {

  error_go_decisions_list <-
    paste("Providing an object of class 'decision_list'",
          "for the argument 'go_decisions_list'")
  error_overall_min_nogos <- paste(
    "Providing either a non-negative integer or the string 'all'",
    "for the argument 'overall_min_nogos'"
  )

  checkmate::assertClass(
    go_decisions_list,
    classes   = "decision_list",
    .var.name = error_go_decisions_list
  )

  checkmate::assert(
    checkmate::checkChoice(overall_min_nogos, choices = "all"),
    checkmate::checkInt(overall_min_nogos,      lower = 0),
    combine   = "or",
    .var.name = error_overall_min_nogos
  )

  ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##

  overall_min_nogos_org <- overall_min_nogos
  nogo_decisions_list   <- go_decisions_list

  for (s in seq_along(go_decisions_list)) {

    for (m in seq_along(go_decisions_list[[s]]$decisions_list)) {

      ## negate decisions
      nogo_decisions_list[[s]]$decisions_list[[m]] <-
        !go_decisions_list[[s]]$decisions_list[[m]]

      ## check for overall decisions
      n_decisions <- ncol(nogo_decisions_list[[s]]$decisions_list[[m]])

      if (n_decisions > 1) {

        if (identical(overall_min_nogos_org, "all")) {

          overall_min_nogos <- n_decisions - 1L

        }

        nogo_decisions_list[[s]]$decisions_list[[m]][, 1] <-
          apply(
            nogo_decisions_list[[s]]$decisions_list[[m]][, -1],
            1,
            function (x) sum(x) >= overall_min_nogos
          )

      }

    }

  }

  return(nogo_decisions_list)

}

# Tests for getEstimates -------------------------------------------------------

scenarios_list <- simulateScenarios(
  n_subjects_list     = list(c(10, 20, 30)),
  response_rates_list = list(c(0.1, 0.2, 3)),  # 3 triggers historic-rate branch
  n_trials            = 10
)

analyses_list <- performAnalyses(
  scenario_list       = scenarios_list,
  target_rates        = c(0.5, 0.5, 0.5),
  calc_differences    = matrix(c(3, 2, 2, 1), ncol = 2),  # ensures diff cohorts exist
  n_mcmc_iterations   = 100,
  method_names        = "berry"
)

outcome <- createTrial(
  n_subjects   = c(10, 20, 30),
  n_responders = c( 1,  2,  3)
)

outcome_analysis <- performAnalyses(
  scenario_list     = outcome,
  target_rates      = c(0.5, 0.5, 0.5),
  n_mcmc_iterations = 100
)


# -------------------------------------------------------------------
# Test: getEstimates basic structure and contents for simulated scenarios
# Input:
#   - analyses_list: analysis_list with 10 simulated trials, 3 cohorts, diff-cohorts.
# Behaviour:
#   - getEstimates() should return list-per-method of matrices with posterior summaries,
#     Bias and MSE for response-rate parameters and diff parameters.
# Expectations:
#   - Output is a non-empty list.
#   - First inner element is a matrix with columns Mean, SD, 2.5%, 50%, 97.5%, Bias, MSE.
#   - Row names contain p_* for cohorts and "diff" rows for differences.
# Why:
#   - This is the main “happy path”; other tests rely on this shape.
# -------------------------------------------------------------------
test_that("getEstimates works for simulated scenarios and returns sensible structure", {
  res <- getEstimates(analyses_list)

  expect_type(res, "list")
  expect_true(length(res) > 0)

  first_obj <- res[[1]]
  if (is.list(first_obj)) {
    first_mat <- first_obj[[1]]
  } else {
    first_mat <- first_obj
  }

  expect_true(is.matrix(first_mat))
  expect_true(all(c("Mean", "SD", "2.5%", "50%", "97.5%", "Bias", "MSE") %in% colnames(first_mat)))
  expect_true(any(grepl("^p_",   rownames(first_mat))))
  expect_true(any(grepl("diff", rownames(first_mat))))
})


# -------------------------------------------------------------------
# Test: getEstimates with add_parameters and historic-rate handling
# Input:
#   - Same analyses_list as above.
#   - add_parameters = c("mu", "tau", "w_1", "w_2", "w_3").
# Behaviour:
#   - Additional model parameters are appended as extra rows.
#   - For historic cohorts (rr <= 0 or >= 1) true_rr is derived as responders/n_subjects.
#   - Bias and MSE are computed only for p_* rows, not for mu/tau/w_*.
# Expectations:
#   - p_3 row exists and its Bias + true_rr_used equals the median ("50%") estimate.
#   - Column structure identical with and without add_parameters.
#   - Extra rows correspond to mu/tau/w_* and have NA for Bias/MSE.
# Why:
#   - Ensures correct handling of historic cohorts and extra parameters.
# -------------------------------------------------------------------
test_that("additional parameters are added and have NA bias/MSE", {
  res_base <- getEstimates(analyses_list)
  base_obj <- res_base[[1]]
  base_mat <- if (is.list(base_obj)) base_obj[[1]] else base_obj

  res_add <- getEstimates(
    analyses_list   = analyses_list,
    add_parameters  = c("mu", "tau", "w_1", "w_2", "w_3")
  )
  add_obj <- res_add[[1]]
  add_mat <- if (is.list(add_obj)) add_obj[[1]] else add_obj

  p3_row <- grep("^p_3$", rownames(add_mat))
  expect_length(p3_row, 1)

  bias_p3   <- add_mat[p3_row, "Bias"]
  median_p3 <- add_mat[p3_row, "50%"]

  true_rr_raw       <- analyses_list[[1]]$scenario_data$response_rates
  n_subj            <- analyses_list[[1]]$scenario_data$n_subjects[1, ]
  expected_true_rr3 <- true_rr_raw[3] / n_subj[[3]]

  point_estimate3 <- bias_p3 + expected_true_rr3
  expect_equal(point_estimate3, median_p3)

  expect_identical(colnames(base_mat), colnames(add_mat))

  extra_rows <- setdiff(rownames(add_mat), rownames(base_mat))
  expect_true(length(extra_rows) > 0)
  expect_true(all(grepl("mu|tau|w_", extra_rows)))

  expect_true(all(is.na(add_mat[extra_rows, "Bias"])))
  expect_true(all(is.na(add_mat[extra_rows, "MSE"])))
})


# -------------------------------------------------------------------
# Test: getEstimates for single-trial outcome
# Input:
#   - outcome_analysis: analysis_list from createTrial() (one trial only).
# Behaviour:
#   - With n_trials = 1, getEstimates() should *not* compute Bias/MSE,
#     only posterior summaries.
# Expectations:
#   - Returned inner matrix has columns exactly: Mean, SD, 2.5%, 50%, 97.5%.
# Why:
#   - Distinguishes between simulation-based metrics and outcome-only analyses.
# -------------------------------------------------------------------
test_that("single-trial outcome returns only posterior summaries (no bias/MSE)", {
  res_single <- getEstimates(outcome_analysis)
  single_obj <- res_single[[1]]
  single_mat <- if (is.list(single_obj)) single_obj[[1]] else single_obj

  expect_true(is.matrix(single_mat))
  expect_identical(
    colnames(single_mat),
    c("Mean", "SD", "2.5%", "50%", "97.5%")
  )
})


# -------------------------------------------------------------------
# Test: validation of alpha_level and add_parameters in getEstimates
# Input:
#   - analyses_list from the main fixture.
# Behaviour:
#   - alpha_level must be in (0, 1) and must correspond to stored quantiles.
#   - add_parameters must exist in at least one method’s posterior draws.
# Expectations:
#   - alpha_level = 0.07 (not among stored quantiles) -> error.
#   - add_parameters = "totally_unknown_param" -> error with appropriate message.
# Why:
#   - Ensures defensive checks for user-supplied quantiles and parameter names.
# -------------------------------------------------------------------
test_that("alpha_level and add_parameters are validated correctly", {
  expect_error(
    getEstimates(analyses_list, alpha_level = 0.07)
  )

  expect_error(
    getEstimates(analyses_list, add_parameters = c("totally_unknown_param"))
  )
})


# -------------------------------------------------------------------
# Test: point_estimator behaviour and basic type checks in getEstimates
# Input:
#   - analyses_list, point_estimator = "median" and "mean".
# Behaviour:
#   - Both estimators should work and yield same matrix shape.
#   - An invalid point_estimator should raise an error.
#   - Non-analysis_list input should raise an error.
# Expectations:
#   - dim() and colnames() identical for median vs mean.
#   - point_estimator = "mode" -> error.
#   - analyses_list = list(a = 1) -> error on class.
# Why:
#   - Confirms configurability of the estimator and robust input validation.
# -------------------------------------------------------------------
test_that("point_estimator argument is respected and input type is validated", {
  res_median <- getEstimates(analyses_list, point_estimator = "median")
  res_mean   <- getEstimates(analyses_list, point_estimator = "mean")

  med_obj  <- res_median[[1]]
  mean_obj <- res_mean[[1]]

  med_mat  <- if (is.list(med_obj))  med_obj[[1]]  else med_obj
  mean_mat <- if (is.list(mean_obj)) mean_obj[[1]] else mean_obj

  expect_identical(dim(med_mat),   dim(mean_mat))
  expect_identical(colnames(med_mat), colnames(mean_mat))

  expect_error(
    getEstimates(analyses_list, point_estimator = "mode")
  )

  expect_error(
    getEstimates(list(a = 1))
  )
})


# -------------------------------------------------------------------
# Additional explicit validation tests for getEstimates:
# - alpha_level outside (0,1)
# - alpha_level with unavailable quantiles
# - add_parameters not found
# - basic structure for single-trial
# - duplicate explicit checks for invalid point_estimator and analyses_list
# -------------------------------------------------------------------

test_that("throws error for invalid alpha_level (outside 0,1)", {
  expect_error(getEstimates(analyses_list, alpha_level = 1.5), "alpha_level")
})

test_that("throws error if alpha_level quantiles not available", {
  expect_error(
    getEstimates(analyses_list, alpha_level = 0.123),
    "must be among the stored quantiles"
  )
})

test_that("throws error if add_parameters not found in any method", {
  expect_error(
    getEstimates(analyses_list, add_parameters = c("nonexistent")),
    "do not occur"
  )
})

test_that("works for single trial outcome (basic check)", {
  result <- getEstimates(outcome_analysis)
  expect_type(result, "list")
})

test_that("invalid point_estimator throws error explicitly", {
  expect_error(
    getEstimates(analyses_list, point_estimator = "mode"),
    "point_estimator"
  )
})

test_that("invalid analyses_list class throws error explicitly", {
  expect_error(
    getEstimates(list()),
    "analyses_list"
  )
})


# Tests for getGoDecisions / getGoProbabilities / print.decision_list ----------

set.seed(123)

scenarios_list <- simulateScenarios(
  n_subjects_list     = list(c(10, 20, 30)),
  response_rates_list = list(c(0.3, 0.5, 0.7)),
  n_trials            = 10
)

analyses_list <- performAnalyses(
  scenario_list     = scenarios_list,
  target_rates      = c(0.3, 0.3, 0.3),
  n_mcmc_iterations = 100
)

default_cohorts <- c("p_1", "p_2", "p_3")


# -------------------------------------------------------------------
# Test: getGoDecisions – analyses_list must have correct class
# Input:
#   - analyses_list = list() (wrong type)
# Behaviour:
#   - checkmate::assertClass enforces class 'analysis_list'.
# Expectations:
#   - Error mentioning 'analysis_list' is thrown.
# Why:
#   - Prevents silent misuse with arbitrary lists.
# -------------------------------------------------------------------
test_that("getGoDecisions: errors if analyses_list is not of class 'analysis_list'", {
  expect_error(
    getGoDecisions(
      analyses_list   = list(),
      cohort_names    = "p_1",
      evidence_levels = 0.5,
      boundary_rules  = quote(c(TRUE))
    ),
    "analysis_list"
  )
})


# -------------------------------------------------------------------
# Test: getGoDecisions – evidence_levels and boundary_rules required
# Input:
#   - Missing evidence_levels or missing boundary_rules.
# Behaviour:
#   - Custom error messages if either argument is missing.
# Expectations:
#   - Missing evidence_levels -> error message refers to evidence_levels.
#   - Missing boundary_rules  -> error message refers to boundary_rules.
# Why:
#   - These are essential pieces of the decision rule definition.
# -------------------------------------------------------------------
test_that("getGoDecisions: errors when evidence_levels or boundary_rules are missing", {
  expect_error(
    getGoDecisions(
      analyses_list  = analyses_list,
      cohort_names   = "p_1",
      boundary_rules = quote(c(TRUE))
    ),
    "evidence_levels"
  )

  expect_error(
    getGoDecisions(
      analyses_list   = analyses_list,
      cohort_names    = "p_1",
      evidence_levels = 0.5
    ),
    "boundary_rules"
  )
})


# -------------------------------------------------------------------
# Test: getGoDecisions – invalid cohort_names
# Input:
#   - cohort_names = "invalid" (not present in posterior columns).
# Behaviour:
#   - assertSubset ensures cohort_names correspond to posterior parameters.
# Expectations:
#   - Error referencing 'cohort_names'.
# Why:
#   - Prevents mis-specification of cohort parameter names.
# -------------------------------------------------------------------
test_that("getGoDecisions: errors if cohort_names are not valid posterior parameters", {
  expect_error(
    getGoDecisions(
      analyses_list   = analyses_list,
      cohort_names    = "invalid",
      evidence_levels = 0.5,
      boundary_rules  = quote(c(TRUE))
    ),
    "cohort_names"
  )
})


# -------------------------------------------------------------------
# Test: getGoDecisions – overall_min_gos must be positive integer
# Input:
#   - overall_min_gos = 0L.
# Behaviour:
#   - assertInt(overall_min_gos, lower = 1) rejects non-positive values.
# Expectations:
#   - Error referencing 'overall_min_gos'.
# Why:
#   - The minimum number of cohort-wise Go decisions must be >= 1.
# -------------------------------------------------------------------
test_that("getGoDecisions: errors if overall_min_gos is not a positive integer", {
  expect_error(
    getGoDecisions(
      analyses_list   = analyses_list,
      cohort_names    = "p_1",
      evidence_levels = 0.5,
      boundary_rules  = quote(c(TRUE)),
      overall_min_gos = 0L
    ),
    "overall_min_gos"
  )
})


# -------------------------------------------------------------------
# Test: getGoDecisions – numeric evidence_levels in (0,1)
# Input:
#   - evidence_levels = c(-0.1, 0.5, 1.3).
# Behaviour:
#   - check.evidence.levels enforces all levels are in (0,1).
# Expectations:
#   - Error is thrown for invalid vector.
# Why:
#   - Evidence thresholds must be valid probabilities.
# -------------------------------------------------------------------
test_that("getGoDecisions: numeric evidence_levels outside (0,1) cause an error", {
  expect_error(
    getGoDecisions(
      analyses_list   = analyses_list,
      cohort_names    = default_cohorts,
      evidence_levels = c(-0.1, 0.5, 1.3),
      boundary_rules  = quote(c(TRUE, TRUE, TRUE))
    )
  )
})


# -------------------------------------------------------------------
# Test: getGoDecisions – list evidence_levels with valid entries
# Input:
#   - evidence_levels as list of numeric vectors in (0,1).
# Behaviour:
#   - check.evidence.levels is applied to each list element.
# Expectations:
#   - No error is thrown for valid list structure.
# Why:
#   - Supports method-specific evidence levels.
# -------------------------------------------------------------------
test_that("getGoDecisions: list evidence_levels all elements valid", {
  ev_list_ok <- list(
    rep(0.5, length(default_cohorts)),
    rep(0.5, length(default_cohorts))
  )

  expect_silent(
    getGoDecisions(
      analyses_list   = analyses_list,
      cohort_names    = default_cohorts,
      evidence_levels = ev_list_ok,
      boundary_rules  = quote(c(TRUE, TRUE, TRUE))
    )
  )
})


# -------------------------------------------------------------------
# Test: getGoDecisions – non-list boundary_rules must be a language object
# Input:
#   - boundary_rules = 1 (numeric, not a call).
# Behaviour:
#   - check_boundary_rules rejects non-language boundary rules.
# Expectations:
#   - Error referencing 'boundary_rules'.
# Why:
#   - Decision rules must be expressed as quoted R expressions.
# -------------------------------------------------------------------
test_that("getGoDecisions: non-list boundary_rules must be a language object", {
  expect_error(
    getGoDecisions(
      analyses_list   = analyses_list,
      cohort_names    = default_cohorts,
      evidence_levels = c(0.5, 0.5, 0.8),
      boundary_rules  = 1
    ),
    "boundary_rules"
  )
})


# -------------------------------------------------------------------
# Test: getGoDecisions – non-list boundary_rules must use c(...)
# Input:
#   - boundary_rules = quote(list(TRUE, TRUE, TRUE)).
# Behaviour:
#   - The call head must be `c`, so list(...) is rejected.
# Expectations:
#   - Error referencing 'boundary_rules'.
# Why:
#   - The internal logic expects a c(...) vector of logical decisions.
# -------------------------------------------------------------------
test_that("getGoDecisions: non-list boundary_rules must start with c()", {
  expect_error(
    getGoDecisions(
      analyses_list   = analyses_list,
      cohort_names    = default_cohorts,
      evidence_levels = c(0.5, 0.5, 0.8),
      boundary_rules  = quote(list(TRUE, TRUE, TRUE))
    ),
    "boundary_rules"
  )
})


# -------------------------------------------------------------------
# Test: getGoDecisions – non-list boundary_rules length must match #cohorts
# Input:
#   - 3 cohorts but boundary_rules = quote(c(TRUE, TRUE)).
# Behaviour:
#   - Length mismatch triggers error in check_boundary_rules.
# Expectations:
#   - Error referencing 'boundary_rules'.
# Why:
#   - Each cohort must have one logical decision rule.
# -------------------------------------------------------------------
test_that("getGoDecisions: non-list boundary_rules must have one entry per cohort", {
  expect_error(
    getGoDecisions(
      analyses_list   = analyses_list,
      cohort_names    = default_cohorts,
      evidence_levels = c(0.5, 0.5, 0.8),
      boundary_rules  = quote(c(TRUE, TRUE))
    ),
    "boundary_rules"
  )
})


# -------------------------------------------------------------------
# Test: getGoDecisions – list boundary_rules elements must be language
# Input:
#   - boundary_rules = list(123).
# Behaviour:
#   - Each element of the list must be a call; a numeric element is invalid.
# Expectations:
#   - Error referencing 'boundary_rules'.
# Why:
#   - Supports multiple methods with method-specific decision rules.
# -------------------------------------------------------------------
test_that("getGoDecisions: list boundary_rules each element must be a language object", {
  br <- list(123)

  expect_error(
    getGoDecisions(
      analyses_list   = analyses_list,
      cohort_names    = default_cohorts,
      evidence_levels = c(0.5, 0.5, 0.8),
      boundary_rules  = br
    ),
    "boundary_rules"
  )
})


# -------------------------------------------------------------------
# Test: getGoDecisions – list boundary_rules head must be c()
# Input:
#   - boundary_rules = list(quote(list(TRUE, TRUE, TRUE))).
# Behaviour:
#   - Each element must start with c(...); list(...) is rejected.
# Expectations:
#   - Error referencing 'boundary_rules'.
# Why:
#   - Same reasoning as the non-list case, but in list form.
# -------------------------------------------------------------------
test_that("getGoDecisions: list boundary_rules each element must start with c()", {
  br <- list(quote(list(TRUE, TRUE, TRUE)))

  expect_error(
    getGoDecisions(
      analyses_list   = analyses_list,
      cohort_names    = default_cohorts,
      evidence_levels = c(0.5, 0.5, 0.8),
      boundary_rules  = br
    ),
    "boundary_rules"
  )
})


# -------------------------------------------------------------------
# Test: getGoDecisions – list boundary_rules length must match #cohorts
# Input:
#   - boundary_rules = list(quote(c(TRUE, TRUE))) for 3 cohorts.
# Behaviour:
#   - Length mismatch per element triggers error.
# Expectations:
#   - Error referencing 'boundary_rules'.
# Why:
#   - Ensures consistent vector length across all cohorts.
# -------------------------------------------------------------------
test_that("getGoDecisions: list boundary_rules must match number of cohorts", {
  br <- list(quote(c(TRUE, TRUE)))

  expect_error(
    getGoDecisions(
      analyses_list   = analyses_list,
      cohort_names    = default_cohorts,
      evidence_levels = c(0.5, 0.5, 0.8),
      boundary_rules  = br
    ),
    "boundary_rules"
  )
})


# -------------------------------------------------------------------
# Test: getGoDecisions – boundary_rules list length vs method_names
# Input:
#   - boundary_rules list longer than number of methods.
# Behaviour:
#   - Length check should reject boundary_rules with length > length(method_names).
# Expectations:
#   - Error referencing 'boundary_rules'.
# Why:
#   - Each method can have at most one decision rule expression.
# -------------------------------------------------------------------
test_that("errors if boundary_rules list is longer than method_names", {
  m    <- analyses_list[[1]]$analysis_parameters$method_names
  coh  <- c("p_1", "p_2", "p_3")
  ev   <- c(0.5, 0.5, 0.5)
  br   <- rep(list(quote(c(TRUE, TRUE, TRUE))), length(m) + 1L)

  expect_error(
    getGoDecisions(
      analyses_list   = analyses_list,
      cohort_names    = coh,
      evidence_levels = ev,
      boundary_rules  = br
    ),
    "boundary_rules"
  )
})


# -------------------------------------------------------------------
# Test: getGoDecisions – evidence_levels list length vs method_names
# Input:
#   - evidence_levels list longer than number of methods.
# Behaviour:
#   - Length check should reject gamma_levels with length > length(method_names).
# Expectations:
#   - Error referencing 'evidence_levels'.
# Why:
#   - Each method can have at most one vector of evidence thresholds.
# -------------------------------------------------------------------
test_that("errors if evidence_levels list is longer than method_names", {
  m    <- analyses_list[[1]]$analysis_parameters$method_names
  coh  <- c("p_1", "p_2", "p_3")

  ev_vec <- c(0.5, 0.5, 0.5)
  ev     <- rep(list(ev_vec), length(m) + 1L)

  expect_error(
    getGoDecisions(
      analyses_list   = analyses_list,
      cohort_names    = coh,
      evidence_levels = ev,
      boundary_rules  = quote(c(TRUE, TRUE, TRUE))
    ),
    "evidence_levels"
  )
})


# -------------------------------------------------------------------
# Test: getGoDecisions – single boundary_rules expression recycled
# Input:
#   - boundary_rules = quote(c(TRUE, TRUE, TRUE)) and multiple methods.
# Behaviour:
#   - Single expression is replicated for each method internally.
# Expectations:
#   - decision_rules$boundary_rules has length = #methods,
#     each element identical to the original rule.
# Why:
#   - Confirms convenience recycling logic.
# -------------------------------------------------------------------
test_that("single boundary_rules expression is recycled to all methods", {
  m    <- analyses_list[[1]]$analysis_parameters$method_names
  coh  <- c("p_1", "p_2", "p_3")
  ev   <- c(0.5, 0.5, 0.5)
  rule <- quote(c(TRUE, TRUE, TRUE))

  dec <- getGoDecisions(
    analyses_list   = analyses_list,
    cohort_names    = coh,
    evidence_levels = ev,
    boundary_rules  = rule
  )

  br <- dec[[1]]$decision_rules$boundary_rules

  expect_identical(length(br), length(m))
  for (i in seq_along(br)) {
    expect_true(identical(br[[i]], rule))
  }
})


# -------------------------------------------------------------------
# Test: getGoDecisions – single evidence_levels vector recycled
# Input:
#   - evidence_levels = c(0.5, 0.5, 0.5) and multiple methods.
# Behaviour:
#   - Single numeric vector is replicated per method.
# Expectations:
#   - decision_rules$gamma_levels has length = #methods,
#     each element equal to the original vector.
# Why:
#   - Confirms convenience recycling for evidence thresholds.
# -------------------------------------------------------------------
test_that("single evidence_levels vector is recycled to all methods", {
  m    <- analyses_list[[1]]$analysis_parameters$method_names
  coh  <- c("p_1", "p_2", "p_3")
  ev   <- c(0.5, 0.5, 0.5)

  dec <- getGoDecisions(
    analyses_list   = analyses_list,
    cohort_names    = coh,
    evidence_levels = ev,
    boundary_rules  = quote(c(TRUE, TRUE, TRUE))
  )

  gamma <- dec[[1]]$decision_rules$gamma_levels

  expect_identical(length(gamma), length(m))
  for (i in seq_along(gamma)) {
    expect_equal(gamma[[i]], ev)
  }
})


# -------------------------------------------------------------------
# Test: getGoDecisions – method names consistent across scenarios
# Input:
#   - analyses_list with identical method_names per scenario.
# Behaviour:
#   - Method name matrix check passes; decisions are computed.
# Expectations:
#   - Returned object is of class decision_list.
# Why:
#   - Sanity check for the main use-case of multiple scenarios with same methods.
# -------------------------------------------------------------------
test_that("getGoDecisions: succeeds when all scenarios use identical method_names", {
  res <- getGoDecisions(
    analyses_list   = analyses_list,
    cohort_names    = default_cohorts,
    evidence_levels = c(0.5, 0.5, 0.8),
    boundary_rules  = quote(c(TRUE, TRUE, TRUE))
  )

  expect_s3_class(res, "decision_list")
})


# -------------------------------------------------------------------
# Test: getGoDecisions – error when scenarios use different methods
# Input:
#   - analyses_list where scenario_2 has method_names reversed vs scenario_1.
# Behaviour:
#   - Method-name consistency check fails.
# Expectations:
#   - Error with message “analysed with different methods”.
# Why:
#   - Mixed-method setups per scenario are not supported.
# -------------------------------------------------------------------
test_that("getGoDecisions: errors when scenarios were analysed with different methods", {
  bad <- analyses_list
  bad$scenario_2 <- bad$scenario_1
  bad$scenario_2$analysis_parameters$method_names <-
    rev(bad$scenario_2$analysis_parameters$method_names)

  expect_error(
    getGoDecisions(
      analyses_list   = bad,
      cohort_names    = default_cohorts,
      evidence_levels = c(0.5, 0.5, 0.8),
      boundary_rules  = quote(c(TRUE, TRUE, TRUE))
    ),
    "analysed with different methods"
  )
})


# -------------------------------------------------------------------
# Test: getGoDecisions – structure of decision_list and 'overall' column
# Input:
#   - Valid analyses_list, three cohorts, evidence levels.
# Behaviour:
#   - getGoDecisions returns decision_list with:
#     - decisions_list per method/scenario including 'overall' and cohort columns.
#     - decision_rules storing cohort_names and gamma_levels.
# Expectations:
#   - 'overall' column exists.
#   - At least one cohort-level decision column exists.
#   - decision_rules$cohort_names identical to input.
#   - stored gamma levels contain the supplied values.
# Why:
#   - Verifies structural contract of decision_list objects.
# -------------------------------------------------------------------
test_that("getGoDecisions: returns decision_list with overall and cohort decisions", {
  decisions <- getGoDecisions(
    analyses_list   = analyses_list,
    cohort_names    = default_cohorts,
    evidence_levels = c(0.5, 0.5, 0.8),
    boundary_rules  = quote(c(TRUE, TRUE, TRUE))
  )

  scen1  <- decisions[[1]]
  m_dec  <- as.matrix(scen1$decisions_list[[1]])

  expect_true("overall" %in% colnames(m_dec))
  cohort_cols <- setdiff(colnames(m_dec), "overall")
  expect_true(length(cohort_cols) >= 1)

  expect_identical(scen1$decision_rules$cohort_names, default_cohorts)

  stored_gamma_flat <- unlist(scen1$decision_rules$gamma_levels, use.names = FALSE)
  expect_true(all(c(0.5, 0.5, 0.8) %in% stored_gamma_flat))
})


# -------------------------------------------------------------------
# Test: getGoDecisions – semantics of overall_min_gos = 1
# Input:
#   - overall_min_gos = 1, boundary_rules always TRUE.
# Behaviour:
#   - overall decision is TRUE if at least one cohort is TRUE.
# Expectations:
#   - 'overall' column equals row-wise indicator of ≥1 TRUE among cohorts.
# Why:
#   - Checks threshold behaviour for overall Go decision.
# -------------------------------------------------------------------
test_that("overall_min_gos = 1 means overall Go if at least one cohort is Go", {
  dec <- getGoDecisions(
    analyses_list   = analyses_list,
    cohort_names    = default_cohorts,
    evidence_levels = rep(0.5, length(default_cohorts)),
    boundary_rules  = quote(c(TRUE, TRUE, TRUE)),
    overall_min_gos = 1L
  )

  m <- as.matrix(dec[[1]]$decisions_list[[1]])
  coh_cols <- setdiff(colnames(m), "overall")
  expect_true(length(coh_cols) >= 1)

  coh <- m[, coh_cols, drop = FALSE] > 0
  overall_calc <- apply(coh, 1, function(x) sum(x) >= 1L)

  overall <- as.logical(m[, "overall"])
  expect_identical(overall, overall_calc)
})


# -------------------------------------------------------------------
# Test: getGoDecisions – semantics of overall_min_gos = 2
# Input:
#   - overall_min_gos = 2, boundary_rules always TRUE.
# Behaviour:
#   - overall decision is TRUE if at least two cohorts are TRUE.
# Expectations:
#   - 'overall' column equals row-wise indicator of ≥2 TRUE among cohorts.
# Why:
#   - Confirms threshold is correctly applied for higher overall minimum.
# -------------------------------------------------------------------
test_that("overall_min_gos = 2 means overall Go if at least two cohorts are Go", {
  dec <- getGoDecisions(
    analyses_list   = analyses_list,
    cohort_names    = default_cohorts,
    evidence_levels = rep(0.5, length(default_cohorts)),
    boundary_rules  = quote(c(TRUE, TRUE, TRUE)),
    overall_min_gos = 2L
  )

  m <- as.matrix(dec[[1]]$decisions_list[[1]])
  coh_cols <- setdiff(colnames(m), "overall")
  expect_true(length(coh_cols) >= 1)

  coh <- m[, coh_cols, drop = FALSE] > 0
  overall_calc <- apply(coh, 1, function(x) sum(x) >= 2L)

  overall <- as.logical(m[, "overall"])
  expect_identical(overall, overall_calc)
})


# -------------------------------------------------------------------
# Test: getGoDecisions – no resurrection of previously stopped cohorts
# Input:
#   - analyses_list with scenario_data$previous_analyses$go_decisions present.
# Behaviour:
#   - New go_decisions are multiplied by previous_gos, so a previously stopped
#     cohort cannot become Go again.
# Expectations:
#   - No cell where new_go == TRUE and previous_go == FALSE.
# Why:
#   - Implements the intended “no resurrection” rule across analyses.
# -------------------------------------------------------------------
test_that("getGoDecisions: previous go_decisions prevent resurrection of stopped cohorts", {
  if (is.null(analyses_list[[1]]$scenario_data$previous_analyses)) {
    skip("previous_analyses not available in scenario_data")
  }

  decisions <- getGoDecisions(
    analyses_list   = analyses_list,
    cohort_names    = default_cohorts,
    evidence_levels = c(0.5, 0.5, 0.8),
    boundary_rules  = quote(c(TRUE, TRUE, TRUE))
  )

  new_mat     <- as.matrix(decisions[[1]]$decisions_list[[1]])
  new_cohcols <- setdiff(colnames(new_mat), "overall")
  new_gos     <- new_mat[, new_cohcols, drop = FALSE] > 0

  prev_mat <- analyses_list[[1]]$scenario_data$previous_analyses$go_decisions
  prev_gos <- as.matrix(prev_mat[, -1, drop = FALSE]) > 0

  expect_identical(dim(prev_gos), dim(new_gos))

  resurrected <- new_gos & !prev_gos
  expect_false(any(resurrected))
})


# Tests for getGoProbabilities -------------------------------------------------

set.seed(456)

scenarios_list_prob <- simulateScenarios(
  n_subjects_list     = list(c(10, 20)),
  response_rates_list = list(rep(0.9, 2)),
  n_trials            = 10
)

analyses_list_prob <- performAnalyses(
  scenario_list     = scenarios_list_prob,
  target_rates      = rep(0.5, 2),
  n_mcmc_iterations = 100
)

prob_cohorts <- c("p_1", "p_2")

go_decisions_list <- getGoDecisions(
  analyses_list   = analyses_list_prob,
  cohort_names    = prob_cohorts,
  evidence_levels = c(0.5, 0.5),
  boundary_rules  = quote(c(x[1] > 0.8, x[2] > 0.6))
)

nogo_decisions_list <- getGoDecisions(
  analyses_list   = analyses_list_prob,
  cohort_names    = prob_cohorts,
  evidence_levels = c(0.5, 0.5),
  boundary_rules  = quote(c(x[1] < 0.5, x[2] < 0.3))
)


# -------------------------------------------------------------------
# Test: getGoProbabilities – go_decisions_list must be decision_list
# Input:
#   - go_decisions_list = list().
# Behaviour:
#   - assertClass(go_decisions_list, "decision_list") should fail.
# Expectations:
#   - Error mentioning 'go_decisions_list'.
# Why:
#   - Ensures user passes the correct object type.
# -------------------------------------------------------------------
test_that("getGoProbabilities: errors if go_decisions_list is not a decision_list", {
  expect_error(
    getGoProbabilities(go_decisions_list = list()),
    "go_decisions_list"
  )
})


# -------------------------------------------------------------------
# Test: getGoProbabilities – nogo_decisions_list must be decision_list or NULL
# Input:
#   - nogo_decisions_list = list().
# Behaviour:
#   - assertClass(..., null.ok = TRUE) fails when object is not decision_list.
# Expectations:
#   - Error mentioning 'decision_list'.
# Why:
#   - Prevents malformed NoGo objects.
# -------------------------------------------------------------------
test_that("getGoProbabilities: errors if nogo_decisions_list has wrong class", {
  expect_error(
    getGoProbabilities(
      go_decisions_list   = go_decisions_list,
      nogo_decisions_list = list()
    ),
    "decision_list"
  )
})


# -------------------------------------------------------------------
# Test: getGoProbabilities – Go/NoGo matrices must have same dimensions
# Input:
#   - bad_nogo with first method’s matrix missing a column.
# Behaviour:
#   - Dimension check should fail if the shapes differ.
# Expectations:
#   - Error is thrown (no specific regex needed).
# Why:
#   - Go and NoGo must align element-wise.
# -------------------------------------------------------------------
test_that("getGoProbabilities: errors if Go and NoGo matrices have different dimensions", {
  bad_nogo <- nogo_decisions_list
  bad_nogo[[1]]$decisions_list[[1]] <-
    bad_nogo[[1]]$decisions_list[[1]][, -1, drop = FALSE]

  expect_error(
    getGoProbabilities(
      go_decisions_list   = go_decisions_list,
      nogo_decisions_list = bad_nogo
    )
  )
})


# -------------------------------------------------------------------
# Test: getGoProbabilities – Go-only case structure
# Input:
#   - go_decisions_list, nogo_decisions_list = NULL.
# Behaviour:
#   - Returns list-per-method of matrices with single row "Go".
# Expectations:
#   - probs is a list of lists of matrices.
#   - First matrix has rownames "Go" and at least one column.
# Why:
#   - Checks basic output shape in the simpler Go-only mode.
# -------------------------------------------------------------------
test_that("getGoProbabilities: Go-only call returns list-of-lists of matrices", {
  probs <- getGoProbabilities(go_decisions_list)

  expect_type(probs, "list")
  expect_true(all(sapply(probs, is.list)))

  first_method   <- names(probs)[1]
  first_scenario <- names(probs[[first_method]])[1]
  mat            <- probs[[first_method]][[first_scenario]]

  expect_true(is.matrix(mat))
  expect_identical(rownames(mat), "Go")
  expect_true(ncol(mat) >= 1)
})


# -------------------------------------------------------------------
# Test: getGoProbabilities – Go row equals column means of go_decisions
# Input:
#   - go_decisions_list with full decision matrices.
# Behaviour:
#   - Row "Go" is defined as colMeans(go_decisions).
# Expectations:
#   - For each method/scenario, mat_prob["Go", ] == colMeans(go_mat).
# Why:
#   - Validates the probability computation logic.
# -------------------------------------------------------------------
test_that("getGoProbabilities: Go row equals column means of go_decisions", {
  probs <- getGoProbabilities(go_decisions_list)

  method_names   <- names(go_decisions_list[[1]]$decisions_list)
  scenario_names <- names(go_decisions_list)

  for (m in method_names) {
    for (s in scenario_names) {
      mat_prob <- probs[[m]][[s]]
      go_mat   <- go_decisions_list[[s]]$decisions_list[[m]]

      expected <- colMeans(go_mat)
      expect_equal(
        mat_prob["Go", ],
        expected,
        info = paste("Mismatch in method", m, "scenario", s)
      )
    }
  }
})


# -------------------------------------------------------------------
# Test: getGoProbabilities – Go and NoGo rows match colMeans of inputs
# Input:
#   - go_decisions_list and nogo_decisions_list for same scenarios.
# Behaviour:
#   - "Go" row = colMeans(go_mat), "NoGo" row = colMeans(nogo_mat).
# Expectations:
#   - Equality holds for all methods and scenarios.
# Why:
#   - Confirms correct aggregation of both Go and NoGo decisions.
# -------------------------------------------------------------------
test_that("getGoProbabilities: Go and NoGo rows match colMeans of input decisions", {
  probs <- getGoProbabilities(
    go_decisions_list   = go_decisions_list,
    nogo_decisions_list = nogo_decisions_list
  )

  method_names   <- names(go_decisions_list[[1]]$decisions_list)
  scenario_names <- names(go_decisions_list)

  for (m in method_names) {
    for (s in scenario_names) {
      mat_prob <- probs[[m]][[s]]

      go_mat   <- go_decisions_list[[s]]$decisions_list[[m]]
      nogo_mat <- nogo_decisions_list[[s]]$decisions_list[[m]]

      expected_go   <- colMeans(go_mat)
      expected_nogo <- colMeans(nogo_mat)

      expect_equal(
        mat_prob["Go", ],
        expected_go,
        info = paste("Go row mismatch in method", m, "scenario", s)
      )

      expect_equal(
        mat_prob["NoGo", ],
        expected_nogo,
        info = paste("NoGo row mismatch in method", m, "scenario", s)
      )
    }
  }
})


# -------------------------------------------------------------------
# Test: getGoProbabilities – column sums to 1 with Go/NoGo/Consider
# Input:
#   - go_decisions_list and nogo_decisions_list.
# Behaviour:
#   - "Consider" row is computed as 1 - p(Go) - p(NoGo), so column sums = 1.
# Expectations:
#   - For each scenario, colSums(matrix) == 1 exactly.
# Why:
#   - Sanity check that probabilities are normalized.
# -------------------------------------------------------------------
test_that("getGoProbabilities: columns sum to 1 when NoGo decisions are provided", {
  probs <- getGoProbabilities(
    go_decisions_list   = go_decisions_list,
    nogo_decisions_list = nogo_decisions_list
  )

  first_method <- names(probs)[1]
  for (scen_name in names(probs[[first_method]])) {
    mat <- probs[[first_method]][[scen_name]]

    col_sums <- colSums(mat)
    expect_true(
      all(abs(col_sums - 1) == 0),
      info = paste("Column sums not 1 for scenario", scen_name)
    )
  }
})


# -------------------------------------------------------------------
# Test: getGoProbabilities – overlapping Go and NoGo decisions
# Input:
#   - Modified go_decisions_list and nogo_decisions_list where at least one
#     cell is TRUE in both.
# Behaviour:
#   - Overlap check should throw an error.
# Expectations:
#   - Error containing “both go and nogo decisions”.
# Why:
#   - A trial cannot be simultaneously Go and NoGo in the same cell.
# -------------------------------------------------------------------
test_that("getGoProbabilities: errors if any decision is both Go and NoGo", {
  overlap_go   <- go_decisions_list
  overlap_nogo <- nogo_decisions_list

  scen1_name <- names(overlap_go)[1]
  meth1_name <- names(overlap_go[[scen1_name]]$decisions_list)[1]

  overlap_go[[scen1_name]]$decisions_list[[meth1_name]][1, 1]   <- TRUE
  overlap_nogo[[scen1_name]]$decisions_list[[meth1_name]][1, 1] <- TRUE

  expect_error(
    getGoProbabilities(
      go_decisions_list   = overlap_go,
      nogo_decisions_list = overlap_nogo
    ),
    "both go and nogo decisions"
  )
})


# Tests for print.decision_list ------------------------------------------------

set.seed(789)

scenarios_print <- simulateScenarios(
  n_subjects_list     = list(c(10, 20), c(10, 20)),
  response_rates_list = list(c(0.3, 0.5), c(0.4, 0.6)),
  n_trials            = 5
)

analyses_print <- performAnalyses(
  scenario_list     = scenarios_print,
  target_rates      = c(0.3, 0.3),
  n_mcmc_iterations = 80
)

coh_print <- c("p_1", "p_2")

decisions_dl <- getGoDecisions(
  analyses_list   = analyses_print,
  cohort_names    = coh_print,
  evidence_levels = c(0.5, 0.5),
  boundary_rules  = quote(c(TRUE, TRUE))
)

n_scenarios  <- length(decisions_dl)
n_methods    <- length(decisions_dl[[1]]$decisions_list)


# -------------------------------------------------------------------
# Test: print.decision_list – header shows scenario/method counts
# Input:
#   - decisions_dl with known number of scenarios and methods.
# Behaviour:
#   - Header line prints “decision_list of N scenario(s) with M method(s)”.
# Expectations:
#   - Output contains the constructed header pattern.
# Why:
#   - Confirms informative summary in the print method.
# -------------------------------------------------------------------
test_that("print.decision_list: header shows correct number of scenarios and methods", {
  header_pattern <- sprintf(
    "decision_list of %d scenario%s with %d method%s",
    n_scenarios,
    ifelse(n_scenarios == 1, "", "s"),
    n_methods,
    ifelse(n_methods == 1, "", "s")
  )

  expect_output(
    print(decisions_dl),
    header_pattern
  )
})


# -------------------------------------------------------------------
# Test: print.decision_list – scenario sections appear
# Input:
#   - decisions_dl with scenario names.
# Behaviour:
#   - For each scenario, the print method outputs “  - <scenario_name>”.
# Expectations:
#   - Each scenario name appears at least once in captured output.
# Why:
#   - Confirms per-scenario blocks are printed.
# -------------------------------------------------------------------
test_that("print.decision_list: prints a section for each scenario", {
  scen_names <- names(decisions_dl)
  expect_true(length(scen_names) >= 1)

  output <- capture.output(print(decisions_dl))

  for (nm in scen_names) {
    expect_true(
      any(grepl(paste0("\\b", nm, "\\b"), output)),
      info = paste("Scenario name", nm, "not found in printed output")
    )
  }
})


# -------------------------------------------------------------------
# Test: print.decision_list – method labels appear
# Input:
#   - decisions_dl with method_names present.
# Behaviour:
#   - Row labels are built from method_names via firstUpper().
# Expectations:
#   - Each method name (firstUpper) appears somewhere in output.
# Why:
#   - Verifies mapping from internal method_names to printed labels.
# -------------------------------------------------------------------
test_that("print.decision_list: each method name appears in printed matrix rows", {
  method_names <- names(decisions_dl[[1]]$decisions_list)
  expect_true(length(method_names) >= 1)

  output <- capture.output(print(decisions_dl))

  for (m in method_names) {
    m_upper <- paste0(toupper(substr(m, 1, 1)), substr(m, 2, nchar(m)))
    expect_true(
      any(grepl(m_upper, output, fixed = TRUE)),
      info = paste("Method label", m_upper, "not found in printed output")
    )
  }
})


# -------------------------------------------------------------------
# Test: print.decision_list – digits argument works
# Input:
#   - digits = 1 and digits = 4.
# Behaviour:
#   - Numeric entries are rounded with the given digits; no error.
# Expectations:
#   - Both calls produce output containing “decision_list of”.
# Why:
#   - Ensures digits parameter is respected but does not break printing.
# -------------------------------------------------------------------
test_that("print.decision_list: digits argument is accepted and does not error", {
  expect_output(
    print(decisions_dl, digits = 1),
    "decision_list of"
  )

  expect_output(
    print(decisions_dl, digits = 4),
    "decision_list of"
  )
})


# -------------------------------------------------------------------
# Test: print.decision_list – handles non-NULL decision_rules
# Input:
#   - decisions_dl with decision_rules present.
# Behaviour:
#   - Internal block rewrites boundary_rules for pretty printing.
# Expectations:
#   - Printing completes without error and includes header.
# Why:
#   - Confirms the complex decision_rules formatting branch works.
# -------------------------------------------------------------------
test_that("print.decision_list: handles non-NULL decision_rules without error", {
  expect_output(
    print(decisions_dl),
    "decision_list of"
  )
})


# -------------------------------------------------------------------
# Test: print.decision_list – works when decision_rules are NULL
# Input:
#   - decisions_dl with decision_rules removed.
# Behaviour:
#   - Skips decision_rules formatting and prints summary based only on decisions.
# Expectations:
#   - Header is printed as usual, no error occurs.
# Why:
#   - Ensures robust printing even if decision_rules slot is absent.
# -------------------------------------------------------------------
test_that("print.decision_list: works when decision_rules are NULL", {
  dl_no_rules <- decisions_dl
  for (i in seq_along(dl_no_rules)) {
    dl_no_rules[[i]]$decision_rules <- NULL
  }

  header_pattern <- sprintf(
    "decision_list of %d scenario%s with %d method%s",
    n_scenarios,
    ifelse(n_scenarios == 1, "", "s"),
    n_methods,
    ifelse(n_methods == 1, "", "s")
  )

  expect_output(
    print(dl_no_rules),
    header_pattern
  )
})


# -------------------------------------------------------------------
# Test: print.decision_list – one row per method per scenario
# Input:
#   - decisions_dl with >1 scenario and potentially multiple methods.
# Behaviour:
#   - For each scenario, mat_out = rbind over methods; rownames become
#     “    - <method>”.
# Expectations:
#   - Total number of lines starting with “    - ” equals n_scenarios * #methods.
# Why:
#   - Confirms the aggregation over methods matches the printed labels.
# -------------------------------------------------------------------
test_that("print.decision_list: for multiple methods, one row per method is printed per scenario", {
  go_probs <- getGoProbabilities(decisions_dl)

  scenario_index <- 1L
  mat_ref <- do.call(
    rbind,
    lapply(go_probs, function(y) y[[scenario_index]])
  )

  output <- capture.output(print(decisions_dl))

  method_label_lines <- grep("^    - ", output, value = TRUE)
  expected_total_rows <- n_scenarios * nrow(mat_ref)

  expect_equal(
    length(method_label_lines),
    expected_total_rows,
    info = paste(
      "Number of printed method label lines (", length(method_label_lines),
      ") does not match expected total rows (", expected_total_rows, ")"
    )
  )
})


# ===================================================================
# Tests for getAverageNSubjects -------------------------------------
# ===================================================================

scenarios_list <- simulateScenarios(
  n_subjects_list     = list(c(10, 20, 30), c(15, 25, 35)),
  response_rates_list = list(c(0.1, 0.2, 0.3), c(0.15, 0.25, 0.35)),
  n_trials            = 2
)


# -------------------------------------------------------------------
# Test: getAverageNSubjects – structure for multiple scenarios
# Input:
#   - scenario_list with two scenarios.
# Behaviour:
#   - Returns list of numeric vectors, one per scenario.
# Expectations:
#   - Result is a list with length equal to number of scenarios.
#   - Each element is numeric.
# Why:
#   - Basic contract of getAverageNSubjects.
# -------------------------------------------------------------------
test_that("returns correct structure for multiple scenarios", {
  result <- getAverageNSubjects(scenarios_list)
  expect_type(result, "list")
  expect_equal(length(result), length(scenarios_list))
  expect_true(all(sapply(result, is.numeric)))
})


# -------------------------------------------------------------------
# Test: getAverageNSubjects – computes column means of n_subjects
# Input:
#   - scenarios_list with known n_subjects for scenario_1.
# Behaviour:
#   - For each scenario, colMeans() of n_subjects is returned.
# Expectations:
#   - result[[1]] equals colMeans(scenarios_list[[1]]$n_subjects).
# Why:
#   - Verifies the actual calculation performed.
# -------------------------------------------------------------------
test_that("computes correct column means", {
  expected <- colMeans(scenarios_list[[1]]$n_subjects)
  result <- getAverageNSubjects(scenarios_list)
  expect_equal(result[[1]], expected)
})


# -------------------------------------------------------------------
# Test: getAverageNSubjects – works for single scenario
# Input:
#   - single_scenario scenario_list with one element.
# Behaviour:
#   - Returns list of length 1, named as input, with column means.
# Expectations:
#   - names(result) match names(single_scenario).
#   - result[[1]] equals colMeans(single_scenario[[1]]$n_subjects).
# Why:
#   - Handles degenerate one-scenario case correctly.
# -------------------------------------------------------------------
test_that("works for single scenario", {
  single_scenario <- simulateScenarios(
    n_subjects_list     = list(c(10, 20, 30)),
    response_rates_list = list(c(0.1, 0.2, 0.3)),
    n_trials            = 3
  )
  result <- getAverageNSubjects(single_scenario)
  expect_equal(names(result), names(single_scenario))
  expect_equal(result[[1]], colMeans(single_scenario[[1]]$n_subjects))
})


# -------------------------------------------------------------------
# Test: getAverageNSubjects – invalid input class
# Input:
#   - list() without class "scenario_list".
# Behaviour:
#   - assertClass(scenario_list, "scenario_list") fails.
# Expectations:
#   - Error mentioning 'scenario_list'.
# Why:
#   - Prevents accidental misuse with arbitrary lists.
# -------------------------------------------------------------------
test_that("throws error for invalid input class", {
  expect_error(getAverageNSubjects(list()), "scenario_list")
})


# -------------------------------------------------------------------
# Test: getAverageNSubjects – empty scenario_list
# Input:
#   - empty list with class "scenario_list".
# Behaviour:
#   - lapply over empty list produces empty list.
# Expectations:
#   - result is list().
# Why:
#   - Defines behaviour for degenerate empty input gracefully.
# -------------------------------------------------------------------
test_that("returns empty list for empty scenario_list", {
  empty_list <- structure(list(), class = "scenario_list")
  result <- getAverageNSubjects(empty_list)
  expect_equal(result, list())
})


# ===================================================================
# Tests for negateGoDecisions ---------------------------------------
# ===================================================================

set.seed(101)

scenarios_neg <- simulateScenarios(
  n_subjects_list     = list(c(10, 15)),
  response_rates_list = list(c(0.4, 0.7)),
  n_trials            = 6
)

analyses_neg <- performAnalyses(
  scenario_list     = scenarios_neg,
  target_rates      = c(0.4, 0.4),
  n_mcmc_iterations = 80
)

coh_neg <- c("p_1", "p_2")

go_decisions_list_neg <- getGoDecisions(
  analyses_list   = analyses_neg,
  cohort_names    = coh_neg,
  evidence_levels = c(0.5, 0.5),
  boundary_rules  = quote(c(TRUE, TRUE))
)

n_scen_neg <- length(go_decisions_list_neg)
n_meth_neg <- length(go_decisions_list_neg[[1]]$decisions_list)


# -------------------------------------------------------------------
# Test: negateGoDecisions – go_decisions_list must be decision_list
# Input:
#   - go_decisions_list = list().
# Behaviour:
#   - assertClass(go_decisions_list, "decision_list") fails.
# Expectations:
#   - Error mentioning 'go_decisions_list'.
# Why:
#   - Ensures correct object type as input.
# -------------------------------------------------------------------
test_that("negateGoDecisions: go_decisions_list must be a decision_list", {
  expect_error(
    negateGoDecisions(go_decisions_list = list()),
    "go_decisions_list"
  )
})


# -------------------------------------------------------------------
# Test: negateGoDecisions – overall_min_nogos validation
# Input:
#   - overall_min_nogos = -1L, and overall_min_nogos = "foo".
# Behaviour:
#   - Combined check: either 'all' or non-negative integer.
# Expectations:
#   - Both invalid values trigger error mentioning 'overall_min_nogos'.
# Why:
#   - Ensures threshold is meaningful or 'all'.
# -------------------------------------------------------------------
test_that("negateGoDecisions: overall_min_nogos must be 'all' or non-negative integer", {
  expect_error(
    negateGoDecisions(
      go_decisions_list = go_decisions_list_neg,
      overall_min_nogos = -1L
    ),
    "overall_min_nogos"
  )

  expect_error(
    negateGoDecisions(
      go_decisions_list = go_decisions_list_neg,
      overall_min_nogos = "foo"
    ),
    "overall_min_nogos"
  )
})


# -------------------------------------------------------------------
# Test: negateGoDecisions – cohort-level entries are logically negated
# Input:
#   - go_decisions_list_neg with multiple scenarios/methods.
# Behaviour:
#   - All non-overall columns are negated (TRUE->FALSE, FALSE->TRUE).
# Expectations:
#   - For each scenario/method, nogo_coh == !go_coh.
# Why:
#   - Core semantics of “negating” a decision_list.
# -------------------------------------------------------------------
test_that("negateGoDecisions: cohort-level entries are logically negated", {
  nogo_list <- negateGoDecisions(
    go_decisions_list = go_decisions_list_neg,
    overall_min_nogos = "all"
  )

  for (s in seq_len(n_scen_neg)) {
    for (m in seq_len(n_meth_neg)) {
      go_mat   <- go_decisions_list_neg[[s]]$decisions_list[[m]]
      nogo_mat <- nogo_list[[s]]$decisions_list[[m]]

      if (ncol(go_mat) > 1L) {
        go_coh   <- go_mat[, -1, drop = FALSE]
        nogo_coh <- nogo_mat[, -1, drop = FALSE]

        expect_identical(
          nogo_coh,
          !go_coh,
          info = paste("Cohort decisions not negated in scenario", s, "method", m)
        )
      }
    }
  }
})


# -------------------------------------------------------------------
# Test: negateGoDecisions – overall_min_nogos = "all"
# Input:
#   - overall_min_nogos = "all".
# Behaviour:
#   - Threshold internally becomes n_decisions - 1 (all cohort columns).
#   - overall TRUE iff all cohort NoGo columns are TRUE.
# Expectations:
#   - overall column equals row-wise indicator of “all cohorts NoGo”.
# Why:
#   - Confirms the semantics of 'all' for overall NoGo decision.
# -------------------------------------------------------------------
test_that("negateGoDecisions: overall_min_nogos = 'all' means 'all cohorts NoGo'", {
  nogo_list <- negateGoDecisions(
    go_decisions_list = go_decisions_list_neg,
    overall_min_nogos = "all"
  )

  for (s in seq_len(n_scen_neg)) {
    for (m in seq_len(n_meth_neg)) {
      nogo_mat <- nogo_list[[s]]$decisions_list[[m]]

      if (ncol(nogo_mat) <= 1L) next

      n_decisions   <- ncol(nogo_mat)
      nogo_coh      <- nogo_mat[, -1, drop = FALSE]
      expected_over <- apply(
        nogo_coh, 1,
        function(x) sum(x) >= (n_decisions - 1L)
      )
      actual_over   <- as.logical(nogo_mat[, 1])

      expect_identical(
        actual_over,
        expected_over
      )
    }
  }
})


# -------------------------------------------------------------------
# Test: negateGoDecisions – overall_min_nogos = 0
# Input:
#   - overall_min_nogos = 0L.
# Behaviour:
#   - sum(x) >= 0 is always TRUE, so overall should be TRUE in all rows.
# Expectations:
#   - overall column is entirely TRUE for all scenarios/methods.
# Why:
#   - Edge-case semantics of minimal threshold.
# -------------------------------------------------------------------
test_that("negateGoDecisions: overall_min_nogos = 0 makes overall always TRUE", {
  nogo_list <- negateGoDecisions(
    go_decisions_list = go_decisions_list_neg,
    overall_min_nogos = 0L
  )

  for (s in seq_len(n_scen_neg)) {
    for (m in seq_len(n_meth_neg)) {
      nogo_mat <- nogo_list[[s]]$decisions_list[[m]]

      if (ncol(nogo_mat) <= 1L) next

      overall_col <- as.logical(nogo_mat[, 1])
      expect_true(
        all(overall_col)
      )
    }
  }
})


# -------------------------------------------------------------------
# Test: negateGoDecisions – numeric overall_min_nogos = k
# Input:
#   - overall_min_nogos = 1L.
# Behaviour:
#   - overall TRUE iff at least k cohort NoGo decisions are TRUE.
# Expectations:
#   - overall column equals row-wise indicator of sum(no-go) >= k.
# Why:
#   - Confirms general threshold logic for k ≥ 1.
# -------------------------------------------------------------------
test_that("negateGoDecisions: numeric overall_min_nogos = k means 'at least k cohorts NoGo'", {
  k <- 1L
  nogo_list <- negateGoDecisions(
    go_decisions_list = go_decisions_list_neg,
    overall_min_nogos = k
  )

  for (s in seq_len(n_scen_neg)) {
    for (m in seq_len(n_meth_neg)) {
      nogo_mat <- nogo_list[[s]]$decisions_list[[m]]
      if (ncol(nogo_mat) <= 1L) next

      nogo_coh      <- nogo_mat[, -1, drop = FALSE]
      expected_over <- apply(nogo_coh, 1, function(x) sum(x) >= k)
      actual_over   <- as.logical(nogo_mat[, 1])

      expect_identical(
        actual_over,
        expected_over,
        info = paste("Overall NoGo != '>= k cohorts' for k =", k,
                     "in scenario", s, "method", m)
      )
    }
  }
})


# -------------------------------------------------------------------
# Test: negateGoDecisions – single-column decision matrices
# Input:
#   - decision_list with only one column and no explicit overall logic.
# Behaviour:
#   - All entries are simply negated; no overall recomputation.
# Expectations:
#   - new_mat == !orig_mat.
# Why:
#   - Confirms graceful behaviour when there is no explicit overall column.
# -------------------------------------------------------------------
test_that("negateGoDecisions: single-column decisions are just negated, no overall logic", {
  simple_dl <- list(
    scenario_1 = list(
      decisions_list = list(
        method_1 = matrix(c(TRUE, FALSE, TRUE), ncol = 1)
      )
    )
  )
  class(simple_dl) <- "decision_list"

  nogo_simple <- negateGoDecisions(
    go_decisions_list = simple_dl,
    overall_min_nogos = "all"
  )

  orig_mat <- simple_dl$scenario_1$decisions_list$method_1
  new_mat  <- nogo_simple$scenario_1$decisions_list$method_1

  expect_identical(
    new_mat,
    !orig_mat,
    info = "Single-column decision matrix not fully negated"
  )
})

#' @title getPriorParameters
#' @md
#' @description This function provides default prior parameters for the analysis methods
#' that can be used in \code{\link[bhmbasket]{performAnalyses}}.
#' @param method_names A vector of strings for the names of the methods to be used.
#' Available methods: `c("berry", "exnex", "exnex_adj", "exnex_mix", "exnex_adj_mix", "pooled", "stratified", "stratified_mix)`
#' @param target_rates A vector of numerics in `(0, 1)` for the
#' target rate of each cohort
#' @param n_worth An integer for the number of subjects the variability of the prior should reflect
#' response rate scale, Default: `1`
#' @param tau_scale A numeric for the scale parameter of the Half-normal distribution of \eqn{\tau}
#' in the methods `"berry"`, `"exnex"`, `"exnex"_mix`, `"exnex_adj"` and `"exnex_adj_mix"`, Default: `1`
#' @param w_j A numeric in `(0, 1)` for the weight of the Ex component in the methods `"exnex"`
#' and `"exnex_adj"`, Default: `0.5`
#' @return A list with prior parameters of class `prior_parameters_list`
#' @details
#' Regarding the default prior parameters for `"berry"`, `"exnex"`, `"exnex_mix"`, `"exnex_adj"` and `"exnex_adj_mix"`:
#' \itemize{
#'   \item `"berry"`: The mean of \eqn{\mu} is set to `0`.
#'   Its variance is calculated as proposed in "Robust exchangeability designs for early
#'   phase clinical trials with multiple strata" (Neuenschwander et al. (2016))
#'   with regard to `n_worth`.
#'   The scale parameter of \eqn{\tau} is set to `tau_scale`.
#'   \item `"exnex"`: The weight of the Ex component is set to `w_j`.
#'   For the Ex component:
#'   The target rate that results in the greatest variance is determined.
#'   The mean of \eqn{\mu} is set to that target rate.
#'   The variance of \eqn{\mu} is calculated as proposed in "Robust exchangeability designs for early
#'   phase clinical trials with multiple strata" (Neuenschwander et al. (2016))
#'   with regard to `n_worth`.
#'   The scale parameter of \eqn{\tau} is set to `tau_scale`.
#'   
#'   For the Nex components:
#'   The means of \eqn{\mu_j} are set to the respective target rates.
#'   The variances of \eqn{\tau_j} are calculated as proposed in "Robust exchangeability designs for early
#'   phase clinical trials with multiple strata" (Neuenschwander et al. (2016))
#'   with regard to `n_worth`, see also \code{\link[bhmbasket]{getMuVar}}.
#'   \item `"exnex_adj"`: The weight of the Ex component is set to `w_j`.
#'   For the Ex component:
#'   The target rate that results in the greatest variance is determined.
#'   The mean of \eqn{\mu} is set to `0`.
#'   The variance of \eqn{\mu} is calculated as proposed in "Robust exchangeability designs for early
#'   phase clinical trials with multiple strata" (Neuenschwander et al. (2016))
#'   with regard to `n_worth`, see also \code{\link[bhmbasket]{getMuVar}}.
#'   The scale parameter of \eqn{\tau} is set to `tau_scale`.
#'   
#'   For the Nex components:
#'   The means of \eqn{\mu_j} are set to the `0`.
#'   The variances of \eqn{\tau_j} are calculated as proposed in "Robust exchangeability designs for early
#'   phase clinical trials with multiple strata" (Neuenschwander et al. (2016))
#'   with regard to `n_worth`, see also \code{\link[bhmbasket]{getMuVar}}.
#'   \item `"exnex_mix"`: Uses the same default Ex prior construction as `"exnex"`.
#'   The Nex part default parameters are specified as a one-component mixture prior with
#'   `w_nex = 1`, `mean_nex = matrix(logit(target_rates), nrow = 1)`,
#'   and `sd_nex = matrix(sqrt(getMuVar(target_rates, 0, n_worth)), nrow = 1)`.
#'   This keeps the default mixture representation compatible with the `getPriorParameter()`'s input.
#'   \item `"exnex_adj_mix"`: Uses the same default Ex prior construction as `"exnex_adj"`.
#'   The Nex part is specified as a one-component mixture prior with
#'   `w_nex = 1`, `mean_nex = matrix(logit(target_rates), nrow = 1)`,
#'   and `sd_nex = matrix(sqrt(getMuVar(target_rates, 0, n_worth)), nrow = 1)`.
#'   The Ex component is centered as in `"exnex_adj"`.
#'   \item `"pooled"`: The target rate that results in the greatest variance is determined.
#'   The scale parameter \eqn{\alpha} is set to that target rate times `n_worth`.
#'   The scale parameter \eqn{\beta} is set to 1 - that target rate times `n_worth`.
#'   \item `"stratified"`:
#'   The scale parameters \eqn{\alpha_j} are set to `target_rates * n_worth`.
#'   The scale parameters \eqn{\beta_j} are set to `(1 - target_rates) * n_worth`.
#'   \item `"stratified_mix"`:
#'   A two-component beta mixture prior is created by default for each cohort.
#'   The first component uses `a_j = target_rates * n_worth` and
#'   `b_j = (1 - target_rates) * n_worth`.
#'   The second component is a vague prior with `a_j = 1` and `b_j = 1`.
#'   The default mixture weights are `c(0.8, 0.2)`.
#' }
#' @author Stephan Wojciekowski
#' @examples
#' prior_parameters_list <- getPriorParameters(
#'   method_names = c("berry", "exnex", "exnex_mix", "exnex_adj",
#'                    "exnex_adj_mix", "pooled", "stratified", "stratified_mix"),
#'   target_rates = c(0.1, 0.2, 0.3))
#' @rdname getPriorParameters
#' @seealso
#'  \code{\link[bhmbasket]{performAnalyses}}
#'  \code{\link[bhmbasket]{setPriorParametersBerry}}
#'  \code{\link[bhmbasket]{setPriorParametersExNex}}
#'  \code{\link[bhmbasket]{setPriorParametersExNexAdj}}
#'  \code{\link[bhmbasket]{setPriorParametersPooled}}
#'  \code{\link[bhmbasket]{setPriorParametersStratified}}
#'  \code{\link[bhmbasket]{setPriorParametersStratifiedMix}}
#'  \code{\link[bhmbasket]{combinePriorParameters}}
#'  \code{\link[bhmbasket]{getMuVar}}
#' @references Berry, Scott M., et al. "Bayesian hierarchical modeling of patient subpopulations:
#' efficient designs of phase II oncology clinical trials."
#' \emph{Clinical Trials} 10.5 (2013): 720-734.
#' @references Neuenschwander, Beat, et al. "Robust exchangeability designs
#' for early phase clinical trials with multiple strata."
#' \emph{Pharmaceutical statistics} 15.2 (2016): 123-134.
#' @export
getPriorParameters <- function (

  method_names,
  target_rates,
  n_worth   = 1,
  tau_scale = 1,
  w_j       = 0.5

) {

  error_method_names <-
    paste("Providing a (vector of) strings for the argument 'method_names'\n",
          "Must be one of 'berry', 'exnex', 'exnex_mix', 'exnex_adj', 'exnex_adj_mix',",
          "'pooled', 'stratified', 'stratified_mix'")
  error_target_rates <-
    "Providing a vector of numerics in (0, 1) for the argument 'target_rates'"
  error_tau_scale    <-
    "Providing a positive numeric for the argument 'tau_scale'"
  error_n_worth      <- "Providing a positive integer for the argument 'n_worth'"
  error_w_j          <- "Providing a numeric in (0, 1) for the argument 'w_j'"

  checkmate::assertCharacter(method_names, any.missing = FALSE, .var.name = error_method_names)

  method_names <- tryCatch({

    match.arg(
      method_names,
      choices = c(
        "berry", "exnex", "exnex_adj", "pooled", "stratified", "stratified_mix",
        "exnex_mix", "exnex_adj_mix"
      ),
      several.ok = TRUE
    )

  }, error = function (e) {
    stop(error_method_names, call. = FALSE)
  })

  checkmate::assertNumeric(target_rates, any.missing = FALSE, .var.name = error_target_rates)
  checkmate::assertTRUE(all(target_rates > 0 & target_rates < 1), .var.name = error_target_rates)

  checkmate::assertNumber(tau_scale, .var.name = error_tau_scale)
  checkmate::assertTRUE(tau_scale > 0, .var.name = error_tau_scale)

  checkmate::assertInt(n_worth, lower = 1, .var.name = error_n_worth)

  checkmate::assertNumeric(w_j, lower = 0, upper = 1, len = 1, .var.name = error_w_j)


  ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ###

  method_names <- sort(method_names)

  prior_parameters_list <- lapply(method_names, function (method_name) {

    if (method_name == "berry") {

      getPriorParametersBerry(
        target_rates = target_rates,
        n_worth      = n_worth,
        tau_scale    = tau_scale)[[1]]

    } else if (method_name == "exnex") {

      getPriorParametersExNex(
        target_rates = target_rates,
        n_worth      = n_worth,
        tau_scale    = tau_scale,
        w_j          = w_j)[[1]]

    } else if (method_name == "exnex_adj") {

      getPriorParametersExNexAdj(
        target_rates = target_rates,
        n_worth      = n_worth,
        tau_scale    = tau_scale,
        w_j          = w_j)[[1]]

    } else if (method_name == "exnex_mix") {

      getPriorParametersExNex(
        target_rates = target_rates,
        n_worth      = n_worth,
        tau_scale    = tau_scale,
        w_j          = w_j,
        w_nex        = 1,
        mean_nex     = matrix(logit(target_rates), nrow = 1),
        sd_nex       = matrix(sqrt(getMuVar(target_rates, 0, n_worth)), nrow = 1))[[1]]

    } else if (method_name == "exnex_adj_mix") {

      getPriorParametersExNexAdj(
        target_rates = target_rates,
        n_worth      = n_worth,
        tau_scale    = tau_scale,
        w_j          = w_j,
        w_nex        = 1,
        mean_nex     = matrix(logit(target_rates), nrow = 1),
        sd_nex       = matrix(sqrt(getMuVar(target_rates, 0, n_worth)), nrow = 1))[[1]]

    } else if (method_name == "pooled") {

      getPriorParametersPooled(
        target_rates = target_rates,
        n_worth      = n_worth)[[1]]

    } else if (method_name == "stratified") {

      getPriorParametersStratified(
        target_rates = target_rates,
        n_worth      = n_worth)[[1]]

    }

    else if (method_name == "stratified_mix") {

      getPriorParametersStratifiedMix(
        target_rates = target_rates,
        n_worth      = n_worth)[[1]]

    }

  })

  names(prior_parameters_list) <- method_names
  class(prior_parameters_list) <- "prior_parameters_list"

  return (prior_parameters_list)

}

#' @title combinePriorParameters
#' @md
#' @description This function combines prior parameters from different sources and returns them
#' for use in \code{\link[bhmbasket]{performAnalyses}}.
#' @param list_of_prior_parameters A list of items with class `prior_parameters_list`
#' @return A list with prior parameters of class `prior_parameters_list`
#' @details
#' This function is intended to combine the prior parameters set with the functions
#' \code{\link[bhmbasket]{setPriorParametersBerry}},
#' \code{\link[bhmbasket]{setPriorParametersExNex}},
#' \code{\link[bhmbasket]{setPriorParametersExNexAdj}},
#' \code{\link[bhmbasket]{setPriorParametersPooled}}, or
#' \code{\link[bhmbasket]{setPriorParametersStratified}},
#' in case more than one analysis method should be applied with
#' \code{\link[bhmbasket]{performAnalyses}}.
#' @author Stephan Wojciekowski
#' @examples
#'  prior_parameters_stratified <- setPriorParametersStratified(c(1, 2), c(3, 4))
#'  prior_parameters_berry      <- setPriorParametersBerry(0, 1, 2)
#'
#'  prior_parameters_list       <- combinePriorParameters(
#'    list(prior_parameters_berry,
#'         prior_parameters_stratified))
#' @rdname combinePriorParameters
#' @seealso
#'  \code{\link[bhmbasket]{performAnalyses}}
#'  \code{\link[bhmbasket]{setPriorParametersBerry}}
#'  \code{\link[bhmbasket]{setPriorParametersExNex}}
#'  \code{\link[bhmbasket]{setPriorParametersExNexAdj}}
#'  \code{\link[bhmbasket]{setPriorParametersPooled}}
#'  \code{\link[bhmbasket]{setPriorParametersStratified}}
#'  \code{\link[bhmbasket]{getPriorParameters}}
#'  \code{\link[bhmbasket]{getMuVar}}
#' @export
combinePriorParameters <- function (

  list_of_prior_parameters

) {

  error_list <- "Providing a list of items with class 'prior_parameters_list'"

  checkmate::assertList(
    list_of_prior_parameters, types = "list", any.missing = FALSE, .var.name = error_list
    )

  checkmate::assertTRUE(
    all(vapply(list_of_prior_parameters, function(x) inherits(x, "prior_parameters_list"), logical(1))),
    .var.name = error_list
  )

  ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ###

  method_names <- sapply(list_of_prior_parameters, names)

  checkmate::assertTRUE(
    length(unique(method_names)) == length(method_names),
  )

  prior_parameters_list <- vector(mode = "list", length(method_names))
  names(prior_parameters_list) <- method_names

  for (n in seq_along(method_names)) {

    prior_parameters_list[[n]] <- list_of_prior_parameters[[n]][[1]]

  }

  prior_parameters_list <- prior_parameters_list[sort(names(prior_parameters_list))]
  class(prior_parameters_list) <- "prior_parameters_list"

  return (prior_parameters_list)

}


is.prior_parameters_list <- function(x) {

  if (missing(x)) stop ("Please provide an object for the argument 'x'")

  inherits(x, "prior_parameters_list")

}

#' @title getMuVar
#' @md
#' @description This function returns the variance of \eqn{\mu} that is worth a certain number
#' of subjects for the distribution of the response rates.
#' @param response_rate A numeric for the response rate
#' @param tau_scale A numeric for the scale parameter of the Half-normal distribution of \eqn{\tau}
#' @param n_worth An integer for the number of subjects the variance of \eqn{\mu} should be worth
#' with regard to the variability of the distribution of the response rate, Default: `1`
#' @return Returns a numeric for the variance of \eqn{\mu}
#' @details Calculates the variance `mu_var` in
#' \deqn{logit(p) = \theta ~ N(\mu, \tau),
#' \mu ~ N(mu_mean, mu_var), \tau ~ HN(tau_scale),}
#' for `n_worth` number of observations, as in Neuenschwander et al. (2016).
#' @rdname getMuVar
#' @author Stephan Wojciekowski
#' @examples
#'   getMuVar(response_rate = 0.3,
#'            tau_scale     = 1)
#'   getMuVar(response_rate = 0.3,
#'            tau_scale     = 1,
#'            n_worth       = 2)
#' @references Neuenschwander, Beat, et al. "Robust exchangeability designs
#' for early phase clinical trials with multiple strata."
#' \emph{Pharmaceutical statistics} 15.2 (2016): 123-134.
#' @export
getMuVar <- function (

  response_rate,
  tau_scale,
  n_worth = 1

) {

  error_response_rate <-
    "Providing a numeric in (0, 1) for the argument 'response_rate'"
  error_tau_scale <-
    "Providing a positive numeric for the argument 'tau_scale'"
  error_n_worth <-
    "Providing a positive integer for the argument 'n_worth'"

  checkmate::assertNumeric(response_rate, .var.name = error_response_rate)
  checkmate::assertTRUE(all(response_rate > 0 & response_rate < 1), .var.name = error_response_rate)

  checkmate::assertNumber(tau_scale, lower = 0, .var.name = error_tau_scale)

  checkmate::assertInt(n_worth, lower = 1, .var.name = error_n_worth)

  ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ###

  mu_var <- (n_worth * response_rate * (1 - response_rate))^-1 - tau_scale^2

  return (mu_var)

}

## berry ####

getPriorParametersBerry <- function (

  target_rates,
  tau_scale = 1,
  n_worth   = 1

) {

  error_target_rates <-
    "Providing a vector of numerics in (0, 1) for the argument 'target_rates'"
  error_tau_scale    <-
    "Providing a positive numeric for the argument 'tau_scale'"
  error_n_worth      <-
    "Providing a positive integer for the argument 'n_worth'"

  checkmate::assertNumeric(target_rates, any.missing = FALSE, .var.name = error_target_rates)
  checkmate::assertTRUE(all(target_rates > 0 & target_rates < 1), .var.name = error_target_rates)

  checkmate::assertNumber(tau_scale, .var.name = error_tau_scale)
  checkmate::assertTRUE(tau_scale > 0, .var.name = error_tau_scale)

  checkmate::assertInt(n_worth, lower = 1, .var.name = error_n_worth)

  ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ###

  target_rate_max_var <- target_rates[abs(target_rates - 0.5) == max(abs(target_rates - 0.5))][1]

  mu_var <- getMuVar(target_rate_max_var, tau_scale, n_worth)

  if (mu_var <= 0) stop(simpleError(paste(
    "The provided input parameters lead to a variance of mu <= 0.",
    "Consider to decrease 'tau_scale' or 'n_worth'")))

  prior_parameters <- list(
    mu_mean   = 0,
    mu_sd     = mu_var^0.5,
    tau_scale = tau_scale)

  prior_parameters_list <- list (berry = prior_parameters)
  class(prior_parameters_list) <- "prior_parameters_list"

  return (prior_parameters_list)

}

#' @title setPriorParametersBerry
#' @md
#' @description This function sets prior parameters for the analysis method `"berry"`
#' for use in \code{\link[bhmbasket]{performAnalyses}}.
#' @param mu_mean A numeric for the mean of \eqn{\mu}
#' @param mu_sd A positive numeric for the standard deviation of \eqn{\mu}
#' @param tau_scale A positive numeric for the scale parameter of \eqn{\tau}
#' @return A list with prior parameters of class `prior_parameters_list`
#' @details
#' This function sets the prior parameters for the method proposed by Berry et al. (2013).
#' Note that the implemented distribution of \eqn{\tau} is half-normal.
#' @author Stephan Wojciekowski
#' @examples
#'  prior_parameters_berry <- setPriorParametersBerry(0, 1, 2)
#' @rdname setPriorParametersBerry
#' @seealso
#'  \code{\link[bhmbasket]{performAnalyses}}
#'  \code{\link[bhmbasket]{getPriorParameters}}
#'  \code{\link[bhmbasket]{combinePriorParameters}}
#'  \code{\link[bhmbasket]{setPriorParametersExNex}}
#'  \code{\link[bhmbasket]{setPriorParametersExNexAdj}}
#'  \code{\link[bhmbasket]{setPriorParametersPooled}}
#'  \code{\link[bhmbasket]{setPriorParametersStratified}}
#'  \code{\link[bhmbasket]{getMuVar}}
#' @references Berry, Scott M., et al. "Bayesian hierarchical modeling of patient subpopulations:
#' efficient designs of phase II oncology clinical trials."
#' \emph{Clinical Trials} 10.5 (2013): 720-734.
#' @export
setPriorParametersBerry <- function (

  mu_mean,
  mu_sd,
  tau_scale

) {

  error_mu_mean  <-
    "Providing a numeric for the argument 'mu_mean'"
  error_mu_sd  <-
    "Providing a positive numeric for the argument 'mu_sd'"
  error_tau_scale <-
    "Providing a positive numeric for the argument 'tau_scale'"


  checkmate::assertNumber(mu_mean, .var.name = error_mu_mean)

  checkmate::assertNumber(mu_sd, .var.name = error_mu_sd)
  checkmate::assertTRUE(mu_sd > 0, .var.name = error_mu_sd)

  checkmate::assertNumber(tau_scale, .var.name = error_tau_scale)
  checkmate::assertTRUE(tau_scale > 0, .var.name = error_tau_scale)

  ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ###

  prior_parameters <- list(
    mu_mean   = mu_mean,
    mu_sd     = mu_sd,
    tau_scale = tau_scale)

  prior_parameters_list <- list (berry = prior_parameters)
  class(prior_parameters_list) <- "prior_parameters_list"

  return (prior_parameters_list)

}

## exnex ####

getPriorParametersExNex <- function (

  target_rates,
  tau_scale = 1,
  n_worth   = 1,
  w_j       = 0.5,

  w_nex     = NULL,
  mean_nex  = NULL,
  sd_nex    = NULL

) {

  error_target_rates <- "Providing a vector of numerics in (0, 1) for the argument 'target_rates'"
  error_tau_scale    <- "Providing a positive numeric for the argument 'tau_scale'"
  error_n_worth      <- "Providing a positive integer for the argument 'n_worth'"
  error_w_j          <- "Providing a numeric in (0, 1) for the argument 'w_j'"
  error_w_nex        <- "Providing a numeric vector of weights in [0, 1] summing to 1 for the argument 'w_nex'"
  error_mean_nex     <- "Providing a numeric matrix for the argument 'mean_nex'"
  error_sd_nex       <- "Providing a positive numeric matrix for the argument 'sd_nex'"
  error_dim          <- "'mean_nex' and 'sd_nex' must have the same dimensions, nrow(mean_nex) must equal length(w_nex), and ncol(mean_nex) must equal length(target_rates)"

  checkmate::assertNumeric(target_rates, any.missing = FALSE, .var.name = error_target_rates)
  checkmate::assertTRUE(all(target_rates > 0 & target_rates < 1), .var.name = error_target_rates)

  checkmate::assertNumber(tau_scale, .var.name = error_tau_scale)
  checkmate::assertTRUE(tau_scale > 0, .var.name = error_tau_scale)

  checkmate::assertInt(n_worth, lower = 1, .var.name = error_n_worth)

  checkmate::assertNumeric(w_j, lower = 0, upper = 1, len = 1, .var.name = error_w_j)

  target_rate_max_var <- target_rates[abs(target_rates - 0.5) == max(abs(target_rates - 0.5))][1]
  mu_var <- getMuVar(target_rate_max_var, tau_scale, n_worth)

  if (mu_var <= 0) {
    stop(simpleError(paste(
      "The provided input parameters lead to a variance of mu <= 0.",
      "Consider to decrease 'tau_scale' or 'n_worth'"
    )))
  }

  prior_base <- list(
    mu_mean   = logit(target_rate_max_var),
    mu_sd     = sqrt(mu_var),
    tau_scale = tau_scale,
    w_j       = w_j
  )

  if (is.null(w_nex) && is.null(mean_nex) && is.null(sd_nex)) {

    prior_extra <- list(
      mu_j  = logit(target_rates),
      tau_j = sqrt(getMuVar(target_rates, 0, n_worth))
    )

  } else {

    checkmate::assertNumeric(w_nex, any.missing = FALSE, lower = 0, upper = 1, .var.name = error_w_nex)
    checkmate::assertTRUE(isTRUE(all.equal(sum(w_nex), 1)), .var.name = error_w_nex)

    checkmate::assertTRUE(is.matrix(mean_nex), .var.name = error_mean_nex)
    checkmate::assertTRUE(is.matrix(sd_nex), .var.name = error_sd_nex)
    checkmate::assertTRUE(all(sd_nex > 0), .var.name = error_sd_nex)
    checkmate::assertTRUE(all(dim(mean_nex) == dim(sd_nex)), .var.name = error_dim)
    checkmate::assertTRUE(nrow(mean_nex) == length(w_nex), .var.name = error_dim)
    checkmate::assertTRUE(ncol(mean_nex) == length(target_rates), .var.name = error_dim)

    prior_extra <- list(
      w_nex    = as.numeric(w_nex),
      mean_nex = mean_nex,
      sd_nex   = sd_nex
    )
  }

  prior_parameters_list <- list(exnex = c(prior_base, prior_extra))
  class(prior_parameters_list) <- "prior_parameters_list"

  return(prior_parameters_list)
}

#' @title setPriorParametersExNex
#' @md
#' @description This function sets prior parameters for the analysis method `"exnex"`
#' for use in \code{\link[bhmbasket]{performAnalyses}}.
#'
#' It supports two specifications for the Nex part:
#' \itemize{
#'   \item the standard ExNex specification with cohort-specific Nex priors via `mu_j` and `tau_j`
#'   \item an extended specification with a mixture prior on the Nex part via `w_nex`, `mean_nex`, and `sd_nex`
#' }
#'
#' @param mu_mean A numeric for the mean of \eqn{\mu}
#' @param mu_sd A positive numeric for the standard deviation of \eqn{\mu}
#' @param tau_scale A positive numeric for the scale parameter of \eqn{\tau}
#' @param mu_j A vector of numerics for the means \eqn{\mu_j} of the standard Nex priors.
#'   Ignored if `w_nex`, `mean_nex`, and `sd_nex` are provided.
#' @param tau_j A vector of positive numerics for the standard deviations \eqn{\tau_j}
#'   of the standard Nex priors. Ignored if `w_nex`, `mean_nex`, and `sd_nex` are provided.
#' @param w_j A numeric in `(0, 1)` for the weight of the Ex component, or a numeric vector
#'   of mixture weights summing to 1 if multiple Ex components are specified.
#' @param w_nex An optional numeric vector of mixture weights in \eqn{[0,1]} summing to 1
#'   for the Nex mixture prior.
#' @param mean_nex An optional numeric matrix of Nex mixture means with one row per Nex
#'   mixture component and one column per cohort.
#' @param sd_nex An optional positive numeric matrix of Nex mixture standard deviations with
#'   one row per Nex mixture component and one column per cohort.
#'
#' @return A list with prior parameters of class `prior_parameters_list`
#'
#' @details
#' This function sets the prior parameters for the method proposed by Neuenschwander et al. (2016).
#' If `w_nex`, `mean_nex`, and `sd_nex` are all `NULL`, the standard ExNex formulation is used.
#' Otherwise, the Nex part is specified as a finite mixture prior.
#'
#' @author Stephan Wojciekowski
#'
#' @examples
#' ## standard ExNex
#' prior_parameters_exnex <- setPriorParametersExNex(
#'   mu_mean   = 0,
#'   mu_sd     = 1,
#'   tau_scale = 2,
#'   mu_j      = c(4, 5),
#'   tau_j     = c(6, 7),
#'   w_j       = 0.8
#' )
#'
#' ## ExNex with Nex mixture prior
#' prior_parameters_exnex_mix <- setPriorParametersExNex(
#'   mu_mean   = 0,
#'   mu_sd     = 1,
#'   tau_scale = 2,
#'   w_j       = 0.8,
#'   w_nex     = c(0.7, 0.3),
#'   mean_nex  = rbind(c(4, 5), c(2, 3)),
#'   sd_nex    = rbind(c(6, 7), c(8, 9))
#' )
#'
#' @rdname setPriorParametersExNex
#' @seealso
#'  \code{\link[bhmbasket]{performAnalyses}}
#'  \code{\link[bhmbasket]{getPriorParameters}}
#'  \code{\link[bhmbasket]{combinePriorParameters}}
#'  \code{\link[bhmbasket]{setPriorParametersBerry}}
#'  \code{\link[bhmbasket]{setPriorParametersExNexAdj}}
#'  \code{\link[bhmbasket]{setPriorParametersPooled}}
#'  \code{\link[bhmbasket]{setPriorParametersStratified}}
#'  \code{\link[bhmbasket]{getMuVar}}
#' @references Neuenschwander, Beat, et al. "Robust exchangeability designs
#' for early phase clinical trials with multiple strata."
#' \emph{Pharmaceutical statistics} 15.2 (2016): 123-134.
#' @export
setPriorParametersExNex <- function (

  mu_mean,
  mu_sd,
  tau_scale,

  mu_j      = NULL,
  tau_j     = NULL,

  w_j,

  w_nex     = NULL,
  mean_nex  = NULL,
  sd_nex    = NULL

) {

  error_mu_mean    <- "Providing a numeric for the argument 'mu_mean'"
  error_mu_sd      <- "Providing a positive numeric for the argument 'mu_sd'"
  error_tau_scale  <- "Providing a positive numeric for the argument 'tau_scale'"
  error_mu_j       <- "Providing a (vector of) numeric(s) for the argument 'mu_j'"
  error_tau_j      <- "Providing a (vector of) positive numeric(s) for the argument 'tau_j'"
  error_w_j        <- "Providing a numeric in (0, 1) for the argument 'w_j'"
  error_w_nex      <- "Providing a numeric vector of weights in [0, 1] summing to 1 for the argument 'w_nex'"
  error_mean_nex   <- "Providing a numeric matrix for the argument 'mean_nex'"
  error_sd_nex     <- "Providing a positive numeric matrix for the argument 'sd_nex'"
  error_dim        <- "'mean_nex' and 'sd_nex' must have the same dimensions, and nrow(mean_nex) must equal length(w_nex)"
  error_mu_mean_sd <- "'mu_mean' and 'mu_sd' must have same length"
  error_mu_j_tau_j <- "'mu_j' and 'tau_j' must have the same length"
  error_w_j_long   <- "'w_j' must have length equal to length(mu_mean) + 1 if length(mu_mean) > 1"
  error_w_j_short  <- "'w_j' must have length 1 or 2 if length(mu_mean) = 1"
  error_w_j_sum    <- "Sum over items in 'w_j' must equal 1 if length(w_j) > 1"

  checkmate::assert_numeric(mu_mean, any.missing = FALSE, .var.name = error_mu_mean)
  checkmate::assert_numeric(mu_sd, lower = 0, any.missing = FALSE, .var.name = error_mu_sd)
  checkmate::assertTRUE(all(mu_sd > 0), .var.name = error_mu_sd)

  checkmate::assertNumber(tau_scale, .var.name = error_tau_scale)
  checkmate::assertTRUE(tau_scale > 0, .var.name = error_tau_scale)

  checkmate::assert_numeric(w_j, lower = 0, upper = 1, any.missing = FALSE, .var.name = error_w_j)

  checkmate::assert_true(length(mu_mean) == length(mu_sd), .var.name = error_mu_mean_sd)

  if (length(mu_mean) > 1) {
    checkmate::assert_true(length(w_j) == length(mu_mean) + 1, .var.name = error_w_j_long)
  } else {
    checkmate::assert_true(length(w_j) %in% c(1, 2), .var.name = error_w_j_short)
  }

  if (length(w_j) > 1) {
    checkmate::assert_true(isTRUE(all.equal(sum(w_j), 1)), .var.name = error_w_j_sum)
  }

  prior_base <- list(
    mu_mean   = mu_mean,
    mu_sd     = mu_sd,
    tau_scale = tau_scale,
    w_j       = w_j
  )

  if (is.null(w_nex) && is.null(mean_nex) && is.null(sd_nex)) {

    checkmate::assert_numeric(mu_j, any.missing = FALSE, .var.name = error_mu_j)
    checkmate::assert_numeric(tau_j, any.missing = FALSE, .var.name = error_tau_j)
    checkmate::assertTRUE(all(tau_j > 0), .var.name = error_tau_j)
    checkmate::assert_true(length(mu_j) == length(tau_j), .var.name = error_mu_j_tau_j)

    prior_extra <- list(
      mu_j  = mu_j,
      tau_j = tau_j
    )

  } else {

    checkmate::assertNumeric(w_nex, any.missing = FALSE, lower = 0, upper = 1, .var.name = error_w_nex)
    checkmate::assertTRUE(isTRUE(all.equal(sum(w_nex), 1)), .var.name = error_w_nex)

    checkmate::assertTRUE(is.matrix(mean_nex), .var.name = error_mean_nex)
    checkmate::assertTRUE(is.matrix(sd_nex), .var.name = error_sd_nex)
    checkmate::assertTRUE(all(sd_nex > 0), .var.name = error_sd_nex)
    checkmate::assertTRUE(all(dim(mean_nex) == dim(sd_nex)), .var.name = error_dim)
    checkmate::assertTRUE(nrow(mean_nex) == length(w_nex), .var.name = error_dim)

    prior_extra <- list(
      w_nex    = as.numeric(w_nex),
      mean_nex = mean_nex,
      sd_nex   = sd_nex
    )
  }

  prior_parameters_list <- list(exnex = c(prior_base, prior_extra))
  class(prior_parameters_list) <- "prior_parameters_list"

  return(prior_parameters_list)
}

## exnex_adj ####

getPriorParametersExNexAdj <- function (

  target_rates,
  tau_scale = 1,
  n_worth   = 1,

  w_j       = 0.5,

  w_nex     = NULL,
  mean_nex  = NULL,
  sd_nex    = NULL

) {

  error_target_rates <-
    "Providing a vector of numerics in (0, 1) for the argument 'target_rates'"
  error_tau_scale    <-
    "Providing a positive numeric for the argument 'tau_scale'"
  error_n_worth      <-
    "Providing a positive integer for the argument 'n_worth'"
  error_w_j          <-
    "Providing a numeric in (0, 1) for the argument 'w_j'"
  error_w_nex        <-
    "Providing a numeric vector of weights in [0, 1] summing to 1 for the argument 'w_nex'"
  error_mean_nex     <-
    "Providing a numeric matrix for the argument 'mean_nex'"
  error_sd_nex       <-
    "Providing a positive numeric matrix for the argument 'sd_nex'"
  error_dim          <-
    "'mean_nex' and 'sd_nex' must have the same dimensions, nrow(mean_nex) must equal length(w_nex), and ncol(mean_nex) must equal length(target_rates)"

  checkmate::assertNumeric(
    target_rates, any.missing = FALSE, .var.name = error_target_rates
  )
  checkmate::assertTRUE(
    all(target_rates > 0 & target_rates < 1), .var.name = error_target_rates
  )

  checkmate::assertNumber(
    tau_scale, .var.name = error_tau_scale
  )
  checkmate::assertTRUE(
    tau_scale > 0, .var.name = error_tau_scale
  )

  checkmate::assertInt(
    n_worth, lower = 1, .var.name = error_n_worth
  )

  checkmate::assertNumeric(
    w_j, lower = 0, upper = 1, len = 1, .var.name = error_w_j
  )

  ## mixed Nex validation only if mixture inputs are supplied
  if (!(is.null(w_nex) && is.null(mean_nex) && is.null(sd_nex))) {
    checkmate::assertNumeric(
      w_nex, any.missing = FALSE, lower = 0, upper = 1, .var.name = error_w_nex
    )
    checkmate::assertTRUE(
      isTRUE(all.equal(sum(w_nex), 1)), .var.name = error_w_nex
    )

    checkmate::assertTRUE(
      is.matrix(mean_nex), .var.name = error_mean_nex
    )
    checkmate::assertTRUE(
      is.matrix(sd_nex), .var.name = error_sd_nex
    )
    checkmate::assertTRUE(
      all(sd_nex > 0), .var.name = error_sd_nex
    )
    checkmate::assertTRUE(
      all(dim(mean_nex) == dim(sd_nex)), .var.name = error_dim
    )
    checkmate::assertTRUE(
      nrow(mean_nex) == length(w_nex), .var.name = error_dim
    )
    checkmate::assertTRUE(
      ncol(mean_nex) == length(target_rates), .var.name = error_dim
    )
  }

  ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ###

  prior_parameters <- getPriorParametersExNex(
    target_rates = target_rates,
    tau_scale    = tau_scale,
    n_worth      = n_worth,
    w_j          = w_j,
    w_nex        = w_nex,
    mean_nex     = mean_nex,
    sd_nex       = sd_nex
  )[[1]]

  prior_parameters$mu_mean <- 0

  if (is.null(w_nex) && is.null(mean_nex) && is.null(sd_nex)) {
    prior_parameters$mu_j <- rep(0, length(target_rates))
  }

  prior_parameters_list <- list(exnex_adj = prior_parameters)
  class(prior_parameters_list) <- "prior_parameters_list"

  return(prior_parameters_list)
}

#' @title setPriorParametersExNexAdj
#' @md
#' @description This function sets prior parameters for the analysis method `"exnex_adj"`
#' for use in \code{\link[bhmbasket]{performAnalyses}}.
#'
#' It supports two specifications for the Nex part:
#' \itemize{
#'   \item the standard ExNex Adjusted specification with cohort-specific Nex priors via `mu_j` and `tau_j`
#'   \item an extended specification with a mixture prior on the Nex part via `w_nex`, `mean_nex`, and `sd_nex`
#' }
#'
#' @param mu_mean [numeric] Mean of \eqn{\mu}
#' @param mu_sd [numeric] Positive standard deviation of \eqn{\mu}
#' @param tau_scale [numeric] Positive scale parameter of \eqn{\tau}
#' @param mu_j [numeric] Vector of means \eqn{\mu_j} for the standard Nex priors.
#'   Ignored if `w_nex`, `mean_nex`, and `sd_nex` are provided.
#' @param tau_j [numeric] Vector of positive standard deviations \eqn{\tau_j}
#'   for the standard Nex priors. Ignored if `w_nex`, `mean_nex`, and `sd_nex` are provided.
#' @param w_j [numeric] Weight of the Ex component in `(0, 1)`, or a numeric vector
#'   of mixture weights summing to 1 if multiple Ex components are specified.
#' @param w_nex [numeric] Optional vector of mixture weights in \eqn{[0,1]} summing to 1
#'   for the Nex mixture prior.
#' @param mean_nex [numeric] Optional matrix of Nex mixture means with one row per Nex
#'   mixture component and one column per cohort.
#' @param sd_nex [numeric] Optional positive matrix of Nex mixture standard deviations with
#'   one row per Nex mixture component and one column per cohort.
#'
#' @return A list with prior parameters of class `prior_parameters_list`
#'
#' @details
#' This function sets prior parameters for the ExNex Adjusted method, which combines
#' the approach proposed by Neuenschwander et al. (2016) and the approach proposed by
#' Berry et al. (2013). If `w_nex`, `mean_nex`, and `sd_nex` are all `NULL`, the standard
#' ExNex Adjusted formulation is used. Otherwise, the Nex part is specified as a finite
#' mixture prior.
#'
#' @author Stephan Wojciekowski
#'
#' @examples
#' ## standard ExNex Adjusted
#' prior_parameters_exnex_adj <- setPriorParametersExNexAdj(
#'   mu_mean   = 0,
#'   mu_sd     = 1,
#'   tau_scale = 2,
#'   mu_j      = c(4, 5),
#'   tau_j     = c(6, 7),
#'   w_j       = 0.8
#' )
#'
#' ## ExNex Adjusted with Nex mixture prior
#' prior_parameters_exnex_adj_mix <- setPriorParametersExNexAdj(
#'   mu_mean   = 0,
#'   mu_sd     = 1,
#'   tau_scale = 2,
#'   w_j       = 0.8,
#'   w_nex     = c(0.7, 0.3),
#'   mean_nex  = rbind(c(0, 0), c(-1, -1)),
#'   sd_nex    = rbind(c(6, 7), c(8, 9))
#' )
#'
#' @rdname setPriorParametersExNexAdj
#' @seealso
#'  \code{\link[bhmbasket]{performAnalyses}}
#'  \code{\link[bhmbasket]{getPriorParameters}}
#'  \code{\link[bhmbasket]{combinePriorParameters}}
#'  \code{\link[bhmbasket]{setPriorParametersBerry}}
#'  \code{\link[bhmbasket]{setPriorParametersExNex}}
#'  \code{\link[bhmbasket]{setPriorParametersPooled}}
#'  \code{\link[bhmbasket]{setPriorParametersStratified}}
#'  \code{\link[bhmbasket]{getMuVar}}
#' @references Neuenschwander, Beat, et al. "Robust exchangeability designs
#' for early phase clinical trials with multiple strata."
#' \emph{Pharmaceutical statistics} 15.2 (2016): 123-134.
#' @export
setPriorParametersExNexAdj <- function (

  mu_mean,
  mu_sd,
  tau_scale,

  mu_j      = NULL,
  tau_j     = NULL,

  w_j,

  w_nex     = NULL,
  mean_nex  = NULL,
  sd_nex    = NULL

) {

  error_mu_mean    <- "Providing a numeric for the argument 'mu_mean'"
  error_mu_sd      <- "Providing a positive numeric for the argument 'mu_sd'"
  error_tau_scale  <- "Providing a positive numeric for the argument 'tau_scale'"
  error_mu_j       <- "Providing a (vector of) numeric(s) for the argument 'mu_j'"
  error_tau_j      <- "Providing a (vector of) positive numeric(s) for the argument 'tau_j'"
  error_w_j        <- "Providing a numeric in (0, 1) for the argument 'w_j'"
  error_w_nex      <- "Providing a numeric vector of weights in [0, 1] summing to 1 for the argument 'w_nex'"
  error_mean_nex   <- "Providing a numeric matrix for the argument 'mean_nex'"
  error_sd_nex     <- "Providing a positive numeric matrix for the argument 'sd_nex'"
  error_dim        <- "'mean_nex' and 'sd_nex' must have the same dimensions, and nrow(mean_nex) must equal length(w_nex)"
  error_mu_mean_sd <- "'mu_mean' and 'mu_sd' must have same length"
  error_mu_j_tau_j <- "'mu_j' and 'tau_j' must have the same length"
  error_w_j_long   <- "'w_j' must have length equal to length(mu_mean) + 1 if length(mu_mean) > 1"
  error_w_j_short  <- "'w_j' must have length 1 or 2 if length(mu_mean) = 1"
  error_w_j_sum    <- "Sum over items in 'w_j' must equal 1 if length(w_j) > 1"

  checkmate::assert_numeric(
    mu_mean, any.missing = FALSE, .var.name = error_mu_mean
  )

  checkmate::assert_numeric(
    mu_sd, lower = 0, any.missing = FALSE, .var.name = error_mu_sd
  )
  checkmate::assertTRUE(
    all(mu_sd > 0), .var.name = error_mu_sd
  )

  checkmate::assertNumber(
    tau_scale, .var.name = error_tau_scale
  )
  checkmate::assertTRUE(
    tau_scale > 0, .var.name = error_tau_scale
  )

  checkmate::assert_numeric(
    w_j, lower = 0, upper = 1, any.missing = FALSE, .var.name = error_w_j
  )

  checkmate::assert_true(
    length(mu_mean) == length(mu_sd), .var.name = error_mu_mean_sd
  )

  if (length(mu_mean) > 1) {
    checkmate::assert_true(
      length(w_j) == length(mu_mean) + 1, .var.name = error_w_j_long
    )
  } else {
    checkmate::assert_true(
      length(w_j) %in% c(1, 2), .var.name = error_w_j_short
    )
  }

  if (length(w_j) > 1) {
    checkmate::assert_true(
      isTRUE(all.equal(sum(w_j), 1)), .var.name = error_w_j_sum
    )
  }

  ## standard ExNex Adjusted
  if (is.null(w_nex) && is.null(mean_nex) && is.null(sd_nex)) {

    checkmate::assert_numeric(
      mu_j, any.missing = FALSE, .var.name = error_mu_j
    )
    checkmate::assert_numeric(
      tau_j, any.missing = FALSE, .var.name = error_tau_j
    )
    checkmate::assertTRUE(
      all(tau_j > 0), .var.name = error_tau_j
    )
    checkmate::assert_true(
      length(mu_j) == length(tau_j), .var.name = error_mu_j_tau_j
    )

  } else {

    ## mixed Nex version
    checkmate::assertNumeric(
      w_nex, any.missing = FALSE, lower = 0, upper = 1, .var.name = error_w_nex
    )
    checkmate::assertTRUE(
      isTRUE(all.equal(sum(w_nex), 1)), .var.name = error_w_nex
    )

    checkmate::assertTRUE(
      is.matrix(mean_nex), .var.name = error_mean_nex
    )
    checkmate::assertTRUE(
      is.matrix(sd_nex), .var.name = error_sd_nex
    )
    checkmate::assertTRUE(
      all(sd_nex > 0), .var.name = error_sd_nex
    )
    checkmate::assertTRUE(
      all(dim(mean_nex) == dim(sd_nex)), .var.name = error_dim
    )
    checkmate::assertTRUE(
      nrow(mean_nex) == length(w_nex), .var.name = error_dim
    )
  }

  ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ###

  prior_parameters_list <- setPriorParametersExNex(
    mu_mean   = mu_mean,
    mu_sd     = mu_sd,
    tau_scale = tau_scale,
    mu_j      = mu_j,
    tau_j     = tau_j,
    w_j       = w_j,
    w_nex     = w_nex,
    mean_nex  = mean_nex,
    sd_nex    = sd_nex
  )

  names(prior_parameters_list) <- "exnex_adj"

  return(prior_parameters_list)
}


## pooled ####

getPriorParametersPooled <- function (

  target_rates,
  n_worth = 1

) {

  error_target_rates <-
    "Providing a vector of numerics in (0, 1) for the argument 'target_rates'"
  error_n_worth      <-
    "Providing a positive integer for the argument 'n_worth'"

  checkmate::assertNumeric(
    target_rates, any.missing = FALSE, .var.name = error_target_rates
  )
  checkmate::assertTRUE(
    all(target_rates > 0 & target_rates < 1), .var.name = error_target_rates
  )

  checkmate::assertInt(
    n_worth, lower = 1, .var.name = error_n_worth
  )

  ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ###

  target_rate_min_var <- target_rates[abs(target_rates - 0.5) == min(abs(target_rates - 0.5))][1]

  a <- target_rate_min_var * n_worth
  b <- (1 - target_rate_min_var) * n_worth

  prior_parameters <- list(a = a, b = b)

  prior_parameters_list        <- list(pooled = prior_parameters)
  class(prior_parameters_list) <- "prior_parameters_list"

  return (prior_parameters_list)

}

#' @title setPriorParametersPooled
#' @md
#' @description This function sets prior parameters for the analysis method `"pooled"`
#' for use in \code{\link[bhmbasket]{performAnalyses}}.
#' @param a A positive numeric for \eqn{\alpha}
#' @param b A positive numeric for \eqn{\beta}
#' @return A list with prior parameters of class `prior_parameters_list`
#' @details
#' The method `"pooled"` is a beta-binomial model that pools all cohorts.
#' The prior parameters are the scale parameters of the beta prior distribution.
#' @author Stephan Wojciekowski
#' @examples
#'  prior_parameters_pooled <- setPriorParametersPooled(1, 2)
#' @rdname setPriorParametersPooled
#' @seealso
#'  \code{\link[bhmbasket]{performAnalyses}}
#'  \code{\link[bhmbasket]{getPriorParameters}}
#'  \code{\link[bhmbasket]{combinePriorParameters}}
#'  \code{\link[bhmbasket]{setPriorParametersBerry}}
#'  \code{\link[bhmbasket]{setPriorParametersExNex}}
#'  \code{\link[bhmbasket]{setPriorParametersExNexAdj}}
#'  \code{\link[bhmbasket]{setPriorParametersStratified}}
#'  \code{\link[bhmbasket]{getMuVar}}
#' @export
setPriorParametersPooled <- function (

  a,
  b

) {

  error_a <- "Providing a positive numeric for the argument 'a'"
  error_b <- "Providing a positive numeric for the argument 'b'"

  checkmate::assertNumber(
    a, .var.name = error_a
  )
  checkmate::assertTRUE(
    a > 0, .var.name = error_a
  )

  checkmate::assertNumber(
    b, .var.name = error_b
  )
  checkmate::assertTRUE(
    b > 0, .var.name = error_b
  )

  ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ###

  prior_parameters <- list(a = a, b = b)

  prior_parameters_list        <- list(pooled = prior_parameters)
  class(prior_parameters_list) <- "prior_parameters_list"

  return (prior_parameters_list)

}

## stratified ####

getPriorParametersStratified <- function (

  target_rates,
  n_worth = 1

) {

  error_target_rates <-
    "Providing a vector of numerics in (0, 1) for the argument 'target_rates'"
  error_n_worth      <-
    "Providing a positive integer for the argument 'n_worth'"

  checkmate::assertNumeric(
    target_rates, any.missing = FALSE, .var.name = error_target_rates
  )
  checkmate::assertTRUE(
    all(target_rates > 0 & target_rates < 1), .var.name = error_target_rates
  )

  checkmate::assertInt(
    n_worth, lower = 1, .var.name = error_n_worth
  )

  ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ###

  a_j <- target_rates * n_worth
  b_j <- (1 - target_rates) * n_worth

  prior_parameters <- list(a_j = a_j, b_j = b_j)

  prior_parameters_list        <- list(stratified = prior_parameters)
  class(prior_parameters_list) <- "prior_parameters_list"

  return (prior_parameters_list)

}

#' @title setPriorParametersStratified
#' @md
#' @description This function sets prior parameters for the analysis method `"stratified"`
#' for use in \code{\link[bhmbasket]{performAnalyses}}.
#' @param a_j A vector of positive numerics for \eqn{\alpha}
#' @param b_j A vector of positive numerics for \eqn{\beta}
#' @return A list with prior parameters of class `prior_parameters_list`
#' @details
#' The method `"stratified"` is a beta-binomial model that assesses each cohort individually.
#' The prior parameters are the scale parameters of the beta prior distributions.
#' @author Stephan Wojciekowski
#' @examples
#'  prior_parameters_pooled <- setPriorParametersStratified(c(1, 2), c(3, 4))
#' @rdname setPriorParametersStratified
#' @seealso
#'  \code{\link[bhmbasket]{performAnalyses}}
#'  \code{\link[bhmbasket]{getPriorParameters}}
#'  \code{\link[bhmbasket]{combinePriorParameters}}
#'  \code{\link[bhmbasket]{setPriorParametersBerry}}
#'  \code{\link[bhmbasket]{setPriorParametersExNex}}
#'  \code{\link[bhmbasket]{setPriorParametersExNexAdj}}
#'  \code{\link[bhmbasket]{setPriorParametersPooled}}
#'  \code{\link[bhmbasket]{getMuVar}}
#' @export
setPriorParametersStratified <- function (

  a_j,
  b_j

) {

  error_a_j          <- "Providing a (vector of) positive numeric(s) for the argument 'a_j'"
  error_b_j          <- "Providing a (vector of) positive numeric(s) for the argument 'b_j'"
  error_a_j_b_j_len  <- "a_j and b_j must have the same length"

  checkmate::assertNumeric(
    a_j, any.missing = FALSE, .var.name = error_a_j
  )
  checkmate::assertTRUE(
    all(a_j > 0), .var.name = error_a_j
  )

  checkmate::assertNumeric(
    b_j, any.missing = FALSE, .var.name = error_b_j
  )
  checkmate::assertTRUE(
    all(b_j > 0), .var.name = error_b_j
  )

  checkmate::assertTRUE(
    length(a_j) == length(b_j), .var.name = error_a_j_b_j_len
  )

  ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ###

  prior_parameters <- list(a_j = a_j, b_j = b_j)

  prior_parameters_list        <- list(stratified = prior_parameters)
  class(prior_parameters_list) <- "prior_parameters_list"

  return (prior_parameters_list)

}

## stratified_mix ####

getPriorParametersStratifiedMix <- function (

  target_rates,
  n_worth = 1,
  w       = c(0.8, 0.2)

) {

  error_target_rates <-
    "Providing a vector of numerics in (0, 1) for the argument 'target_rates'"
  error_n_worth      <-
    "Providing a positive integer for the argument 'n_worth'"
  error_w            <-
    "Providing a numeric vector of weights in [0, 1] summing to 1 for the argument 'w'"

  checkmate::assertNumeric(
    target_rates, any.missing = FALSE, .var.name = error_target_rates
  )
  checkmate::assertTRUE(
    all(target_rates > 0 & target_rates < 1), .var.name = error_target_rates
  )

  checkmate::assertInt(
    n_worth, lower = 1, .var.name = error_n_worth
  )

  checkmate::assertNumeric(
    w, any.missing = FALSE, lower = 0, upper = 1, .var.name = error_w
  )
  checkmate::assertTRUE(
    isTRUE(all.equal(sum(w), 1)), .var.name = error_w
  )

  ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ###

  K <- length(w)
  J <- length(target_rates)

  if (K != 2) {
    stop(simpleError("The default implementation of 'getPriorParametersStratifiedMix()' currently expects a weight vector of length 2."))
  }

  a_j <- matrix(NA_real_, nrow = K, ncol = J)
  b_j <- matrix(NA_real_, nrow = K, ncol = J)

  ## component 1: informative stratified prior
  a_j[1, ] <- target_rates * n_worth
  b_j[1, ] <- (1 - target_rates) * n_worth

  ## component 2: vague prior
  a_j[2, ] <- rep(1, J)
  b_j[2, ] <- rep(1, J)

  prior_parameters <- list(
    w   = w,
    a_j = a_j,
    b_j = b_j
  )

  prior_parameters_list <- list(stratified_mix = prior_parameters)
  class(prior_parameters_list) <- "prior_parameters_list"

  return (prior_parameters_list)

}

#' @title setPriorParametersStratifiedMix
#' @md
#' @description This function sets prior parameters for the analysis method `"stratified_mix"`
#' for use in \code{\link[bhmbasket]{performAnalyses}}.
#' @param w A numeric vector of mixture weights in \eqn{[0,1]} summing to 1.
#' @param a_j A positive numeric matrix of beta shape parameters \eqn{\alpha_j},
#' with rows per mixture component and columns per cohort.
#' @param b_j A positive numeric matrix of beta shape parameters \eqn{\beta_j},
#' with rows per mixture component and columns per cohort.
#' @return A list with prior parameters of class `prior_parameters_list`
#' @details
#' The method `"stratified_mix"` is a beta-binomial model that assesses each cohort
#' individually with a finite mixture beta prior.
#' See also the R package `RBesT`.
#' @author Stephan Wojciekowski
#' @examples
#' prior_parameters_stratified_mix <- setPriorParametersStratifiedMix(
#'   w   = c(0.8, 0.2),
#'   a_j = rbind(c(2, 3), c(1, 1)),
#'   b_j = rbind(c(8, 7), c(1, 1))
#' )
#' @rdname setPriorParametersStratifiedMix
#' @seealso
#'  \code{\link[bhmbasket]{performAnalyses}}
#'  \code{\link[bhmbasket]{getPriorParameters}}
#'  \code{\link[bhmbasket]{combinePriorParameters}}
#'  \code{\link[bhmbasket]{setPriorParametersStratified}}
#' @export
setPriorParametersStratifiedMix <- function(
    w,
    a_j,
    b_j
) {

  error_w    <- "Providing a numeric vector of weights in [0,1] summing to 1 for argument 'w'"
  error_a_j  <- "Providing a positive numeric matrix for argument 'a_j'"
  error_b_j  <- "Providing a positive numeric matrix for argument 'b_j'"
  error_dim  <- "a_j and b_j must have the same dimensions, and nrow(a_j) must equal length(w)"

  checkmate::assertNumeric(w, any.missing = FALSE, lower = 0, upper = 1, .var.name = error_w)
  checkmate::assertTRUE(isTRUE(all.equal(sum(w), 1)), .var.name = error_w)

  checkmate::assertTRUE(is.matrix(a_j), .var.name = error_a_j)
  checkmate::assertTRUE(is.matrix(b_j), .var.name = error_b_j)
  checkmate::assertTRUE(all(a_j > 0), .var.name = error_a_j)
  checkmate::assertTRUE(all(b_j > 0), .var.name = error_b_j)

  checkmate::assertTRUE(all(dim(a_j) == dim(b_j)), .var.name = error_dim)
  checkmate::assertTRUE(nrow(a_j) == length(w), .var.name = error_dim)

  prior_parameters <- list(
    w   = as.numeric(w),
    a_j = a_j,
    b_j = b_j
  )

  prior_parameters_list <- list(stratified_mix = prior_parameters)
  class(prior_parameters_list) <- "prior_parameters_list"

  return(prior_parameters_list)
}

# Tests for getMuVar -----------------------------------------------------------

# ------------------------------------------------------------------
# Test: correct mu_var for simple reference case
# Input:
#   - response_rate = 0.5, tau_scale = 1, n_worth = 2
# Behaviour:
#   - getMuVar() implements: mu_var = (n_worth * p * (1 - p))^-1 - tau_scale^2
# Expectations:
#   - Result equals manual calculation, is a finite double.
# Why:
#   - Sanity-checks the core formula and numeric type for a simple case.
# ------------------------------------------------------------------
test_that("mu_var calculation is correct for simple case", {

  # Manual calculation: (n_worth * p * (1 - p))^-1 - tau_scale^2
  expected <- (2 * 0.5 * (1 - 0.5))^-1 - 1^2

  result <- getMuVar(response_rate = 0.5, tau_scale = 1, n_worth = 2)

  expect_equal(result, expected)

  expect_type(result, "double")

  expect_true(is.finite(result))

})

# ------------------------------------------------------------------
# Test: response_rate must lie strictly between 0 and 1
# Input:
#   - response_rate = -0.1 and 1.5, tau_scale = 1
# Behaviour:
#   - Function should reject probabilities outside (0,1).
# Expectations:
#   - Errors for each invalid response_rate.
# Why:
#   - The variance formula assumes a valid Bernoulli probability.
# ------------------------------------------------------------------
test_that("error if response_rate out of bounds", {

  expect_error(getMuVar(response_rate = -0.1, tau_scale = 1))

  expect_error(getMuVar(response_rate = 1.5, tau_scale = 1))

})

# ------------------------------------------------------------------
# Test: tau_scale must be non-negative
# Input:
#   - response_rate = 0.5, tau_scale = -1
# Behaviour:
#   - Negative scale parameter is invalid and must trigger an error.
# Expectations:
#   - Error for tau_scale < 0.
# Why:
#   - tau_scale represents a scale for a prior; negative values are not meaningful.
# ------------------------------------------------------------------
test_that("error if tau_scale negative", {

  expect_error(getMuVar(response_rate = 0.5, tau_scale = -1))

})

# ------------------------------------------------------------------
# Test: n_worth must be a positive integer
# Input:
#   - n_worth = 0 and 1.5
# Behaviour:
#   - Non-positive or non-integer n_worth should fail validation.
# Expectations:
#   - Errors in both cases.
# Why:
#   - n_worth describes an “effective sample size”; must be >= 1 and integer.
# ------------------------------------------------------------------
test_that("error if n_worth not positive integer", {

  expect_error(getMuVar(response_rate = 0.5, tau_scale = 1, n_worth = 0))

  expect_error(getMuVar(response_rate = 0.5, tau_scale = 1, n_worth = 1.5))

})


# Tests for getPriorParametersBerry --------------------------------------------

# ------------------------------------------------------------------
# Test: getPriorParametersBerry returns correctly structured object
# Input:
#   - target_rates = c(0.2, 0.8), tau_scale = 1, n_worth = 2
# Behaviour:
#   - Function builds a prior_parameters_list for "berry" with fields
#     mu_mean, mu_sd, tau_scale.
# Expectations:
#   - Class "prior_parameters_list", name "berry", inner list with
#     mu_mean, mu_sd, tau_scale.
# Why:
#   - Verifies the returned container type and its internal naming.
# ------------------------------------------------------------------
test_that("valid input returns prior_parameters_list with correct structure", {
  result <- getPriorParametersBerry(target_rates = c(0.2, 0.8), tau_scale = 1, n_worth = 2)

  expect_s3_class(result, "prior_parameters_list")
  expect_named(result, "berry")

  # Check inner structure
  expect_true(is.list(result$berry))
  expect_named(result$berry, c("mu_mean", "mu_sd", "tau_scale"))
})

# ------------------------------------------------------------------
# Test: mu_sd is computed from mu_var as defined by max-variance target rate
# Input:
#   - target_rates = c(0.2, 0.8), tau_scale = 1, n_worth = 2
# Behaviour:
#   - Pick target_rate furthest from 0.5, compute
#     mu_var = (n_worth * p * (1 - p))^-1 - tau_scale^2, then mu_sd = sqrt(mu_var).
# Expectations:
#   - result$berry$mu_sd equals the manually computed value.
# Why:
#   - Confirms the link between target_rates, mu_var and mu_sd.
# ------------------------------------------------------------------
test_that("mu_sd is computed correctly for simple case", {
  target_rates <- c(0.2, 0.8)
  tau_scale <- 1
  n_worth <- 2

  target_rate_max_var <- target_rates[abs(target_rates - 0.5) == max(abs(target_rates - 0.5))][1]
  expected_mu_var <- (n_worth * target_rate_max_var * (1 - target_rate_max_var))^-1 - tau_scale^2
  expected_mu_sd <- sqrt(expected_mu_var)

  result <- getPriorParametersBerry(target_rates, tau_scale, n_worth)
  expect_equal(result$berry$mu_sd, expected_mu_sd)
})

# ------------------------------------------------------------------
# Test: getPriorParametersBerry errors if mu_var <= 0
# Input:
#   - target_rates = 0.5, tau_scale = 100, n_worth = 1
# Behaviour:
#   - Very large tau_scale leads to mu_var <= 0, which is invalid.
# Expectations:
#   - Error is thrown.
# Why:
#   - A non-positive mu_var cannot produce a real standard deviation.
# ------------------------------------------------------------------
test_that("error if mu_var <= 0", {
  # Large tau_scale will make mu_var negative
  expect_error(
    getPriorParametersBerry(target_rates = c(0.5), tau_scale = 100, n_worth = 1)
  )
})

# ------------------------------------------------------------------
# Test: single target_rate works and returns a berry prior
# Input:
#   - target_rates = 0.5, tau_scale = 1, n_worth = 1
# Behaviour:
#   - Function should still construct a valid prior_parameters_list.
# Expectations:
#   - Class "prior_parameters_list", name "berry".
# Why:
#   - Ensures single-cohort use is supported.
# ------------------------------------------------------------------
test_that("single target_rate works", {
  result <- getPriorParametersBerry(target_rates = c(0.5), tau_scale = 1, n_worth = 1)
  expect_s3_class(result, "prior_parameters_list")
  expect_named(result, "berry")
})

# Tests for setPriorParametersBerry --------------------------------------------

# ------------------------------------------------------------------
# Test: setPriorParametersBerry creates correct structure
# Input:
#   - mu_mean = 0.1, mu_sd = 0.5, tau_scale = 1
# Behaviour:
#   - Function wraps values into a "berry" prior_parameters_list.
# Expectations:
#   - Class "prior_parameters_list", name "berry",
#     berry list contains mu_mean, mu_sd, tau_scale.
# Why:
#   - Tests manual prior specification path for Berry.
# ------------------------------------------------------------------
test_that("valid input returns prior_parameters_list with correct structure", {
  result <- setPriorParametersBerry(
    mu_mean = 0.1,
    mu_sd = 0.5,
    tau_scale = 1
  )

  expect_s3_class(result, "prior_parameters_list")
  expect_named(result, "berry")
  expect_true(is.list(result$berry))
  expect_named(result$berry, c("mu_mean", "mu_sd", "tau_scale"))
})

# ------------------------------------------------------------------
# Test: mu_sd must be strictly positive
# Input:
#   - mu_sd = 0
# Behaviour:
#   - Function should reject non-positive standard deviations.
# Expectations:
#   - Error mentioning "mu_sd".
# Why:
#   - Standard deviations must be > 0.
# ------------------------------------------------------------------
test_that("error if mu_sd is non-positive", {
  expect_error(
    setPriorParametersBerry(mu_mean = 0.1, mu_sd = 0, tau_scale = 1),
    "mu_sd"
  )
})

# ------------------------------------------------------------------
# Test: tau_scale must be strictly positive
# Input:
#   - tau_scale = 0
# Behaviour:
#   - Function should reject non-positive scale values.
# Expectations:
#   - Error mentioning "tau_scale".
# Why:
#   - Scale parameter for tau must be > 0.
# ------------------------------------------------------------------
test_that("error if tau_scale is non-positive", {
  expect_error(
    setPriorParametersBerry(mu_mean = 0.1, mu_sd = 0.5, tau_scale = 0),
    "tau_scale"
  )
})

# ------------------------------------------------------------------
# Test: mu_mean must be numeric
# Input:
#   - mu_mean = "a" (character)
# Behaviour:
#   - Non-numeric mu_mean should be rejected.
# Expectations:
#   - Error mentioning "mu_mean".
# Why:
#   - Prior mean is a numeric parameter on the link scale.
# ------------------------------------------------------------------
test_that("error if mu_mean is not numeric", {
  expect_error(
    setPriorParametersBerry(mu_mean = "a", mu_sd = 0.5, tau_scale = 1),
    "mu_mean"
  )
})

# Tests for getPriorParametersExNeX --------------------------------------------

# ------------------------------------------------------------------
# Test: getPriorParametersExNex returns correctly structured object
# Input:
#   - target_rates = c(0.3, 0.9), tau_scale = 1, n_worth = 2, w_j = 0.5
# Behaviour:
#   - Builds an "exnex" prior with fields mu_mean, mu_sd, tau_scale,
#     mu_j, tau_j, w_j.
# Expectations:
#   - Class "prior_parameters_list", name "exnex",
#     exnex list has all expected fields.
# Why:
#   - Checks the container structure for the ExNex prior definition.
# ------------------------------------------------------------------
test_that("valid input returns prior_parameters_list with correct structure", {
  result <- getPriorParametersExNex(target_rates = c(0.3, 0.9), tau_scale = 1, n_worth = 2, w_j = 0.5)

  expect_s3_class(result, "prior_parameters_list")
  expect_named(result, "exnex")

  # Check inner structure
  expect_true(is.list(result$exnex))
  expect_named(result$exnex, c("mu_mean", "mu_sd", "tau_scale", "w_j", "mu_j", "tau_j"))
})

# ------------------------------------------------------------------
# Test: ExNex prior uses logit transform for mu_mean and mu_j
# Input:
#   - target_rates = 0.8, tau_scale = 1, n_worth = 2, w_j = 0.5
# Behaviour:
#   - mu_mean = logit(p_max_var), mu_j = logit(target_rates).
# Expectations:
#   - result$exnex$mu_mean and mu_j match manual logit calculations.
# Why:
#   - Confirms link-scale parameterisation is correctly applied.
# ------------------------------------------------------------------
test_that("mu_mean and mu_j use logit transform", {
  target_rates <- 0.8
  tau_scale <- 1
  n_worth <- 2
  w_j <- 0.5

  target_rate_max_var <- target_rates[abs(target_rates - 0.5) == max(abs(target_rates - 0.5))][1]
  expected_mu_mean <- logit(target_rate_max_var)
  expected_mu_j <- logit(target_rates)
  result <- getPriorParametersExNex(target_rates, tau_scale, n_worth, w_j)

  expect_equal(result$exnex$mu_mean, expected_mu_mean)
  expect_equal(result$exnex$mu_j, expected_mu_j)
})

# ------------------------------------------------------------------
# Test: getPriorParametersExNex errors if mu_var <= 0
# Input:
#   - target_rates = 0.5, tau_scale = 100, n_worth = 1
# Behaviour:
#   - Large tau_scale makes mu_var non-positive.
# Expectations:
#   - Error thrown.
# Why:
#   - Same reason as Berry: must have positive variance for mu.
# ------------------------------------------------------------------
test_that("error if mu_var <= 0", {
  # Large tau_scale will make mu_var negative
  expect_error(
    getPriorParametersExNex(target_rates = c(0.5), tau_scale = 100, n_worth = 1)
  )
})

# ------------------------------------------------------------------
# Test: mu_sd and tau_j computations follow theoretical formulas
# Input:
#   - target_rates = 0.8, tau_scale = 1, n_worth = 2
# Behaviour:
#   - mu_sd from max-variance target, tau_j from per-cohort variance formula.
# Expectations:
#   - result$exnex$mu_sd and tau_j match manual calculations.
# Why:
#   - Ensures ExNex prior scale parameters are derived correctly.
# ------------------------------------------------------------------
test_that("mu_sd and tau_j are computed correctly", {
  target_rates <- 0.8
  tau_scale <- 1
  n_worth <- 2

  target_rate_max_var <- target_rates[abs(target_rates - 0.5) == max(abs(target_rates - 0.5))][1]
  expected_mu_var <- (n_worth * target_rate_max_var * (1 - target_rate_max_var))^-1 - tau_scale^2
  expected_mu_sd <- sqrt(expected_mu_var)

  expected_tau_j <- sqrt((n_worth * target_rates * (1 - target_rates))^-1)

  result <- getPriorParametersExNex(target_rates, tau_scale, n_worth, w_j = 0.5)

  expect_equal(result$exnex$mu_sd, expected_mu_sd)
  expect_equal(result$exnex$tau_j, expected_tau_j)
})

# Tests for setPriorParametersExNeX --------------------------------------------

# ------------------------------------------------------------------
# Test: setPriorParametersExNeX creates full exnex structure
# Input:
#   - Vectors for mu_mean, mu_sd, tau_scale, mu_j, tau_j, w_j.
# Behaviour:
#   - Returns a prior_parameters_list named "exnex" with all components.
# Expectations:
#   - Class "prior_parameters_list", name "exnex",
#     exnex element has mu_mean, mu_sd, tau_scale, mu_j, tau_j, w_j.
# Why:
#   - Validates the manual ExNex prior construction path.
# ------------------------------------------------------------------
test_that("valid input returns prior_parameters_list with correct structure", {
  result <- setPriorParametersExNex(
    mu_mean = c(0.1, 0.2),
    mu_sd = c(0.5, 0.6),
    tau_scale = 1,
    mu_j = c(0.1, 0.2),
    tau_j = c(0.3, 0.4),
    w_j = c(0.3, 0.3, 0.4)
  )

  expect_s3_class(result, "prior_parameters_list")
  expect_named(result, "exnex")
  expect_true(is.list(result$exnex))
  expect_named(result$exnex, c("mu_mean", "mu_sd", "tau_scale", "w_j", "mu_j", "tau_j"))
})

# ------------------------------------------------------------------
# Test: w_j length rules when mu_mean length = 1
# Input:
#   - mu_mean length 1, w_j length 3
# Behaviour:
#   - For a single mu_mean, w_j length 3 is not allowed.
# Expectations:
#   - Error.
# Why:
#   - Tests internal consistency checks between mixing weights and mean length.
# ------------------------------------------------------------------
test_that("error if w_j length is invalid when mu_mean length = 1", {
  expect_error(
    setPriorParametersExNex(
      mu_mean = c(0.1),
      mu_sd = c(0.5),
      tau_scale = 1,
      mu_j = c(0.1),
      tau_j = c(0.3),
      w_j = c(0.3, 0.3, 0.4)  # length 3 is invalid for mu_mean length 1
    )
  )
})

# Tests for getPriorParametersExNeXAdj -----------------------------------------

# ------------------------------------------------------------------
# Test: getPriorParametersExNeXAdj returns exnex_adj with centered mu's
# Input:
#   - target_rates = c(0.2, 0.3), tau_scale = 1, n_worth = 2, w_j = 0.5
# Behaviour:
#   - Returns "exnex_adj" prior with mu_mean and mu_j shifted to 0.
# Expectations:
#   - Class "prior_parameters_list", name "exnex_adj",
#     mu_mean = 0, mu_j = 0-vector.
# Why:
#   - Confirms adjusted ExNex prior is properly centered.
# ------------------------------------------------------------------
test_that("valid input returns prior_parameters_list with correct structure", {
  result <- getPriorParametersExNexAdj(
    target_rates = c(0.2, 0.3),
    tau_scale = 1,
    n_worth = 2,
    w_j = 0.5
  )

  expect_s3_class(result, "prior_parameters_list")
  expect_named(result, "exnex_adj")
  expect_true(is.list(result$exnex_adj))
  expect_named(result$exnex_adj, c("mu_mean", "mu_sd", "tau_scale", "w_j", "mu_j", "tau_j"))

  # Check adjusted values
  expect_equal(result$exnex_adj$mu_mean, 0)
  expect_equal(result$exnex_adj$mu_j, rep(0, length(c(0.2, 0.3))))
})

# Tests for setPriorParametersExNeXAdj -----------------------------------------

# ------------------------------------------------------------------
# Test: setPriorParametersExNeXAdj creates correct exnex_adj structure
# Input:
#   - Vectors for mu_mean, mu_sd, tau_scale, mu_j, tau_j, w_j.
# Behaviour:
#   - Returns "exnex_adj" prior_parameters_list with all fields.
# Expectations:
#   - Class "prior_parameters_list", name "exnex_adj",
#     exnex_adj has mu_mean, mu_sd, tau_scale, mu_j, tau_j, w_j.
# Why:
#   - Checks manual specification of adjusted ExNex prior.
# ------------------------------------------------------------------
test_that("valid input returns prior_parameters_list with correct structure", {
  result <- setPriorParametersExNexAdj(
    mu_mean = c(0.1, 0.2),
    mu_sd = c(0.5, 0.6),
    tau_scale = 1,
    mu_j = c(0.1, 0.2),
    tau_j = c(0.3, 0.4),
    w_j = c(0.3, 0.3, 0.4)
  )

  expect_s3_class(result, "prior_parameters_list")
  expect_named(result, "exnex_adj")
  expect_true(is.list(result$exnex_adj))
  expect_named(result$exnex_adj, c("mu_mean", "mu_sd", "tau_scale",  "w_j", "mu_j", "tau_j"))
})

# ------------------------------------------------------------------
# Test: acceptable w_j lengths when mu_mean length = 1
# Input:
#   - mu_mean length 1, w_j length 1 (valid).
# Behaviour:
#   - Function should accept such configuration silently.
# Expectations:
#   - No error raised.
# Why:
#   - Sanity-checks "short" w_j is allowed in simple case.
# ------------------------------------------------------------------
test_that("'w_j' length rule works when mu_mean length = 1", {
  # Valid cases: w_j length 1 or 2
  expect_silent(
    setPriorParametersExNexAdj(
      mu_mean = 0.1,
      mu_sd = 0.5,
      tau_scale = 1,
      mu_j = c(0.1),
      tau_j = c(0.3),
      w_j = c(0.5)  # length 1
    )
  )
})

# Tests for getPriorParametersPooled --------------------------------------------

# ------------------------------------------------------------------
# Test: getPriorParametersPooled returns pooled Beta parameters
# Input:
#   - target_rates = c(0.2, 0.4, 0.6), n_worth = 2
# Behaviour:
#   - Selects a target_rate (internally defined rule) and computes
#     a = p * n_worth, b = (1 - p) * n_worth.
# Expectations:
#   - Class "prior_parameters_list", name "pooled", with positive a,b.
# Why:
#   - Validates the pooled Beta( a, b ) construction.
# ------------------------------------------------------------------
test_that("valid input returns prior_parameters_list with correct structure", {
  result <- getPriorParametersPooled(
    target_rates = c(0.2, 0.4, 0.6),
    n_worth = 2
  )

  expect_s3_class(result, "prior_parameters_list")
  expect_named(result, "pooled")
  expect_true(is.list(result$pooled))
  expect_named(result$pooled, c("a", "b"))
  expect_true(result$pooled$a > 0)
  expect_true(result$pooled$b > 0)
})

# ------------------------------------------------------------------
# Test: pooled prior requires valid target_rates in (0,1)
# Input:
#   - target_rates including 0 or 1.
# Behaviour:
#   - Must error on invalid probabilities.
# Expectations:
#   - Errors mentioning "target_rates".
# Why:
#   - Beta prior requires positive support away from 0 and 1.
# ------------------------------------------------------------------
test_that("error if target_rates contains invalid values", {
  expect_error(getPriorParametersPooled(target_rates = c(0, 0.5), n_worth = 1), "target_rates")
  expect_error(getPriorParametersPooled(target_rates = c(1, 0.5), n_worth = 1), "target_rates")
})

# ------------------------------------------------------------------
# Test: n_worth must be positive for pooled prior
# Input:
#   - n_worth = 0
# Behaviour:
#   - Function should reject non-positive n_worth.
# Expectations:
#   - Error mentioning "n_worth".
# Why:
#   - n_worth again represents effective sample size.
# ------------------------------------------------------------------
test_that("error if n_worth is non-positive", {
  expect_error(getPriorParametersPooled(target_rates = c(0.3, 0.4), n_worth = 0), "n_worth")
})

# ------------------------------------------------------------------
# Test: pooled prior selects target_rate closest to 0.5
# Input:
#   - target_rates = c(0.1, 0.4, 0.8), n_worth = 3
# Behaviour:
#   - Among candidates, p = 0.4 is closest to 0.5, so a = 0.4*3, b = 0.6*3.
# Expectations:
#   - result$pooled$a == 0.4*3, result$pooled$b == 0.6*3.
# Why:
#   - Checks selection logic for the central rate.
# ------------------------------------------------------------------
test_that("selects target_rate closest to 0.5", {
  result <- getPriorParametersPooled(target_rates = c(0.1, 0.4, 0.8), n_worth = 3)
  expect_equal(result$pooled$a, 0.4 * 3)
  expect_equal(result$pooled$b, (1 - 0.4) * 3)
})

# Tests for setPriorParametersPooled -------------------------------------------

# ------------------------------------------------------------------
# Test: setPriorParametersPooled builds pooled Beta structure
# Input:
#   - a = 2, b = 3
# Behaviour:
#   - Wraps into "pooled" prior_parameters_list.
# Expectations:
#   - Class "prior_parameters_list", name "pooled", a and b preserved.
# Why:
#   - Tests manual specification path for pooled prior.
# ------------------------------------------------------------------
test_that("valid input returns prior_parameters_list with correct structure", {
  result <- setPriorParametersPooled(a = 2, b = 3)

  expect_s3_class(result, "prior_parameters_list")
  expect_named(result, "pooled")
  expect_true(is.list(result$pooled))
  expect_named(result$pooled, c("a", "b"))
  expect_equal(result$pooled$a, 2)
  expect_equal(result$pooled$b, 3)
})

# Tests for getPriorParametersStratified ---------------------------------------

# ------------------------------------------------------------------
# Test: stratified prior computes a_j and b_j per target rate
# Input:
#   - target_rates = c(0.2, 0.4, 0.6), n_worth = 2
# Behaviour:
#   - a_j = p * n_worth, b_j = (1 - p) * n_worth elementwise.
# Expectations:
#   - result$stratified$a_j and b_j match manual multiplication.
# Why:
#   - Verifies Beta-prior parameters for each stratum.
# ------------------------------------------------------------------
test_that("computes a_j and b_j correctly for multiple target rates", {
  target_rates <- c(0.2, 0.4, 0.6)
  n_worth <- 2
  result <- getPriorParametersStratified(target_rates, n_worth)

  expect_equal(result$stratified$a_j, target_rates * n_worth)
  expect_equal(result$stratified$b_j, (1 - target_rates) * n_worth)
})

# ------------------------------------------------------------------
# Test: stratified prior handles a single target rate
# Input:
#   - target_rates = 0.3, n_worth = 5
# Behaviour:
#   - Computes scalar a_j and b_j.
# Expectations:
#   - a_j = 0.3*5, b_j = 0.7*5.
# Why:
#   - Ensures single-stratum usage is supported.
# ------------------------------------------------------------------
test_that("handles single target rate correctly", {
  result <- getPriorParametersStratified(target_rates = 0.3, n_worth = 5)
  expect_equal(result$stratified$a_j, 0.3 * 5)
  expect_equal(result$stratified$b_j, (1 - 0.3) * 5)
})

# ------------------------------------------------------------------
# Test: stratified result has stable structure
# Input:
#   - target_rates = c(0.25, 0.75), n_worth = 3
# Behaviour:
#   - Returns "stratified" prior with a_j and b_j fields.
# Expectations:
#   - Class "prior_parameters_list", name "stratified", fields a_j,b_j.
# Why:
#   - Confirms naming and type are consistent.
# ------------------------------------------------------------------
test_that("result structure is consistent", {
  result <- getPriorParametersStratified(c(0.25, 0.75), 3)
  expect_s3_class(result, "prior_parameters_list")
  expect_named(result, "stratified")
  expect_named(result$stratified, c("a_j", "b_j"))
})

# Tests for setPriorParametersStratified ---------------------------------------

# ------------------------------------------------------------------
# Test: setPriorParametersStratified preserves a_j,b_j and structure
# Input:
#   - a_j = c(1,2,3), b_j = c(4,5,6)
# Behaviour:
#   - Returns "stratified" prior with these vectors.
# Expectations:
#   - a_j,b_j unchanged; class prior_parameters_list.
# Why:
#   - Checks manual specification path for stratified Beta priors.
# ------------------------------------------------------------------
test_that("returns correct structure and preserves input values", {
  a_j <- c(1, 2, 3)
  b_j <- c(4, 5, 6)
  result <- setPriorParametersStratified(a_j, b_j)

  expect_equal(result$stratified$a_j, a_j)
  expect_equal(result$stratified$b_j, b_j)
  expect_s3_class(result, "prior_parameters_list")
})

# ------------------------------------------------------------------
# Test: setPriorParametersStratified works with single-element vectors
# Input:
#   - a_j = 10, b_j = 20
# Behaviour:
#   - Returns scalar a_j,b_j under "stratified".
# Expectations:
#   - Values preserved.
# Why:
#   - Confirms scalar case is supported smoothly.
# ------------------------------------------------------------------
test_that("works with single-element vectors", {
  result <- setPriorParametersStratified(a_j = 10, b_j = 20)
  expect_equal(result$stratified$a_j, 10)
  expect_equal(result$stratified$b_j, 20)
})

# ------------------------------------------------------------------
# Test: length consistency for a_j and b_j
# Input:
#   - a_j length 2, b_j length 1
# Behaviour:
#   - Function should error on mismatched lengths.
# Expectations:
#   - Error mentioning "a_j and b_j".
# Why:
#   - Each stratum must have both a and b.
# ------------------------------------------------------------------
test_that("length consistency check works", {
  expect_error(setPriorParametersStratified(a_j = c(1, 2), b_j = c(3)), "a_j and b_j")
})

# Tests for getPriorParameters -------------------------------------------------

# ------------------------------------------------------------------
# Test: getPriorParameters returns combined prior_parameters_list
# Input:
#   - method_names = c("berry", "pooled", "stratified", "exnex", "exnex_adj"),
#     plus reasonable target_rates, n_worth, tau_scale, w_j.
# Behaviour:
#   - Constructs all requested method priors and merges them.
# Expectations:
#   - Class "prior_parameters_list", 5 elements, names sorted as in implementation.
# Why:
#   - Validates the high-level prior factory for multiple methods.
# ------------------------------------------------------------------
test_that("valid input returns a prior_parameters_list with correct names and class", {
  result <- getPriorParameters(
    method_names = c("berry", "pooled", "stratified", "exnex", "exnex_adj"),
    target_rates = c(0.2, 0.1, 0.3, 0.2, 0.4, 0.6),
    n_worth = 2,
    tau_scale = 1,
    w_j = 0.5
  )
  expect_s3_class(result, "prior_parameters_list")
  expect_named(result, c("berry", "exnex", "exnex_adj", "pooled", "stratified"))
  expect_type(result, "list")
  expect_length(result, 5)

})

# ------------------------------------------------------------------
# Test: getPriorParameters errors on invalid method_names
# Input:
#   - method_names = "invalid"
# Behaviour:
#   - Function must reject unknown methods.
# Expectations:
#   - Error.
# Why:
#   - Guards against typos or unsupported priors.
# ------------------------------------------------------------------
test_that("invalid method_names throws error", {
  expect_error(getPriorParameters(
    method_names = "invalid",
    target_rates = c(0.2, 0.3)
  ))
})

# ------------------------------------------------------------------
# Test: getPriorParameters errors on invalid target_rates
# Input:
#   - target_rates outside [0,1]
# Behaviour:
#   - Should reject probabilities <=0 or >=1.
# Expectations:
#   - Error.
# Why:
#   - All underlying constructions assume valid rates.
# ------------------------------------------------------------------
test_that("invalid target_rates throws error", {
  expect_error(getPriorParameters(
    method_names = "berry",
    target_rates = c(-0.1, 1.2)
  ))
})

# ------------------------------------------------------------------
# Test: getPriorParameters errors on invalid tau_scale
# Input:
#   - tau_scale = -1
# Behaviour:
#   - Negative tau_scale should fail.
# Expectations:
#   - Error.
# Why:
#   - Same constraint as lower-level prior builders.
# ------------------------------------------------------------------
test_that("invalid tau_scale throws error", {
  expect_error(getPriorParameters(
    method_names = "berry",
    target_rates = c(0.2, 0.3),
    tau_scale = -1
  ))
})

# ------------------------------------------------------------------
# Test: getPriorParameters errors on invalid n_worth
# Input:
#   - n_worth = 0
# Behaviour:
#   - Must reject non-positive effective sample size.
# Expectations:
#   - Error.
# Why:
#   - Consistency with other n_worth checks.
# ------------------------------------------------------------------
test_that("invalid n_worth throws error", {
  expect_error(getPriorParameters(
    method_names = "berry",
    target_rates = c(0.2, 0.3),
    n_worth = 0
  ))
})

# ------------------------------------------------------------------
# Test: getPriorParameters errors on invalid w_j
# Input:
#   - w_j = 2 (outside [0,1] or invalid)
# Behaviour:
#   - Weight must be in acceptable range.
# Expectations:
#   - Error.
# Why:
#   - Mixture weights need to be interpretable probabilities.
# ------------------------------------------------------------------
test_that("invalid w_j throws error", {
  expect_error(getPriorParameters(
    method_names = "berry",
    target_rates = c(0.2, 0.3),
    w_j = 2
  ))
})

# Tests for combinePriorParameters ---------------------------------------------

# ------------------------------------------------------------------
# Test: combinePriorParameters merges multiple prior_parameter objects
# Input:
#   - prior_parameters_stratified + prior_parameters_berry
# Behaviour:
#   - Flattens and merges, returning a single prior_parameters_list.
# Expectations:
#   - Class "prior_parameters_list", length 2, names are union of input names,
#     each element is a list.
# Why:
#   - Confirms high-level combination for multi-method priors.
# ------------------------------------------------------------------
test_that("combinePriorParameters returns correct structure with real objects", {
  prior_parameters_stratified <- setPriorParametersStratified(c(1, 2), c(3, 4))
  prior_parameters_berry      <- setPriorParametersBerry(1, 1, 2)

  result <- combinePriorParameters(list(prior_parameters_berry, prior_parameters_stratified))

  # Check class and type
  expect_s3_class(result, "prior_parameters_list")
  expect_type(result, "list")

  # Names should match method names from input
  expect_named(result, sort(c(names(prior_parameters_berry), names(prior_parameters_stratified))))

  # Length should equal number of input elements
  expect_length(result, 2)

  # Each element should be a list (inner structure)
  expect_true(all(vapply(result, is.list, logical(1))))
})

# ------------------------------------------------------------------
# Test: combinePriorParameters sorts method names alphabetically
# Input:
#   - [stratified, berry] in that order
# Behaviour:
#   - Resulting list names should be sorted.
# Expectations:
#   - names(result) == sort of input method names.
# Why:
#   - Ensures deterministic ordering in combined prior.
# ------------------------------------------------------------------
test_that("combinePriorParameters sorts names alphabetically", {
  prior_parameters_stratified <- setPriorParametersStratified(c(1, 2), c(3, 4))
  prior_parameters_berry      <- setPriorParametersBerry(0, 1, 2)

  result <- combinePriorParameters(list(prior_parameters_stratified, prior_parameters_berry))

  expect_equal(names(result), sort(c(names(prior_parameters_stratified), names(prior_parameters_berry))))
})

# ------------------------------------------------------------------
# Test: combinePriorParameters errors on duplicate method names
# Input:
#   - Two separate prior_parameters_berry objects.
# Behaviour:
#   - Duplicate "berry" name should be rejected.
# Expectations:
#   - Error.
# Why:
#   - Each method may appear only once in the combined object.
# ------------------------------------------------------------------
test_that("combinePriorParameters errors on duplicate method names", {
  prior_parameters_berry1 <- setPriorParametersBerry(0, 1, 2)
  prior_parameters_berry2 <- setPriorParametersBerry(0, 1, 2)

  expect_error(
    combinePriorParameters(list(prior_parameters_berry1, prior_parameters_berry2))
  )
})

# ------------------------------------------------------------------
# Test: combinePriorParameters errors when input is not a list
# Input:
#   - "not_a_list"
# Behaviour:
#   - Type check should fail for non-list input.
# Expectations:
#   - Error.
# Why:
#   - Enforces input contract: a list of prior_parameters_list objects.
# ------------------------------------------------------------------
test_that("combinePriorParameters errors if input is not a list", {
  expect_error(combinePriorParameters("not_a_list"))
})

# ------------------------------------------------------------------
# Test: combinePriorParameters errors if any element has wrong class
# Input:
#   - One prior_parameters_berry, one plain list(dummy = TRUE)
# Behaviour:
#   - Mixed-class list must be rejected.
# Expectations:
#   - Error.
# Why:
#   - Every element must be a proper prior_parameters_list.
# ------------------------------------------------------------------
test_that("combinePriorParameters errors if any element is not prior_parameters_list", {
  prior_parameters_berry <- setPriorParametersBerry(0, 1, 2)
  bad_input <- list(prior_parameters_berry, list(dummy = TRUE))

  expect_error(combinePriorParameters(bad_input))
})

applicablePreviousTrials <- function(

  scenario_list,
  method_names,
  quantiles,
  n_cohorts,
  calc_differences

) {

  ## analyze only unique trials that have not been previously analyzed,
  ## i.e. trials that were updated with continueRecruitment(),
  ## i.e. trials that have an overall go decision from a previous decision rule
  ## i.e. trials that have quantiles stored for each cohort that is to be analysed
  applicable_previous_trials <-
    ## check that in each scenario the same analysis methods were analyzed previously
    all(sapply(seq_along(scenario_list), function (i) {
      isTRUE(all.equal(names(scenario_list[[i]]$previous_analyses$post_quantiles),
                       names(scenario_list[[1]]$previous_analyses$post_quantiles)))
    })) &
    ## check that the current analysis method names match the method names of the previous analyses
    isTRUE(all.equal(names(scenario_list[[1]]$previous_analyses$post_quantiles), method_names)) &
    ## check that the stored quantiles are the same across all scenarios
    all(sapply(seq_along(scenario_list), function (i) {
      isTRUE(all.equal(rownames(scenario_list[[i]]$previous_analyses$post_quantiles[[1]][[1]]),
                       rownames(scenario_list[[1]]$previous_analyses$post_quantiles[[1]][[1]])))
    })) &
    ## check that the new quantiles are within the stored quantiles
    all(paste0(as.character(quantiles * 100), "%") %in%
          rownames(scenario_list[[1]]$previous_analyses$post_quantiles[[1]][[1]])) &
    ## check that there are stored quantiles for each cohort that is to be analysed
    all(paste0("p_", seq_len(n_cohorts)) %in%
          colnames(scenario_list[[1]]$previous_analyses$post_quantiles[[1]][[1]]))

  ## check that all differences have been previously calculated
  if (!is.null(calc_differences)) {

    applicable_previous_trials <- applicable_previous_trials &
      all(apply(calc_differences, 1, function (x) {
        paste0("p_diff_", paste0(as.character(x), collapse = ""))
      }) %in% colnames(scenario_list[[1]]$previous_analyses$post_quantiles[[1]][[1]]))

  }

  return (applicable_previous_trials)

}

calcDiffsMCMC <- function (

  posterior_samples,
  calc_differences

) {

  org_names <- colnames(posterior_samples)

  diffs <- apply(calc_differences, 1, function (x) {

    matrix(posterior_samples[, grepl(x[1], org_names) & grepl("p", org_names)] -
             posterior_samples[, grepl(x[2], org_names) & grepl("p", org_names)],
           ncol = 1)

  })

  diff_names <- apply(calc_differences, 1, function (x) {

    paste0("p_diff_", paste0(as.character(x), collapse = ""))

  })

  colnames(diffs) <- diff_names

  posterior_samples <- cbind(posterior_samples, diffs)

  return (posterior_samples)

}

getModelFile <- function(method_name) {

  model_map <- c(
    berry         = "berry.txt",
    exnex         = "exnex.txt",
    exnex_mix     = "exnex_mix.txt",
    exnex_adj     = "exnex_adj.txt",
    exnex_adj_mix = "exnex_adj_mix.txt"
  )

  if (!method_name %in% names(model_map)) {
    stop("method_name must be one of berry, exnex, exnex_mix, exnex_adj, exnex_adj_mix")
  }

  system.file(package = "bhmbasket", "jags_models", model_map[[method_name]], mustWork = TRUE)
}

getPosteriors <- function (

  j_parameters,
  j_model_file,
  j_data,

  n_mcmc_iterations

) {

  ## Adaption and burn-in not included
  posterior_samples <- performJags(
    data               = j_data,
    parameters_to_save = j_parameters,
    model_file         = j_model_file,
    n_iter             = n_mcmc_iterations)

  ## replace squarebrackets provided by rjags with workable characters
  colnames(posterior_samples) <- gsub("\\[", "_", colnames(posterior_samples))
  colnames(posterior_samples) <- gsub("\\]", "", colnames(posterior_samples))

  weights_indices <- grepl("exch", colnames(posterior_samples))
  if (any(weights_indices)) {

    superfluous_weights <- !grepl(",1", colnames(posterior_samples))

    colnames(posterior_samples)[weights_indices] <-
      paste0("w_", seq_along(j_data$n))

    posterior_samples <- posterior_samples[, !(weights_indices & superfluous_weights)]

  }

  return (posterior_samples)

}

getPostQuantiles <- function (

  ## The method to be applied to the likelihood and the quantiles of the posterior
  method_name,
  quantiles,

  ## Scenario data
  scenario_data,

  ## Differences between cohorts
  calc_differences  = NULL,

  ## JAGS parameters
  j_parameters,
  j_model_file,
  j_data,

  ## MCMC Parameters
  n_mcmc_iterations = 1e4,

  ## Where to save one of the posterior response rates approximations provided by JAGS
  save_path         = NULL,
  save_trial        = NULL

) {

  if (is.null(dim(scenario_data$n_responders))) {

    scenario_data$n_responders <- t(convertVector2Matrix(scenario_data$n_responders))
    scenario_data$n_subjects   <- t(convertVector2Matrix(scenario_data$n_subjects))

  }

  n_analyses <- nrow(scenario_data$n_responders)

  ## Create random index for saving one of the posterior response rates
  if (is.null(save_trial) && !is.null(save_path)) {
    # set.seed(seed)
    save_trial <- sample(seq_len(n_analyses), size = 1)
  }

  ## Run parallel loops
  ## prepare foreach loop over

  exported_stuff <- c(
    "posteriors2Quantiles", "getPosteriors", "getPostQuantilesOfTrial",
    "qbetaDiff", "chunkVector")

  ## prepare chunking
  chunks_outer <- chunkVector(seq_len(n_analyses), foreach::getDoParWorkers())

  "%dorng%" <- doRNG::"%dorng%"
  "%dopar%" <- foreach::"%dopar%"
  posterior_quantiles_list <- suppressMessages(
    foreach::foreach(
      k = chunks_outer,
      .combine  = c,
      .verbose  = FALSE,
      .packages = c("rjags"),
      .export   = exported_stuff) %dorng% {

        chunks_inner <- chunkVector(k, foreach::getDoParWorkers())

        foreach::foreach(i = chunks_inner, .combine = c) %dorng% {

          lapply(i, function (j) {

            ## Calculate the posterior quantiles for the kth unique trial outcome
            getPostQuantilesOfTrial(
              n_responders      = as.numeric(scenario_data$n_responders[j, ]),
              n_subjects        = as.numeric(scenario_data$n_subjects[j, ]),
              trial_index       = j,
              j_data            = j_data,
              j_parameters      = j_parameters,
              j_model_file      = j_model_file,
              method_name       = method_name,
              quantiles         = quantiles,
              calc_differences  = calc_differences,
              n_mcmc_iterations = n_mcmc_iterations,
              save_path         = save_path,
              save_trial        = save_trial)

          })

        }

      })

  return (posterior_quantiles_list)

}

getPostQuantilesOfTrial <- function (

  n_responders,
  n_subjects,
  trial_index,

  j_data,
  j_parameters,
  j_model_file,
  method_name,
  quantiles,
  calc_differences,
  n_mcmc_iterations,

  save_path,
  save_trial

) {

  j_data$r <- n_responders
  j_data$n <- n_subjects

  if (method_name %in% c("stratified", "stratified_mix")) {

    posterior_quantiles <- getPostQuantilesStratified(
      j_data            = j_data,
      quantiles         = quantiles,
      calc_differences  = calc_differences,
      n_mcmc_iterations = n_mcmc_iterations
    )

  } else if (method_name == "pooled") {

    posterior_quantiles <- getPostQuantilesPooled(
      j_data           = j_data,
      quantiles        = quantiles,
      calc_differences = calc_differences
    )

  } else {

    ## Get posterior response rates per indication
    posterior_samples <- getPosteriors(
      j_parameters      = j_parameters,
      j_model_file      = j_model_file,
      j_data            = j_data,
      n_mcmc_iterations = n_mcmc_iterations)

    ## Calculate differences between response rates of cohorts
    if (!is.null(calc_differences)) {

      posterior_samples <- calcDiffsMCMC(
        posterior_samples = posterior_samples,
        calc_differences  = calc_differences)

    }

    ## Save posterior response rates per indication for one randomly selected simulation,
    ## due to time and storage space constraints only one simulation
    if (!is.null(save_path)) {
      if (trial_index == save_trial) {
        saveRDS(posterior_samples,
                file = file.path(save_path, paste0("posterior_samples_",
                                                   trial_index, "_", method_name, ".rds")))
      }
    }

    ## Calculate the required quantiles for the decision rules
    posterior_quantiles <- posteriors2Quantiles(
      quantiles  = quantiles,
      posteriors = posterior_samples)

  }

  return (posterior_quantiles)

}

getPostQuantilesPooled <- function(

  j_data,
  quantiles,
  calc_differences

) {

  shape_1 <- j_data$a + sum(j_data$r)
  shape_2 <- j_data$b + sum(j_data$n) - sum(j_data$r)

  posterior_quantiles <- stats::qbeta(quantiles, shape1 = shape_1, shape2 = shape_2)

  posterior_quantiles <- matrix(posterior_quantiles,
                                ncol = length(j_data$r), nrow = length(quantiles))

  colnames(posterior_quantiles) <- paste0("p_", seq_along(j_data$r))
  rownames(posterior_quantiles) <- paste0(quantiles * 100, "%")

  posterior_mean      <- shape_1 / (shape_1 + shape_2)
  posterior_sd        <- ((shape_1 * shape_2) / ((shape_1 + shape_2)^2 * (shape_1 + shape_2 + 1)))^0.5
  posterior_quantiles <- rbind(posterior_quantiles,
                               Mean = posterior_mean,
                               SD   = posterior_sd)

  if (!is.null(calc_differences)) {

    diff_quantiles <- apply(calc_differences, 1, function (x) {

      matrix(rep(0, nrow(posterior_quantiles)), ncol = 1)

    })

    diff_names <- apply(calc_differences, 1, function (x) {

      paste0("p_diff_", paste0(as.character(x), collapse = ""))

    })

    colnames(diff_quantiles) <- diff_names

    posterior_quantiles <- cbind(posterior_quantiles, diff_quantiles)

  }

  return (posterior_quantiles)

}

getPostQuantilesStratified <- function(

  j_data,
  quantiles,
  calc_differences,
  n_mcmc_iterations

) {

  is_mix <- is.matrix(j_data$a_j) && nrow(j_data$a_j) > 1L

  if (!is_mix) {

    shape_1 <- j_data$a_j + j_data$r
    shape_2 <- j_data$b_j + j_data$n - j_data$r

    posterior_quantiles <- t(sapply(quantiles, function(x)
      stats::qbeta(x, shape1 = shape_1, shape2 = shape_2)))

    if (nrow(posterior_quantiles) == 1) {
      posterior_quantiles <- t(posterior_quantiles)
    }

    colnames(posterior_quantiles) <- paste0("p_", seq_along(j_data$a_j))
    rownames(posterior_quantiles) <- paste0(quantiles * 100, "%")

    posterior_mean <- shape_1 / (shape_1 + shape_2)
    posterior_sd <- ((shape_1 * shape_2) /
                       ((shape_1 + shape_2)^2 * (shape_1 + shape_2 + 1)))^0.5

    posterior_quantiles <- rbind(
      posterior_quantiles,
      Mean = posterior_mean,
      SD   = posterior_sd
    )

    if (!is.null(calc_differences)) {

      diff_quantiles <- apply(calc_differences, 1, function(x) {
        matrix(
          qbetaDiff(
            quantiles  = quantiles,
            x_1_shape1 = shape_1[x[1]],
            x_1_shape2 = shape_2[x[1]],
            x_2_shape1 = shape_1[x[2]],
            x_2_shape2 = shape_2[x[2]],
            n_mcmc     = n_mcmc_iterations
          ),
          ncol = 1
        )
      })

      diff_names <- apply(calc_differences, 1, function(x) {
        paste0("p_diff_", paste0(as.character(x), collapse = ""))
      })

      colnames(diff_quantiles) <- diff_names
      posterior_quantiles <- cbind(posterior_quantiles, diff_quantiles)
    }

    return(posterior_quantiles)
  }

  shape_1 <- j_data$a_j + matrix(
    j_data$r,
    nrow = nrow(j_data$a_j),
    ncol = ncol(j_data$a_j),
    byrow = TRUE
  )

  shape_2 <- j_data$b_j + matrix(
    j_data$n - j_data$r,
    nrow = nrow(j_data$b_j),
    ncol = ncol(j_data$b_j),
    byrow = TRUE
  )

  posterior_quantiles <- matrix(
    NA_real_,
    nrow = length(quantiles),
    ncol = ncol(j_data$a_j)
  )

  w_post <- matrix(
    NA_real_,
    nrow = length(j_data$w),
    ncol = ncol(j_data$a_j)
  )

  for (j in seq_len(ncol(j_data$a_j))) {

    log_fk <- lbeta(shape_1[, j], shape_2[, j]) -
      lbeta(j_data$a_j[, j], j_data$b_j[, j])

    log_w <- log(j_data$w) + log_fk
    w_post[, j] <- exp(log_w - max(log_w))
    w_post[, j] <- w_post[, j] / sum(w_post[, j])

    k <- sample.int(
      length(j_data$w),
      size = n_mcmc_iterations,
      replace = TRUE,
      prob = w_post[, j]
    )

    p_draw <- stats::rbeta(
      n_mcmc_iterations,
      shape1 = shape_1[k, j],
      shape2 = shape_2[k, j]
    )

    posterior_quantiles[, j] <- stats::quantile(p_draw, probs = quantiles)
  }

  if (nrow(posterior_quantiles) == 1) {
    posterior_quantiles <- t(posterior_quantiles)
  }

  colnames(posterior_quantiles) <- paste0("p_", seq_len(ncol(j_data$a_j)))
  rownames(posterior_quantiles) <- paste0(quantiles * 100, "%")

  posterior_mean <- numeric(ncol(j_data$a_j))
  posterior_sd <- numeric(ncol(j_data$a_j))

  for (j in seq_len(ncol(j_data$a_j))) {

    mean_k <- shape_1[, j] / (shape_1[, j] + shape_2[, j])
    var_k <- ((shape_1[, j] * shape_2[, j]) /
                ((shape_1[, j] + shape_2[, j])^2 * (shape_1[, j] + shape_2[, j] + 1)))

    posterior_mean[j] <- sum(w_post[, j] * mean_k)
    posterior_sd[j] <- sqrt(sum(w_post[, j] * (var_k + (mean_k - posterior_mean[j])^2)))
  }

  posterior_quantiles <- rbind(
    posterior_quantiles,
    Mean = posterior_mean,
    SD   = posterior_sd
  )

  if (!is.null(calc_differences)) {

    diff_quantiles <- apply(calc_differences, 1, function(x) {

      log_fk_1 <- lbeta(shape_1[, x[1]], shape_2[, x[1]]) -
        lbeta(j_data$a_j[, x[1]], j_data$b_j[, x[1]])
      log_w_1 <- log(j_data$w) + log_fk_1
      w_post_1 <- exp(log_w_1 - max(log_w_1))
      w_post_1 <- w_post_1 / sum(w_post_1)

      log_fk_2 <- lbeta(shape_1[, x[2]], shape_2[, x[2]]) -
        lbeta(j_data$a_j[, x[2]], j_data$b_j[, x[2]])
      log_w_2 <- log(j_data$w) + log_fk_2
      w_post_2 <- exp(log_w_2 - max(log_w_2))
      w_post_2 <- w_post_2 / sum(w_post_2)

      comp_1 <- sample.int(
        length(j_data$w),
        size = n_mcmc_iterations,
        replace = TRUE,
        prob = w_post_1
      )

      comp_2 <- sample.int(
        length(j_data$w),
        size = n_mcmc_iterations,
        replace = TRUE,
        prob = w_post_2
      )

      draw_1 <- stats::rbeta(
        n_mcmc_iterations,
        shape1 = shape_1[comp_1, x[1]],
        shape2 = shape_2[comp_1, x[1]]
      )

      draw_2 <- stats::rbeta(
        n_mcmc_iterations,
        shape1 = shape_1[comp_2, x[2]],
        shape2 = shape_2[comp_2, x[2]]
      )

      matrix(
        c(
          stats::quantile(draw_1 - draw_2, probs = quantiles),
          Mean = mean(draw_1 - draw_2),
          SD   = stats::sd(draw_1 - draw_2)
        ),
        ncol = 1
      )
    })

    diff_names <- apply(calc_differences, 1, function(x) {
      paste0("p_diff_", paste0(as.character(x), collapse = ""))
    })

    colnames(diff_quantiles) <- diff_names
    posterior_quantiles <- cbind(posterior_quantiles, diff_quantiles)
  }

  posterior_quantiles
}

getUniqueRows <- function (

  matrix

) {

  n_rows      <- nrow(matrix)
  n_cols      <- ncol(matrix)

  unique_rows <- stats::aggregate(id ~ .,
                                  data = cbind(id = seq_along(n_rows), matrix),
                                  FUN  = length)

  return (unique_rows[, seq_len(n_cols)])

}

getUniqueTrials <- function (

  scenario_list

) {

  all_scenarios_n_responders <- do.call(rbind, lapply(scenario_list, function (x) x$n_responders))
  all_scenarios_n_subjects   <- do.call(rbind, lapply(scenario_list, function (x) x$n_subjects))
  all_scenarios_overall_gos  <- do.call(rbind, lapply(scenario_list, function (x)
    x$previous_analyses$go_decisions))[, 1]

  return (getUniqueRows(cbind(all_scenarios_n_responders,
                              all_scenarios_n_subjects,
                              go_flag = all_scenarios_overall_gos)))

}

is.analysis_list <- function (x) {

  if (missing(x)) stop ("Please provide an object for the argument 'x'")

  inherits(x, "analysis_list")

}

#' @title loadAnalyses
#' @md
#' @description This function loads an analysis performed with
#' \code{\link[bhmbasket]{performAnalyses}}
#' @param load_path A string providing a path where the scenarios are being stored,
#' Default: \code{\link[base]{tempfile}}
#' @param scenario_numbers A (vector of) positive integer(s) for the scenario number(s)
#' @param analysis_numbers A (vector of) positive integer(s) for the analysis number(s),
#' Default: `rep(1, length(scenario_numbers))`
#' @return Returns an object of class `analysis_list`
#' @seealso
#'  \code{\link[bhmbasket]{performAnalyses}}
#'  \code{\link[bhmbasket]{saveAnalyses}}
#'  \code{\link[base]{tempfile}}
#' @rdname loadAnalyses
#' @examples
#'   trial_data <- createTrial(
#'     n_subjects   = c(10, 20, 30),
#'     n_responders = c(1, 2, 3))
#'
#'   analysis_list <- performAnalyses(
#'     scenario_list      = trial_data,
#'     target_rates       = rep(0.5, 3),
#'     n_mcmc_iterations  = 100)
#'
#'   save_info     <- saveAnalyses(analysis_list)
#'   analysis_list <- loadAnalyses(scenario_numbers = save_info$scenario_numbers,
#'                                 analysis_numbers = save_info$analysis_numbers,
#'                                 load_path        = save_info$path)
#' @author Stephan Wojciekowski
#' @export
loadAnalyses <- function (

  scenario_numbers,
  analysis_numbers = rep(1, length(scenario_numbers)),
  load_path        = tempdir()

) {

  error_scenario_numbers <-
    "Please provide a vector of positive integers for the argument 'scenario_numbers'"
  error_analysis_numbers <-
    "Please provide a vector of positive integers for the argument 'analysis_numbers'"
  error_load_path <-
    "Please provide a string containing a path for the argument 'load_path'"
  error_len_match <-
    "'scenario_numbers' and 'analysis_numbers' must have equal length"

  checkmate::assertNumeric(
    scenario_numbers,
    any.missing = FALSE,
    .var.name   = error_scenario_numbers
  )
  checkmate::assertIntegerish(
    scenario_numbers,
    lower     = 1,
    .var.name = error_scenario_numbers
  )

  checkmate::assertIntegerish(
    analysis_numbers,
    lower       = 1,
    any.missing = FALSE,
    .var.name   = error_analysis_numbers
  )

  checkmate::assertTRUE(
    identical(length(scenario_numbers), length(analysis_numbers)),
    .var.name = error_len_match
  )

  checkmate::assertCharacter(
    load_path,
    len         = 1,
    any.missing = FALSE,
    .var.name   = error_load_path
  )

  ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ###

  analyses_list <- vector(mode = "list", length = length(scenario_numbers))

  for (s in seq_along(scenario_numbers)) {

    analyses_list[[s]] <- readRDS(
      file.path(load_path,
                paste0("analysis_data_", scenario_numbers[s], "_", analysis_numbers[s], ".rds"))
    )

  }

  names(analyses_list) <- paste0("scenario_", scenario_numbers)
  class(analyses_list) <- "analysis_list"

  return (analyses_list)

}

mapUniqueTrials <- function (

  scenario_list,
  method_quantiles_list,
  trials_unique_calc,
  applicable_previous_trials

) {

  method_names     <- names(method_quantiles_list)
  scenario_numbers <- sapply(scenario_list, function (x) x$scenario_number)

  ## Create hash tables for results for easy retrieval

  hash_keys        <- getHashKeys(trials_unique_calc)
  hash_tables_list <- vector(mode = "list", length = length(method_quantiles_list))

  for (n in seq_along(hash_tables_list)) {

    hash_tables_list[[n]] <- createHashTable(hash_keys, method_quantiles_list[[n]])

  }

  ## prepare foreach
  exported_stuff <- c("convertVector2Matrix")

  ## run foreach
  "%do%" <- foreach::"%do%"
  scenario_method_quantiles_list <- foreach::foreach(k = seq_along(scenario_numbers),
                                                     .verbose  = FALSE,
                                                     .export   = exported_stuff
  ) %do% {

    ## Find the indices of the trials of a specific scenario for go trials
    scenario_data_matrix <- cbind(scenario_list[[k]]$n_responders,
                                  scenario_list[[k]]$n_subjects)

    ## check whether there where previous analyses
    if (applicable_previous_trials) {

      scenario_go_flags         <- scenario_list[[k]]$previous_analyses$go_decisions[, 1] > 0
      scenario_method_quantiles <- scenario_list[[k]]$previous_analyses$post_quantiles

    } else {

      scenario_go_flags         <- rep(TRUE, length = nrow(scenario_data_matrix))
      scenario_method_quantiles <- vector(mode = "list", length = length(method_names))
      names(scenario_method_quantiles) <- method_names

    }

    ## In case there are trial realizations that need updating
    ## This should only not be the case if all trial realizations of a scenario have a NoGo decision
    ## and there are applicable previous trials.
    if (any(scenario_go_flags)) {

      ## Get search keys
      scenario_data_matrix_go <- convertVector2Matrix(scenario_data_matrix[scenario_go_flags, ])
      search_keys             <- getHashKeys(scenario_data_matrix_go)

      ## Save scenario specific posterior quantiles for each method
      for (n in seq_along(method_names)) {

        scenario_method_quantiles[[method_names[n]]][scenario_go_flags] <-
          getHashValues(search_keys, hash_tables_list[[n]])

      }

    }

    return (scenario_method_quantiles)

  }

  names(scenario_method_quantiles_list) <- paste0("scenario_", scenario_numbers)

  return (scenario_method_quantiles_list)

}

## Wrapper of getPostQuantiles() for specifying several scenarios
## Returns a list of quantiles of posterior distributions according to supplied methods.
#' @title performAnalyses
#' @md
#' @description This function performs the analysis of simulated or observed trial data with the
#' specified methods
#' and returns the quantiles of the posterior response rates
#' @param scenario_list An object of class `scenario_list`,
#' as e.g. created with \code{\link[bhmbasket]{simulateScenarios}}
#' @param evidence_levels A vector of numerics in `(0, 1)` for the
#' `1-evidence_levels`-quantiles of the posterior response rates to be saved.
#' Default: `c(0.025, 0.05, 0.5, 0.8, 0.9, 0.95, 0.975)`
#' @param method_names A vector of strings for the names of the methods to be used. Must
#' be one of the default values, Default:
#' `c("berry", "exnex", "exnex_mix", "exnex_adj", "exnex_adj_mix", "pooled", "stratified", "stratified_mix")`
#' @param target_rates A vector of numerics in `(0, 1)` for the
#' target rates of each cohort, Default: `NULL`
#' @param prior_parameters_list An object of class `prior_parameters_list`,
#' as e.g. created with \code{\link[bhmbasket]{getPriorParameters}}
#' @param calc_differences A matrix of positive integers with 2 columns.
#' For each row the differences will be calculated.
#' Also a vector of positive integers can be provided for a single difference.
#' The integers are the numbers for the cohorts to be subtracted from one another.
#' E.g. providing `c(2, 1)` calculates the difference between cohort `2` and cohort `1`.
#' If `NULL`, no subtractions are performed, Default: `NULL`
#' @param n_mcmc_iterations A positive integer for the number of MCMC iterations,
#' see Details, Default: `10000`.
#' If `n_mcmc_iterations` is present in `.GlobalEnv` and `missing(n_mcmc_iterations)`,
#' the globally available value will be used.
#' @param n_cores Argument is deprecated and does nothing as of version 0.9.3.
#' A positive integer for the number of cores for the parallelization,
#' Default: `1`
#' @param seed Argument is deprecated and does nothing as of version 0.9.3.
#' A numeric for the random seed, Default: `1`
#' @param verbose A logical indicating whether messages should be printed, Default: `TRUE`
#' @return An object of class `analysis_list`.
#' @details
#' This function applies the following analysis models to (simulated) scenarios of class
#' `scenario_list`:
#' \itemize{
#'   \item Bayesian hierarchical model (BHM) proposed by Berry et al. (2013): `"berry"`
#'   \item BHM proposed by Neuenschwander et al. (2016): `"exnex"`
#'   \item BHM that combines above approaches: `"exnex_adj"`
#'   \item Pooled beta-binomial approach: `"pooled"`
#'   \item Stratified beta-binomial approach: `"stratified"`
#'   \item BHM with mixture prior on the Nex component: `"exnex_mix"`
#'   \item Adjusted BHM with mixture prior on the Nex component: `"exnex_adj_mix"`
#'   \item Stratified beta-binomial approach with mixture beta prior: `"stratified_mix"`
#' }
#' The posterior distributions of the BHMs are approximated with Markov chain Monte Carlo (MCMC)
#' methods implemented in JAGS.
#' Two independent chains are used with each `n_mcmc_iterations` number of MCMC iterations.
#' The first `floor(n_mcmc_iterations / 3)` number of iterations are discarded as burn-in period.
#' No thinning is applied.
#'
#' Note that the value for `n_mcmc_iterations` required for a good approximation of the posterior
#' distributions depends on the analysis model, the investigated scenarios, and the use case.
#' The default value might be a good compromise between run-time and approximation for
#' the estimation of decision probabilities, but
#' it should definitively be increased for the analysis of a single trial's outcome.
#'
#' The analysis models will only be applied to the unique trial realizations across
#' all simulated scenarios.
#' The models can be applied in parallel by registering a parallel backend for the 'foreach'
#' framework, e.g. with `doFuture::registerDoFuture()` and `future::plan(future::multisession)`.
#' The parallelization is nested, so that the resources of a HPC environment can be used
#' efficiently.
#' For more on this topic, kindly see the respective vignette.
#' The tasks that are to be performed in parallel are chunked according to the number of workers
#' determined with `foreach::getDoParWorkers()`.
#'
#' The JAGS code for the BHM `"exnex"` was taken from Neuenschwander et al. (2016).
#' The JAGS code for the BHM `"exnex_adj"` is based on the JAGS code for `"exnex"`.
#' The JAGS code for the BHM `"exnex_mix"` extends the `"exnex"` model by using
#' a mixture prior for the Nex component.
#' The JAGS code for the BHM `"exnex_adj_mix"` extends the `"exnex_adj"` model by using
#' a mixture prior for the Nex component.
#' @seealso
#'  \code{\link[bhmbasket]{simulateScenarios}}
#'  \code{\link[bhmbasket]{createTrial}}
#'  \code{\link[bhmbasket]{getPriorParameters}}
#' @rdname performAnalyses
#' @author Stephan Wojciekowski
#' @examples
#'  trial_data <- createTrial(
#'    n_subjects   = c(10, 20, 30),
#'    n_responders = c(1, 2, 3))
#'
#'  analysis_list <- performAnalyses(
#'    scenario_list      = trial_data,
#'    target_rates       = rep(0.5, 3),
#'    calc_differences   = matrix(c(3, 2, 1, 1), ncol = 2),
#'    n_mcmc_iterations  = 100)
#' @references Berry, Scott M., et al. "Bayesian hierarchical modeling of patient subpopulations:
#' efficient designs of phase II oncology clinical trials."
#' \emph{Clinical Trials} 10.5 (2013): 720-734.
#' @references Neuenschwander, Beat, et al. "Robust exchangeability designs
#' for early phase clinical trials with multiple strata."
#' \emph{Pharmaceutical statistics} 15.2 (2016): 123-134.
#' @references Plummer, Martyn. "JAGS: A program for analysis of Bayesian graphical models
#' using Gibbs sampling."
#' \emph{Proceedings of the 3rd international workshop on distributed statistical computing.}
#' Vol. 124. No. 125.10. 2003.
#' @export
performAnalyses <- function (

  scenario_list,
  evidence_levels       = c(0.025, 0.05, 0.5, 0.8, 0.9, 0.95, 0.975),

  method_names          = c("berry", "exnex", "exnex_mix", "exnex_adj", "exnex_adj_mix", "pooled", "stratified", "stratified_mix"),
  target_rates          = NULL,
  prior_parameters_list = NULL,

  calc_differences      = NULL,

  n_mcmc_iterations     = 1e4,
  n_cores               = 1,
  seed                  = 1,
  verbose               = TRUE

) {

  error_scenario_list <-
    "Please provide an object of class scenario_list for the argument 'scenario_list'"
  error_evidence_levels <-
    "Please provide a vector of numerics in (0, 1) for the argument 'evidence_levels'"
  error_method_names <-
    paste("Please provide a (vector of) strings for the argument 'method_names'\n",
          "Must be one of 'berry', 'exnex', 'exnex_adj', 'exnex_mix', 'exnex_adj_mix', 'pooled', 'stratified', 'stratified_mix'")
  error_target_rates <-
    paste("Please provide either 'NULL' or a vector of numerics in (0, 1)",
          "for the argument 'target_rates'")
  error_prior_parameters_list <-
    "Please provide either 'NULL' or an object of class 'prior_parameters_list'"
  error_calc_differences <-
    paste("Please provide either 'NULL' or a matrix of integers with ncol = 2.",
          "The values of the integers must be less than or equal to the number",
          "of cohorts")
  error_n_mcmc_iterations <-
    "Please provide a positive integer for the argument 'n_mcmc_iterations'"
  error_verbose <-
    "Please provide a logical for the argument 'verbose'"

  error_need_target_rates_for_methods <-
    "Please provide 'target_rates' when using the methods 'berry', 'exnex_adj' and/or 'exnex_adj_mix'"
  error_need_one_of_prior_or_target <-
    "Please provide at least one of 'prior_parameters_list' or 'target_rates'"
  error_target_length <-
    "The length of 'target_rates' does not match the number of cohorts"
  error_prior_methods_mismatch <-
    paste("Not all specified methods in 'method_names'",
          "have prior parameters specified in 'prior_parameters_list'")
  error_prior_cohorts_mismatch <-
    paste("The number of cohorts specified in 'prior_parameters_list' does not match",
          "the number of cohorts specified in 'scenario_list'")

  warning_n_cores <- "The argument 'n_cores' is deprecated as of version 0.9.3."
  warning_seed    <- "The argument 'seed' is deprecated as of version 0.9.3."

  if (!missing(n_cores)) warning(warning_n_cores)
  if (!missing(seed))    warning(warning_seed)

  # scenario_list must be supplied and have the correct class
  checkmate::assertClass(
    scenario_list,
    "scenario_list",
    .var.name = error_scenario_list
  )

  # evidence_levels: numeric, all in (0, 1)
  checkmate::assertNumeric(
    evidence_levels,
    any.missing = FALSE,
    .var.name   = error_evidence_levels
  )
  checkmate::assertTRUE(
    all(evidence_levels > 0 & evidence_levels < 1),
    .var.name = error_evidence_levels
  )

  # method_names: character and must be one of the allowed methods
  checkmate::assertCharacter(
    method_names,
    any.missing = FALSE,
    min.len     = 1,
    .var.name   = error_method_names
  )
  method_names <- tryCatch(
    match.arg(
      method_names,
      choices    = c("berry", "exnex", "exnex_mix", "exnex_adj", "exnex_adj_mix", "pooled", "stratified", "stratified_mix"),
      several.ok = TRUE
    ),
    error = function(e) {
      stop(error_method_names, call. = FALSE)
    }
  )

  # target_rates: either NULL or numeric in (0, 1)
  if (!is.null(target_rates)) {
    checkmate::assertNumeric(
      target_rates,
      any.missing = FALSE,
      .var.name   = error_target_rates
    )
    checkmate::assertTRUE(
      all(target_rates > 0 & target_rates < 1),
      .var.name = error_target_rates
    )
  }

  # prior_parameters_list: either NULL or object with class 'prior_parameters_list'
  if (!is.null(prior_parameters_list)) {
    checkmate::assertClass(
      prior_parameters_list,
      "prior_parameters_list",
      .var.name = error_prior_parameters_list
    )
  }


  if (is.null(target_rates)) {

    # If target_rates are missing, berry, exnex_adj and exnex_adj_mix are not allowed
    checkmate::assertTRUE(
      !any(c("berry", "exnex_adj", "exnex_adj_mix") %in% method_names),
      .var.name = error_need_target_rates_for_methods
    )

    # Need at least one of prior_parameters_list or target_rates
    checkmate::assertTRUE(
      !is.null(prior_parameters_list),
      .var.name = error_need_one_of_prior_or_target
    )

  } else {
    # If target_rates is present, its length must match the number of cohorts
    n_coh <- ncol(scenario_list[[1]]$n_subjects)
    checkmate::assertTRUE(
      identical(length(target_rates), n_coh),
      .var.name = error_target_length
    )
  }


  if (!is.null(prior_parameters_list)) {

    # All methods used must have entries in prior_parameters_list
    checkmate::assertTRUE(
      all(method_names %in% names(prior_parameters_list)),
      .var.name = error_prior_methods_mismatch
    )

    # For exnex / exnex_adj / stratified, per-cohort prior lengths must
    # match the number of cohorts in scenario_list

    n_coh <- ncol(scenario_list[[1]]$n_subjects)

    inconsistent_cohorts <- any(
      sapply(
        intersect(names(prior_parameters_list),
                  c("exnex", "exnex_adj", "stratified")),
        function(name) {
          max(sapply(prior_parameters_list[[name]], length)) != n_coh
        }
      )
    )

    if ("stratified_mix" %in% names(prior_parameters_list)) {
      pp <- prior_parameters_list[["stratified_mix"]]

      inconsistent_stratified_mix <-
        !is.matrix(pp$a_j) ||
        !is.matrix(pp$b_j) ||
        any(dim(pp$a_j) != dim(pp$b_j)) ||
        ncol(pp$a_j) != n_coh ||
        length(pp$w) != nrow(pp$a_j)

      inconsistent_cohorts <- inconsistent_cohorts || inconsistent_stratified_mix
    }

    if (any(c("exnex_mix", "exnex_adj_mix") %in% names(prior_parameters_list))) {
      inconsistent_mix <- any(
        sapply(
          intersect(names(prior_parameters_list), c("exnex_mix", "exnex_adj_mix")),
          function(name) {
            pp <- prior_parameters_list[[name]]
            !is.matrix(pp$mean_nex) ||
              !is.matrix(pp$sd_nex) ||
              ncol(pp$mean_nex) != n_coh ||
              ncol(pp$sd_nex) != n_coh ||
              any(dim(pp$mean_nex) != dim(pp$sd_nex))
          }
        )
      )
      inconsistent_cohorts <- inconsistent_cohorts || inconsistent_mix
    }

    checkmate::assertFALSE(
      inconsistent_cohorts,
      .var.name = error_prior_cohorts_mismatch
    )
  }

  n_cohorts_min <- min(sapply(scenario_list, function(x) ncol(x$n_responders)))

  if (!is.null(calc_differences)) {

    checkmate::assertNumeric(
      calc_differences,
      any.missing = FALSE,
      .var.name   = error_calc_differences
    )

    is_len2   <- identical(length(calc_differences), 2L)
    has_2cols <- !is.null(dim(calc_differences)) &&
      identical(ncol(calc_differences), 2L)

    checkmate::assertTRUE(
      is_len2 || has_2cols,
      .var.name = error_calc_differences
    )

    checkmate::assertIntegerish(
      calc_differences,
      lower       = 1,
      any.missing = FALSE,
      .var.name   = error_calc_differences
    )

    checkmate::assertTRUE(
      max(calc_differences) <= n_cohorts_min,
      .var.name = error_calc_differences
    )
  }
  rm(n_cohorts_min)

  # If n_mcmc_iterations exists in .GlobalEnv and argument is missing, reuse it
  if ("n_mcmc_iterations" %in% ls(envir = .GlobalEnv) && missing(n_mcmc_iterations)) {
    n_mcmc_iterations <- get("n_mcmc_iterations", envir = .GlobalEnv)
  }

  checkmate::assertInt(
    n_mcmc_iterations,
    lower     = 1,
    .var.name = error_n_mcmc_iterations
  )

  checkmate::assertLogical(
    verbose,
    len         = 1L,
    any.missing = FALSE,
    .var.name   = error_verbose
  )

  # Only need a parallel backend if there is at least one method that is not pooled/stratified
  if (!all(method_names %in% c("stratified", "pooled", "stratified_mix"))) {
    checkForParallelBackend()
  }

  ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ###

  ## message to user
  if (verbose) message(format(Sys.time(), "%d-%h-%Y"), " Performing Analyses")

  ## some housekeeping
  method_names <- sort(method_names)
  quantiles    <- sort(unique(round(1 - c(0.025, 0.05, 0.5, 0.8, 0.9, 0.95, 0.975,
                                          evidence_levels), 9)))
  if (!is.null(calc_differences)) {
    calc_differences <- convertVector2Matrix(calc_differences)
  }

  ## get scenario numbers
  scenario_numbers <- sapply(scenario_list, function (x) x$scenario_number)

  ## get unique trials over all scenarios
  trials_unique <- getUniqueTrials(scenario_list)
  n_cohorts     <- (ncol(trials_unique) - 1L) / 2

  ## analyze only unique trials that have not been previously analyzed
  applicable_previous_trials <- applicablePreviousTrials(
    scenario_list    = scenario_list,
    method_names     = method_names,
    quantiles        = quantiles,
    n_cohorts        = n_cohorts,
    calc_differences = calc_differences)

  ## only get previous go indices if all conditions for previous trials are met
  if (applicable_previous_trials) {
    calc_trial_indices <- trials_unique[, ncol(trials_unique)] > 0
  } else {
    calc_trial_indices <- rep(TRUE, nrow(trials_unique))
  }

  ## get resulting unique number of responders and number of subjects
  trials_unique_calc <- trials_unique[calc_trial_indices, -ncol(trials_unique)]
  n_responders       <- trials_unique_calc[, seq_len(n_cohorts)]
  n_subjects         <- trials_unique_calc[, seq_len(n_cohorts) + n_cohorts]

  ## message to user
  if (verbose) {

    message("         Analyzing ", length(scenario_numbers) ," scenario", ifelse(length(scenario_numbers) == 1, "", "s")," ",
            "(", nrow(n_responders), " unique", ifelse(applicable_previous_trials, " updated ", " "),
            "trial realization", ifelse(nrow(trials_unique) == 1, "", "s"),")")

  }

  ## get default prior parameters if needed
  if(is.null(prior_parameters_list)) {

    prior_parameters_list <- getPriorParameters(
      method_names = method_names,
      target_rates = target_rates)

  }

  ## Create lists to save all results of the unique trials
  method_quantiles_list        <- vector(mode = "list", length = length(method_names))
  names(method_quantiles_list) <- method_names

  ## For each method
  for (n in seq_along(method_names)) {

    ## message to user
    if (verbose) {
      start_time  <- Sys.time()
      out_message <- paste0(format(start_time, "   %H:%M", digits = 1),
                            " - with ", firstUpper(method_names[n]), " ...")
      message(out_message, rep(".", max(0, 33 - nchar(out_message))))
    }

    ## prepare analysis
    prepare_analysis <- prepareAnalysis(
      method_name       = method_names[n],
      target_rates      = target_rates,
      prior_parameters  = prior_parameters_list[[method_names[n]]])

    ## run analysis
    method_quantiles_list[[method_names[n]]] <- getPostQuantiles(
      method_name       = method_names[n],
      quantiles         = quantiles,
      scenario_data     = list(n_subjects   = n_subjects,
                               n_responders = n_responders),
      calc_differences  = calc_differences,
      j_parameters      = prepare_analysis$j_parameters,
      j_model_file      = prepare_analysis$j_model_file,
      j_data            = prepare_analysis$j_data,
      n_mcmc_iterations = n_mcmc_iterations,
      save_path         = NULL,
      save_trial        = NULL)

    ## message to user
    if (verbose) {
      message("             finished after ", round(Sys.time() - start_time, 1), " ",
              units(Sys.time() - start_time), ".")
      rm(start_time)
    }

  }

  ## message to user
  if (verbose) {
    start_time  <- Sys.time()
    message("         Processing scenarios ...")
  }

  ## Process scenarios
  scenario_method_quantiles_list <- mapUniqueTrials(
    scenario_list              = scenario_list,
    method_quantiles_list      = method_quantiles_list,
    trials_unique_calc         = trials_unique_calc,
    applicable_previous_trials = applicable_previous_trials)

  ## message to user
  if (verbose) {
    message("             finished after ", round(Sys.time() - start_time, 1), " ",
            units(Sys.time() - start_time), ".")
    rm(start_time)
  }

  ## combine results from all scenarios & return
  analyses_list        <- vector(mode = "list", length = length(scenario_numbers))
  names(analyses_list) <- paste0("scenario_", scenario_numbers)

  for (s in seq_along(scenario_numbers)) {

    analyses_list[[s]] <- list(
      quantiles_list      = scenario_method_quantiles_list[[s]],
      scenario_data       = scenario_list[[s]],
      analysis_parameters = list(
        quantiles             = quantiles,
        method_names          = method_names,
        prior_parameters_list = prior_parameters_list,
        n_mcmc_iterations     = n_mcmc_iterations))

  }

  class(analyses_list) <- "analysis_list"

  return (analyses_list)

}

## based on R2jags::jags
## stripped down to improve performance
performJags <- function (

  data,
  parameters_to_save,
  model_file,
  n_chains = 2,
  n_iter   = 1e4,
  n_burnin = floor(n_iter/3)

) {

  n_adapt <- ifelse(n_burnin > 0, n_burnin, 100)

  inits <- vector("list", n_chains)
  for (i in 1:n_chains) {
    inits[[i]]$.RNG.name <- "base::Wichmann-Hill"
    inits[[i]]$.RNG.seed <- stats::runif(1, 0, 2^31)
  }

  j_model <- rjags::jags.model(file     = model_file,
                               data     = data,
                               inits    = inits,
                               n.chains = n_chains,
                               n.adapt  = 0,
                               quiet    = TRUE)

  rjags::adapt(object         = j_model,
               n.iter         = n_adapt,
               progress.bar   = "none",
               end.adaptation = TRUE)

  samples <- rjags::coda.samples(model          = j_model,
                                 variable.names = parameters_to_save,
                                 n.iter         = n_iter - n_burnin,
                                 thin           = 1,
                                 progress.bar   = "none")

  return(do.call(rbind, samples))

}

posteriors2Quantiles <- function (

  quantiles,
  posteriors

) {

  posterior_quantiles <- apply(posteriors, 2, function (x) stats::quantile(x, probs = quantiles))

  posterior_mean      <- apply(posteriors, 2, mean)
  posterior_sd        <- apply(posteriors, 2, stats::sd)
  posterior_quantiles <- rbind(posterior_quantiles,
                               Mean = posterior_mean,
                               SD   = posterior_sd)

  return (posterior_quantiles)

}

prepareAnalysis <- function (

  method_name,

  prior_parameters = NULL,
  target_rates     = NULL

) {

  if (method_name == "berry") {

    j_data <- list(
      mean_mu       = prior_parameters$mu_mean,
      precision_mu  = prior_parameters$mu_sd^-2,
      precision_tau = prior_parameters$tau_scale^-2,
      p_t           = target_rates,
      J             = length(target_rates)
    )

    j_model_file <- getModelFile(method_name = "berry")
    j_parameters <- c("p", "mu", "tau")

  } else if (method_name %in% c("exnex", "exnex_adj", "exnex_mix", "exnex_adj_mix")) {

    ## shared ExNex data
    j_data <- list(
      Nexch        = length(prior_parameters$mu_mean),
      Nmix         = length(prior_parameters$mu_mean) + 1L,
      mu_mean      = prior_parameters$mu_mean,
      mu_prec      = prior_parameters$mu_sd^-2,
      tau_HN_scale = rep(prior_parameters$tau_scale, length(prior_parameters$mu_mean))
    )

    ## shared pMix
    if (identical(length(prior_parameters$w_j), 1L)) {
      j_data$pMix <- c(prior_parameters$w_j, 1 - prior_parameters$w_j)
    } else {
      j_data$pMix <- prior_parameters$w_j
    }

    ## standard ExNex / ExNex Adj
    if (method_name %in% c("exnex", "exnex_adj")) {

      j_data$Nstrata  <- length(prior_parameters$mu_j)
      j_data$nex_mean <- prior_parameters$mu_j
      j_data$nex_prec <- prior_parameters$tau_j^-2
      
    }

    ## ExNex mix / ExNex Adj mix
    if (method_name %in% c("exnex_mix", "exnex_adj_mix")) {

      j_data$Nstrata  <- ncol(prior_parameters$mean_nex)
      j_data$K_nex    <- length(prior_parameters$w_nex)
      j_data$w_nex    <- prior_parameters$w_nex
      j_data$mean_nex <- prior_parameters$mean_nex
      j_data$prec_nex <- prior_parameters$sd_nex^-2
      
    }

    ## adjusted models
    if (method_name %in% c("exnex_adj", "exnex_adj_mix")) {
      j_data$p_target <- target_rates
    }

    j_model_file <- getModelFile(method_name = method_name)
    j_parameters <- c("p", "mu", "tau", "exch")

  } else if (method_name %in% c("stratified", "pooled", "stratified_mix")) {

    ## For methods "stratified", "pooled" and "stratified_mix" no MCMC simulations are necessary,
    ## as the posterior response rates of the cohorts follow known beta distributions for stratified
    ## and pooled and mixture beta distribution for stratified mix.
    j_model_file <- "dummy path to JAGS model"
    j_parameters <- "dummy JAGS parameters"
    j_data       <- prior_parameters

  } else {

    stop(
      "method_name must be one of berry, exnex, exnex_adj, exnex_mix, exnex_adj_mix, stratified, pooled, stratified_mix"
    )

  }

  return (list(
    j_parameters = j_parameters,
    j_model_file = j_model_file,
    j_data       = j_data
  ))
  
}

#' @export
print.analysis_list <- function (x, digits = 2, ...) {

  n_scenarios    <- length(x)
  scenario_names <- names(x)

  n_methods      <- length(x[[1]]$quantiles_list)
  method_names   <- names(x[[1]]$quantiles_list)

  estimates          <- getEstimates(x)
  n_mcmc_interations <- x[[1]]$analysis_parameters$n_mcmc_iterations

  evidence_levels <- sort(1 - x[[1]]$analysis_parameters$quantiles)

  cat("analysis_list of ", n_scenarios, " scenario", ifelse(n_scenarios == 1, "", "s"),
      " with ", n_methods, " method", ifelse(n_methods == 1, "", "s"),"\n\n", sep = "")

  for (n in seq_along(scenario_names)) {

    if (n_scenarios == 1L) {

      expr <- quote(t(y[, 1:2]))

    } else {

      expr <- quote(t(y[[n]][, 1:2]))

    }

    mat_out <- do.call(rbind, lapply(estimates, function (y) eval(expr)))

    rownames(mat_out) <-  paste0(
      c("    - ", "      "),
      c(rbind(
        paste0(
          firstUpper(method_names),
          sapply(method_names, function (y) {
            getBlankString(max(nchar(method_names)) - nchar(y) + 1)
          })),
        rep(getBlankString(max(nchar(method_names)) + 3),
            length = length(method_names))
      )),
      rownames(mat_out))

    cat("  -", scenario_names[n], "\n")
    print(round(mat_out, digits = digits))

    cat("\n")

  }

  cat("  -", n_mcmc_interations, "MCMC iterationns per BHM method\n")
  cat("  - Available evidence levels:", evidence_levels, "\n")

}

qbetaDiff <- function (

  quantiles,

  x_1_shape1,
  x_1_shape2,

  x_2_shape1,
  x_2_shape2,

  n_mcmc = 1e6

) {

  sample_1   <- stats::rbeta(n_mcmc, shape1 = x_1_shape1, shape2 = x_1_shape2)
  sample_2   <- stats::rbeta(n_mcmc, shape1 = x_2_shape1, shape2 = x_2_shape2)
  difference <- sample_1 - sample_2

  quantiles_diff <- stats::quantile(difference, probs = quantiles)

  mean_diff      <- mean(difference)
  sd_diff        <- stats::sd(difference)
  quantiles_diff <- c(quantiles_diff, mean_diff, sd_diff)

  return (quantiles_diff)

}


#' @title saveAnalyses
#' @md
#' @description This function saves an object of class `analysis_list`
#' @param analyses_list An object of class `analysis_list`,
#' as created with \code{\link[bhmbasket]{performAnalyses}}
#' @param save_path A string for the path where the scenarios are being stored,
#' Default: \code{\link[base]{tempfile}}
#' @param analysis_numbers A positive integer naming the analysis number.
#' If `NULL`, the function will look for the number of saved analyses of the scenario
#' in the directory and add 1, Default: `NULL`
#' @return A named list of length 3 of vectors with scenario and analysis numbers and
#' the `save_path`
#' @seealso
#'  \code{\link[bhmbasket]{performAnalyses}}
#'  \code{\link[bhmbasket]{loadAnalyses}}
#'  \code{\link[base]{tempfile}}
#' @rdname saveAnalyses
#' @examples
#'   trial_data <- createTrial(
#'     n_subjects   = c(10, 20, 30),
#'     n_responders = c(1, 2, 3))
#'
#'   analysis_list <- performAnalyses(
#'     scenario_list      = trial_data,
#'     target_rates       = rep(0.5, 3),
#'     n_mcmc_iterations  = 100)
#'
#'   save_info     <- saveAnalyses(analysis_list)
#'   analysis_list <- loadAnalyses(scenario_numbers = save_info$scenario_numbers,
#'                                 analysis_numbers = save_info$analysis_numbers,
#'                                 load_path        = save_info$path)
#' @author Stephan Wojciekowski
#' @export
saveAnalyses <- function (

  analyses_list,
  save_path        = tempdir(),
  analysis_numbers = NULL

) {

  error_analyses_list <-
    "Please provide an object of class analysis_list for the argument 'analyses_list'"
  error_save_path     <-
    "Please provide a string containing a path for the argument 'save_path'"
  error_analysis_numbers <- paste(
    "Please provide a vector of positive integers for the argument 'analysis_numbers'",
    "with length equal to the length of 'analyses_list'"
  )

  checkmate::assertClass(
    analyses_list,
    "analysis_list",
    .var.name = error_analyses_list
  )
  checkmate::assertCharacter(
    save_path,
    len         = 1,
    any.missing = FALSE,
    .var.name   = error_save_path
  )

  if (!is.null(analysis_numbers)) {
    checkmate::assertIntegerish(
      analysis_numbers,
      lower       = 1,
      any.missing = FALSE,
      .var.name   = error_analysis_numbers
    )

    checkmate::assertTRUE(
      identical(length(analyses_list), length(analysis_numbers)),
      .var.name = error_analysis_numbers
    )
  }

  ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ###

  scenario_numbers <- sapply(analyses_list, function (x) x$scenario_data$scenario_number)

  if (is.null(analysis_numbers)) {
    analysis_numbers <- rep(0, length(analyses_list))
  }

  for (s in seq_along(analyses_list)) {

    ## Get analysis number
    if (identical(analysis_numbers[s], 0)) {

      analysis_numbers[s] <- sum(grepl(paste0("analysis_data_", scenario_numbers[s], "_"),
                                       list.files(save_path))) + 1L

    }

    ## Save the analysis
    file_name <- paste0("analysis_data_", scenario_numbers[s], "_", analysis_numbers[s],".rds")
    saveRDS(analyses_list[[s]], file = file.path(save_path, file_name),
            compress = "xz")

  }

  return (list(scenario_numbers = scenario_numbers,
               analysis_numbers = analysis_numbers,
               path             = save_path))

}

# Tests for applicablePreviousTrials -------------------------------------------

app_prev <- setup_applicable_prev_trials()

scen_initial       <- app_prev$scen_initial
analysis_initial   <- app_prev$analysis_initial
go_initial         <- app_prev$go_initial
scen_next          <- app_prev$scen_next
analysis_with_diff <- app_prev$analysis_with_diff
go_with_diff       <- app_prev$go_with_diff
scen_next_diff     <- app_prev$scen_next_diff
method_names_prev  <- app_prev$method_names_prev
quantiles_prev     <- app_prev$quantiles_prev
n_coh_prev         <- app_prev$n_coh_prev

# ------------------------------------------------------------------
# Test: TRUE when all preconditions for reusing previous trials are met
# Input:
#   - scenario_list = scen_next
#   - method_names = method_names_prev
#   - quantiles = quantiles_prev
#   - n_cohorts = n_coh_prev
#   - calc_differences = NULL
# Behaviour:
#   - No diff columns requested; structure and methods consistent.
# Expectations:
#   - Returns a logical scalar TRUE.
# Why:
#   - Baseline positive case for applicablePreviousTrials().
# ------------------------------------------------------------------
test_that("applicablePreviousTrials returns TRUE when all conditions are met (no differences)", {
  res <- applicablePreviousTrials(
    scenario_list    = scen_next,
    method_names     = method_names_prev,  # "pooled"
    quantiles        = quantiles_prev,
    n_cohorts        = n_coh_prev,
    calc_differences = NULL
  )

  expect_true(is.logical(res))
  expect_length(res, 1L)
  expect_true(res)
})

# ------------------------------------------------------------------
# Test: TRUE when diff columns are requested and present in previous_analyses
# Input:
#   - scenario_list = scen_next_diff (contains p_diff_* columns)
#   - calc_differences = matrix(c(3,2)) -> p_diff_32
# Behaviour:
#   - Checks that all requested differences are available in previous quantiles.
# Expectations:
#   - Returns TRUE.
# Why:
#   - Confirms that the function correctly recognises usable previous diffs.
# ------------------------------------------------------------------
test_that("applicablePreviousTrials returns TRUE when required diff columns exist", {
  quantiles_diff <- analysis_with_diff[[1]]$analysis_parameters$quantiles
  calc_diff      <- matrix(c(3, 2), ncol = 2)   # corresponds to "p_diff_32"

  res <- applicablePreviousTrials(
    scenario_list    = scen_next_diff,
    method_names     = method_names_prev,
    quantiles        = quantiles_diff,
    n_cohorts        = n_coh_prev,
    calc_differences = calc_diff
  )

  expect_true(is.logical(res))
  expect_length(res, 1L)
  expect_true(res)
})

# ------------------------------------------------------------------
# Test: FALSE when at least one required diff column is missing
# Input:
#   - scenario_list = scen_next_diff but with a p_diff_* column removed
#   - calc_differences still asks for that difference
# Behaviour:
#   - Detection of missing p_diff_ij should invalidate reuse.
# Expectations:
#   - Returns FALSE.
# Why:
#   - Ensures robustness to incomplete previous quantile objects.
# ------------------------------------------------------------------
test_that("applicablePreviousTrials returns FALSE when required diff columns are missing", {
  scen_broken <- scen_next_diff

  # Remove the diff column from the first method's first trial matrix
  pq      <- scen_broken[[1]]$previous_analyses$post_quantiles
  method1 <- names(pq)[1]
  mat1    <- pq[[method1]][[1]]
  diff_cols <- grepl("^p_diff_", colnames(mat1))

  if (any(diff_cols)) {
    mat1 <- mat1[, !diff_cols, drop = FALSE]
    pq[[method1]][[1]] <- mat1
    scen_broken[[1]]$previous_analyses$post_quantiles <- pq
  } else {
    skip("No p_diff_* column to remove")
  }

  quantiles_diff <- analysis_with_diff[[1]]$analysis_parameters$quantiles
  calc_diff      <- matrix(c(3, 2), ncol = 2)   # still asking for p_diff_32

  res <- applicablePreviousTrials(
    scenario_list    = scen_broken,
    method_names     = method_names_prev,
    quantiles        = quantiles_diff,
    n_cohorts        = n_coh_prev,
    calc_differences = calc_diff
  )

  expect_true(is.logical(res))
  expect_length(res, 1L)
  expect_false(res)
})

# ------------------------------------------------------------------
# Test: FALSE when method_names differ across scenarios
# Input:
#   - Two scenarios with different names in post_quantiles
# Behaviour:
#   - Applicable only if all scenarios share identical method sets.
# Expectations:
#   - Returns FALSE.
# Why:
#   - Prevents reusing previous trials when method structure is inconsistent.
# ------------------------------------------------------------------
test_that("applicablePreviousTrials returns FALSE when method_names differ across scenarios", {
  # Copy scen_next to two scenarios, then break method-name consistency in scenario_2
  scen_multi <- scen_next
  scen_multi$scenario_2 <- scen_multi$scenario_1
  names(scen_multi) <- c("scenario_1", "scenario_2")
  class(scen_multi) <- "scenario_list"

  names(scen_multi$scenario_2$previous_analyses$post_quantiles) <-
    paste0("alt_", names(scen_multi$scenario_2$previous_analyses$post_quantiles))

  res <- applicablePreviousTrials(
    scenario_list    = scen_multi,
    method_names     = method_names_prev,
    quantiles        = quantiles_prev,
    n_cohorts        = n_coh_prev,
    calc_differences = NULL
  )

  expect_true(is.logical(res))
  expect_length(res, 1L)
  expect_false(res)
})


# Tests for calcDiffsMCMC ------------------------------------------------------

# ------------------------------------------------------------------
# Test: calcDiffsMCMC adds p_diff_ij columns with correct values
# Input:
#   - posterior_samples with p_1, p_2, p_3, mu
#   - calc_differences = (1,2) and (3,1)
# Behaviour:
#   - Adds p_diff_12 = p_1 - p_2, p_diff_31 = p_3 - p_1.
# Expectations:
#   - Original columns retained; new diff columns added with correct numeric values.
# Why:
#   - Validates difference computation used downstream in decision rules.
# ------------------------------------------------------------------
test_that("calcDiffsMCMC: adds correctly named difference columns with correct values", {
  posterior_samples <- cbind(
    p_1 = c(0.1, 0.2),
    p_2 = c(0.5, 0.6),
    p_3 = c(0.8, 0.9),
    mu  = c(1.0, 2.0)
  )

  calc_differences <- rbind(
    c(1, 2),
    c(3, 1)
  )

  out <- calcDiffsMCMC(
    posterior_samples = posterior_samples,
    calc_differences  = calc_differences
  )

  expect_true(all(c("p_1", "p_2", "p_3", "mu") %in% colnames(out)))
  expect_true(all(c("p_diff_12", "p_diff_31") %in% colnames(out)))

  expected_diff_12 <- posterior_samples[, "p_1"] - posterior_samples[, "p_2"]
  expected_diff_31 <- posterior_samples[, "p_3"] - posterior_samples[, "p_1"]

  expect_equal(out[, "p_diff_12"], expected_diff_12)
  expect_equal(out[, "p_diff_31"], expected_diff_31)
})


# Tests for performJags --------------------------------------------------------

model_file <- tempfile(fileext = ".bug")
writeLines(
  c(
    "model {",
    "  theta ~ dbeta(1, 1)",
    "  for (i in 1:N) {",
    "    y[i] ~ dbern(theta)",
    "  }",
    "}"
  ),
  con = model_file
)

data_list <- list(
  N = 20L,
  y = c(rep(1L, 14), rep(0L, 6))
)

# ------------------------------------------------------------------
# Test: performJags runs a simple Bernoulli model and returns reasonable samples
# Input:
#   - Beta(1,1) prior, 14 successes / 6 failures
# Behaviour:
#   - Posterior is Beta(15,7); posterior mean ~ 15/22.
# Expectations:
#   - Returns matrix with column "theta", expected number of rows,
#     and sample mean close to 15/22.
# Why:
#   - Integration test for JAGS wrapper and posterior extraction.
# ------------------------------------------------------------------
test_that("performJags runs a simple Bernoulli model and returns a sensible sample matrix", {
  skip_if_not_installed("rjags")

  n_chains <- 2L
  n_iter   <- 1000L
  n_burnin <- 100L

  samples <- performJags(
    data               = data_list,
    parameters_to_save = "theta",
    model_file         = model_file,
    n_chains           = n_chains,
    n_iter             = n_iter,
    n_burnin           = n_burnin
  )

  expect_true(is.matrix(samples))
  expect_identical(colnames(samples), "theta")

  expected_rows <- n_chains * (n_iter - n_burnin)
  expect_equal(nrow(samples), expected_rows)

  # Given 14 successes and 6 failures with a Beta(1,1) prior, the posterior is Beta(15, 7).
  # The analytic posterior mean is 15 / (15 + 7) = 15/22.
  expect_equal(mean(samples[, "theta"]), 15/22, tolerance = 0.01)
})

# ------------------------------------------------------------------
# Test: performJags handles n_burnin = 0
# Input:
#   - Same model/data, with n_burnin = 0.
# Behaviour:
#   - Uses all n_iter samples in the posterior.
# Expectations:
#   - Number of rows equals n_chains * n_iter;
#     sample mean still close to 15/22.
# Why:
#   - Verifies edge case where no explicit burn-in is requested.
# ------------------------------------------------------------------
test_that("performJags handles n_burning equals 0 correctly", {
  skip_if_not_installed("rjags")

  n_chains <- 2L
  n_iter   <- 1000L
  n_burnin <- 0

  samples <- performJags(
    data               = data_list,
    parameters_to_save = "theta",
    model_file         = model_file,
    n_chains           = n_chains,
    n_iter             = n_iter,
    n_burnin           = n_burnin
  )

  expect_true(is.matrix(samples))
  expect_identical(colnames(samples), "theta")

  expected_rows <- n_chains * (n_iter - n_burnin)
  expect_equal(nrow(samples), expected_rows)

  expect_equal(mean(samples[, "theta"]), 15/22, tolerance = 0.01)
})


# Tests for getPosteriors ------------------------------------------------------

# ------------------------------------------------------------------
# Test: getPosteriors basic behaviour and name cleaning (no exch weights)
# Input:
#   - Berry model prepared via prepareAnalysis, single trial scenario.
# Behaviour:
#   - Calls JAGS, returns posterior samples matrix with cleaned column names
#     (no brackets, no w_*, no "exch").
# Expectations:
#   - Matrix result, finite entries, column names stripped of indices,
#     no exNEX-specific columns.
# Why:
#   - Ensures generic posterior extraction cleans names consistently.
# ------------------------------------------------------------------
test_that("getPosteriors: basic posterior sampling and name cleaning (no exch weights)", {
  skip_if_not_installed("rjags")

  set.seed(123)

  scen_list <- simulateScenarios(
    n_subjects_list     = list(c(10, 20)),
    response_rates_list = list(c(0.5, 0.6)),
    n_trials            = 1
  )
  scen1 <- scen_list$scenario_1

  target_rates <- c(0.5, 0.5)
  priors       <- getPriorParameters(
    method_names = "berry",
    target_rates = target_rates
  )

  prep <- prepareAnalysis(
    method_name      = "berry",
    target_rates     = target_rates,
    prior_parameters = priors[["berry"]]
  )

  j_data <- prep$j_data
  j_data$r <- as.numeric(scen1$n_responders[1, ])
  j_data$n <- as.numeric(scen1$n_subjects[1, ])

  post <- getPosteriors(
    j_parameters      = prep$j_parameters,
    j_model_file      = prep$j_model_file,
    j_data            = j_data,
    n_mcmc_iterations = 20
  )

  expect_true(is.matrix(post))
  expect_gt(nrow(post), 0)
  expect_true(all(is.finite(post)))

  expect_false(any(grepl("\\[", colnames(post))))
  expect_false(any(grepl("\\]", colnames(post))))

  expect_false(any(grepl("^w_", colnames(post))))
  expect_false(any(grepl("exch", colnames(post))))
})

# ------------------------------------------------------------------
# Test: getPosteriors for exNEX renames exch weights and strips extras
# Input:
#   - exNEX analysis prepared via prepareAnalysis on a small scenario.
# Behaviour:
#   - Columns for exchangeability weights are renamed to w_j*;
#     exNEX internal parameters and redundant labels removed.
# Expectations:
#   - Posterior matrix, finite; no "exch" in names;
#     exactly J cohort w_* columns; no ",1" suffixes.
# Why:
#   - Confirms special handling of mixture weights in exNEX models.
# ------------------------------------------------------------------
test_that("getPosteriors: exNEX exchangeability weights are renamed to w_j and extras dropped", {
  skip_if_not_installed("rjags")

  set.seed(456)

  scen_list <- simulateScenarios(
    n_subjects_list     = list(c(10, 20)),
    response_rates_list = list(c(0.6, 0.7)),
    n_trials            = 1
  )
  scen1 <- scen_list$scenario_1

  target_rates <- c(0.5, 0.5)
  priors       <- getPriorParameters(
    method_names = "exnex",
    target_rates = target_rates
  )

  prep <- prepareAnalysis(
    method_name      = "exnex",
    target_rates     = target_rates,
    prior_parameters = priors[["exnex"]]
  )

  j_data <- prep$j_data
  j_data$r <- as.numeric(scen1$n_responders[1, ])
  j_data$n <- as.numeric(scen1$n_subjects[1, ])

  post <- getPosteriors(
    j_parameters      = prep$j_parameters,
    j_model_file      = prep$j_model_file,
    j_data            = j_data,
    n_mcmc_iterations = 20
  )

  expect_true(is.matrix(post))
  expect_gt(nrow(post), 0)
  expect_true(all(is.finite(post)))

  expect_false(any(grepl("exch", colnames(post))))

  w_cols <- grep("^w_", colnames(post))
  expect_equal(length(w_cols), length(j_data$n))

  expect_false(any(grepl(",1", colnames(post))))
})


# Tests for getPostQuantiles ---------------------------------------------------

# ------------------------------------------------------------------
# Test: pooled backend with calc_differences adds zero p_diff_* columns
# Input:
#   - scenario_data with one trial / 2 cohorts
#   - method_name = "pooled", j_data = Beta prior, calc_differences for 1↔2
# Behaviour:
#   - Pooled backend duplicates same marginal posterior for p_1 and p_2;
#     p_diff_* columns are filled with zeros by design.
# Expectations:
#   - One matrix in list; exact column/row names;
#     p_diff_* = 0; pooled Beta quantiles/mean/SD match analytic forms.
# Why:
#   - Verifies deterministic pooled implementation and diff-column handling.
# ------------------------------------------------------------------
test_that("getPostQuantiles (pooled): calc_differences adds zero columns with correct names and shapes", {
  n_subjects_mat   <- matrix(c(10, 20), nrow = 1)
  n_responders_mat <- matrix(c(3,  5),  nrow = 1)

  scenario_data <- list(
    n_subjects   = n_subjects_mat,
    n_responders = n_responders_mat
  )

  j_data    <- list(a = 1, b = 1)
  quantiles <- c(0.025, 0.5, 0.975)

  # Differences to compute (pooled backend fills them with zeros)
  calc_differences <- rbind(c(1, 2), c(2, 1))

  out <- getPostQuantiles(
    method_name       = "pooled",
    quantiles         = quantiles,
    scenario_data     = scenario_data,
    calc_differences  = calc_differences,
    j_parameters      = NULL,
    j_model_file      = NULL,
    j_data            = j_data,
    n_mcmc_iterations = 20,
    save_path         = NULL,
    save_trial        = NULL
  )

  expect_type(out, "list")
  expect_length(out, 1)

  mat <- out[[1]]
  expect_true(is.matrix(mat))

  expect_identical(
    colnames(mat),
    c("p_1", "p_2", "p_diff_12", "p_diff_21")
  )
  expect_identical(
    rownames(mat),
    c("2.5%", "50%", "97.5%", "Mean", "SD")
  )

  expect_true(all(mat[, c("p_diff_12", "p_diff_21")] == 0))

  r_tot   <- sum(n_responders_mat[1, ])
  n_tot   <- sum(n_subjects_mat[1, ])
  shape_1 <- j_data$a + r_tot
  shape_2 <- j_data$b + (n_tot - r_tot)

  expected_q  <- stats::qbeta(quantiles, shape1 = shape_1, shape2 = shape_2)
  expected_mu <- shape_1 / (shape_1 + shape_2)
  expected_sd <- sqrt((shape_1 * shape_2) / ((shape_1 + shape_2)^2 * (shape_1 + shape_2 + 1)))

  expect_equal(unname(mat[c("2.5%", "50%", "97.5%"), "p_1"]), expected_q, tolerance = 1e-12)
  expect_equal(unname(mat[c("2.5%", "50%", "97.5%"), "p_2"]), expected_q, tolerance = 1e-12)
  expect_equal(unname(mat["Mean", c("p_1", "p_2")]), rep(expected_mu, 2), tolerance = 1e-12)
  expect_equal(unname(mat["SD",   c("p_1", "p_2")]), rep(expected_sd, 2), tolerance = 1e-12)
})

# ------------------------------------------------------------------
# Test: stratified backend with multiple trials and calc_differences
# Input:
#   - 2 trials x 2 cohorts; method_name = "stratified"
#   - calc_differences = (1,2)
# Behaviour:
#   - Returns per-trial quantile matrices with p_1, p_2 and p_diff_12,
#     finite values for marginal posterior summaries.
# Expectations:
#   - Length of out equals number of trials; each matrix has expected
#     row/column labels and finite p_j entries.
# Why:
#   - Confirms stratified Beta backend and diff logic over multiple trials.
# ------------------------------------------------------------------
test_that("getPostQuantiles (stratified): multiple trials and calc_differences produce p_j and p_diff_*", {
  skip_if_not_installed("foreach")
  skip_if_not_installed("doRNG")

  foreach::registerDoSEQ()

  n_subjects_mat <- rbind(
    c(10, 10),
    c(20, 20)
  )
  n_responders_mat <- rbind(
    c(3, 5),
    c(6, 10)
  )

  scenario_data <- list(
    n_subjects   = n_subjects_mat,
    n_responders = n_responders_mat
  )

  j_data <- list(
    a_j = c(1, 1),
    b_j = c(1, 1)
  )

  quantiles        <- c(0.25, 0.5, 0.75)
  calc_differences <- matrix(c(1, 2), ncol = 2)

  out <- getPostQuantiles(
    method_name       = "stratified",
    quantiles         = quantiles,
    scenario_data     = scenario_data,
    calc_differences  = calc_differences,
    j_parameters      = NULL,
    j_model_file      = NULL,
    j_data            = j_data,
    n_mcmc_iterations = 20,
    save_path         = NULL,
    save_trial        = NULL
  )

  expect_type(out, "list")
  expect_length(out, 2)

  for (mat in out) {
    expect_true(is.matrix(mat))

    expect_setequal(
      colnames(mat),
      c("p_1", "p_2", "p_diff_12")
    )

    expect_setequal(
      rownames(mat),
      c(paste0(quantiles * 100, "%"), "Mean", "SD")
    )

    expect_true(all(is.finite(mat[, c("p_1", "p_2")])))

    diff_col <- mat[, "p_diff_12"]
    expect_true(is.numeric(diff_col))
  }
})


# Tests for prepareAnalysis ----------------------------------------------------

# ------------------------------------------------------------------
# Test: berry branch builds j_data and parameter names
# Input:
#   - target_rates = c(0.5, 0.7); priors from getPriorParameters("berry")
# Behaviour:
#   - j_data fields populated from priors and target_rates;
#     J equals number of cohorts; JAGS parameters and model file exist.
# Expectations:
#   - mean_mu, precision_mu, precision_tau, p_t, J consistent;
#     j_parameters = c("p","mu","tau"); model file exists.
# Why:
#   - Validates mapping from high-level Berry prior to JAGS inputs.
# ------------------------------------------------------------------
test_that("prepareAnalysis: berry branch builds correct j_data and parameters", {
  target_rates <- c(0.5, 0.7)
  priors_list  <- getPriorParameters(method_names = "berry", target_rates = target_rates)
  priors_berry <- priors_list[["berry"]]

  prep <- prepareAnalysis(
    method_name      = "berry",
    prior_parameters = priors_berry,
    target_rates     = target_rates
  )

  j_data <- prep$j_data

  expect_equal(j_data$mean_mu,       priors_berry$mu_mean)
  expect_equal(j_data$precision_mu,  priors_berry$mu_sd^-2)
  expect_equal(j_data$precision_tau, priors_berry$tau_scale^-2)
  expect_equal(j_data$p_t,           target_rates)
  expect_equal(j_data$J,             length(target_rates))

  expect_equal(prep$j_parameters, c("p", "mu", "tau"))
  expect_true(is.character(prep$j_model_file))
  expect_true(file.exists(prep$j_model_file))
})

# ------------------------------------------------------------------
# Test: exnex branch constructs mixture hyperparameters and pMix
# Input:
#   - target_rates = c(0.4,0.6,0.8); priors from getPriorParameters("exnex")
# Behaviour:
#   - j_data contains Nexch, Nmix, Nstrata, mu_mean, mu_prec,
#     tau_HN_scale, nex_mean, nex_prec and pMix that sums to 1.
# Expectations:
#   - Dimensions consistent with priors; pMix length Nexch+1 and sums to 1.
# Why:
#   - Ensures exNEX prior is correctly translated into JAGS hyperparameters.
# ------------------------------------------------------------------
test_that("prepareAnalysis: exnex branch builds mixture priors and pMix", {
  target_rates <- c(0.4, 0.6, 0.8)
  priors_list  <- getPriorParameters(method_names = "exnex", target_rates = target_rates)
  priors_exnex <- priors_list[["exnex"]]

  prep <- prepareAnalysis(
    method_name      = "exnex",
    prior_parameters = priors_exnex,
    target_rates     = target_rates
  )

  j_data <- prep$j_data

  Nexch   <- length(priors_exnex$mu_mean)
  Nstrata <- length(priors_exnex$mu_j)

  expect_equal(j_data$Nexch,   Nexch)
  expect_equal(j_data$Nmix,    Nexch + 1L)
  expect_equal(j_data$Nstrata, Nstrata)

  expect_equal(j_data$mu_mean, priors_exnex$mu_mean)
  expect_equal(j_data$mu_prec, priors_exnex$mu_sd^-2)

  expect_equal(j_data$tau_HN_scale,
               rep(priors_exnex$tau_scale, Nexch))

  expect_equal(j_data$nex_mean, priors_exnex$mu_j)
  expect_equal(j_data$nex_prec, priors_exnex$tau_j^-2)

  expect_length(j_data$pMix, Nexch + 1L)
  expect_equal(sum(j_data$pMix), 1, tolerance = 1e-12)

  expect_equal(prep$j_parameters, c("p", "mu", "tau", "exch"))
  expect_true(file.exists(prep$j_model_file))
})

# ------------------------------------------------------------------
# Test: exnex_adj branch includes p_target and uses adjusted model
# Input:
#   - target_rates = c(0.5,0.6); priors from getPriorParameters("exnex_adj")
# Behaviour:
#   - j_data$p_target equals target_rates; exnex_adj JAGS model used;
#     parameters include "exch".
# Expectations:
#   - p_target set; j_parameters = c("p","mu","tau","exch");
#     model file path contains "exnex_adj".
# Why:
#   - Checks adjusted exNEX prior (centered) is wired correctly.
# ------------------------------------------------------------------
test_that("prepareAnalysis: exnex_adj branch includes p_target and uses exnex_adj model", {
  target_rates <- c(0.5, 0.6)
  priors_list  <- getPriorParameters(method_names = "exnex_adj", target_rates = target_rates)
  priors       <- priors_list[["exnex_adj"]]

  prep <- prepareAnalysis(
    method_name      = "exnex_adj",
    prior_parameters = priors,
    target_rates     = target_rates
  )

  j_data <- prep$j_data

  expect_equal(j_data$p_target, target_rates)
  expect_equal(prep$j_parameters, c("p", "mu", "tau", "exch"))
  expect_true(grepl("exnex_adj", prep$j_model_file))
  expect_true(file.exists(prep$j_model_file))
})

# ------------------------------------------------------------------
# Test: stratified and pooled branches pass priors through and use dummy JAGS info
# Input:
#   - target_rates = (0.5,0.5); priors for "stratified" and "pooled"
# Behaviour:
#   - j_data equals prior structures;
#     j_model_file and j_parameters are placeholders.
# Expectations:
#   - Dummy values returned; priors unchanged.
# Why:
#   - These methods do not require JAGS; the interface must still be consistent.
# ------------------------------------------------------------------
test_that("prepareAnalysis: stratified and pooled use dummy JAGS info and pass priors through", {
  target_rates <- c(0.5, 0.5)
  priors_list  <- getPriorParameters(
    method_names = c("stratified", "pooled"),
    target_rates = target_rates
  )

  prep_strat <- prepareAnalysis(
    method_name      = "stratified",
    prior_parameters = priors_list[["stratified"]],
    target_rates     = target_rates
  )
  expect_equal(prep_strat$j_model_file, "dummy path to JAGS model")
  expect_equal(prep_strat$j_parameters, "dummy JAGS parameters")
  expect_identical(prep_strat$j_data, priors_list[["stratified"]])

  prep_pooled <- prepareAnalysis(
    method_name      = "pooled",
    prior_parameters = priors_list[["pooled"]],
    target_rates     = target_rates
  )
  expect_equal(prep_pooled$j_model_file, "dummy path to JAGS model")
  expect_equal(prep_pooled$j_parameters, "dummy JAGS parameters")
  expect_identical(prep_pooled$j_data, priors_list[["pooled"]])
})

# ------------------------------------------------------------------
# Test: prepareAnalysis errors for unsupported method_name
# Input:
#   - method_name = "invalid_method"
# Behaviour:
#   - Guard clause should trigger with informative message.
# Expectations:
#   - Error containing "method_name must be one of".
# Why:
#   - Protects against typos / unsupported analysis methods.
# ------------------------------------------------------------------
test_that("prepareAnalysis: invalid method_name throws an error", {
  expect_error(
    prepareAnalysis(method_name = "invalid_method"),
    "method_name must be one of"
  )
})


# Tests for getUniqueRows ------------------------------------------------------

# ------------------------------------------------------------------
# Test: getUniqueRows returns unique row combinations with full column set
# Input:
#   - Matrix with duplicate rows intermixed.
# Behaviour:
#   - Removes duplicates, preserving all columns.
# Expectations:
#   - ncol(out) == ncol(input); rows correspond to unique() of data.frame(mat).
# Why:
#   - Low-level helper for deduplicating trial patterns.
# ------------------------------------------------------------------
test_that("getUniqueRows: returns unique row combinations with correct columns", {
  mat <- rbind(
    c(1, 2),
    c(1, 2),
    c(2, 3),
    c(2, 3),
    c(4, 5)
  )

  out <- getUniqueRows(mat)

  expect_equal(ncol(out), ncol(mat))

  out_df <- as.data.frame(out)
  expect_equal(nrow(out_df), nrow(unique(out_df)))

  exp_df <- as.data.frame(unique(mat))

  out_ord <- out_df[do.call(order, out_df), , drop = FALSE]
  exp_ord <- exp_df[do.call(order, exp_df), , drop = FALSE]

  expect_equal(unname(out_ord), unname(exp_ord))
})


# Tests for getUniqueTrials ----------------------------------------------------

# ------------------------------------------------------------------
# Test: getUniqueTrials combines scenarios and returns unique responder/subject/go rows
# Input:
#   - scenario_list with controlled n_responders/n_subjects and go_decisions.
# Behaviour:
#   - Stacks all trials, attaches go_flag, and returns unique rows.
# Expectations:
#   - 5 columns (2 responders + 2 subjects + 1 go_flag);
#     no duplicate rows; equal to unique() of combined matrix.
# Why:
#   - Verifies mapping from scenario_list to a unique-trial representation.
# ------------------------------------------------------------------
test_that("getUniqueTrials: combines scenarios and returns unique responder/subject/go rows", {
  set.seed(123)

  scenario_list <- simulateScenarios(
    n_subjects_list     = list(c(10, 10),
                               c(10, 10)),
    response_rates_list = list(c(0.1, 0.2),
                               c(0.1, 0.2)),
    n_trials            = 2
  )

  # scenario_1: two identical trials (1,2 | 10,10)
  scenario_list$scenario_1$n_responders <- rbind(
    c(1, 2),
    c(1, 2)
  )
  scenario_list$scenario_1$n_subjects <- rbind(
    c(10, 10),
    c(10, 10)
  )

  # scenario_2: one duplicate row (1,2 | 10,10) and one distinct row (2,1 | 10,10)
  scenario_list$scenario_2$n_responders <- rbind(
    c(1, 2),
    c(2, 1)
  )
  scenario_list$scenario_2$n_subjects <- rbind(
    c(10, 10),
    c(10, 10)
  )

  scenario_list$scenario_1$previous_analyses <- list(
    go_decisions = cbind(c(1, 0))
  )
  scenario_list$scenario_2$previous_analyses <- list(
    go_decisions = cbind(c(1, 1))
  )

  class(scenario_list) <- "scenario_list"

  out <- getUniqueTrials(scenario_list)

  expect_equal(ncol(out), 5)

  out_df <- as.data.frame(out)
  expect_equal(nrow(out_df), nrow(unique(out_df)))

  all_resp <- do.call(rbind, lapply(scenario_list, function(x) x$n_responders))
  all_subj <- do.call(rbind, lapply(scenario_list, function(x) x$n_subjects))
  all_go   <- do.call(
    rbind,
    lapply(scenario_list, function(x) x$previous_analyses$go_decisions)
  )[, 1]

  combined <- cbind(all_resp, all_subj, go_flag = all_go)

  exp_df <- as.data.frame(unique(combined))

  out_ord <- out_df[do.call(order, out_df), , drop = FALSE]
  exp_ord <- exp_df[do.call(order, exp_df), , drop = FALSE]

  expect_equal(
    unname(as.matrix(out_ord)),
    unname(as.matrix(exp_ord))
  )
})


# Tests for mapUniqueTrials ----------------------------------------------------

set.seed(123)

trial_data <- createTrial(
  n_subjects   = c(10, 10),
  n_responders = c(1, 1))

analysis_list <- performAnalyses(
  scenario_list      = trial_data,
  target_rates       = rep(0.5, 2),
  method_names       = "berry",
  n_mcmc_iterations  = 20)

trials_unique <- getUniqueTrials(trial_data)
n_cohorts     <- (ncol(trials_unique) - 1L) / 2

applicable_previous_trials <- applicablePreviousTrials(
  scenario_list    = trial_data,
  method_names     = "berry",
  quantiles        = analysis_list$scenario_1$analysis_parameters$quantiles,
  n_cohorts        = n_cohorts,
  calc_differences = NULL)

expect_false(applicable_previous_trials)

if (applicable_previous_trials) {
  calc_trial_indices <- trials_unique[, ncol(trials_unique)] > 0
} else {
  calc_trial_indices <- rep(TRUE, nrow(trials_unique))
}

trials_unique_calc <- trials_unique[calc_trial_indices, -ncol(trials_unique)]
n_responders       <- trials_unique_calc[, seq_len(n_cohorts)]
n_subjects         <- trials_unique_calc[, seq_len(n_cohorts) + n_cohorts]

method_quantiles_list        <- vector(mode = "list", length = 1)
names(method_quantiles_list) <- "berry"

prior_parameters_list <- getPriorParameters(
  method_names = "berry",
  target_rates = rep(0.5, 2))

prepare_analysis <- prepareAnalysis(
  method_name       = "berry",
  target_rates      = rep(0.5, 2),
  prior_parameters  = prior_parameters_list[["berry"]])

method_quantiles_list[["berry"]] <- getPostQuantiles(
  method_name       = "berry",
  quantiles         = analysis_list$scenario_1$analysis_parameters$quantiles,
  scenario_data     = list(n_subjects   = n_subjects,
                           n_responders = n_responders),
  calc_differences  = NULL,
  j_parameters      = prepare_analysis$j_parameters,
  j_model_file      = prepare_analysis$j_model_file,
  j_data            = prepare_analysis$j_data,
  n_mcmc_iterations = 20,
  save_path         = NULL,
  save_trial        = NULL)

scenario_method_quantiles_list <- mapUniqueTrials(
  scenario_list              = trial_data,
  method_quantiles_list      = method_quantiles_list,
  trials_unique_calc         = trials_unique_calc,
  applicable_previous_trials = applicable_previous_trials)

# ------------------------------------------------------------------
# Test: with no previous trials, mapUniqueTrials copies unique quantiles per scenario
# Input:
#   - trial_data with a single scenario and no applicable_previous_trials.
# Behaviour:
#   - All trials are treated as new; their matrices equal method_quantiles_list entries.
# Expectations:
#   - scenario_1 exists; "berry" method present; each trial's matrix identical
#     to precomputed method_quantiles_list$berry[[i]].
# Why:
#   - Baseline behaviour for the mapping when no reuse is possible.
# ------------------------------------------------------------------
test_that("mapUniqueTrials: without previous trials, maps unique trial quantiles back per scenario", {
  foreach::registerDoSEQ()

  out <- scenario_method_quantiles_list

  expect_type(out, "list")
  expect_identical(names(out), "scenario_1")

  scen1 <- out[["scenario_1"]]
  expect_true(is.list(scen1))
  expect_true("berry" %in% names(scen1))

  for (i in seq_along(method_quantiles_list$berry)) {
    expect_identical(scen1$berry[[i]], method_quantiles_list$berry[[i]])
  }
})

# ------------------------------------------------------------------
# Test: with previous trials, only GO trials are updated from hash tables
# Input:
#   - One scenario with 2 trials: trial 1 GO, trial 2 NoGo.
#   - previous_analyses$post_quantiles present; new values provided for GO pattern.
# Behaviour:
#   - GO trial gets updated quantiles; NoGo trial retains its old ones.
# Expectations:
#   - scen1$berry[[1]] == new_q_go; scen1$berry[[2]] == original old_q_list[[2]].
# Why:
#   - Validates fine-grained mapping logic between unique patterns and trials.
# ------------------------------------------------------------------
test_that("mapUniqueTrials: with previous trials, only GO trials are updated from hash tables", {
  foreach::registerDoSEQ()
  set.seed(456)

  scene <- getScenario(
    n_subjects     = c(10, 10),
    response_rates = c(0.1, 0.2),
    n_trials       = 2
  )

  trial_data <- list(scenario_1 = scene)
  class(trial_data) <- "scenario_list"
  if (is.null(trial_data$scenario_1$scenario_number)) {
    trial_data$scenario_1$scenario_number <- 1L
  }

  go_mat <- trial_data$scenario_1$previous_analyses$go_decisions
  go_mat[, 1] <- c(TRUE, FALSE)
  trial_data$scenario_1$previous_analyses$go_decisions <- go_mat

  n_trials   <- trial_data$scenario_1$n_trials
  n_cohorts  <- ncol(trial_data$scenario_1$n_subjects)
  quantiles  <- c(0.025, 0.5, 0.975)

  prior_parameters_list <- getPriorParameters(
    method_names = "berry",
    target_rates = rep(0.5, n_cohorts)
  )

  prep <- prepareAnalysis(
    method_name      = "berry",
    prior_parameters = prior_parameters_list[["berry"]],
    target_rates     = rep(0.5, n_cohorts)
  )

  old_q_list <- vector("list", length = n_trials)
  for (i in seq_len(n_trials)) {
    n_subj_i <- trial_data$scenario_1$n_subjects[i, , drop = FALSE]
    n_resp_i <- trial_data$scenario_1$n_responders[i, , drop = FALSE]

    old_q_list[[i]] <- getPostQuantiles(
      method_name       = "berry",
      quantiles         = quantiles,
      scenario_data     = list(
        n_subjects   = n_subj_i,
        n_responders = n_resp_i
      ),
      calc_differences  = NULL,
      j_parameters      = prep$j_parameters,
      j_model_file      = prep$j_model_file,
      j_data            = prep$j_data,
      n_mcmc_iterations = 20,
      save_path         = NULL,
      save_trial        = NULL
    )[[1]]
  }

  trial_data$scenario_1$previous_analyses$post_quantiles <- list(
    berry = old_q_list
  )

  trials_unique <- getUniqueTrials(trial_data)
  n_cohorts_u   <- (ncol(trials_unique) - 1L) / 2L

  applicable_previous_trials <- TRUE

  calc_trial_indices <- trials_unique[, ncol(trials_unique)] > 0
  trials_unique_calc <- trials_unique[calc_trial_indices, -ncol(trials_unique), drop = FALSE]

  new_q_go <- old_q_list[[1]] + 1

  method_quantiles_list <- list(
    berry = list(new_q_go)
  )

  out <- mapUniqueTrials(
    scenario_list              = trial_data,
    method_quantiles_list      = method_quantiles_list,
    trials_unique_calc         = trials_unique_calc,
    applicable_previous_trials = applicable_previous_trials
  )

  expect_type(out, "list")
  expect_identical(names(out), "scenario_1")

  scen1 <- out[["scenario_1"]]
  expect_true(is.list(scen1))
  expect_true("berry" %in% names(scen1))

  expect_identical(scen1$berry[[1]], new_q_go)
  expect_identical(scen1$berry[[2]], old_q_list[[2]])
})

## Test: mapUniqueTrials with pooled backend preserves naive per-trial quantiles
## Input:
##   - One scenario with several trials and 2 cohorts (generated via getScenario()).
##   - trials_unique_calc derived from getUniqueTrials(), with
##     applicable_previous_trials = FALSE so that all unique patterns are treated
##     as newly analysed.
##   - method_quantiles_list for method "pooled" containing posterior quantile
##     matrices computed only for the unique trial patterns using getPostQuantiles().
##
## Naive reference:
##   - For each original trial i (row i of n_responders, n_subjects), we call
##     getPostQuantiles(method_name = "pooled", ...) directly and store the
##     resulting single-trial quantile matrix in naive_list[[i]].
##
## Expected output:
##   - mapUniqueTrials returns a list with one element named "scenario_1".
##   - scenario_1$pooled is a list of length equal to the original number of trials.
##   - For each trial i, scenario_1$pooled[[i]] is exactly equal to
##     naive_list[[i]] (the per-trial quantile matrix from the naive analysis).
##
## Why this test:
##   - For the pooled design, getPostQuantiles() is deterministic given
##     (n_responders, n_subjects) and j_data.
##   - mapUniqueTrials must use hashing + unique trial patterns to avoid
##     redundant computations, but still reproduce the same per-trial quantiles
##     as the naive approach that analyses each trial separately.

test_that("mapUniqueTrials: pooled backend preserves naive per-trial Beta quantiles", {
  skip_if_not_installed("foreach")
  foreach::registerDoSEQ()
  set.seed(123)

  scene <- getScenario(
    n_subjects     = c(10, 20),
    response_rates = c(0.4, 0.6),
    n_trials       = 10
  )

  if (is.null(scene$scenario_number)) {
    scene$scenario_number <- 1L
  }

  scenario_list <- list(scenario_1 = scene)
  class(scenario_list) <- "scenario_list"

  n_subj   <- scene$n_subjects
  n_resp   <- scene$n_responders
  n_trials <- nrow(n_subj)
  n_coh    <- ncol(n_subj)

  trials_unique <- getUniqueTrials(scenario_list)
  n_coh_u       <- (ncol(trials_unique) - 1L) / 2L
  expect_equal(n_coh_u, n_coh)

  applicable_previous_trials <- FALSE

  calc_trial_indices <- rep(TRUE, nrow(trials_unique))
  trials_unique_calc <- trials_unique[calc_trial_indices, -ncol(trials_unique), drop = FALSE]

  n_resp_unique <- trials_unique_calc[, seq_len(n_coh),         drop = FALSE]
  n_subj_unique <- trials_unique_calc[, seq_len(n_coh) + n_coh, drop = FALSE]

  j_data    <- list(a = 1, b = 1)
  quantiles <- c(0.025, 0.5, 0.975)

  unique_quantiles <- getPostQuantiles(
    method_name       = "pooled",
    quantiles         = quantiles,
    scenario_data     = list(
      n_subjects   = n_subj_unique,
      n_responders = n_resp_unique
    ),
    calc_differences  = NULL,
    j_parameters      = NULL,
    j_model_file      = NULL,
    j_data            = j_data,
    n_mcmc_iterations = 20,
    save_path         = NULL,
    save_trial        = NULL
  )

  naive_list <- vector("list", length = n_trials)
  for (i in seq_len(n_trials)) {
    n_subj_i <- n_subj[i, , drop = FALSE]
    n_resp_i <- n_resp[i, , drop = FALSE]

    naive_list[[i]] <- getPostQuantiles(
      method_name       = "pooled",
      quantiles         = quantiles,
      scenario_data     = list(
        n_subjects   = n_subj_i,
        n_responders = n_resp_i
      ),
      calc_differences  = NULL,
      j_parameters      = NULL,
      j_model_file      = NULL,
      j_data            = j_data,
      n_mcmc_iterations = 20,
      save_path         = NULL,
      save_trial        = NULL
    )[[1]]
  }

  method_quantiles_list <- list(
    pooled = unique_quantiles
  )

  out <- mapUniqueTrials(
    scenario_list              = scenario_list,
    method_quantiles_list      = method_quantiles_list,
    trials_unique_calc         = trials_unique_calc,
    applicable_previous_trials = applicable_previous_trials
  )

  expect_type(out, "list")
  expect_identical(names(out), "scenario_1")

  scen1_out <- out[["scenario_1"]]
  expect_true(is.list(scen1_out))
  expect_true("pooled" %in% names(scen1_out))
  expect_equal(length(scen1_out$pooled), n_trials)

  for (i in seq_len(n_trials)) {
    expect_equal(
      scen1_out$pooled[[i]],
      naive_list[[i]],
      tolerance = 1e-12
    )
  }
})

# Tests for posteriors2Quantiles ----------------------------------------------

# ------------------------------------------------------------------
# Test: posteriors2Quantiles computes requested quantiles, mean and SD
# Input:
#   - Normal samples theta ~ N(2, 3^2); 5000 draws.
# Behaviour:
#   - Returns a matrix with rows: quantiles, Mean, SD.
# Expectations:
#   - Row/column names as specified; values close to empirical summaries.
# Why:
#   - Core summarisation helper for posterior draws.
# ------------------------------------------------------------------
test_that("posteriors2Quantiles: computes quantiles, mean, and sd for each column", {
  set.seed(123)

  theta <- rnorm(5000, mean = 2, sd = 3)
  post  <- cbind(theta = theta)

  quantiles <- c(0.25, 0.5, 0.75)

  out <- posteriors2Quantiles(
    quantiles  = quantiles,
    posteriors = post
  )

  expect_true(is.matrix(out))
  expect_identical(colnames(out), "theta")
  expect_identical(rownames(out),
                   c("25%", "50%", "75%", "Mean", "SD"))

  exp_q <- stats::quantile(theta, probs = quantiles)
  expect_equal(out[c("25%", "50%", "75%"), "theta"],
               exp_q,
               tolerance = 1e-12)

  expect_equal(out["Mean", "theta"], mean(theta), tolerance = 1e-12)
  expect_equal(out["SD",   "theta"], stats::sd(theta), tolerance = 1e-12)
})


# Tests for performAnalyses ----------------------------------------------------

# Shared scenario_list for performAnalyses tests
scenario_list_pa <- make_scenario_list_pa()

prior_parameters_list <- getPriorParameters(
  method_names = c("pooled", "exnex"),
  target_rates = c(0.5, 0.5, 0.5)
)

# ------------------------------------------------------------------
# Test: performAnalyses returns a well-formed analysis_list
# Input:
#   - scenario_list_pa; methods pooled + exnex; supplied priors.
# Behaviour:
#   - One analysis entry per scenario with quantiles_list, scenario_data,
#     and analysis_parameters.
# Expectations:
#   - S3 class "analysis_list"; names "scenario_k"; scenario_data unchanged.
# Why:
#   - High-level sanity check of analysis pipeline.
# ------------------------------------------------------------------
test_that("performAnalyses returns a well-formed analysis_list", {
  res <- performAnalyses(
    scenario_list         = scenario_list_pa,
    target_rates          = c(0.5, 0.5, 0.5),
    method_names          = c("pooled", "exnex"),
    prior_parameters_list = prior_parameters_list,
    n_mcmc_iterations     = 20,
    verbose               = FALSE
  )

  expect_s3_class(res, "analysis_list")
  expect_equal(length(res), length(scenario_list_pa))
  expect_identical(names(res), paste0("scenario_", sapply(scenario_list_pa, `[[`, "scenario_number")))

  scen1 <- res[[1]]
  expect_true(is.list(scen1))
  expect_true(all(c("quantiles_list", "scenario_data", "analysis_parameters") %in% names(scen1)))

  expect_identical(scen1$scenario_data, scenario_list_pa[[1]])
})

# ------------------------------------------------------------------
# Test: performAnalyses sorts method_names and stores in analysis_parameters
# Input:
#   - method_names in unsorted order: stratified, berry, pooled.
# Behaviour:
#   - Internally sorts method_names; quantiles_list and analysis_parameters
#     follow sorted order.
# Expectations:
#   - scen1$analysis_parameters$method_names sorted; quantiles_list names match.
# Why:
#   - Ensures deterministic method ordering across analyses.
# ------------------------------------------------------------------
test_that("performAnalyses sorts method_names and stores them in analysis_parameters", {
  res <- performAnalyses(
    scenario_list       = scenario_list_pa,
    target_rates        = c(0.5, 0.5, 0.5),
    method_names        = c("stratified", "berry", "pooled"),
    n_mcmc_iterations   = 20,
    verbose             = FALSE
  )

  scen1 <- res[[1]]

  expected_sorted <- sort(c("stratified", "berry", "pooled"))
  expect_identical(scen1$analysis_parameters$method_names, expected_sorted)

  expect_true(is.list(scen1$quantiles_list))
  expect_identical(names(scen1$quantiles_list), expected_sorted)
})

# ------------------------------------------------------------------
# Test: performAnalyses builds quantiles from defaults + evidence_levels
# Input:
#   - evidence_levels = c(0.1,0.2,0.3)
# Behaviour:
#   - analysis_parameters$quantiles is union of reversed evidence_levels
#     and default quantiles, sorted.
# Expectations:
#   - Numeric, sorted; contains all default and evidence-based quantiles.
# Why:
#   - Confirms evidence_levels are incorporated into posterior summaries.
# ------------------------------------------------------------------
test_that("performAnalyses constructs quantiles from defaults and evidence_levels", {
  ev_levels <- c(0.1, 0.2, 0.3)

  res <- performAnalyses(
    scenario_list       = scenario_list_pa,
    evidence_levels     = ev_levels,
    target_rates        = c(0.5, 0.5, 0.5),
    method_names        = c("pooled"),
    n_mcmc_iterations   = 20,
    verbose             = FALSE
  )

  q <- res[[1]]$analysis_parameters$quantiles
  expect_true(is.numeric(q))
  expect_false(is.unsorted(q))

  defaults <- c(0.025, 0.05, 0.5, 0.8, 0.9, 0.95, 0.975)
  expected_q <- sort(unique(round(1 - c(defaults, ev_levels), 9)))

  expect_true(all(expected_q %in% q))
})

# ------------------------------------------------------------------
# Test: performAnalyses auto-fills prior_parameters_list when omitted
# Input:
#   - method_names = c("berry","pooled"), prior_parameters_list = NULL.
# Behaviour:
#   - Internally calls getPriorParameters and stores result in analysis_parameters.
# Expectations:
#   - Non-null priors; list; contains both methods as names.
# Why:
#   - Convenience behaviour for users not providing priors explicitly.
# ------------------------------------------------------------------
test_that("performAnalyses fills prior_parameters_list when not supplied", {
  methods <- c("berry", "pooled")

  res <- performAnalyses(
    scenario_list         = scenario_list_pa,
    target_rates          = c(0.5, 0.5, 0.5),
    method_names          = methods,
    prior_parameters_list = NULL,
    n_mcmc_iterations     = 20,
    verbose               = FALSE
  )

  priors <- res[[1]]$analysis_parameters$prior_parameters_list

  expect_false(is.null(priors))
  expect_true(is.list(priors))
  expect_true(all(methods %in% names(priors)))
})

# ------------------------------------------------------------------
# Test: performAnalyses prints a progress message when verbose = TRUE
# Input:
#   - verbose = TRUE
# Behaviour:
#   - Progress message "Performing Analyses" printed to console.
# Expectations:
#   - expect_message detects that substring.
# Why:
#   - Confirms user-facing verbosity option works.
# ------------------------------------------------------------------
test_that("performAnalyses prints a progress message when verbose = TRUE", {
  expect_message(
    performAnalyses(
      scenario_list       = scenario_list_pa,
      target_rates        = c(0.5, 0.5, 0.5),
      method_names        = c("pooled"),
      n_mcmc_iterations   = 20,
      verbose             = TRUE
    ),
    "Performing Analyses"
  )
})


# Tests for loadAnalyses -------------------------------------------------------

# ------------------------------------------------------------------
# Test: loadAnalyses reads saved analyses and sets class/names
# Input:
#   - Two RDS files matching analysis_data_1_1 and analysis_data_2_2.
# Behaviour:
#   - Reads into an analysis_list, named scenario_1, scenario_2.
# Expectations:
#   - S3 class "analysis_list"; names c("scenario_1","scenario_2");
#     contents identical to original dummies.
# Why:
#   - Confirms correct file naming convention and object reconstruction.
# ------------------------------------------------------------------
test_that("loadAnalyses: loads saved analyses and sets class/names correctly", {
  tmpdir <- tempdir()

  scen_nums <- c(1, 2)
  anal_nums <- c(1, 2)

  dummy1 <- list(foo = 1)
  dummy2 <- list(bar = 2)

  saveRDS(dummy1, file = file.path(tmpdir, "analysis_data_1_1.rds"))
  saveRDS(dummy2, file = file.path(tmpdir, "analysis_data_2_2.rds"))

  loaded <- loadAnalyses(
    scenario_numbers = scen_nums,
    analysis_numbers = anal_nums,
    load_path        = tmpdir
  )

  expect_s3_class(loaded, "analysis_list")
  expect_identical(names(loaded), c("scenario_1", "scenario_2"))

  expect_identical(loaded[[1]], dummy1)
  expect_identical(loaded[[2]], dummy2)
})

# ------------------------------------------------------------------
# Test: loadAnalyses validates scenario_numbers as positive integers
# Input:
#   - scenario_numbers = "a" or c(0,1).
# Behaviour:
#   - Non-integer or non-positive scenario_numbers should fail.
# Expectations:
#   - Errors mentioning "scenario_numbers".
# Why:
#   - Enforces basic integrity on index arguments.
# ------------------------------------------------------------------
test_that("loadAnalyses: scenario_numbers must be positive integers", {
  tmpdir <- tempdir()

  expect_error(
    loadAnalyses(scenario_numbers = "a", load_path = tmpdir),
    "scenario_numbers"
  )

  expect_error(
    loadAnalyses(scenario_numbers = c(0, 1), load_path = tmpdir),
    "scenario_numbers"
  )
})

# ------------------------------------------------------------------
# Test: loadAnalyses validates analysis_numbers length and positivity
# Input:
#   - analysis_numbers length mismatched to scenario_numbers or non-positive.
# Behaviour:
#   - Should error with appropriate message.
# Expectations:
#   - Errors mentioning "analysis_numbers".
# Why:
#   - Prevents ambiguous mapping between scenarios and analyses.
# ------------------------------------------------------------------
test_that("loadAnalyses: analysis_numbers must be positive integers of same length", {
  tmpdir <- tempdir()

  expect_error(
    loadAnalyses(
      scenario_numbers = c(1, 2),
      analysis_numbers = c(1),
      load_path        = tmpdir
    ),
    "analysis_numbers"
  )

  expect_error(
    loadAnalyses(
      scenario_numbers = c(1, 2),
      analysis_numbers = c(1, -1),
      load_path        = tmpdir
    ),
    "analysis_numbers"
  )
})

# ------------------------------------------------------------------
# Test: loadAnalyses requires load_path be a single character string
# Input:
#   - load_path = 123 or c("a","b").
# Behaviour:
#   - Type and length checks must fail.
# Expectations:
#   - Errors mentioning "load_path".
# Why:
#   - Guards against invalid path inputs.
# ------------------------------------------------------------------
test_that("loadAnalyses: load_path must be a single character string", {
  expect_error(
    loadAnalyses(scenario_numbers = 1, load_path = 123),
    "load_path"
  )
  expect_error(
    loadAnalyses(scenario_numbers = 1, load_path = c("a", "b")),
    "load_path"
  )
})


# Tests for printAnalyses ------------------------------------------------------

# ------------------------------------------------------------------
# Test: print.analysis_list prints header, scenario blocks, labels, and estimates
# Input:
#   - analyses from pooled method on 2 scenarios; estimates via getEstimates().
# Behaviour:
#   - print() shows header, per-scenario listing, "Pooled" label,
#     and numeric summary values like mean and SD.
# Expectations:
#   - Output contains correct header; scenario lines; method label;
#     and numeric values matching rounded estimates.
# Why:
#   - Verifies user-facing summary of analysis_list.
# ------------------------------------------------------------------
test_that("print.analysis_list: prints header, scenario blocks, method label, and numeric estimates", {
  set.seed(123)

  scen <- simulateScenarios(
    n_subjects_list     = list(c(10, 20),
                               c(10, 20)),
    response_rates_list = list(c(0.3, 0.6),
                               c(0.4, 0.7)),
    n_trials            = 50
  )

  analyses <- performAnalyses(
    scenario_list      = scen,
    method_names       = "pooled",
    target_rates       = c(0.5, 0.5),
    n_mcmc_iterations  = 20,
    verbose            = FALSE
  )

  est_list <- getEstimates(analyses)

  expect_true(is.list(est_list[[1]]))
  est_mat1 <- est_list[[1]][[1]]

  mean_p1 <- round(est_mat1["p_1", "Mean"], digits = 2)
  sd_p1   <- round(est_mat1["p_1", "SD"],   digits = 2)

  out <- capture.output(print(analyses))
  flat_out <- paste(out, collapse = " ")

  expect_true(
    any(grepl("analysis_list of 2 scenarios with 1 method", out))
  )

  expect_true(
    any(grepl("^  - scenario_1", out))
  )
  expect_true(
    any(grepl("^  - scenario_2", out))
  )

  expect_true(
    any(grepl("Pooled", out))
  )

  mean_str <- sprintf("%.2f", mean_p1)
  sd_str   <- sprintf("%.2f", sd_p1)

  expect_true(
    grepl(mean_str, flat_out)
  )
  expect_true(
    grepl(sd_str, flat_out)
  )

  expect_true(
    any(grepl("MCMC iterationns per BHM method", out))
  )
  expect_true(
    any(grepl("Available evidence levels:", out))
  )
})

# ------------------------------------------------------------------
# Test: print.analysis_list shows distinct scenario-specific estimates
# Input:
#   - Two scenarios with different response rates, pooled method.
# Behaviour:
#   - Printed output must include different means for p_1 per scenario.
# Expectations:
#   - Output includes both rounded means; these means differ numerically.
# Why:
#   - Ensures print summarises scenario-specific posterior differences.
# ------------------------------------------------------------------
test_that("print.analysis_list: multiple scenarios print distinct scenario-specific estimates", {
  set.seed(456)

  n_subj <- c(10, 20)
  rr1    <- c(0.3, 0.6)
  rr2    <- c(0.7, 0.2)

  scen_multi <- simulateScenarios(
    n_subjects_list     = list(n_subj, n_subj),
    response_rates_list = list(rr1, rr2),
    n_trials            = 30
  )

  analyses_multi <- performAnalyses(
    scenario_list      = scen_multi,
    method_names       = "pooled",
    target_rates       = c(0.5, 0.5),
    n_mcmc_iterations  = 20,
    verbose            = FALSE
  )

  est_multi <- getEstimates(analyses_multi)

  expect_true(is.list(est_multi[[1]]))
  m1 <- est_multi[[1]][[1]]
  m2 <- est_multi[[1]][[2]]

  mean1 <- round(m1["p_1", "Mean"], 2)
  mean2 <- round(m2["p_1", "Mean"], 2)

  out <- capture.output(print(analyses_multi))
  flat_out <- paste(out, collapse = " ")

  expect_true(
    any(grepl("analysis_list of 2 scenarios with 1 method", out))
  )

  expect_true(any(grepl("^  - scenario_1", out)))
  expect_true(any(grepl("^  - scenario_2", out)))

  mean1_str <- sprintf("%.2f", mean1)
  mean2_str <- sprintf("%.2f", mean2)

  expect_true(
    grepl(mean1_str, flat_out)
  )
  expect_true(
    grepl(mean2_str, flat_out)
  )

  expect_false(
    isTRUE(all.equal(mean1, mean2))
  )
})

# ------------------------------------------------------------------
# Test: digits argument controls numeric precision in printed output
# Input:
#   - digits = 2 and digits = 4; same analyses object.
# Behaviour:
#   - Numeric summaries printed with specified precision.
# Expectations:
#   - Output with digits=2 contains 2-decimal format; digits=4 contains 4-decimal format.
# Why:
#   - Confirms users can tune displayed precision.
# ------------------------------------------------------------------
test_that("print.analysis_list: digits argument controls printed numeric precision", {
  set.seed(789)

  scen <- simulateScenarios(
    n_subjects_list     = list(c(10, 20),
                               c(10, 20)),
    response_rates_list = list(c(0.33, 0.66),
                               c(0.33, 0.66)),
    n_trials            = 40
  )

  analyses <- performAnalyses(
    scenario_list      = scen,
    method_names       = "pooled",
    target_rates       = c(0.5, 0.5),
    n_mcmc_iterations  = 20,
    verbose            = FALSE
  )

  est_list <- getEstimates(analyses)
  est_obj  <- est_list[[1]]
  expect_true(is.list(est_obj))
  est_mat1 <- est_obj[[1]]

  mean_p1 <- est_mat1["p_1", "Mean"]

  out_2 <- capture.output(print(analyses, digits = 2))
  out_4 <- capture.output(print(analyses, digits = 4))

  flat_2 <- paste(out_2, collapse = " ")
  flat_4 <- paste(out_4, collapse = " ")

  fmt2 <- sprintf("%.2f", round(mean_p1, 2))
  fmt4 <- sprintf("%.4f", round(mean_p1, 4))

  expect_true(
    grepl(fmt2, flat_2)
  )

  expect_true(
    grepl(fmt4, flat_4)
  )
})


# Tests for saveAnalyses -------------------------------------------------------

# ------------------------------------------------------------------
# Test: saveAnalyses writes analysis_list to disk and loadAnalyses reads it back
# Input:
#   - analyses from pooled method, multiple scenarios; temp directory.
# Behaviour:
#   - saveAnalyses returns path and scenario/analysis numbers;
#     files exist; loadAnalyses reproduces original object.
# Expectations:
#   - info list correctly populated; RDS files present;
#     loaded analysis_list identical in length/names to original.
# Why:
#   - Round-trip test for persistence layer.
# ------------------------------------------------------------------
test_that("saveAnalyses: saves analysis_list to disk and loadAnalyses can read it back", {
  set.seed(101)

  scen <- simulateScenarios(
    n_subjects_list     = list(c(10, 20),
                               c(15, 25)),
    response_rates_list = list(c(0.3, 0.6),
                               c(0.5, 0.8)),
    n_trials            = 20
  )

  analyses <- performAnalyses(
    scenario_list      = scen,
    method_names       = "pooled",
    target_rates       = c(0.5, 0.5),
    n_mcmc_iterations  = 20,
    verbose            = FALSE
  )

  tmp_dir <- tempdir()

  info <- saveAnalyses(analyses, save_path = tmp_dir)

  expect_true(is.list(info))
  expect_equal(length(info$scenario_numbers), length(analyses))
  expect_equal(length(info$analysis_numbers), length(analyses))
  expect_identical(info$path, tmp_dir)

  internal_scen <- vapply(
    analyses,
    function(x) x$scenario_data$scenario_number,
    FUN.VALUE = numeric(1)
  )
  expect_equal(info$scenario_numbers, internal_scen)

  for (i in seq_along(info$scenario_numbers)) {
    f <- file.path(
      tmp_dir,
      paste0("analysis_data_",
             info$scenario_numbers[i], "_",
             info$analysis_numbers[i], ".rds")
    )
    expect_true(file.exists(f))
  }

  loaded <- loadAnalyses(
    scenario_numbers = info$scenario_numbers,
    analysis_numbers = info$analysis_numbers,
    load_path        = info$path
  )

  expect_s3_class(loaded, "analysis_list")
  expect_identical(length(loaded), length(analyses))
  expect_identical(names(loaded), names(analyses))
})

# ------------------------------------------------------------------
# Test: saveAnalyses requires an object of class 'analysis_list'
# Input:
#   - list(a = 1) instead of analysis_list.
# Behaviour:
#   - Class check must fail with informative message.
# Expectations:
#   - Error mentioning "analysis_list".
# Why:
#   - Prevents misuse with arbitrary lists.
# ------------------------------------------------------------------
test_that("saveAnalyses: non-analysis_list input triggers a class error", {
  expect_error(
    saveAnalyses(list(a = 1)),
    "analysis_list",
    ignore.case = TRUE
  )
})

# ------------------------------------------------------------------
# Test: saveAnalyses requires save_path to be a length-1 character vector
# Input:
#   - save_path = c("a","b").
# Behaviour:
#   - Argument validation should fail.
# Expectations:
#   - Error mentioning "save_path".
# Why:
#   - Enforces a single destination directory.
# ------------------------------------------------------------------
test_that("saveAnalyses: save_path must be a character vector of length 1", {
  set.seed(202)

  scen <- simulateScenarios(
    n_subjects_list     = list(c(10, 20)),
    response_rates_list = list(c(0.3, 0.6)),
    n_trials            = 10
  )

  analyses <- performAnalyses(
    scenario_list      = scen,
    method_names       = "pooled",
    target_rates       = c(0.5, 0.5),
    n_mcmc_iterations  = 20,
    verbose            = FALSE
  )

  expect_error(
    saveAnalyses(analyses, save_path = c("a", "b")),
    "save_path",
    ignore.case = TRUE
  )
})

# ------------------------------------------------------------------
# Test: RBesT comparison for mixture models
# Input:
#   - Fixed 2-component beta-mixture prior from RBesT
#   - Two identical cohorts with same data
# Behaviour:
#   - stratified_mix should match RBesT closely (same beta-mixture updating)
#   - exnex_mix and exnex_adj_mix use a logit-normal NEX mixture derived
#     from the same RBesT prior via rmix() + normal-mixture fitting
#   - identical cohorts should produce identical posteriors
# Expectations:
#   - stratified_mix close to RBesT
#   - exnex_mix / exnex_adj_mix finite and symmetric across the 2 cohorts
# Why:
#   - Validates the three implemented mixture branches against a common
#     external reference prior and checks structural symmetry.
# ------------------------------------------------------------------
test_that("mixture branches agree with RBesT reference prior and preserve symmetry", {
  skip_if_not_installed("RBesT")

  set.seed(123)

  ## symmetric 2-cohort setup
  n <- c(20, 20)
  r <- c(6, 6)

  trial_data <- createTrial(
    n_subjects   = n,
    n_responders = r
  )

  ## fixed RBesT beta-mixture prior
  w <- c(0.8, 0.2)

  a_one <- c(4, 1)
  b_one <- c(12, 1)

  a_j <- cbind(a_one, a_one)
  b_j <- cbind(b_one, b_one)

  ## 1) stratified_mix vs RBesT
  pp_strat_mix <- setPriorParametersStratifiedMix(
    w   = w,
    a_j = a_j,
    b_j = b_j
  )

  res_strat <- performAnalyses(
    scenario_list         = trial_data,
    method_names          = "stratified_mix",
    prior_parameters_list = pp_strat_mix,
    n_mcmc_iterations     = 2000,
    verbose               = FALSE
  )

  bhm_strat <- res_strat[[1]]$quantiles_list$stratified_mix[[1]][
    c("2.5%", "50%", "97.5%", "Mean", "SD"),
    c("p_1", "p_2"),
    drop = FALSE
  ]
  rownames(bhm_strat) <- c("q025", "q500", "q975", "Mean", "SD")

  prior_rbest <- RBesT::mixbeta(
    c(w[1], a_one[1], b_one[1]),
    c(w[2], a_one[2], b_one[2])
  )

  post_rbest <- RBesT::postmix(prior_rbest, n = n[1], r = r[1])
  post_rbest_summary <- summary(post_rbest)

  rbest_ref <- c(
    q025 = unname(post_rbest_summary[" 2.5%"]),
    q500 = unname(post_rbest_summary["50.0%"]),
    q975 = unname(post_rbest_summary["97.5%"]),
    Mean = unname(post_rbest_summary["mean"]),
    SD   = unname(post_rbest_summary["sd"])
  )
  
  bhm_strat_combined <- (bhm_strat[, "p_1"] + bhm_strat[, "p_2"]) / 2
  
  expect_equal(
    unname(bhm_strat_combined),
    unname(rbest_ref),
    tolerance = 0.01
  )

  ## 2) derive logit-normal mixture for exnex*
  p_draws <- RBesT::rmix(prior_rbest, 20000)

  eps <- 1e-8
  p_draws <- pmin(pmax(p_draws, eps), 1 - eps)
  eta_draws <- qlogis(p_draws)

  mix_fit <- RBesT::automixfit(eta_draws)

  w_nex_derived <- as.numeric(mix_fit["w", ])
  mean_one      <- as.numeric(mix_fit["m", ])
  sd_one        <- as.numeric(mix_fit["s", ])

  expect_equal(sum(w_nex_derived), 1, tolerance = 1e-8)
  expect_true(all(sd_one > 0))

  mean_nex <- cbind(mean_one, mean_one)
  sd_nex   <- cbind(sd_one, sd_one)

  ## 3) exnex_mix with EX off
  pp_exnex_mix <- setPriorParametersExNex(
    mu_mean   = 0,
    mu_sd     = 1,
    tau_scale = 0.5,
    w_j       = 0,
    w_nex     = w_nex_derived,
    mean_nex  = mean_nex,
    sd_nex    = sd_nex
  )
  names(pp_exnex_mix) <- "exnex_mix"

  res_exnex_mix <- performAnalyses(
    scenario_list         = trial_data,
    method_names          = "exnex_mix",
    prior_parameters_list = pp_exnex_mix,
    n_mcmc_iterations     = 1e4,
    verbose               = FALSE
  )

  bhm_exnex_mix <- res_exnex_mix[[1]]$quantiles_list$exnex_mix[[1]][
    c("2.5%", "50%", "97.5%", "Mean", "SD"),
    c("p_1", "p_2"),
    drop = FALSE
  ]
  rownames(bhm_exnex_mix) <- c("q025", "q500", "q975", "Mean", "SD")

  expect_true(all(is.finite(bhm_exnex_mix)))
  
  bhm_exnex_mix_combined <- (bhm_exnex_mix[, "p_1"] +
                               bhm_exnex_mix[, "p_2"]) / 2

  expect_equal(
    unname(bhm_exnex_mix_combined),
    unname(rbest_ref),
    tolerance = 0.01
  )

  ## 4) exnex_adj_mix with EX off
  pp_exnex_adj_mix <- setPriorParametersExNexAdj(
    mu_mean   = 0,
    mu_sd     = 1,
    tau_scale = 0.5,
    w_j       = 0,
    w_nex     = w_nex_derived,
    mean_nex  = mean_nex,
    sd_nex    = sd_nex
  )
  names(pp_exnex_adj_mix) <- "exnex_adj_mix"

  res_exnex_adj_mix <- performAnalyses(
    scenario_list         = trial_data,
    method_names          = "exnex_adj_mix",
    target_rates          = c(0.5, 0.5),
    prior_parameters_list = pp_exnex_adj_mix,
    n_mcmc_iterations     = 1e4,
    verbose               = FALSE
  )

  bhm_exnex_adj_mix <- res_exnex_adj_mix[[1]]$quantiles_list$exnex_adj_mix[[1]][
    c("2.5%", "50%", "97.5%", "Mean", "SD"),
    c("p_1", "p_2"),
    drop = FALSE
  ]
  rownames(bhm_exnex_adj_mix) <- c("q025", "q500", "q975", "Mean", "SD")

  expect_true(all(is.finite(bhm_exnex_adj_mix)))
  
  bhm_exnex_adj_mix_combined <- (bhm_exnex_adj_mix[, "p_1"] +
                                   bhm_exnex_adj_mix[, "p_2"]) / 2

  expect_equal(
    unname(bhm_exnex_adj_mix_combined),
    unname(rbest_ref),
    tolerance = 0.01
  )

})

#' @title continueRecruitment
#' @md
#' @description This function continues the recruitment of subjects for a set of scenarios
#' based on the Go / NoGo decisions in the simulated trial outcomes of said scenarios.
#' @param n_subjects_add_list A list that contains for each scenario an integer vector for
#' the number of subjects per cohort to be additionally recruited.
#' @param decisions_list A list with decisions per scenario created with
#' \code{\link[bhmbasket]{getGoDecisions}}
#' @param method_name A string for the method name of the analysis the decisions are based on.
#' Can be `NULL` if only one method has been used for analysis, Default: `NULL`
#' @return An object of class `scenario_list` with the scenario data for each specified scenario.
#' @details
#' This function is intended to be used for analyses with the following work flow:\cr
#' `simulateScenarios()` -> `performAnalyses()` -> `getGoDecisions()`-> \cr
#' `continueRecruitment()` -> `performAnalyses()` -> `getGoDecisions()`-> \cr
#' `continueRecruitment()` -> ...
#'
#' Note that `n_subjects_add_list` takes the additional number of subjects to be recruited,
#' not the overall number of subjects.
#' This way the work flow can be repeated as often as
#' required, which can be useful e.g. for interim analyses.
#' @examples
#' interim_scenarios <- simulateScenarios(
#'   n_subjects_list     = list(c(10, 20, 30)),
#'   response_rates_list = list(rep(0.9, 3)),
#'   n_trials            = 10)
#'
#' interim_analyses <- performAnalyses(
#'   scenario_list       = interim_scenarios,
#'   target_rates        = rep(0.5, 3),
#'   n_mcmc_iterations   = 100)
#'
#' interim_gos <- getGoDecisions(
#'   analyses_list       = interim_analyses,
#'   cohort_names        = c("p_1", "p_2", "p_3"),
#'   evidence_levels     = c(0.5, 0.8, 0.5),
#'   boundary_rules      = quote(c(x[1] > 0.8, x[2] > 0.6, x[3] > 0.7)))
#'
#' scenarios_list <- continueRecruitment(
#'   n_subjects_add_list = list(c(30, 20, 10)),
#'   decisions_list      = interim_gos,
#'   method_name         = "exnex_adj")
#' @seealso
#'  \code{\link[bhmbasket]{simulateScenarios}}
#'  \code{\link[bhmbasket]{performAnalyses}}
#'  \code{\link[bhmbasket]{getGoDecisions}}
#' @rdname continueRecruitment
#' @author Stephan Wojciekowski
#' @export
continueRecruitment <- function (

  n_subjects_add_list,
  decisions_list,

  method_name = NULL

) {

  error_n_subjects_add_list <-
    "Providing a list of vectors of positive integers for the argument 'n_subjects_add_list'"
  error_decisions_list <-
    "Providing an object of class decision_list for the argument 'decisions_list'"
  error_method_name <- simpleError(paste(
    "Please provide a string naming an analysis method for the argument 'method_name'",
    "Must be one of 'berry', 'exnex', 'exnex_mix', 'exnex_adj', 'exnex_adj_mix', 'pooled', 'stratified', 'stratified_mix'"))

  checkmate::assert_true(!missing(n_subjects_add_list), .var.name = error_n_subjects_add_list)

  checkmate::assert_true(!missing(decisions_list),      .var.name = error_decisions_list)

  checkmate::assert_class(decisions_list, "decision_list", .var.name = error_decisions_list)

  if (is.null(method_name)) {

    n_methods <- length(decisions_list$scenario_1$decisions_list)

    if (n_methods > 1) {

      stop (error_method_name)

    } else {

      method_name <- names(decisions_list$scenario_1$decisions_list)

    }

  } else {

    method_name <- tryCatch({

      match.arg(
        method_name,
        choices    = c('berry', 'exnex', 'exnex_mix', 'exnex_adj', 'exnex_adj_mix', 'pooled', 'stratified', 'stratified_mix'),
        several.ok = FALSE)

    }, error = function (e) e)

    if (inherits(method_name, "error"))          stop (error_method_name)

  }

  if (!is.list(n_subjects_add_list)) {

    n_subjects_add_list <- rep(list(n_subjects_add_list), length(decisions_list))

  }


  checkmate::assert_list(

    n_subjects_add_list, types = c("integer", "numeric"),
    any.missing = FALSE,
    .var.name = error_n_subjects_add_list
  )

  checkmate::assert_list(

    n_subjects_add_list, len = length(decisions_list), any.missing = FALSE,
    .var.name = "The lengths of 'n_subjects_add_list' and 'decisions_list' must be equal"
  )

  checkmate::assert_true(

    all(vapply(n_subjects_add_list,
               checkmate::test_integerish,
               logical(1),
               lower = 0, any.missing = FALSE)
    ),
    .var.name = error_n_subjects_add_list
  )

  ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ###

  scenario_numbers <- as.numeric(sub("scenario_", "", names(decisions_list)))

  scenario_list <- vector(mode = "list", length = length(decisions_list))
  names(scenario_list) <- paste0("scenario_", scenario_numbers)
  for (s in seq_along(scenario_list)) {

    if (!(method_name %in%
          decisions_list[[s]]$analysis_data$analysis_parameters$method_names)) {

      stop (simpleError("Selected method_name not analyzed"))

    }

    n_subjects_add <- n_subjects_add_list[[s]]
    response_rates <- decisions_list[[s]]$scenario_data$response_rates
    cohort_names   <- sub("rr_", "", colnames(response_rates))

    if (any(response_rates > 0 & response_rates < 1)) {

      index_new <- which(response_rates > 0 & response_rates < 1)

    } else {

      stop (simpleError(paste0(
        "Only historical cohorts in scenario ",
        decisions_list[[s]]$scenario_data$scenario_number)))

    }

    response_rates_new <- response_rates[, index_new]
    cohort_names_new   <- cohort_names[index_new]


    checkmate::assert_true(
      length(n_subjects_add) == length(response_rates_new),
      .var.name = "The length of n_subjects_add must be equal ,
      to the length of the response rates that are in (0, 1)"
    )

    n_trials <- decisions_list[[s]]$scenario_data$n_trials

    add_scenario <- getScenario(
      n_subjects     = n_subjects_add,
      response_rates = response_rates_new,
      cohort_names   = cohort_names_new,
      n_trials       = n_trials)

    go_decisions <- decisions_list[[s]]$decisions_list[[method_name]]
    previous_gos <- go_decisions

    if ("overall" %in% colnames(go_decisions)) {
      overall_gos  <- go_decisions[, which(colnames(go_decisions) == "overall")]
      go_decisions <- go_decisions[, -which(colnames(go_decisions) == "overall")]
    } else {
      overall_gos <- rep(TRUE, nrow(go_decisions))
    }
    if (!all(index_new %in% as.numeric(sub("decision_", "", colnames(go_decisions))))) {
      stop (simpleError(
        "There must be a decision for each recruiting cohort in the 'decisions_list'"))
    }

    go_decisions <- go_decisions[overall_gos, index_new]

    n_responders_add <- add_scenario$n_responders[overall_gos, ] * go_decisions
    n_subjects_add   <- add_scenario$n_subjects[overall_gos, ] * go_decisions

    n_responders <- decisions_list[[s]]$scenario_data$n_responders
    n_subjects   <- decisions_list[[s]]$scenario_data$n_subjects

    n_responders[overall_gos, index_new] <- n_responders[overall_gos, index_new] + n_responders_add
    n_subjects[overall_gos, index_new]   <- n_subjects[overall_gos, index_new] + n_subjects_add


    scenario_list[[s]] <- list(
      n_subjects        = n_subjects,
      n_responders      = n_responders,
      response_rates    = response_rates,
      previous_analyses = list(
        go_decisions   = previous_gos,
        post_quantiles = decisions_list[[s]]$analysis_data$quantiles_list),
      n_trials          = n_trials)

    scenario_list[[s]]$scenario_number <-
      decisions_list[[s]]$scenario_data$scenario_number

  }

  class(scenario_list) <- "scenario_list"

  return (scenario_list)

}

#' @title createTrial
#' @description This function creates an object of class `scenario_list`
#' for a single trial outcome, which can subsequently be analyzed with other functions of
#' `bhmbasket`, e.g. \code{\link[bhmbasket]{performAnalyses}}
#' @param n_subjects A vector of integers for the number of subjects in the trial outcome
#' @param n_responders A vector of integers for the number of responders in the trial outcome
#' @return An object of class `scenario_list` with the scenario data for a single trial outcome.
#' @details This function is a wrapper for \code{\link[bhmbasket]{simulateScenarios}} with
#' ```
#' simulateScenarios(
#'  n_subjects_list     = list(n_subjects),
#'  response_rates_list = list(n_responders),
#'  n_trials            = 1)
#' ```
#' @seealso
#'  \code{\link[bhmbasket]{simulateScenarios}}
#'  \code{\link[bhmbasket]{performAnalyses}}
#' @author Stephan Wojciekowski
#' @rdname createTrial
#' @examples
#'  trial_outcome <- createTrial(n_subjects   = c(10, 20, 30, 40),
#'                               n_responders = c( 1,  2,  3,  4))
#' @export
#' @md
createTrial <- function (

  n_subjects,
  n_responders

) {

  error_n_subjects <-
    "Providing a vector of integers for the argument 'n_subjects'"
  error_n_responders <-
    "Providing a vector of integers for the argument 'n_responders'"

  checkmate::assert_integerish(n_subjects, any.missing = FALSE, min.len = 1,
                               .var.name = error_n_subjects)
  checkmate::assert_integerish(n_responders, any.missing = FALSE, min.len = 1,
                               .var.name = error_n_responders)

  ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ###

  utils::capture.output({
    trial <- simulateScenarios(
      n_subjects_list     = list(as.integer(n_subjects)),
      response_rates_list = list(as.integer(n_responders)),
      n_trials            = 1
    )
  })

  return (trial)

}

getNSubjects <- function (

  recruitment_per_month,
  start_date,
  analysis_dates,

  date_format = "%m/%d/%Y"

) {

  recruitment_per_day <- recruitment_per_month * (12 / 365)

  start_date     <- as.Date(start_date, format = date_format)
  analysis_dates <- as.Date(analysis_dates, format = date_format)

  n_subjects_matrix <- floor((analysis_dates - start_date) %o% recruitment_per_day)

  rownames(n_subjects_matrix) <- format(analysis_dates, format = date_format)
  colnames(n_subjects_matrix) <- paste0("cohort_", seq_len(ncol(n_subjects_matrix)))

  return (n_subjects_matrix)

}

getRecruitment <- function (

  n_subjects_required,
  recruitment_per_month,
  start_date,

  date_format = "%m/%d/%Y"

) {


  checkmate::assert_numeric(

    n_subjects_required, lower = 0, any.missing = FALSE,
  .var.name = "Providing a matrix of non-negative integers for the argument 'n_subjects_required'"
  )

  checkmate::assert_true(

    checkmate::test_integerish(n_subjects_required, lower = 0),
    .var.name = "Providing a matrix of non-negative integers for the argument 'n_subjects_required'"
    )

  checkmate::assert_numeric(

    recruitment_per_month, lower = 0, any.missing = FALSE,
    .var.name = "Providing a vector of non-negative numerics for the argument 'recruitment_per_month'")

  checkmate::assert_string(

    start_date, pattern = ".+",
    .var.name = "Please provide a string in the format 'date_format' of for the argument 'start_date'")

  ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ###

  hist_index <- recruitment_per_month == 0

  recruitment_per_day <- recruitment_per_month[!hist_index] * (12 / 365)

  start_date <- as.Date(start_date, format = date_format)

  n_subjects_required <- convertVector2Matrix(n_subjects_required)

  checkmate::assert_true(
    length(recruitment_per_month) == ncol(n_subjects_required))

  days_required <- apply(convertVector2Matrix(n_subjects_required[, !hist_index]),
                         1, function (n_subj) {
                           max(ceiling(n_subj * recruitment_per_day^-1))
                         })

  n_subjects_matrix_0 <- getNSubjects(
    recruitment_per_month = recruitment_per_month[!hist_index],
    start_date            = start_date,
    analysis_dates        = start_date + days_required,
    date_format           = date_format)

  n_subjects_matrix <- matrix(NA,
                              ncol = ncol(n_subjects_required),
                              nrow = nrow(n_subjects_required))

  n_subjects_matrix[, !hist_index] <- n_subjects_matrix_0
  n_subjects_matrix[, hist_index]  <- n_subjects_required[, hist_index]

  rownames(n_subjects_matrix) <- rownames(n_subjects_matrix_0)
  colnames(n_subjects_matrix) <- paste0("cohort_", seq_len(ncol(n_subjects_required)))

  return (n_subjects_matrix)

}

getResponders <- function (

  response_rates,
  n_subjects,

  n_trials

) {

  n_responders <- t(replicate(n = n_trials, {

    stats::rbinom(n    = length(n_subjects), # n_cohorts
                  size = n_subjects,
                  prob = response_rates)

  }))

  return (n_responders)

}

getScenario <- function(n_subjects, response_rates,
                        cohort_names = seq_along(n_subjects), n_trials = 1e4
) {

  checkmate::assert_integerish(n_subjects, lower = 0, any.missing = FALSE)

  checkmate::assert_numeric(response_rates, any.missing = FALSE)

  checkmate::assert_integerish(n_trials, lower = 1, any.missing = FALSE, len = 1)

  checkmate::assert_true(length(cohort_names) == length(n_subjects),
                         .var.name = "cohort_names must match length of n_subjects")

  ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ###

  response_rates            <- convertVector2Matrix(response_rates)
  cohort_names_chr          <- as.character(cohort_names)
  colnames(response_rates)  <- paste0("rr_", cohort_names_chr)

  new_cohorts  <- FALSE
  hist_cohorts <- FALSE

  # New cohorts: 0 < rr < 1
  if (any(response_rates < 1 & response_rates > 0)) {
    new_cohorts <- TRUE
    index_new   <- which(response_rates < 1 & response_rates > 0)

    # Match original check (length on sliced matrix/vector)
    checkmate::assert_true(

      length(n_subjects[index_new]) == length(response_rates[, index_new]),

      .var.name = "n_subjects and response_rates must have same length"

    )

    n_responders <- getResponders(
      response_rates = response_rates[, index_new, drop = FALSE],
      n_subjects     = n_subjects[index_new],
      n_trials       = n_trials
    )
  }

  # Historical cohorts: rr >= 1 OR rr == 0
  if (any(response_rates >= 1 | response_rates == 0)) {
    hist_cohorts <- TRUE
    index_hist   <- which(response_rates >= 1 | response_rates == 0)

    checkmate::assert_true(

      length(n_subjects[index_hist]) == length(response_rates[, index_hist]),

      .var.name = "n_subjects and response_rates must have same length"

    )

    if (new_cohorts) {
      n_responders_hist <- matrix(
        rep(response_rates[index_hist], each = nrow(n_responders)),
        nrow = nrow(n_responders)
      )
    } else {
      n_responders_hist <- matrix(response_rates[index_hist], nrow = 1)
    }
  }

  if (isTRUE(new_cohorts) && isTRUE(hist_cohorts)) {

    n_responders <- cbind(n_responders, n_responders_hist)

  } else if (isTRUE(hist_cohorts) && !isTRUE(new_cohorts)) {

    n_responders <- n_responders_hist

  }

  previous_gos <- matrix(
    TRUE, byrow = TRUE,
    ncol = length(n_subjects) + 1L,
    nrow = nrow(n_responders)
  )
  colnames(previous_gos) <- c("overall", paste0("decision_", cohort_names_chr))

  n_subjects_mat <- matrix(
    n_subjects, byrow = TRUE,
    ncol = length(n_subjects),
    nrow = nrow(n_responders)
  )

  colnames(n_subjects_mat) <- paste0("n_", cohort_names_chr)
  colnames(n_responders)   <- paste0("r_", cohort_names_chr)

  scenario_data <- list(
    n_subjects        = n_subjects_mat,
    n_responders      = n_responders,
    response_rates    = response_rates,
    previous_analyses = list(
      go_decisions   = previous_gos,
      post_quantiles = NULL
    ),
    n_trials          = n_trials
  )

  return(scenario_data)
}

is.scenario_list <- function (x) {
  if (missing(x)) stop ("Please provide an object for the argument 'x'")
  inherits(x, "scenario_list")
}

#' @title loadScenarios
#' @md
#' @description This function loads scenarios saved with \code{\link[bhmbasket]{saveScenarios}}
#' @param scenario_numbers A vector of integers naming the scenario to be loaded
#' @param load_path A string for the directory where the scenarios are being stored,
#' Default: \code{\link[base]{tempfile}}
#' @return Returns an object of class `scenario_list`
#' @rdname loadScenarios
#' @author Stephan Wojciekowski
#' @seealso
#'  \code{\link[bhmbasket]{simulateScenarios}}
#'  \code{\link[bhmbasket]{saveScenarios}}
#'  \code{\link[base]{tempfile}}
#' @examples
#'   scenarios_list <- simulateScenarios(
#'     n_subjects_list     = list(c(10, 20, 30)),
#'     response_rates_list = list(rep(0.9, 3)),
#'     n_trials            = 10)
#'
#'   save_info      <- saveScenarios(scenarios_list)
#'   scenarios_list <- loadScenarios(scenario_numbers = save_info$scenario_numbers,
#'                                   load_path        = save_info$path)
#' @export
loadScenarios <- function (

  scenario_numbers,
  load_path = tempdir()

) {

  checkmate::assert_integerish(
    scenario_numbers,
    lower = 1, any.missing = FALSE,
    .var.name = "Providing a vector of positive integers for the argument 'scenario_numbers'"
  )

  checkmate::assert_string(load_path, .var.name = "Providing a string for the argument 'load_path'")

  ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ###

  files <- file.path(load_path, paste0("scenario_data_", scenario_numbers, ".rds"))
  scenario_list <- lapply(files, readRDS)

  names(scenario_list) <- paste0("scenario_", scenario_numbers)
  class(scenario_list) <- "scenario_list"

  return (scenario_list)

}

#' @export
print.scenario_list <- function(x, ...) {

  n_scenarios    <- length(x)
  scenario_names <- names(x)

  n_cohorts      <- length(x[[1]]$response_rates)
  cohort_names   <- paste0("c_", seq_len(n_cohorts))

  response_rates <- lapply(x, function (x) x$response_rates)
  n_subjects     <- getAverageNSubjects(x)

  n_trial_realizations  <- x[[1]]$n_trials
  n_unique_realizations <- nrow(getUniqueTrials(x))

  cat("scenario_list of ", n_scenarios, " scenario", ifelse(n_scenarios == 1, "", "s"),
      " with ", n_cohorts, " cohort", ifelse(n_cohorts == 1, "", "s"),"\n\n", sep = "")
  for (n in seq_along(scenario_names)) {

    df_out <- t(data.frame(c(response_rates[[n]]),
                           n_subjects[[n]]))
    rownames(df_out) <- c("    - true response rates:",
                          "    - average number of subjects:")
    colnames(df_out) <- cohort_names

    cat("  -", scenario_names[n], "\n")
    print(df_out)

    cat("\n")

  }

  cat("  -", n_trial_realizations, "trial realizations per scenario\n")
  cat("  -", n_unique_realizations, "unique trial realizations overall\n")

}

#' @title saveScenarios
#' @md
#' @description Saves the scenario data in a newly created or existing directory
#' @param scenario_list An object of class `scenario_list`, e.g. created with `simulateScenarios()`
#' @param save_path A string providing the path of the directory in which the scenario data
#' should be saved, Default: \code{\link[base]{tempfile}}
#' @return A named list of length 2 with the scenario numbers and the `save_path`
#' @author Stephan Wojciekowski
#' @seealso
#'  \code{\link[bhmbasket]{simulateScenarios}}
#'  \code{\link[bhmbasket]{loadScenarios}}
#'  \code{\link[base]{tempfile}}
#' @examples
#'   scenarios_list <- simulateScenarios(
#'     n_subjects_list     = list(c(10, 20, 30)),
#'     response_rates_list = list(rep(0.9, 3)),
#'     n_trials            = 10)
#'
#'   save_info      <- saveScenarios(scenarios_list)
#'   scenarios_list <- loadScenarios(scenario_numbers = save_info$scenario_numbers,
#'                                   load_path        = save_info$path)
#' @rdname saveScenarios
#' @export
saveScenarios <- function (

  scenario_list,
  save_path = tempdir()

) {

  checkmate::assert(
    checkmate::check_class(scenario_list, "scenario_list"),
    checkmate::check_character(save_path, len = 1),
    combine = "and",
    .var.name = "Please provide an object of class scenario_list for the argument 'scenario_list'"
  )

  ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ###

  if (!dir.exists(save_path)) {
    dir.create(save_path)
  }

  scenario_numbers <- sapply(scenario_list, function (x) x$scenario_number)

  for (s in seq_along(scenario_list)) {
    saveRDS(
      scenario_list[[s]],
      file = paste0(save_path, "/scenario_data_", scenario_numbers[s], ".rds")
    )
  }

  return (list(scenario_numbers = scenario_numbers, path = save_path))
}

#' @title simulateScenarios
#' @description This function creates scenarios for the analysis with
#' \code{\link[bhmbasket]{performAnalyses}}.
#' @param n_subjects_list A list that contains for each scenario a vector for
#' the number of subjects per cohort.
#' A single vector can be provided if all scenarios should have the same number of subjects.
#' @param response_rates_list A list that contains for each scenario a vector for
#' the response rates per cohort.
#' @param scenario_numbers A vector of positive integers naming the scenarios,
#' Default: `seq_along(response_rates_list)`
#' @param n_trials An integer indicating the number of trial simulations per response rates,
#' Default: `10000`. If `n_trials` is present in `.GlobalEnv` and `missing(n_trials)`,
#' the globally available value will be used.
#' @return An object of class `scenario_list` with the scenario data for each specified scenario.
#' @details The function simulates trials with binary outcome for each scenario.
#' Integer values for the response rates will be treated as observed outcomes.
#' @author Stephan Wojciekowski
#' @seealso
#'  \code{\link[bhmbasket]{saveScenarios}}
#'  \code{\link[bhmbasket]{createTrial}}
#'  \code{\link[bhmbasket]{performAnalyses}}
#' @rdname simulateScenarios
#' @examples
#'   n_subjects     <- c(10, 20, 30)
#'
#'   rr_negative    <- rep(0.1, 3)
#'   rr_nugget      <- c(0.9, 0.1, 0.1)
#'   rr_positive    <- rep(0.9, 3)
#'
#'   scenarios_list <- simulateScenarios(
#'     n_subjects_list     = list(n_subjects,
#'                                n_subjects,
#'                                n_subjects),
#'     response_rates_list = list(rr_negative,
#'                                rr_nugget,
#'                                rr_positive))
#' @export
#' @md
simulateScenarios <- function (
    n_subjects_list,
    response_rates_list,
    scenario_numbers = seq_along(response_rates_list),
    n_trials         = 1e4
) {

  error_n_subjects_list     <-
    "Providing a list of vectors of positive integers for the argument 'n_subjects_list'"
  error_response_rates_list <-
    paste("Providing a list of vectors of non-negative numerics for the argument",
          "'response_rates_list'\n", "Values outside of (0, 1) must be integers")
  error_n_trials            <-
    "Providing a positive integer for the argument 'n_trials'"
  error_scenario_numbers    <-
    "Providing a vector of positive integers for the argument 'scenario_numbers'"

  checkmate::assert_list(
    response_rates_list,
    types       = "numeric",
    any.missing = FALSE,
    .var.name   = error_response_rates_list
  )

  if (!is.list(n_subjects_list)) {
    n_subjects_list <- rep(list(n_subjects_list), length(response_rates_list))
  }

  checkmate::assert_list(
    n_subjects_list,
    types       = "numeric",
    any.missing = FALSE,
    .var.name   = error_n_subjects_list
  )

  checkmate::assert_true(
    all(vapply(
      n_subjects_list,
      checkmate::test_integerish,
      logical(1),
      lower        = 1,
      any.missing  = FALSE
    )),
    .var.name = error_n_subjects_list
  )

  checkmate::assert_integerish(
    scenario_numbers,
    lower       = 1,
    any.missing = FALSE,
    .var.name   = error_scenario_numbers
  )

  checkmate::assert_true(
    length(scenario_numbers) == length(n_subjects_list),
    .var.name = "'scenario_numbers' and 'n_subjects_list' must have same length"
  )

  checkmate::assert_true(
    length(n_subjects_list) == length(response_rates_list),
    .var.name = "'n_subjects_list' and 'response_rates_list' must have same length"
  )

  n_cohorts <- length(response_rates_list[[1L]])

  checkmate::assert_true(
    n_cohorts >= 2L,
    .var.name = "Each scenario having at least 2 cohorts"
  )

  cohort_lengths_rr <- vapply(response_rates_list, length, integer(1))
  checkmate::assert_true(
    all(cohort_lengths_rr == n_cohorts),
    .var.name = "All scenarios having same number of cohorts in 'response_rates_list'"
  )

  cohort_lengths_ns <- vapply(n_subjects_list, length, integer(1))
  checkmate::assert_true(
    all(cohort_lengths_ns == n_cohorts),
    .var.name = "All scenarios having same number of cohorts in 'n_subjects_list'"
  )

  for (rates in response_rates_list) {

    is_whole <- abs(rates - round(rates)) < .Machine$double.eps^0.5

    ok <- (rates > 0 & rates < 1) |
      (is_whole & (rates == 0 | rates >= 1))

    checkmate::assert_true(
      all(ok),
      .var.name = error_response_rates_list
    )
  }

  for (i in seq_along(n_subjects_list)) {
    checkmate::assert_true(
      all(n_subjects_list[[i]] >= response_rates_list[[i]]),
      .var.name = "Values in 'response_rates_list' must not exceed 'n_subjects_list'"
    )
  }


  if ("n_trials" %in% ls(envir = .GlobalEnv) && missing(n_trials)) {
    n_trials <- get("n_trials", envir = .GlobalEnv)
  }

  checkmate::assert_count(
    n_trials,
    positive = TRUE,
    .var.name = error_n_trials
  )

  ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ###

  scenario_list <- vector(mode = "list", length = length(scenario_numbers))

  for (s in seq_along(scenario_numbers)) {
    scenario_list[[s]] <- getScenario(
      n_subjects     = n_subjects_list[[s]],
      response_rates = response_rates_list[[s]],
      n_trials       = n_trials
    )
    scenario_list[[s]]$scenario_number <- scenario_numbers[s]
  }

  names(scenario_list) <- paste0("scenario_", scenario_numbers)
  class(scenario_list) <- "scenario_list"

  return(scenario_list)
}

## 1. Helper scenario_list and analysis_list for decision tests
## ------------------------------------------------------------------

helper_n_subjects <- c(10, 10)
helper_rr         <- c(0.5, 0.5)

helper_scenarios <- simulateScenarios(
  n_subjects_list     = list(helper_n_subjects),
  response_rates_list = list(helper_rr),
  scenario_numbers    = 1L,
  n_trials            = 5L
)

helper_analyses <- performAnalyses(
  scenario_list     = helper_scenarios,
  target_rates      = c(0.5, 0.5),
  method_names      = "berry",
  n_mcmc_iterations = 100,
  verbose           = FALSE
)

decisions <- getGoDecisions(
  analyses_list   = helper_analyses,
  cohort_names    = c("p_1", "p_2"),
  evidence_levels = c(0.5, 0.5),
  boundary_rules  = quote(c(TRUE, TRUE)),
  overall_min_gos = 1L
)

decisions_list <- decisions


# Tests for simulateScenarios --------------------------------------------------

# ------------------------------------------------------------------
# Test: basic structure and handling of historical cohorts
# Input:
#   - n_subjects = c(10,20,30), multiple response rate vectors (rr_1..rr_4),
#     n_trials = 25 for 4 scenarios.
# Behaviour:
#   - simulateScenarios() returns a scenario_list with expected matrices,
#     names, dimensions, and fixed responders for historical cohorts
#     (rr <= 0 or >= 1).
# Expectations:
#   - Class "scenario_list" with names scenario_1..scenario_4.
#   - n_subjects, response_rates, n_responders are matrices.
#   - Column names are n_*, r_*, rr_*.
#   - For rr_3 historical columns, n_responders are constant across trials.
#   - rr_5 outside [0,1] triggers an error.
# Why:
#   - Verifies core construction and the special handling of historical arms.
# ------------------------------------------------------------------
test_that("simulateScenarios has correct structure", {
  n_subjects <- c(10, 20, 30)
  rr_1 <- c(0.1, 0.1, 0.1)
  rr_2 <- c(0.9, 0.9, 0.9)
  rr_3 <- c(0.9, 0.9, 1)
  rr_4 <- c(0.9, 0.9, 0)
  rr_5 <- c(0.9, 0.9, 1.1)

  scenarios <- simulateScenarios(
    n_subjects_list     = list(n_subjects, n_subjects, n_subjects, n_subjects),
    response_rates_list = list(rr_1, rr_2, rr_3, rr_4),
    scenario_numbers    = c(1, 2, 3, 4),
    n_trials            = 25
  )

  expect_s3_class(scenarios, "scenario_list")
  expect_equal(names(scenarios), paste0("scenario_", 1:4))

  s1 <- scenarios[[1]]
  s3 <- scenarios[[3]]

  expect_true(is.matrix(s1$n_subjects))
  expect_true(is.matrix(s1$response_rates))
  expect_equal(s1$n_trials, 25)

  hist_cols <- which(rr_3 <= 0 | rr_3 >= 1)

  expect_true(
    all(apply(
      s3$n_responders[, hist_cols, drop = FALSE],
      2,
      function(x) length(unique(x)) == 1
    ))
  )

  expect_equal(ncol(s1$n_subjects), length(n_subjects))
  expect_equal(ncol(s1$response_rates), length(n_subjects))

  expect_equal(
    colnames(s1$n_subjects),
    paste0("n_", seq_len(length(n_subjects)))
  )

  expect_equal(
    colnames(s1$n_responders),
    paste0("r_", seq_len(length(n_subjects)))
  )

  expect_equal(
    colnames(s1$response_rates),
    paste0("rr_", seq_len(length(n_subjects)))
  )

  expect_error(
    simulateScenarios(
      n_subjects_list     = list(n_subjects),
      response_rates_list = list(rr_5),
      scenario_numbers    = 5,
      n_trials            = 5
    )
  )
})


# ------------------------------------------------------------------
# Test: argument validation, list consistency, and default n_trials
# Input:
#   - Various invalid n_trials, mismatched list lengths, cohort lengths,
#     missing arguments, and a global n_trials in .GlobalEnv.
# Behaviour:
#   - simulateScenarios() should:
#     • error on missing required args,
#     • error on invalid n_trials values,
#     • error on length mismatches / cohort mismatches,
#     • accept vector n_subjects_list (recycled to list),
#     • use .GlobalEnv$n_trials when argument is omitted.
# Expectations:
#   - Correct errors mentioning relevant argument names.
#   - Positive cases return scenario_list.
#   - When n_trials omitted and global n_trials = 7L, output n_trials = 7L.
# Why:
#   - Ensures robust front-end validation and backward-compatible defaults.
# ------------------------------------------------------------------
test_that("simulateScenarios has no mismatch and gives appropriate error messages", {
  n_subjects <- list(c(10, 20, 30))
  rr_1       <- c(0.1, 0.1, 0.1)
  rr_ok      <- c(0.2, 0.3, 0.4)

  expect_true(checkmate::test_int(5L, lower = 1))

  expect_error(
    simulateScenarios(n_subjects_list = n_subjects),
    "response_rates_list"
  )

  expect_error(
    simulateScenarios(response_rates_list = list(rr_1)),
    "n_subjects_list"
  )

  expect_s3_class(
    simulateScenarios(
      n_subjects_list     = c(10, 20, 30),
      response_rates_list = list(rr_1, rr_1),
      n_trials            = 2
    ),
    "scenario_list"
  )

  invalid_trials <- list(
    zero     = 0L,
    negative = -1L,
    nonint   = 2.5,
    na       = NA_integer_,
    nan      = NaN,
    inf      = Inf,
    ninf     = -Inf,
    length2  = c(2L, 3L)
  )

  for (case in invalid_trials) {
    expect_error(
      simulateScenarios(
        n_subjects_list     = list(c(10, 20, 30)),
        response_rates_list = list(rr_1),
        n_trials            = case
      ),
      "n_trials",
      info = paste("invalid n_trials case:", paste(case, collapse = ","))
    )
  }

  expect_error(
    simulateScenarios(
      n_subjects_list     = list(c(10, 20, 30)),
      response_rates_list = list(rr_1, rr_1)
    ),
    "same length"
  )

  expect_error(
    simulateScenarios(
      n_subjects_list     = list(c(10, 20, 30), c(10, 20, 30)),
      response_rates_list = list(c(0.1, 0.2, 0.3), c(0.2, 0.3, 0.4)),
      scenario_numbers    = c(1)
    ),
    "scenario_numbers"
  )

  expect_error(
    simulateScenarios(
      n_subjects_list     = list(10),
      response_rates_list = list(0.1),
      n_trials            = 2
    ),
    "at least 2 cohorts"
  )

  expect_error(
    simulateScenarios(
      n_subjects_list     = list(c(10, 20), c(10, 20)),
      response_rates_list = list(c(0.1, 0.2), 0.3),
      n_trials            = 2
    ),
    "same number of cohorts"
  )

  expect_error(
    simulateScenarios(
      n_subjects_list     = list(c(10, 20), 10),
      response_rates_list = list(c(0.1, 0.2), c(0.1, 0.3)),
      n_trials            = 2
    ),
    "same number of cohorts"
  )

  old <- if (exists("n_trials", .GlobalEnv)) get("n_trials", .GlobalEnv) else NULL

  on.exit({
    if (is.null(old)) {
      rm(list = "n_trials", envir = .GlobalEnv)
    } else {
      assign("n_trials", old, .GlobalEnv)
    }
  }, add = TRUE)

  assign("n_trials", 7L, envir = .GlobalEnv)

  out <- simulateScenarios(
    n_subjects_list     = list(c(10, 20, 30)),
    response_rates_list = list(rr_ok)
  )

  expect_equal(out[[1]]$n_trials, 7L)
})

# Tests for saveScenarios ------------------------------------------------------

# ------------------------------------------------------------------
# Test: saveScenarios writes RDS files and returns metadata
# Input:
#   - scenario_list with 2 simple mock scenarios (scenario_number 1 and 2),
#     save_path = temp directory.
# Behaviour:
#   - For each scenario, saveScenarios() writes a scenario_data_<num>.rds file
#     and returns a list with the scenario_numbers and the path used.
# Expectations:
#   - Files scenario_data_1.rds and scenario_data_2.rds exist.
#   - Returned scenario_numbers == c(1,2) and path == temp_dir.
# Why:
#   - Checks that serialization of scenarios and meta-return are correct.
# ------------------------------------------------------------------
test_that("saveScenarios saves files correctly", {
  mock_scenario <- function(num) {
    structure(
      list(scenario_number = num, data = paste("Scenario", num)),
      class = "scenario"
    )
  }
  scenario_list <- list(mock_scenario(1), mock_scenario(2))
  class(scenario_list) <- "scenario_list"

  temp_dir <- tempfile()
  dir.create(temp_dir)

  result <- saveScenarios(scenario_list, save_path = temp_dir)

  expect_true(file.exists(file.path(temp_dir, "scenario_data_1.rds")))
  expect_true(file.exists(file.path(temp_dir, "scenario_data_2.rds")))

  expect_equal(result$scenario_numbers, c(1, 2))
  expect_equal(result$path, temp_dir)

  unlink(temp_dir, recursive = TRUE)
})


# ------------------------------------------------------------------
# Test: saveScenarios creates directory if it does not exist
# Input:
#   - scenario_list with one scenario, save_path = non-existent temp path.
# Behaviour:
#   - Function must create the directory, save the file and return path.
# Expectations:
#   - dir.exists(save_path) becomes TRUE.
#   - Returned path equals the requested save_path.
# Why:
#   - Ensures convenience behaviour when user passes a new directory.
# ------------------------------------------------------------------
test_that("saveScenarios creates directory when save_path does not exist", {
  temp_dir <- tempfile()
  expect_false(dir.exists(temp_dir))
  on.exit(unlink(temp_dir, recursive = TRUE), add = TRUE)

  sl <- structure(list(list(scenario_number = 1)), class = "scenario_list")

  res <- saveScenarios(sl, save_path = temp_dir)

  expect_true(dir.exists(temp_dir))
  expect_equal(res$path, temp_dir)
})

# Tests for print.scenario_list ------------------------------------------------

# ------------------------------------------------------------------
# Test: print.scenario_list header shows scenario and cohort count
# Input:
#   - scenario_list with 2 scenarios, each with 2 cohorts.
# Behaviour:
#   - print.scenario_list() prints a header with counts plus scenario names.
# Expectations:
#   - Output contains “scenario_list of 2 scenarios with 2 cohorts”.
#   - Output mentions scenario_1 and scenario_2.
# Why:
#   - Verifies the informative summary header and scenario labels.
# ------------------------------------------------------------------
test_that("print.scenario_list header is correct and shows cohort count", {
  n_subjects <- c(5, 6)
  rr1 <- c(0.1, 0.2)
  rr2 <- c(0.3, 0.4)

  scenarios <- simulateScenarios(
    n_subjects_list     = list(n_subjects, n_subjects),
    response_rates_list = list(rr1, rr2),
    scenario_numbers    = c(1, 2),
    n_trials            = 3
  )

  expect_output(print(scenarios), "scenario_list of 2 scenarios with 2 cohorts")
  expect_output(print(scenarios), "  - scenario_1", fixed = FALSE)
  expect_output(print(scenarios), "  - scenario_2", fixed = FALSE)
})


# ------------------------------------------------------------------
# Test: print.scenario_list prints table section labels
# Input:
#   - scenario_list with 1 scenario and 3 cohorts.
# Behaviour:
#   - print.scenario_list() prints labeled rows for “true response rates”
#     and “average number of subjects”.
# Expectations:
#   - Output contains those labels.
# Why:
#   - Confirms that the summary table describes what is being shown.
# ------------------------------------------------------------------
test_that("print.scenario_list prints table row labels", {
  n_subjects <- c(7, 8, 9)
  rr <- c(0.2, 0.3, 0.4)

  scenarios <- simulateScenarios(
    n_subjects_list     = list(n_subjects),
    response_rates_list = list(rr),
    scenario_numbers    = 1,
    n_trials            = 2
  )

  expect_output(print(scenarios), "true response rates:")
  expect_output(print(scenarios), "average number of subjects:")
})


# ------------------------------------------------------------------
# Test: print.scenario_list uses cohort column names c_1, c_2, ...
# Input:
#   - scenario_list with 1 scenario and 3 cohorts.
# Behaviour:
#   - Columns in the printed tables are labeled c_1, c_2, c_3.
# Expectations:
#   - Output contains literal c_1, c_2, c_3 as word boundaries.
# Why:
#   - Ensures cohort naming scheme is user-friendly and consistent.
# ------------------------------------------------------------------
test_that("print.scenario_list uses cohort column names c_1, c_2, ...", {
  n_subjects <- c(7, 8, 9)
  rr <- c(0.2, 0.3, 0.4)

  scenarios <- simulateScenarios(
    n_subjects_list     = list(n_subjects),
    response_rates_list = list(rr),
    scenario_numbers    = 1,
    n_trials            = 2
  )

  expect_output(print(scenarios), "\\bc_1\\b")
  expect_output(print(scenarios), "\\bc_2\\b")
  expect_output(print(scenarios), "\\bc_3\\b")
})


# ------------------------------------------------------------------
# Test: print.scenario_list footer shows realizations info
# Input:
#   - scenario_list with n_trials = 5.
# Behaviour:
#   - Footer prints number of realizations per scenario and overall.
# Expectations:
#   - Output mentions “5 trial realizations per scenario”.
#   - Output mentions “unique trial realizations overall”.
# Why:
#   - Provides context for how many simulations underpin the summaries.
# ------------------------------------------------------------------
test_that("print.scenario_list prints footer lines with realizations info", {
  n_subjects <- c(10, 10)
  rr <- c(0.1, 0.2)

  scenarios <- simulateScenarios(
    n_subjects_list     = list(n_subjects),
    response_rates_list = list(rr),
    scenario_numbers    = 1,
    n_trials            = 5
  )

  expect_output(print(scenarios), "5 trial realizations per scenario")
  expect_output(print(scenarios), "unique trial realizations overall")
})

# Tests for loadScenarios ------------------------------------------------------

# ------------------------------------------------------------------
# Test: loadScenarios round-trip with saveScenarios
# Input:
#   - Two scenarios generated via simulateScenarios(), saved to temp_dir.
# Behaviour:
#   - loadScenarios() reads back the RDS files into a scenario_list.
# Expectations:
#   - Returned object has class "scenario_list".
#   - Loaded object equals the original scenario_list.
# Why:
#   - Validates persistence and recovery of scenario definitions.
# ------------------------------------------------------------------
test_that("loadScenarios loads saved scenarios and returns proper class/names", {
  temp_dir <- tempfile()
  dir.create(temp_dir)

  scenario_list_to_save <- simulateScenarios(
    n_subjects_list     = list(c(10, 20, 30), c(10, 20, 30)),
    response_rates_list = list(c(0.1, 0.2, 0.3), c(0.5, 0.6, 0.7)),
    scenario_numbers    = c(1, 2),
    n_trials            = 5
  )

  saveScenarios(scenario_list_to_save, save_path = temp_dir)

  loaded <- loadScenarios(c(1, 2), load_path = temp_dir)

  expect_s3_class(loaded, "scenario_list")
  expect_equal(loaded, scenario_list_to_save)

  unlink(temp_dir, recursive = TRUE, force = TRUE)
})


# ------------------------------------------------------------------
# Test: loadScenarios validates load_path argument
# Input:
#   - Incorrect load_path values: vector, numeric, NULL, NA_character_.
# Behaviour:
#   - loadScenarios() should error when load_path is not a single valid string.
# Expectations:
#   - Each invalid call errors with message mentioning 'load_path'.
# Why:
#   - Ensures the file location argument is properly validated.
# ------------------------------------------------------------------
test_that("loadScenarios points out invalid paths", {
  expect_error(loadScenarios(1L, load_path = c(tempdir(), tempdir())), "load_path")
  expect_error(loadScenarios(1L, load_path = 123), "load_path")
  expect_error(loadScenarios(1L, load_path = NULL), "load_path")
  expect_error(loadScenarios(1L, load_path = NA_character_), "load_path")
  # Optional: empty string behaviour may be handled elsewhere
  # expect_error(loadScenarios(1L, load_path = ""), "load_path")
})


# ------------------------------------------------------------------
# Test: loadScenarios errors when requested files are missing
# Input:
#   - Valid directory without any scenario_data_*.rds files,
#     scenario_numbers = c(999).
# Behaviour:
#   - loadScenarios() attempts to read missing files and should error.
# Expectations:
#   - Error is thrown.
# Why:
#   - Confirms that missing files are not silently ignored.
# ------------------------------------------------------------------
test_that("loadScenarios errors when files do not exist", {
  temp_dir <- tempfile()
  dir.create(temp_dir)

  expect_error(
    suppressWarnings(loadScenarios(c(999), load_path = temp_dir))
  )

  unlink(temp_dir, recursive = TRUE, force = TRUE)
})

# Tests for is.scenario_list ---------------------------------------------------

# ------------------------------------------------------------------
# Test: is.scenario_list TRUE for valid scenario_list and FALSE otherwise
# Input:
#   - Valid scenario_list (proper class and scenario elements),
#     plus several invalid objects.
# Behaviour:
#   - is.scenario_list() returns TRUE only for the valid scenario_list.
# Expectations:
#   - TRUE for valid_list.
#   - FALSE for list(), character, NULL.
# Why:
#   - Ensures the helper detects correct class/structure.
# ------------------------------------------------------------------
test_that("is.scenario_list returns TRUE for valid input and FALSE otherwise", {
  mock_scenario <- function(num) {
    structure(list(scenario_number = num), class = "scenario")
  }
  valid_list <- list(mock_scenario(1), mock_scenario(2))
  class(valid_list) <- "scenario_list"

  expect_true(is.scenario_list(valid_list))

  expect_false(is.scenario_list(list()))
  expect_false(is.scenario_list("not a scenario list"))
  expect_false(is.scenario_list(NULL))
})


# ------------------------------------------------------------------
# Test: is.scenario_list errors when x is missing
# Input:
#   - Call is.scenario_list() with no argument.
# Behaviour:
#   - Function should throw an informative error about missing x.
# Expectations:
#   - Error message mentions "argument 'x'" (or similar custom message).
# Why:
#   - Enforces explicit use and avoids accidental calls without input.
# ------------------------------------------------------------------
test_that("is.scenario_list errors when argument x is missing", {
  expect_error(
    is.scenario_list(),
    "Please provide an object for the argument 'x'"
  )
})

# Tests for createTrial --------------------------------------------------------

# ------------------------------------------------------------------
# Test: createTrial returns a trial object for valid input
# Input:
#   - n_subjects = c(10,20), n_responders = c(5,15), with simulateScenarios
#     mocked to avoid heavy computation.
# Behaviour:
#   - createTrial() should call simulateScenarios() internally and
#     return a non-NULL trial object.
# Expectations:
#   - Result is not NULL (shape-specific checks may be done elsewhere).
# Why:
#   - Confirms the wrapper-path from trial inputs to a scenario object.
# ------------------------------------------------------------------
test_that("createTrial returns trial object with valid input", {
  with_mocked_bindings({
    result <- createTrial(n_subjects = c(10, 20), n_responders = c(5, 15))
    expect_true(!is.null(result))
  }, simulateScenarios = function(n_subjects_list, response_rates_list, n_trials) {
    list(trial = "mocked_trial")
  })
})


# ------------------------------------------------------------------
# Test: createTrial errors on invalid n_subjects or n_responders
# Input:
#   - n_subjects containing NA, or n_responders being non-integer.
# Behaviour:
#   - Argument checks should fail and throw informative errors.
# Expectations:
#   - Error mentioning 'n_subjects' for NA in n_subjects.
#   - Error mentioning 'n_responders' for non-integer responders.
# Why:
#   - Ensures basic input validation for trial setup.
# ------------------------------------------------------------------
test_that("createTrial errors with invalid input", {
  expect_error(
    createTrial(n_subjects = c(10, NA), n_responders = c(5, 15)),
    "n_subjects"
  )

  expect_error(
    createTrial(n_subjects = c(10, 20), n_responders = c(5.5, 15)),
    "n_responders"
  )
})

# Tests for continueRecruitment ------------------------------------------------

# ------------------------------------------------------------------
# Test: continueRecruitment errors if n_subjects_add_list is missing
# Input:
#   - decisions_list provided, n_subjects_add_list omitted.
# Behaviour:
#   - Function must complain about missing n_subjects_add_list.
# Expectations:
#   - Error message mentions 'n_subjects_add_list'.
# Why:
#   - Additional recruitment per scenario must be specified explicitly.
# ------------------------------------------------------------------
test_that("Error is thrown if n_subjects_add_list is missing", {
  expect_error(
    continueRecruitment(
      decisions_list = list(scenario_1 = list(decisions_list = list()))
    ),
    "n_subjects_add_list"
  )
})


# ------------------------------------------------------------------
# Test: continueRecruitment errors if decisions_list is missing
# Input:
#   - n_subjects_add_list provided, decisions_list omitted.
# Behaviour:
#   - Function must complain about missing decisions_list.
# Expectations:
#   - Error message mentions 'decisions_list'.
# Why:
#   - Decisions drive which cohorts continue recruiting.
# ------------------------------------------------------------------
test_that("Error is thrown if decisions_list is missing", {
  expect_error(
    continueRecruitment(
      n_subjects_add_list = list(10)
    ),
    "decisions_list"
  )
})


# ------------------------------------------------------------------
# Test: continueRecruitment errors when decisions_list has wrong class
# Input:
#   - decisions_list is a plain list, not class 'decision_list'.
# Behaviour:
#   - Class assertion should fail.
# Expectations:
#   - Error mentioning 'decisions_list'.
# Why:
#   - Ensures only properly constructed decision_list objects are used.
# ------------------------------------------------------------------
test_that("Error is thrown if decisions_list is not of class 'decision_list'", {
  expect_error(
    continueRecruitment(
      n_subjects_add_list = list(10),
      decisions_list      = list()
    ),
    "decisions_list"
  )
})


# ------------------------------------------------------------------
# Test: continueRecruitment errors for invalid method_name
# Input:
#   - method_name = "invalid_method" not present in decisions_list.
# Behaviour:
#   - Function should reject unknown method names.
# Expectations:
#   - Error is thrown.
# Why:
#   - Prevents referencing non-existent analysis methods.
# ------------------------------------------------------------------
test_that("Error for invalid method_name", {
  expect_error(
    continueRecruitment(
      n_subjects_add_list = list(c(10, 10)),
      decisions_list      = decisions,
      method_name         = "invalid_method"
    )
  )
})


# ------------------------------------------------------------------
# Test: continueRecruitment errors when method_name is NULL and multiple methods
# Input:
#   - decisions modified to contain two methods ("berry", "exnex"),
#     method_name not supplied.
# Behaviour:
#   - Function should require method_name when >1 method is available.
# Expectations:
#   - Error is thrown.
# Why:
#   - Disambiguates which method’s decisions to follow.
# ------------------------------------------------------------------
test_that("Error when method_name is NULL and multiple methods are present", {
  decisions$scenario_1$analysis_data$analysis_parameters$method_names <- c("berry", "exnex")
  decisions$scenario_1$decisions_list <- list(berry = data.frame(), exnex = data.frame())

  expect_error(
    continueRecruitment(
      n_subjects_add_list = list(c(10, 10)),
      decisions_list      = decisions
    )
  )
})


# ------------------------------------------------------------------
# Test: continueRecruitment errors if list lengths do not match
# Input:
#   - n_subjects_add_list length 2 vs decisions_list length 1.
# Behaviour:
#   - Length consistency check should fail.
# Expectations:
#   - Error is thrown.
# Why:
#   - Each scenario’s decisions need corresponding n_subjects_add vector.
# ------------------------------------------------------------------
test_that("Error is thrown if n_subjects_add_list and decisions_list lengths do not match", {
  expect_error(
    continueRecruitment(
      n_subjects_add_list = list(10, 20),
      decisions_list      = decisions_list,
      method_name         = "berry"
    )
  )
})


# ------------------------------------------------------------------
# Test: continueRecruitment errors when selected method not analyzed
# Input:
#   - decisions scenario_1 only analysed with 'berry',
#     requested method_name = 'exnex'.
# Behaviour:
#   - Function should detect that exnex is not in method_names.
# Expectations:
#   - Error containing "Selected method_name not analyzed".
# Why:
#   - Prevents accidentally requesting an unsupported method.
# ------------------------------------------------------------------
test_that("Error when selected method_name not analyzed in scenario", {
  decisions$scenario_1$analysis_data$analysis_parameters$method_names <- "berry"
  decisions$scenario_1$decisions_list <- list(berry = data.frame())

  expect_error(
    continueRecruitment(
      n_subjects_add_list = list(c(10, 10)),
      decisions_list      = decisions,
      method_name         = "exnex"
    ),
    "Selected method_name not analyzed"
  )
})


# ------------------------------------------------------------------
# Test: continueRecruitment errors when all cohorts are historical
# Input:
#   - scenario_1 response_rates <= 0 or >= 1, making all cohorts historical.
# Behaviour:
#   - continueRecruitment() must error with "Only historical cohorts in scenario".
# Expectations:
#   - Error message contains that text.
# Why:
#   - There is nothing to recruit if all cohorts are purely historical.
# ------------------------------------------------------------------
test_that("Error when response rate <=0 or >=1", {
  decisions$scenario_1$analysis_data$analysis_parameters$method_names <- "berry"
  decisions$scenario_1$decisions_list <- list()
  decisions$scenario_1$scenario_data$response_rates <-
    matrix(c(0, 1), nrow = 1, dimnames = list(NULL, c("rr_hist1", "rr_hist2")))
  decisions$scenario_1$scenario_data$n_trials <- 1
  decisions$scenario_1$scenario_data$n_responders <-
    matrix(c(0, 1), nrow = 1, dimnames = list(NULL, c("rr_hist1", "rr_hist2")))
  decisions$scenario_1$scenario_data$n_subjects <-
    matrix(c(0, 1), nrow = 1, dimnames = list(NULL, c("rr_hist1", "rr_hist2")))
  decisions$scenario_1$scenario_data$scenario_number <- 1

  expect_error(
    continueRecruitment(
      n_subjects_add_list = list(c(10, 10)),
      decisions_list      = decisions,
      method_name         = "berry"
    ),
    "Only historical cohorts in scenario 1"
  )
})


# ------------------------------------------------------------------
# Test: continueRecruitment errors when n_subjects_add length mismatches
# Input:
#   - Two recruiting cohorts and one historical, but n_subjects_add_list[[1]]
#     has length 1 instead of 2.
# Behaviour:
#   - Function should detect mismatch between recruiting cohorts and
#     n_subjects_add length.
# Expectations:
#   - Error is thrown.
# Why:
#   - Ensures user supplies one recruitment size per recruiting cohort.
# ------------------------------------------------------------------
test_that("Error when n_subjects_add length does not match recruiting cohorts", {
  decisions$scenario_1$analysis_data$analysis_parameters$method_names <- "berry"
  decisions$scenario_1$decisions_list <- list(berry = data.frame())
  decisions$scenario_1$scenario_data$response_rates <-
    matrix(c(0.3, 0.4, 1), nrow = 1,
           dimnames = list(NULL, c("rr_A", "rr_B", "rr_hist")))

  expect_error(
    continueRecruitment(
      n_subjects_add_list = list(c(10)),
      decisions_list      = decisions,
      method_name         = "berry"
    )
  )
})


# ------------------------------------------------------------------
# Test: continueRecruitment updates only rows where overall == TRUE
# Input:
#   - One recruiting cohort rr_A (0.5) and one historical rr_hist (1),
#     baseline n_subjects/n_responders, decisions overall=TRUE,
#     mocked getScenario that returns updated counts for rr_A only.
# Behaviour:
#   - continueRecruitment() should:
#     • add new subjects/responders for rr_A,
#     • leave rr_hist unchanged.
# Expectations:
#   - New n_subjects["rr_A"] == 10, n_responders["rr_A"] == 5.
#   - n_subjects["rr_hist"] and n_responders["rr_hist"] unchanged at 1.
# Why:
#   - Verifies that recruitment only happens in Go rows and only for
#     non-historical cohorts.
# ------------------------------------------------------------------
test_that("Only overall==TRUE rows are updated", {
  decisions$scenario_1$scenario_data$response_rates <-
    matrix(c(0.5, 1), nrow = 1,
           dimnames = list(NULL, c("rr_A", "rr_hist")))

  decisions$scenario_1$scenario_data$n_subjects <-
    matrix(c(0, 1), nrow = 1,
           dimnames = list(NULL, c("rr_A", "rr_hist")))

  decisions$scenario_1$scenario_data$n_responders <-
    matrix(c(0, 1), nrow = 1,
           dimnames = list(NULL, c("rr_A", "rr_hist")))

  decisions$scenario_1$decisions_list <-
    list(berry = data.frame(decision_1 = TRUE, decision_2 = TRUE, overall = TRUE))
  decisions$scenario_1$analysis_data$analysis_parameters$method_names <- "berry"

  testthat::with_mocked_bindings(
    getScenario = function(n_subjects, response_rates, cohort_names, n_trials) {
      list(
        n_subjects   = matrix(10, nrow = 1, dimnames = list(NULL, cohort_names)),
        n_responders = matrix(5, nrow = 1, dimnames = list(NULL, cohort_names))
      )
    }, {
      out <- continueRecruitment(
        n_subjects_add_list = list(c(10)),
        decisions_list      = decisions,
        method_name         = "berry"
      )

      expect_equal(unname(out$scenario_1$n_subjects[, "rr_A"]), 10)
      expect_equal(unname(out$scenario_1$n_responders[, "rr_A", drop = TRUE]), 5)
      expect_equal(unname(out$scenario_1$n_subjects[, "rr_hist"]), 1)
      expect_equal(unname(out$scenario_1$n_responders[, "rr_hist", drop = TRUE]), 1)
    }
  )
})

### evidence levels
check.evidence.levels <- function (
    
  evidence_levels,
  cohort_names,
  analyses_list,
  error_evidence_levels
  
) {
  
  checkmate::assert_true(
    length(evidence_levels) == length(cohort_names),
    .var.name = "The 'evidence_levels' and the 'cohort_names' must have the same length"
  )
  
  
  if (is.character(evidence_levels)) {
    
    mean_index <- evidence_levels == "mean"
    
    evidence_levels_numeric <- tryCatch({
      as.numeric(evidence_levels[!mean_index])
    }, warning = function(w) w)
    
    checkmate::assert_false(
      inherits(evidence_levels_numeric, "warning"),
      .var.name = "The only string allowed for the argument 'evidence_levels' is 'mean'"
    )
    
    
  } else {
    
    evidence_levels_numeric <- evidence_levels
    
  }
  
  checkmate::assert_numeric(
    evidence_levels_numeric,
    lower = 0, upper = 1, any.missing = FALSE,
    .var = error_evidence_levels
  )
  
  available_quantiles <- round(analyses_list[[1]]$analysis_parameters$quantiles, 9)
  asked_quantiles     <- round(1 - evidence_levels_numeric, 9)
  
  checkmate::assert_true(
    all(asked_quantiles %in% available_quantiles),
    .var.name = paste(
      "The 'evidence_levels' must have matches",
      "in the 'evidence_levels' provided to the call performAnalyses()",
      "that created the 'analyses_list'"
    )
  )
}

### functions ####

checkForParallelBackend <- function () {
  
  "%dopar%" <- foreach::"%dopar%"
  if(!foreach::getDoParRegistered()) {
    
    message("\nCaution: No parallel backend detected for the 'foreach' framework.",
            " For execution in parallel, register a parallel backend, e.g. with:\n",
            "   doFuture::registerDoFuture()\n",
            "   future::plan(future::multisession)\n")
    
    # foreach if without parallel backend gives a warning, warning is being
    # displayed only once, should be repeated anyway, message better
    
    tt <- suppressWarnings(foreach::foreach(k = 1:2) %dopar% {k^k^k})
    rm(tt)
    
  }
  
}

# Tests for check.evidence.levels ----------------------------------------------

# ------------------------------------------------------------------
# Test: check.evidence.levels passes with valid numeric input
# Input:
#   - evidence_levels = c(0.8, 0.9) inside [0,1]
#   - cohort_names = c("A","B") with same length
#   - analyses_list providing quantiles (not used for numeric check)
# Behaviour:
#   - Function validates that numeric inputs are in range and lengths match.
# Expectations:
#   - No error or warning (expect_silent).
# Why:
#   - Confirms the functioning for simple numeric evidence levels.
# ------------------------------------------------------------------
test_that("check.evidence.levels passes with valid numeric input", {
  analyses_list <- list(list(analysis_parameters = list(quantiles = c(0.1, 0.2, 0.3))))
  expect_silent(check.evidence.levels(
    evidence_levels = c(0.8, 0.9),
    cohort_names = c("A", "B"),
    analyses_list = analyses_list,
    error_evidence_levels = "Invalid evidence levels"
  ))
})

# ------------------------------------------------------------------
# Test: check.evidence.levels fails when lengths differ
# Input:
#   - evidence_levels length = 1
#   - cohort_names length = 2
# Behaviour:
#   - Function checks that each cohort has a corresponding evidence level.
# Expectations:
#   - Error containing "must have the same length".
# Why:
#   - Ensures consistent one-to-one mapping between cohorts and evidence.
# ------------------------------------------------------------------
test_that("check.evidence.levels fails when lengths differ", {
  analyses_list <- list(list(analysis_parameters = list(quantiles = c(0.1, 0.2))))
  expect_error(check.evidence.levels(
    evidence_levels = c(0.8),
    cohort_names = c("A", "B"),
    analyses_list = analyses_list,
    error_evidence_levels = "Invalid evidence levels"
  ), "must have the same length")
})

# ------------------------------------------------------------------
# Test: check.evidence.levels fails when character input has invalid strings
# Input:
#   - evidence_levels = c("mean", "wrong")
# Behaviour:
#   - Character inputs must match allowed keywords (e.g. "mean" only).
# Expectations:
#   - Error "The only string allowed".
# Why:
#   - Guards against unsupported descriptive evidence level strings.
# ------------------------------------------------------------------
test_that("check.evidence.levels fails when character input has invalid strings", {
  analyses_list <- list(list(analysis_parameters = list(quantiles = c(0.1, 0.2))))
  expect_error(check.evidence.levels(
    evidence_levels = c("mean", "wrong"),
    cohort_names = c("A", "B"),
    analyses_list = analyses_list,
    error_evidence_levels = "Invalid evidence levels"
  ), "The only string allowed")
})

# ------------------------------------------------------------------
# Test: check.evidence.levels fails when numeric values out of [0,1]
# Input:
#   - evidence_levels = c(-0.1, 1.2)
# Behaviour:
#   - Numeric evidence levels must be valid probabilities.
# Expectations:
#   - Error containing "Invalid evidence levels".
# Why:
#   - Rejects out-of-bound probability inputs.
# ------------------------------------------------------------------
test_that("check.evidence.levels fails when numeric values out of [0,1]", {
  analyses_list <- list(list(analysis_parameters = list(quantiles = c(0.1, 0.2))))
  expect_error(check.evidence.levels(
    evidence_levels = c(-0.1, 1.2),
    cohort_names = c("A", "B"),
    analyses_list = analyses_list,
    error_evidence_levels = "Invalid evidence levels"
  ), "Invalid evidence levels")
})

# ------------------------------------------------------------------
# Test: check.evidence.levels fails when quantiles do not match
# Input:
#   - evidence_levels interpreted as quantiles (0.5, 0.6)
#   - analysis quantiles = c(0.1, 0.2)
# Behaviour:
#   - Evidence levels used as quantiles must match analysis quantiles exactly.
# Expectations:
#   - Error "must have matches".
# Why:
#   - Prevents mismatched quantile specifications for analysis.
# ------------------------------------------------------------------
test_that("check.evidence.levels fails when quantiles do not match", {
  analyses_list <- list(list(analysis_parameters = list(quantiles = c(0.1, 0.2))))
  expect_error(check.evidence.levels(
    evidence_levels = c(0.5, 0.6),
    cohort_names = c("A", "B"),
    analyses_list = analyses_list,
    error_evidence_levels = "Invalid evidence levels"
  ), "must have matches")
})

# Tests for checkForParallelBackend --------------------------------------------

# ------------------------------------------------------------------
# Test: checkForParallelBackend is silent when backend registered
# Input:
#   - doFuture backend registered + sequential plan set
# Behaviour:
#   - Function checks availability of a foreach-compatible future backend.
# Expectations:
#   - No error (expect_silent).
# Why:
#   - Confirms that with a valid backend registered the function accepts it.
# ------------------------------------------------------------------
test_that("checkForParallelBackend is silent when backend registered", {
  # Register a dummy backend
  doFuture::registerDoFuture()
  future::plan(future::sequential)
  expect_silent(checkForParallelBackend())
})

---
title: "Running bhmbasket on HPC"
author: "Stephan Wojciekowski"
date: '`r format(Sys.time(), "%B %d, %Y")`'
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Running bhmbasket on HPC}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r knitr_options, include = FALSE}
knitr::opts_chunk$set(
  eval     = FALSE,  # en- / disables R code evaluation globally
  cache    = FALSE,  # en- / disables R code caching globally
  collapse = TRUE,
  comment  = "#>"
)
```

This vignette provides a short example on how to use the R package `bhmbasket` in a high performance computing (HPC) environment using the R packages `doFuture` and `future.batchtools`.

```{r setup}
library(bhmbasket)
library(doFuture)
library(future.batchtools)

rng_seed <- 5440
set.seed(rng_seed)
```

## Setup of the parallel backend
The code below provides an example for specifying a parallel backend using the future framework with the SLURM job scheduler.
Kindly see the documentation of the R packages `doFuture` and `future.batchtools` for further options.
This code is to be run on the master node.
```{r SLURM_Setup}
## Adapt the SLURM template to requirements
job_time  <- 1   # time for job in hours
n_workers <- 24  # number of worker nodes
n_cpus    <- 16  # number of cpus per worker node
gb_memory <- 2   # memory [GB] per cpu

slurm <- tweak(batchtools_slurm,
           template  = system.file('templates/slurm-simple.tmpl',
                                   package = 'batchtools'),
           workers   = n_workers,
           resources = list(
             walltime  = 60 * 60 * job_time,
             ncpus     = n_cpus,
             memory    = 1000 * gb_memory))

## Register the parallel backend 
registerDoFuture()

## Specify how the futures should be resolved
plan(list(slurm, multisession))
```

## Running some bhmbasket code on HPC
The R package `bhmbasket` makes use of the foreach framework and runs with every applicable parallel backend. 
With a parallel backend registered as shown above and running the code on the master node, the job scheduler will automatically distribute the jobs to the worker nodes via `plan(slurm)`, and with the nested parallelization built into `performAnalyses()`, each worker node makes use of its CPUs via `plan(multisession)`.

Below is some example code, which was taken from the examples section of `?bhmbasket::getEstimates`.
Due to the foreach framework, no adjustments to the code are necessary.
Kindly note that running this small example on a HPC environment will most likely not result in a performance improvement.
```{r}
scenarios_list <- simulateScenarios(
    n_subjects_list     = list(c(10, 20, 30)),
    response_rates_list = list(c(0.1, 0.2, 3)),
    n_trials            = 10)

  analyses_list <- performAnalyses(
    scenario_list       = scenarios_list,
    target_rates        = c(0.1, 0.1, 0.1),
    calc_differences    = matrix(c(3, 2, 2, 1), ncol = 2),
    n_mcmc_iterations   = 100)

  getEstimates(analyses_list)
```

---
title: "Reproducing Parts of Neuenschwander et al. (2016)"
author: "Stephan Wojciekowski"
date: '`r format(Sys.time(), "%B %d, %Y")`'
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Reproducing Parts of Neuenschwander et al. (2016)}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
references:
  - id: exnex2016
    title: Robust exchangeability designs for early phase clinical trials with multiple strata
    author: 
      - family: Neuenschwander
        given: Beat
      - family: Wandel
        given: Simon
      - family: Roychoudhury
        given: Satrajit
      - family: Bailey
        given: Stuart
    container-title: Pharmaceutical statistics
    volume: 15
    DOI: 10.1002/pst.1730
    issue: 2
    page: 123-134
    type: article-journal
    issued:
      year: 2016
  - id: chugh2009
    title: Phase II multicenter trial of imatinib in 10 histologic subtypes of sarcoma using a Bayesian hierarchical statistical model
    author: 
      - family: Chugh
        given: Rashmi
      - family: Wathen
        given: J. Kyle
      - family: Maki
        given: Robert G.
      - family: Benjamin
        given: Robert S.
      - family: Patel
        given: Shreyaskumar R.
      - family: Myers
        given: Paul A.
      - family: Priebat
        given: Dennis A.
      - family: Reinke
        given: Denise K.
      - family: Thomas
        given: Dafydd G.
      - family: Keohan
        given: Mary L.
      - family: Samuels
        given: Brian L.
      - family: Baker
        given: Laurence H.
    container-title: Journal of Clinical Oncology
    volume: 27
    DOI: 10.1200/JCO.2008.20.5054
    issue: 19
    publisher: American Society of Clinical Oncology
    page: 3148–3153
    type: article-journal
    issued:
      year: 2009
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  eval     = TRUE,  # en- / disables R code evaluation globally
  cache    = FALSE,  # en- / disables R code caching globally
  collapse = TRUE,
  comment  = "#>"
)
```

This vignette demonstrates how to use the R package `bhmbasket` to reproduce parts of "Robust exchangeability designs for early phase clinical trials with multiple strata" by @exnex2016.

Two examples are shown in this vignette: the analysis of a basket trial's outcome with the ExNex approach, and the assessment of the operating characteristics of a planned basket trial under several scenarios.

```{r setup}
library(bhmbasket)

rng_seed <- 123
set.seed(rng_seed)
```

## Analysis of a Basket Trial's Outcome
Section 2 of @exnex2016 discusses the application of the ExNex approach to the results of the phase II sarcoma basket trial by @chugh2009.
The prior specifications and the estimated response rates per cohort are shown in the Appendix of @exnex2016.

### Creating the trial
A single trial with a given outcome can be created with the function `createTrial()`, which takes as arguments the number of subjects and the number of responders of the trial.
The outcome of the basket trial by @chugh2009 can be set up as follows:
```{r}
chugh_trial <- createTrial(
  n_subjects   = c(15, 13, 12, 28, 29, 29, 26, 5, 2, 20),
  n_responders = c(2, 0, 1, 6, 7, 3, 5, 1, 0, 3))
```

### Running the model
The analysis of the trial is handled by the function `performAnalyses()`, to which to the  `chugh_trial`, the name of the method `"exnex"`, and the prior parameters as shown in Appendix 5.3 of @exnex2016 are provided:
```{r chugh}
chugh_prior_parameters <- setPriorParametersExNex(
  mu_mean   = c(-1.735, 0.847),
  mu_sd     = c(0.146, 0.266)^-0.5,
  tau_scale = 1,
  mu_j      = rep(-1.734, 10),
  tau_j     = rep(0.128^-0.5, 10),
  w_j       = c(0.5, 0, 0.5))

if (file.exists("chugh_analysis.rds")) {
  
  chugh_analysis <- readRDS("chugh_analysis.rds")
  
} else {
  
  chugh_analysis <- performAnalyses(
  scenario_list         = chugh_trial,
  method_names          = "exnex",
  prior_parameters_list = chugh_prior_parameters,
  seed                  = rng_seed,
  n_mcmc_iterations     = 5e4,
  n_cores               = 2L,
  verbose               = FALSE)
  
  saveRDS(chugh_analysis, "chugh_analysis.rds")
  
}
```
### Getting the estimates
One can use the function `getEstimates()` to get point estimates and credible intervals of the response rates:
```{r}
getEstimates(analyses_list = chugh_analysis)
```
By comparing these values to the respective numbers provided by @exnex2016 in Appendix 5.3 one can see that the results are equal to the second decimal place.

## Assessing the Operating Characteristics of a Basket Trial Design {#trialOC}

Section 3.1 of @exnex2016 presents the design and the operating characteristics of a phase II basket trial in four indications.
In their publication, the authors present the scenarios with the respective true response rates per cohort in Table V along with the respective cohort-wise go probabilities, biases, and mean squared errors.
The number of subjects is 20 each for the first two cohorts and 10 each for the last two cohorts.

<!-- As a side note: the number of simulated trials used by @exnex2016 is unclear. -->
<!-- The first two cohorts should yield the same operating characteristics when their true response rates are equal, and this should also hold true for the last two cohorts. -->
<!-- However, the numbers reported by @exnex2016 do not always meet this expectation (see e.g. the reported biases of Scenario 3). -->

### Simulating the scenarios
The scenarios are simulated with the function `simulateScenarios()`.
There are seven different scenarios presented in Table V.
In this example, the Scenarios 1, 3, and 4 are reproduced.
For each scenario, 10000 basket trial realizations are simulated:
```{r simulateScenarios}
scenarios_list <- simulateScenarios(
  n_subjects_list     = list(c(20, 20, 10, 10),
                             c(20, 20, 10, 10),
                             c(20, 20, 10, 10)),
  response_rates_list = list(c(0.1, 0.1, 0.1, 0.1),
                             c(0.1, 0.1, 0.3, 0.3),
                             c(0.1, 0.1, 0.1, 0.5)),
  scenario_numbers    = c(1, 3, 4),
  n_trials            = 1e4)
```
### Running the model {#performAnalysis}
The analysis of each simulated basket trial is performed with the function `performAnalyses()`.
@exnex2016 use the posterior means of the cohorts' response rates to calculate the biases and mean squared errors, as well as to derive cohort-wise decisions.
Setting the argument `post_mean` to `TRUE` will save the posterior means along with the posterior probability thresholds, which are required for the decision making laid out below.
The prior parameters for the ExNex model are taken from Section 3.1 and Appendix 5.1.2 of @exnex2016.
```{r performAnalyses}
prior_parameters <- setPriorParametersExNex(
  mu_mean   = c(logit(0.1), logit(0.3)),
  mu_sd     = c(3.18, 1.94),
  tau_scale = 1,
  mu_j      = rep(logit(0.2), 4),
  tau_j     = rep(2.5, 4),
  w_j       = c(0.25, 0.25, 0.5))

if (file.exists("analyses_list.rds")) {
  
  analyses_list <- readRDS("analyses_list.rds")
  
} else {
  
  analyses_list <- performAnalyses(
    scenario_list         = scenarios_list,
    method_names          = "exnex",
    prior_parameters_list = prior_parameters,
    seed                  = rng_seed,
    n_mcmc_iterations     = 5e4,
    n_cores               = 2L)
  
  saveRDS(analyses_list, "analyses_list.rds")
  
}
```
```{r, include = FALSE}
## check for unnecessary data
if (any(grepl("mu_",
              colnames(analyses_list$scenario_1$quantiles_list$exnex[[1]])))) {
  
  ## remove unnecessary data
  
  ## ...in quantiles_list
  analyses_list$scenario_1$quantiles_list$exnex <-   lapply(analyses_list$scenario_1$quantiles_list$exnex, function   (x) {
    x[-c(2, 6), -c(1:6, 11, 12)]
  })
  analyses_list$scenario_3$quantiles_list$exnex <-   lapply(analyses_list$scenario_3$quantiles_list$exnex, function   (x) {
    x[-c(2, 6), -c(1:6, 11, 12)]
  })
  analyses_list$scenario_4$quantiles_list$exnex <-   lapply(analyses_list$scenario_4$quantiles_list$exnex, function   (x) {
    x[-c(2, 6), -c(1:6, 11, 12)]
  })
  
  ## ... on quantiles
  analyses_list$scenario_1$analysis_parameters$quantiles <-   c(0.025, 0.100, 0.200, 0.500, 0.975)
  analyses_list$scenario_3$analysis_parameters$quantiles <-   c(0.025, 0.100, 0.200, 0.500, 0.975)
  analyses_list$scenario_4$analysis_parameters$quantiles <-   c(0.025, 0.100, 0.200, 0.500, 0.975)
  
  ## ... on subjects and responders
  analyses_list$scenario_1$scenario_data$n_subjects <- NULL
  analyses_list$scenario_3$scenario_data$n_subjects <- NULL
  analyses_list$scenario_4$scenario_data$n_subjects <- NULL
  analyses_list$scenario_1$scenario_data$n_responders <- NULL
  analyses_list$scenario_3$scenario_data$n_responders <- NULL
  analyses_list$scenario_4$scenario_data$n_responders <- NULL
  
  analyses_list$scenario_1$quantiles_list$exnex[[1]]
    
  saveRDS(analyses_list, "analyses_list.rds", compress = "xz")
}
```
Note that although the total number of simulated trials is 3 x 10000, there are only 2314 unique trial realizations among the three scenarios.
As the function `performAnalyses()` takes advantage of this, its run time is reduced when compared to applying the model to all simulated trials.
However, some additional time is required to map the results from the unique trials back to the simulated trials of each scenario.

### Estimating the biases and MSEs
The bias and mean squared error (MSE) per cohort, as well as some posterior quantiles, are estimated with `getEstimates()`.
The argument `point_estimator = "mean"` ensures that the mean instead of the median will be used for calculating the biases and MSEs.
In order to better assess and compare the results with the numbers presented by @exnex2016, scaling and rounding may be applied with `scaleRoundList()`:
```{r getBiasMSE}
estimates <- getEstimates(
    analyses_list   = analyses_list,
    point_estimator = "mean")

scaleRoundList(
  list         = estimates,
  scale_param  = 100,
  round_digits = 2)
```
<!-- Comparing the biases and MSEs with the respective numbers provided in Table V of @exnex2016, one can see that for the scaled MSEs the results are equal to the second decimal place. -->
<!-- However, some scaled biases differ in the first decimal place. -->
<!-- Please note that the first and last two cohorts should yield the same operating characteristics when their true response rates are equal, which is fulfilled by the results provided by `bhmbasket`. -->
Comparing the biases and MSEs with the respective numbers provided in Table V of @exnex2016, one can see that the results are rather similar.

### Estimating the cohort-wise go probabilities
The cohort-wise go decisions are derived with the function `getGoDecisions()`, which applies the specified decision rules to each simulated basket trial for each scenario.
The estimated go probabilities for each scenario can then be derived with the function `getGoProbabilities()`.

The set up of the decision rules in `getGoDecisions()` warrants further explanation.
Generally, a simple decision rule for a go in a single cohort $j$ can be written as 
$$P(p_j|\text{data} > \bar{p}_{j}) > \gamma_j,$$
where $p_j|\text{data}$ is the posterior response rate, $\bar{p}_{j}$ is the is the boundary response rate, and $\gamma_j$ is the posterior evidence level for a go decision.
Thus, the decision rule consists of three parts: the posterior response rate, the boundary response rate, and the posterior probability threshold.
The arguments for setting up the decision rules correspond to these three parts: `cohort_names` picks the posterior response rates, `boundary_rules` specifies the boundary response rates and type of comparison, and `evidence_levels` provides the posterior evidence levels.
For example, the decision rule $P(p_1|\text{data} > 0.1) > 0.9$ would then be implemented as
```{r, eval = FALSE}
cohort_names    = "p_1"
boundary_rules  = quote(c(x[1] > 0.1))
evidence_levels = 0.9
```
It is possible to specify different boundary rules and evidence levels for different analysis methods with `list()`.
In this case, only the ExNex method is applied and no list is needed.
Note that only evidence levels specified in `performAnalyses()` can be utilized. 

It is possible to implement combined decision criteria, such as utilized by @exnex2016 who require in Section 3.1 for a go decision in the first cohort that 
$$P(p_1|\text{data} > 0.1) > 0.9 \quad\land\quad \text{Mean}(p_1|\text{data}) > 0.2$$
holds true.
This combined decision rule would then be implemented as
```{r, eval = FALSE}
cohort_names    = c("p_1", "p_1")
boundary_rules  = quote(c(x[1] > 0.1 & x[2] > 0.2))
evidence_levels = c(0.9, "mean")
```
Thus, the go probabilities for the Scenarios 1, 3, and 4 can be estimated by:
```{r getGoDecisions}
decisions_list <- getGoDecisions(
  analyses_list   = analyses_list,
  cohort_names    = c("p_1", "p_1",
                      "p_2", "p_2",
                      "p_3", "p_3",
                      "p_4", "p_4"),
  evidence_levels = c(0.9, "mean",
                      0.9, "mean",
                      0.8, "mean",
                      0.8, "mean"),
  boundary_rules  = quote(c(x[1] > 0.1 & x[2] > 0.2,
                            x[3] > 0.1 & x[4] > 0.2,
                            x[5] > 0.1 & x[6] > 0.2,
                            x[7] > 0.1 & x[8] > 0.2)))

go_probabilities <- getGoProbabilities(decisions_list)

scaleRoundList(
  list         = go_probabilities,
  scale_param  = 100,
  round_digits = 0)
```
By comparing these go probabilities to the values provided in Table V of @exnex2016 one can see that the results differ at most by 1 percentage point.

## References

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/PriorFunctions.R
\name{setPriorParametersBerry}
\alias{setPriorParametersBerry}
\title{setPriorParametersBerry}
\usage{
setPriorParametersBerry(mu_mean, mu_sd, tau_scale)
}
\arguments{
\item{mu_mean}{A numeric for the mean of \eqn{\mu}}

\item{mu_sd}{A positive numeric for the standard deviation of \eqn{\mu}}

\item{tau_scale}{A positive numeric for the scale parameter of \eqn{\tau}}
}
\value{
A list with prior parameters of class \code{prior_parameters_list}
}
\description{
This function sets prior parameters for the analysis method \code{"berry"}
for use in \code{\link[bhmbasket]{performAnalyses}}.
}
\details{
This function sets the prior parameters for the method proposed by Berry et al. (2013).
Note that the implemented distribution of \eqn{\tau} is half-normal.
}
\examples{
 prior_parameters_berry <- setPriorParametersBerry(0, 1, 2)
}
\references{
Berry, Scott M., et al. "Bayesian hierarchical modeling of patient subpopulations:
efficient designs of phase II oncology clinical trials."
\emph{Clinical Trials} 10.5 (2013): 720-734.
}
\seealso{
\code{\link[bhmbasket]{performAnalyses}}
\code{\link[bhmbasket]{getPriorParameters}}
\code{\link[bhmbasket]{combinePriorParameters}}
\code{\link[bhmbasket]{setPriorParametersExNex}}
\code{\link[bhmbasket]{setPriorParametersExNexAdj}}
\code{\link[bhmbasket]{setPriorParametersPooled}}
\code{\link[bhmbasket]{setPriorParametersStratified}}
\code{\link[bhmbasket]{getMuVar}}
}
\author{
Stephan Wojciekowski
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/PriorFunctions.R
\name{setPriorParametersPooled}
\alias{setPriorParametersPooled}
\title{setPriorParametersPooled}
\usage{
setPriorParametersPooled(a, b)
}
\arguments{
\item{a}{A positive numeric for \eqn{\alpha}}

\item{b}{A positive numeric for \eqn{\beta}}
}
\value{
A list with prior parameters of class \code{prior_parameters_list}
}
\description{
This function sets prior parameters for the analysis method \code{"pooled"}
for use in \code{\link[bhmbasket]{performAnalyses}}.
}
\details{
The method \code{"pooled"} is a beta-binomial model that pools all cohorts.
The prior parameters are the scale parameters of the beta prior distribution.
}
\examples{
 prior_parameters_pooled <- setPriorParametersPooled(1, 2)
}
\seealso{
\code{\link[bhmbasket]{performAnalyses}}
\code{\link[bhmbasket]{getPriorParameters}}
\code{\link[bhmbasket]{combinePriorParameters}}
\code{\link[bhmbasket]{setPriorParametersBerry}}
\code{\link[bhmbasket]{setPriorParametersExNex}}
\code{\link[bhmbasket]{setPriorParametersExNexAdj}}
\code{\link[bhmbasket]{setPriorParametersStratified}}
\code{\link[bhmbasket]{getMuVar}}
}
\author{
Stephan Wojciekowski
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/PriorFunctions.R
\name{setPriorParametersExNex}
\alias{setPriorParametersExNex}
\title{setPriorParametersExNex}
\usage{
setPriorParametersExNex(
  mu_mean,
  mu_sd,
  tau_scale,
  mu_j = NULL,
  tau_j = NULL,
  w_j,
  w_nex = NULL,
  mean_nex = NULL,
  sd_nex = NULL
)
}
\arguments{
\item{mu_mean}{A numeric for the mean of \eqn{\mu}}

\item{mu_sd}{A positive numeric for the standard deviation of \eqn{\mu}}

\item{tau_scale}{A positive numeric for the scale parameter of \eqn{\tau}}

\item{mu_j}{A vector of numerics for the means \eqn{\mu_j} of the standard Nex priors.
Ignored if \code{w_nex}, \code{mean_nex}, and \code{sd_nex} are provided.}

\item{tau_j}{A vector of positive numerics for the standard deviations \eqn{\tau_j}
of the standard Nex priors. Ignored if \code{w_nex}, \code{mean_nex}, and \code{sd_nex} are provided.}

\item{w_j}{A numeric in \verb{(0, 1)} for the weight of the Ex component, or a numeric vector
of mixture weights summing to 1 if multiple Ex components are specified.}

\item{w_nex}{An optional numeric vector of mixture weights in \eqn{[0,1]} summing to 1
for the Nex mixture prior.}

\item{mean_nex}{An optional numeric matrix of Nex mixture means with one row per Nex
mixture component and one column per cohort.}

\item{sd_nex}{An optional positive numeric matrix of Nex mixture standard deviations with
one row per Nex mixture component and one column per cohort.}
}
\value{
A list with prior parameters of class \code{prior_parameters_list}
}
\description{
This function sets prior parameters for the analysis method \code{"exnex"}
for use in \code{\link[bhmbasket]{performAnalyses}}.

It supports two specifications for the Nex part:
\itemize{
\item the standard ExNex specification with cohort-specific Nex priors via \code{mu_j} and \code{tau_j}
\item an extended specification with a mixture prior on the Nex part via \code{w_nex}, \code{mean_nex}, and \code{sd_nex}
}
}
\details{
This function sets the prior parameters for the method proposed by Neuenschwander et al. (2016).
If \code{w_nex}, \code{mean_nex}, and \code{sd_nex} are all \code{NULL}, the standard ExNex formulation is used.
Otherwise, the Nex part is specified as a finite mixture prior.
}
\examples{
## standard ExNex
prior_parameters_exnex <- setPriorParametersExNex(
  mu_mean   = 0,
  mu_sd     = 1,
  tau_scale = 2,
  mu_j      = c(4, 5),
  tau_j     = c(6, 7),
  w_j       = 0.8
)

## ExNex with Nex mixture prior
prior_parameters_exnex_mix <- setPriorParametersExNex(
  mu_mean   = 0,
  mu_sd     = 1,
  tau_scale = 2,
  w_j       = 0.8,
  w_nex     = c(0.7, 0.3),
  mean_nex  = rbind(c(4, 5), c(2, 3)),
  sd_nex    = rbind(c(6, 7), c(8, 9))
)

}
\references{
Neuenschwander, Beat, et al. "Robust exchangeability designs
for early phase clinical trials with multiple strata."
\emph{Pharmaceutical statistics} 15.2 (2016): 123-134.
}
\seealso{
\code{\link[bhmbasket]{performAnalyses}}
\code{\link[bhmbasket]{getPriorParameters}}
\code{\link[bhmbasket]{combinePriorParameters}}
\code{\link[bhmbasket]{setPriorParametersBerry}}
\code{\link[bhmbasket]{setPriorParametersExNexAdj}}
\code{\link[bhmbasket]{setPriorParametersPooled}}
\code{\link[bhmbasket]{setPriorParametersStratified}}
\code{\link[bhmbasket]{getMuVar}}
}
\author{
Stephan Wojciekowski
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/PriorFunctions.R
\name{combinePriorParameters}
\alias{combinePriorParameters}
\title{combinePriorParameters}
\usage{
combinePriorParameters(list_of_prior_parameters)
}
\arguments{
\item{list_of_prior_parameters}{A list of items with class \code{prior_parameters_list}}
}
\value{
A list with prior parameters of class \code{prior_parameters_list}
}
\description{
This function combines prior parameters from different sources and returns them
for use in \code{\link[bhmbasket]{performAnalyses}}.
}
\details{
This function is intended to combine the prior parameters set with the functions
\code{\link[bhmbasket]{setPriorParametersBerry}},
\code{\link[bhmbasket]{setPriorParametersExNex}},
\code{\link[bhmbasket]{setPriorParametersExNexAdj}},
\code{\link[bhmbasket]{setPriorParametersPooled}}, or
\code{\link[bhmbasket]{setPriorParametersStratified}},
in case more than one analysis method should be applied with
\code{\link[bhmbasket]{performAnalyses}}.
}
\examples{
 prior_parameters_stratified <- setPriorParametersStratified(c(1, 2), c(3, 4))
 prior_parameters_berry      <- setPriorParametersBerry(0, 1, 2)

 prior_parameters_list       <- combinePriorParameters(
   list(prior_parameters_berry,
        prior_parameters_stratified))
}
\seealso{
\code{\link[bhmbasket]{performAnalyses}}
\code{\link[bhmbasket]{setPriorParametersBerry}}
\code{\link[bhmbasket]{setPriorParametersExNex}}
\code{\link[bhmbasket]{setPriorParametersExNexAdj}}
\code{\link[bhmbasket]{setPriorParametersPooled}}
\code{\link[bhmbasket]{setPriorParametersStratified}}
\code{\link[bhmbasket]{getPriorParameters}}
\code{\link[bhmbasket]{getMuVar}}
}
\author{
Stephan Wojciekowski
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/PriorFunctions.R
\name{getMuVar}
\alias{getMuVar}
\title{getMuVar}
\usage{
getMuVar(response_rate, tau_scale, n_worth = 1)
}
\arguments{
\item{response_rate}{A numeric for the response rate}

\item{tau_scale}{A numeric for the scale parameter of the Half-normal distribution of \eqn{\tau}}

\item{n_worth}{An integer for the number of subjects the variance of \eqn{\mu} should be worth
with regard to the variability of the distribution of the response rate, Default: \code{1}}
}
\value{
Returns a numeric for the variance of \eqn{\mu}
}
\description{
This function returns the variance of \eqn{\mu} that is worth a certain number
of subjects for the distribution of the response rates.
}
\details{
Calculates the variance \code{mu_var} in
\deqn{logit(p) = \theta ~ N(\mu, \tau),
\mu ~ N(mu_mean, mu_var), \tau ~ HN(tau_scale),}
for \code{n_worth} number of observations, as in Neuenschwander et al. (2016).
}
\examples{
  getMuVar(response_rate = 0.3,
           tau_scale     = 1)
  getMuVar(response_rate = 0.3,
           tau_scale     = 1,
           n_worth       = 2)
}
\references{
Neuenschwander, Beat, et al. "Robust exchangeability designs
for early phase clinical trials with multiple strata."
\emph{Pharmaceutical statistics} 15.2 (2016): 123-134.
}
\author{
Stephan Wojciekowski
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/DataFunctions.R
\name{loadScenarios}
\alias{loadScenarios}
\title{loadScenarios}
\usage{
loadScenarios(scenario_numbers, load_path = tempdir())
}
\arguments{
\item{scenario_numbers}{A vector of integers naming the scenario to be loaded}

\item{load_path}{A string for the directory where the scenarios are being stored,
Default: \code{\link[base]{tempfile}}}
}
\value{
Returns an object of class \code{scenario_list}
}
\description{
This function loads scenarios saved with \code{\link[bhmbasket]{saveScenarios}}
}
\examples{
  scenarios_list <- simulateScenarios(
    n_subjects_list     = list(c(10, 20, 30)),
    response_rates_list = list(rep(0.9, 3)),
    n_trials            = 10)

  save_info      <- saveScenarios(scenarios_list)
  scenarios_list <- loadScenarios(scenario_numbers = save_info$scenario_numbers,
                                  load_path        = save_info$path)
}
\seealso{
\code{\link[bhmbasket]{simulateScenarios}}
\code{\link[bhmbasket]{saveScenarios}}
\code{\link[base]{tempfile}}
}
\author{
Stephan Wojciekowski
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/PriorFunctions.R
\name{setPriorParametersExNexAdj}
\alias{setPriorParametersExNexAdj}
\title{setPriorParametersExNexAdj}
\usage{
setPriorParametersExNexAdj(
  mu_mean,
  mu_sd,
  tau_scale,
  mu_j = NULL,
  tau_j = NULL,
  w_j,
  w_nex = NULL,
  mean_nex = NULL,
  sd_nex = NULL
)
}
\arguments{
\item{mu_mean}{\link{numeric} Mean of \eqn{\mu}}

\item{mu_sd}{\link{numeric} Positive standard deviation of \eqn{\mu}}

\item{tau_scale}{\link{numeric} Positive scale parameter of \eqn{\tau}}

\item{mu_j}{\link{numeric} Vector of means \eqn{\mu_j} for the standard Nex priors.
Ignored if \code{w_nex}, \code{mean_nex}, and \code{sd_nex} are provided.}

\item{tau_j}{\link{numeric} Vector of positive standard deviations \eqn{\tau_j}
for the standard Nex priors. Ignored if \code{w_nex}, \code{mean_nex}, and \code{sd_nex} are provided.}

\item{w_j}{\link{numeric} Weight of the Ex component in \verb{(0, 1)}, or a numeric vector
of mixture weights summing to 1 if multiple Ex components are specified.}

\item{w_nex}{\link{numeric} Optional vector of mixture weights in \eqn{[0,1]} summing to 1
for the Nex mixture prior.}

\item{mean_nex}{\link{numeric} Optional matrix of Nex mixture means with one row per Nex
mixture component and one column per cohort.}

\item{sd_nex}{\link{numeric} Optional positive matrix of Nex mixture standard deviations with
one row per Nex mixture component and one column per cohort.}
}
\value{
A list with prior parameters of class \code{prior_parameters_list}
}
\description{
This function sets prior parameters for the analysis method \code{"exnex_adj"}
for use in \code{\link[bhmbasket]{performAnalyses}}.

It supports two specifications for the Nex part:
\itemize{
\item the standard ExNex Adjusted specification with cohort-specific Nex priors via \code{mu_j} and \code{tau_j}
\item an extended specification with a mixture prior on the Nex part via \code{w_nex}, \code{mean_nex}, and \code{sd_nex}
}
}
\details{
This function sets prior parameters for the ExNex Adjusted method, which combines
the approach proposed by Neuenschwander et al. (2016) and the approach proposed by
Berry et al. (2013). If \code{w_nex}, \code{mean_nex}, and \code{sd_nex} are all \code{NULL}, the standard
ExNex Adjusted formulation is used. Otherwise, the Nex part is specified as a finite
mixture prior.
}
\examples{
## standard ExNex Adjusted
prior_parameters_exnex_adj <- setPriorParametersExNexAdj(
  mu_mean   = 0,
  mu_sd     = 1,
  tau_scale = 2,
  mu_j      = c(4, 5),
  tau_j     = c(6, 7),
  w_j       = 0.8
)

## ExNex Adjusted with Nex mixture prior
prior_parameters_exnex_adj_mix <- setPriorParametersExNexAdj(
  mu_mean   = 0,
  mu_sd     = 1,
  tau_scale = 2,
  w_j       = 0.8,
  w_nex     = c(0.7, 0.3),
  mean_nex  = rbind(c(0, 0), c(-1, -1)),
  sd_nex    = rbind(c(6, 7), c(8, 9))
)

}
\references{
Neuenschwander, Beat, et al. "Robust exchangeability designs
for early phase clinical trials with multiple strata."
\emph{Pharmaceutical statistics} 15.2 (2016): 123-134.
}
\seealso{
\code{\link[bhmbasket]{performAnalyses}}
\code{\link[bhmbasket]{getPriorParameters}}
\code{\link[bhmbasket]{combinePriorParameters}}
\code{\link[bhmbasket]{setPriorParametersBerry}}
\code{\link[bhmbasket]{setPriorParametersExNex}}
\code{\link[bhmbasket]{setPriorParametersPooled}}
\code{\link[bhmbasket]{setPriorParametersStratified}}
\code{\link[bhmbasket]{getMuVar}}
}
\author{
Stephan Wojciekowski
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Misc.R
\name{logit}
\alias{logit}
\title{logit}
\usage{
logit(p)
}
\arguments{
\item{p}{A numeric in (0, 1)}
}
\value{
logit of p
}
\description{
This function returns the logit of the input argument.
}
\details{
This function is an alias for `stats::binomial()$linkfun`
}
\examples{
logit(invLogit(0.3))
logit(c(0, 0.5, 1))
}
\seealso{
\code{\link[stats]{family}}
}
\author{
Stephan Wojciekowski
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Misc.R
\name{invLogit}
\alias{invLogit}
\title{invLogit}
\usage{
invLogit(theta)
}
\arguments{
\item{theta}{A numeric}
}
\value{
Inverse logit of theta
}
\description{
This function returns the inverse logit of the input argument.
}
\details{
This function is an alias for `stats::binomial()$linkinv`
}
\examples{
invLogit(logit(0.3))
invLogit(c(-Inf, 0, Inf))
}
\seealso{
\code{\link[stats]{family}}
}
\author{
Stephan Wojciekowski
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/AnalysisFunctions.R
\name{loadAnalyses}
\alias{loadAnalyses}
\title{loadAnalyses}
\usage{
loadAnalyses(
  scenario_numbers,
  analysis_numbers = rep(1, length(scenario_numbers)),
  load_path = tempdir()
)
}
\arguments{
\item{scenario_numbers}{A (vector of) positive integer(s) for the scenario number(s)}

\item{analysis_numbers}{A (vector of) positive integer(s) for the analysis number(s),
Default: \code{rep(1, length(scenario_numbers))}}

\item{load_path}{A string providing a path where the scenarios are being stored,
Default: \code{\link[base]{tempfile}}}
}
\value{
Returns an object of class \code{analysis_list}
}
\description{
This function loads an analysis performed with
\code{\link[bhmbasket]{performAnalyses}}
}
\examples{
  trial_data <- createTrial(
    n_subjects   = c(10, 20, 30),
    n_responders = c(1, 2, 3))

  analysis_list <- performAnalyses(
    scenario_list      = trial_data,
    target_rates       = rep(0.5, 3),
    n_mcmc_iterations  = 100)

  save_info     <- saveAnalyses(analysis_list)
  analysis_list <- loadAnalyses(scenario_numbers = save_info$scenario_numbers,
                                analysis_numbers = save_info$analysis_numbers,
                                load_path        = save_info$path)
}
\seealso{
\code{\link[bhmbasket]{performAnalyses}}
\code{\link[bhmbasket]{saveAnalyses}}
\code{\link[base]{tempfile}}
}
\author{
Stephan Wojciekowski
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/OCFunctions.R
\name{getGoProbabilities}
\alias{getGoProbabilities}
\title{getGoProbabilities}
\usage{
getGoProbabilities(go_decisions_list, nogo_decisions_list = NULL)
}
\arguments{
\item{go_decisions_list}{An object of class \code{decision_list},
as returned by \code{\link[bhmbasket]{getGoDecisions}}}

\item{nogo_decisions_list}{An object of class \code{decision_list},
as returned by \code{\link[bhmbasket]{getGoDecisions}}, Default: \code{NULL}}
}
\value{
A list of matrices of Go (and Consider and NoGo) probabilities
}
\description{
Calculates the Go probabilities for given decisions
}
\details{
If only \code{go_decisions_list} is provided
(i.e. \code{nogo_decisions_list} is \code{NULL}),
only Go probabilities will be calculated.
If both \code{go_decisions_list} and \code{nogo_decisions_list} are provided,
Go, Consider, and NoGo probabilities will be calculated.
}
\examples{
scenarios_list <- simulateScenarios(
  n_subjects_list     = list(c(10, 20)),
  response_rates_list = list(rep(0.9, 2)),
  n_trials            = 10)

analyses_list <- performAnalyses(
  scenario_list       = scenarios_list,
  target_rates        = rep(0.5, 2),
  n_mcmc_iterations   = 100)

go_decisions_list <- getGoDecisions(
  analyses_list       = analyses_list,
  cohort_names        = c("p_1", "p_2"),
  evidence_levels     = c(0.5, 0.8),
  boundary_rules      = quote(c(x[1] > 0.8, x[2] > 0.6)))

nogo_decisions_list <- getGoDecisions(
  analyses_list       = analyses_list,
  cohort_names        = c("p_1", "p_2"),
  evidence_levels     = c(0.5, 0.8),
  boundary_rules      = quote(c(x[1] < 0.5, x[2] < 0.3)))

getGoProbabilities(go_decisions_list)
getGoProbabilities(go_decisions_list, nogo_decisions_list)
}
\seealso{
\code{\link[bhmbasket]{getGoDecisions}}
}
\author{
Stephan Wojciekowski
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/AnalysisFunctions.R
\name{saveAnalyses}
\alias{saveAnalyses}
\title{saveAnalyses}
\usage{
saveAnalyses(analyses_list, save_path = tempdir(), analysis_numbers = NULL)
}
\arguments{
\item{analyses_list}{An object of class \code{analysis_list},
as created with \code{\link[bhmbasket]{performAnalyses}}}

\item{save_path}{A string for the path where the scenarios are being stored,
Default: \code{\link[base]{tempfile}}}

\item{analysis_numbers}{A positive integer naming the analysis number.
If \code{NULL}, the function will look for the number of saved analyses of the scenario
in the directory and add 1, Default: \code{NULL}}
}
\value{
A named list of length 3 of vectors with scenario and analysis numbers and
the \code{save_path}
}
\description{
This function saves an object of class \code{analysis_list}
}
\examples{
  trial_data <- createTrial(
    n_subjects   = c(10, 20, 30),
    n_responders = c(1, 2, 3))

  analysis_list <- performAnalyses(
    scenario_list      = trial_data,
    target_rates       = rep(0.5, 3),
    n_mcmc_iterations  = 100)

  save_info     <- saveAnalyses(analysis_list)
  analysis_list <- loadAnalyses(scenario_numbers = save_info$scenario_numbers,
                                analysis_numbers = save_info$analysis_numbers,
                                load_path        = save_info$path)
}
\seealso{
\code{\link[bhmbasket]{performAnalyses}}
\code{\link[bhmbasket]{loadAnalyses}}
\code{\link[base]{tempfile}}
}
\author{
Stephan Wojciekowski
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/OCFunctions.R
\name{getGoDecisions}
\alias{getGoDecisions}
\title{getGoDecisions}
\usage{
getGoDecisions(
  analyses_list,
  cohort_names,
  evidence_levels,
  boundary_rules,
  overall_min_gos = 1
)
}
\arguments{
\item{analyses_list}{An object of class \code{analysis_list},
as created with \code{\link[bhmbasket]{performAnalyses}}}

\item{cohort_names}{A vector of strings with the names of the cohorts, e.g.
\code{c('p_1', 'p_2')}}

\item{evidence_levels}{A vector of numerics in \code{(0, 1)} for the
posterior probability thresholds for the cohorts.
Will be recycled to match the number of methods in the \code{analyses_list}}

\item{boundary_rules}{A quote of a vector for the boundary rules,
\code{quote(c(...))}, see details.
The number of decisions to be taken must match the number of cohorts.
Will be recycled to match the number of methods in the \code{analyses_list}}

\item{overall_min_gos}{A positive integer for the minimum number of
cohort-wise go decisions required for an overall go decision
 Default: \code{1}}
}
\value{
An object of class \code{decision_list}
}
\description{
This function applies decision rules to the analyzed trials.
The resulting \code{decision_list} can be further processed with
\code{\link[bhmbasket]{getGoProbabilities}} or
\code{\link[bhmbasket]{continueRecruitment}}.
}
\details{
This function applies decision rules of the following type to the
outcomes of (simulated) basket trials with binary endpoints:
\deqn{P(p_j|data > p_{B,j}) > \gamma,}
where \eqn{p_j|data} is the posterior response rate of cohort \eqn{j},
\eqn{p_{B,j}} is the response rate boundary of cohort \eqn{j},
and \eqn{\gamma} is the evidence level.
This rule can equivalently be written as \deqn{q_{1-\gamma,j} > p_{B,j},}
where \eqn{q_{1-\gamma,j}} is the \eqn{1-\gamma}-quantile of the posterior
response rate of cohort \eqn{j}.

The arguments \code{cohort_names} and \code{evidence_levels} determine
\eqn{q_{1-\gamma,j}}, where the entries of \code{cohort_names} and
\code{evidence_levels} are matched corresponding to their order.

The argument \code{boundary_rules} provides the rules that describe what
should happen with  the posterior quantiles \eqn{q_{1-\gamma,j}}.
The first posterior quantile determined by the first items of
\code{cohort_names} and \code{evidence_levels} is referred to as \code{x[1]},
the second as \code{x[2]}, etc.
Using the \code{quote(c(...))}-notation,
many different rules can be implemented.
A decision rule for only one cohort would be
\code{boundary_rules = quote(c(x[1] > 0.1))},
\code{cohort_names = 'p_1'}, and \code{evidence_levels = 0.5},
which implements the rule \eqn{P(p_1|data > 0.1) > 0.5}.
The number of decisions to be taken must match the number of cohorts, i.e.
for each cohort there must be a decision rule in the vector separated by a comma.
See the example section for a decision rule for more than one cohort and
the example of \code{\link[bhmbasket]{negateGoDecisions}}
for the implementation of a more complex decision rule.
}
\examples{
scenarios_list <- simulateScenarios(
  n_subjects_list     = list(c(10, 20, 30)),
  response_rates_list = list(c(0.1, 0.1, 0.9)),
  n_trials            = 10)

analyses_list <- performAnalyses(
  scenario_list      = scenarios_list,
  target_rates       = rep(0.5, 3),
  n_mcmc_iterations  = 100)

## Decision rule for more than one cohort
decisions_list <- getGoDecisions(
  analyses_list   = analyses_list,
  cohort_names    = c("p_1", "p_2", "p_3"),
  evidence_levels = c(0.5, 0.5, 0.8),
  boundary_rules  = quote(c(x[1] > 0.7, x[2] < 0.3, x[3] < 0.6)))

## Decision rule for only two of the three cohorts
decisions_list <- getGoDecisions(
  analyses_list   = analyses_list,
  cohort_names    = c("p_1", "p_3"),
  evidence_levels = c(0.5, 0.8),
  boundary_rules  = quote(c(x[1] > 0.7, TRUE, x[3] < 0.6)),
  overall_min_gos = 2L)

## Different decision rules for each method
## This works the same way for the different evidence_levels
decisions_list <- getGoDecisions(
  analyses_list   = analyses_list,
  cohort_names    = c("p_1", "p_2", "p_3"),
  evidence_levels = c(0.5, 0.5, 0.8),
  boundary_rules  = list(
    quote(c(x[1] > 0.1,  x[2] < 0.5,  x[3] < 0.1)),   # "berry"
    quote(c(x[1] > 0.2,  x[2] < 0.4,  x[3] < 0.2)),   # "exnex"
    quote(c(x[1] > 0.25, x[2] < 0.35, x[3] < 0.25)),  # "exnex_adj"
    quote(c(x[1] > 0.3,  x[2] < 0.3,  x[3] < 0.3)),   # "exnex_adj_mix"
    quote(c(x[1] > 0.35, x[2] < 0.25, x[3] < 0.35)),  # "exnex_mix"
    quote(c(x[1] > 0.4,  x[2] < 0.2,  x[3] < 0.4)),   # "pooled"
    quote(c(x[1] > 0.45, x[2] < 0.15, x[3] < 0.45)),  # "stratified"
    quote(c(x[1] > 0.5,  x[2] < 0.1,  x[3] < 0.5))    # "stratified_mix"
  ))
}
\seealso{
\code{\link[bhmbasket]{performAnalyses}}
 \code{\link[bhmbasket]{getGoProbabilities}}
 \code{\link[bhmbasket]{negateGoDecisions}}
 \code{\link[bhmbasket]{continueRecruitment}}
}
\author{
Stephan Wojciekowski
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/OCFunctions.R
\name{getEstimates}
\alias{getEstimates}
\title{getEstimates}
\usage{
getEstimates(
  analyses_list,
  add_parameters = NULL,
  point_estimator = "median",
  alpha_level = 0.05
)
}
\arguments{
\item{analyses_list}{An object of class \code{analysis_list},
as created with \code{\link[bhmbasket]{performAnalyses}}}

\item{add_parameters}{A vector of strings naming additional parameters
from the Bayesian hierarchical models, e.g. \code{c('mu', 'tau')}.
If \code{NULL}, no additional parameters will be evaluated,
Default: \code{NULL}}

\item{point_estimator}{A string indicating the type of estimator used for calculation of
bias and MSE. Must be one of \code{'median'} or \code{'mean'}}

\item{alpha_level}{A numeric in (0, 1) for the level of the credible interval.
Only values corresponding to quantiles saved in \code{\link[bhmbasket]{performAnalyses}}
will work, Default: \code{0.05}}
}
\value{
A named list of matrices of estimates of response rates and credible intervals.
Estimates of bias and MSE are included for response rate estimates of simulated trials.
}
\description{
This function calculates the point estimates and credible intervals per cohort,
as well as estimates of the biases and the mean squared errors of the point estimates per cohort.
}
\details{
Bias and MSE will only be calculated for response rate estimates of simulated trials.
For additional parameters, bias and MSE will not be calculated.

Possible additional parameters for the Bayesian hierarchical models are
\code{c('mu', 'tau')} for \code{'berry'}, \code{'exnex'}, \code{'exnex_mix'}, \code{'exnex_adj'},
and \code{'exnex_adj_mix'}.
The ExNex-based models can also access posterior weights.
\code{paste0("w_", seq_len(n_cohorts))}.
}
\examples{
  scenarios_list <- simulateScenarios(
    n_subjects_list     = list(c(10, 20, 30)),
    response_rates_list = list(c(0.1, 0.2, 3)),
    n_trials            = 10)

  analyses_list <- performAnalyses(
    scenario_list       = scenarios_list,
    target_rates        = c(0.1, 0.1, 0.1),
    calc_differences    = matrix(c(3, 2, 2, 1), ncol = 2),
    n_mcmc_iterations   = 100)

  getEstimates(analyses_list)
  getEstimates(analyses_list   = analyses_list,
               add_parameters  = c("mu", "tau", "w_1", "w_2", "w_3"),
               point_estimator = "mean",
               alpha_level     = 0.1)

  outcome <- createTrial(
    n_subjects          = c(10, 20, 30),
    n_responders        = c( 1,  2,  3))

  outcome_analysis <- performAnalyses(
    scenario_list       = outcome,
    target_rates        = c(0.1, 0.1, 0.1),
    n_mcmc_iterations   = 100)

  getEstimates(outcome_analysis)
  getEstimates(analyses_list  = outcome_analysis,
               add_parameters = c("mu", "w_1", "w_2", "w_3"))
}
\seealso{
\code{\link[bhmbasket]{createTrial}}
\code{\link[bhmbasket]{simulateScenarios}}
\code{\link[bhmbasket]{performAnalyses}}
}
\author{
Stephan Wojciekowski
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/PriorFunctions.R
\name{setPriorParametersStratified}
\alias{setPriorParametersStratified}
\title{setPriorParametersStratified}
\usage{
setPriorParametersStratified(a_j, b_j)
}
\arguments{
\item{a_j}{A vector of positive numerics for \eqn{\alpha}}

\item{b_j}{A vector of positive numerics for \eqn{\beta}}
}
\value{
A list with prior parameters of class \code{prior_parameters_list}
}
\description{
This function sets prior parameters for the analysis method \code{"stratified"}
for use in \code{\link[bhmbasket]{performAnalyses}}.
}
\details{
The method \code{"stratified"} is a beta-binomial model that assesses each cohort individually.
The prior parameters are the scale parameters of the beta prior distributions.
}
\examples{
 prior_parameters_pooled <- setPriorParametersStratified(c(1, 2), c(3, 4))
}
\seealso{
\code{\link[bhmbasket]{performAnalyses}}
\code{\link[bhmbasket]{getPriorParameters}}
\code{\link[bhmbasket]{combinePriorParameters}}
\code{\link[bhmbasket]{setPriorParametersBerry}}
\code{\link[bhmbasket]{setPriorParametersExNex}}
\code{\link[bhmbasket]{setPriorParametersExNexAdj}}
\code{\link[bhmbasket]{setPriorParametersPooled}}
\code{\link[bhmbasket]{getMuVar}}
}
\author{
Stephan Wojciekowski
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/PriorFunctions.R
\name{setPriorParametersStratifiedMix}
\alias{setPriorParametersStratifiedMix}
\title{setPriorParametersStratifiedMix}
\usage{
setPriorParametersStratifiedMix(w, a_j, b_j)
}
\arguments{
\item{w}{A numeric vector of mixture weights in \eqn{[0,1]} summing to 1.}

\item{a_j}{A positive numeric matrix of beta shape parameters \eqn{\alpha_j},
with rows per mixture component and columns per cohort.}

\item{b_j}{A positive numeric matrix of beta shape parameters \eqn{\beta_j},
with rows per mixture component and columns per cohort.}
}
\value{
A list with prior parameters of class \code{prior_parameters_list}
}
\description{
This function sets prior parameters for the analysis method \code{"stratified_mix"}
for use in \code{\link[bhmbasket]{performAnalyses}}.
}
\details{
The method \code{"stratified_mix"} is a beta-binomial model that assesses each cohort
individually with a finite mixture beta prior.
See also the R package \code{RBesT}.
}
\examples{
prior_parameters_stratified_mix <- setPriorParametersStratifiedMix(
  w   = c(0.8, 0.2),
  a_j = rbind(c(2, 3), c(1, 1)),
  b_j = rbind(c(8, 7), c(1, 1))
)
}
\seealso{
\code{\link[bhmbasket]{performAnalyses}}
\code{\link[bhmbasket]{getPriorParameters}}
\code{\link[bhmbasket]{combinePriorParameters}}
\code{\link[bhmbasket]{setPriorParametersStratified}}
}
\author{
Stephan Wojciekowski
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Misc.R
\name{scaleRoundList}
\alias{scaleRoundList}
\title{scaleRoundList}
\usage{
scaleRoundList(list, scale_param = 1, round_digits = NULL)
}
\arguments{
\item{list}{The list to which the scaling and rounding should be applied to.}

\item{scale_param}{A numeric for the scaling of each item of the list, Default: `1`}

\item{round_digits}{An integer for the number of digits.
If `NULL`, no rounding will be applied, Default: `NULL`}
}
\value{
A list of scaled and rounded numerics
}
\description{
This function applies scaling and rounding to each item of a list of numerics
}
\examples{
some_list <- as.list(runif(5))
scaleRoundList(some_list, scale_param = 100, round_digits = 2)

scenarios_list <- simulateScenarios(
  n_subjects_list     = list(c(10, 20, 30)),
  response_rates_list = list(c(0.1, 0.2, 0.3)),
  n_trials            = 10)

analyses_list <- performAnalyses(
  scenario_list       = scenarios_list,
  target_rates        = rep(0.5, 3),
  n_mcmc_iterations   = 100)

scaleRoundList(
  list         = getEstimates(analyses_list),
  scale_param  = 100,
  round_digits = 2)
}
\author{
Stephan Wojciekowski
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/OCFunctions.R
\name{negateGoDecisions}
\alias{negateGoDecisions}
\title{negateGoDecisions}
\usage{
negateGoDecisions(go_decisions_list, overall_min_nogos = "all")
}
\arguments{
\item{go_decisions_list}{An object of class \code{decision_list},
as returned by \code{\link[bhmbasket]{getGoDecisions}}}

\item{overall_min_nogos}{Either a non-negative integer or the string \code{all}
for the minimum number of cohort-level NoGo decisions required for
an overall NoGo decision, Default: \code{all}}
}
\value{
A list of NoGo decisions of class \code{decision_list}
}
\description{
Negates the go decisions derived with
\code{\link[bhmbasket]{getGoDecisions}}.
}
\details{
This function is intended for implementing decision rules with a
consider zone as
e.g. proposed in "Bayesian design of proof-of-concept trials" by
Fisch et al. (2015).
This approach involves two criteria, Significance and Relevance.
\itemize{
\item Significance: high evidence that the treatment effect is greater
than some smaller value (e.g. treatment effect under H0)
\item Relevance: moderate evidence that the treatment effect is greater
than some larger value (e.g. treatment effect under a certain alternative)
}
The decision for a cohort is then taken as follows:
\itemize{
\item Go decision: Significance and Relevance
\item Consider decision: either Significance, or Relevance, but not both
\item NoGo decision: no Significance and no Relevance
}
In the example below, the following criteria for are implemented for each of
the three cohorts:
\itemize{
\item Significance: \eqn{P(p_j > 0.4) > 0.95}
\item Relevance: \eqn{P(p_j > 0.8) > 0.5}
}
}
\examples{
scenarios_list <- simulateScenarios(
  n_subjects_list     = list(c(10, 20, 30)),
  response_rates_list = list(rep(0.9, 3)),
  n_trials            = 10)

analysis_list <- performAnalyses(
  scenario_list      = scenarios_list,
  target_rates       = rep(0.5, 3),
  n_mcmc_iterations  = 100)

go_decisions_list <- getGoDecisions(
  analyses_list   = analysis_list,
  cohort_names    = c("p_1", "p_2", "p_3",
                      "p_1", "p_2", "p_3"),
  evidence_levels = c(0.5,  0.5,  0.5,
                      0.95, 0.95, 0.95),
  boundary_rules  = quote(c(x[1] > 0.8 & x[4] > 0.4,
                            x[2] > 0.8 & x[5] > 0.4,
                            x[3] > 0.8 & x[6] > 0.4)))

nogo_decisions <- negateGoDecisions(getGoDecisions(
  analyses_list   = analysis_list,
  cohort_names    = c("p_1", "p_2", "p_3",
                      "p_1", "p_2", "p_3"),
  evidence_levels = c(0.5,  0.5,  0.5,
                      0.95, 0.95, 0.95),
  boundary_rules  = quote(c(x[1] > 0.8 | x[4] > 0.4,
                            x[2] > 0.8 | x[5] > 0.4,
                            x[3] > 0.8 | x[6] > 0.4))))

getGoProbabilities(go_decisions_list, nogo_decisions)
}
\references{
Fisch, Roland, et al.
"Bayesian design of proof-of-concept trials."
\emph{Therapeutic innovation & regulatory science} 49.1 (2015): 155-162.
}
\seealso{
\code{\link[bhmbasket]{getGoDecisions}}
}
\author{
Stephan Wojciekowski
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/AnalysisFunctions.R
\name{performAnalyses}
\alias{performAnalyses}
\title{performAnalyses}
\usage{
performAnalyses(
  scenario_list,
  evidence_levels = c(0.025, 0.05, 0.5, 0.8, 0.9, 0.95, 0.975),
  method_names = c("berry", "exnex", "exnex_mix", "exnex_adj", "exnex_adj_mix", "pooled",
    "stratified", "stratified_mix"),
  target_rates = NULL,
  prior_parameters_list = NULL,
  calc_differences = NULL,
  n_mcmc_iterations = 10000,
  n_cores = 1,
  seed = 1,
  verbose = TRUE
)
}
\arguments{
\item{scenario_list}{An object of class \code{scenario_list},
as e.g. created with \code{\link[bhmbasket]{simulateScenarios}}}

\item{evidence_levels}{A vector of numerics in \verb{(0, 1)} for the
\code{1-evidence_levels}-quantiles of the posterior response rates to be saved.
Default: \code{c(0.025, 0.05, 0.5, 0.8, 0.9, 0.95, 0.975)}}

\item{method_names}{A vector of strings for the names of the methods to be used. Must
be one of the default values, Default:
\code{c("berry", "exnex", "exnex_mix", "exnex_adj", "exnex_adj_mix", "pooled", "stratified", "stratified_mix")}}

\item{target_rates}{A vector of numerics in \verb{(0, 1)} for the
target rates of each cohort, Default: \code{NULL}}

\item{prior_parameters_list}{An object of class \code{prior_parameters_list},
as e.g. created with \code{\link[bhmbasket]{getPriorParameters}}}

\item{calc_differences}{A matrix of positive integers with 2 columns.
For each row the differences will be calculated.
Also a vector of positive integers can be provided for a single difference.
The integers are the numbers for the cohorts to be subtracted from one another.
E.g. providing \code{c(2, 1)} calculates the difference between cohort \code{2} and cohort \code{1}.
If \code{NULL}, no subtractions are performed, Default: \code{NULL}}

\item{n_mcmc_iterations}{A positive integer for the number of MCMC iterations,
see Details, Default: \code{10000}.
If \code{n_mcmc_iterations} is present in \code{.GlobalEnv} and \code{missing(n_mcmc_iterations)},
the globally available value will be used.}

\item{n_cores}{Argument is deprecated and does nothing as of version 0.9.3.
A positive integer for the number of cores for the parallelization,
Default: \code{1}}

\item{seed}{Argument is deprecated and does nothing as of version 0.9.3.
A numeric for the random seed, Default: \code{1}}

\item{verbose}{A logical indicating whether messages should be printed, Default: \code{TRUE}}
}
\value{
An object of class \code{analysis_list}.
}
\description{
This function performs the analysis of simulated or observed trial data with the
specified methods
and returns the quantiles of the posterior response rates
}
\details{
This function applies the following analysis models to (simulated) scenarios of class
\code{scenario_list}:
\itemize{
\item Bayesian hierarchical model (BHM) proposed by Berry et al. (2013): \code{"berry"}
\item BHM proposed by Neuenschwander et al. (2016): \code{"exnex"}
\item BHM that combines above approaches: \code{"exnex_adj"}
\item Pooled beta-binomial approach: \code{"pooled"}
\item Stratified beta-binomial approach: \code{"stratified"}
\item BHM with mixture prior on the Nex component: \code{"exnex_mix"}
\item Adjusted BHM with mixture prior on the Nex component: \code{"exnex_adj_mix"}
\item Stratified beta-binomial approach with mixture beta prior: \code{"stratified_mix"}
}
The posterior distributions of the BHMs are approximated with Markov chain Monte Carlo (MCMC)
methods implemented in JAGS.
Two independent chains are used with each \code{n_mcmc_iterations} number of MCMC iterations.
The first \code{floor(n_mcmc_iterations / 3)} number of iterations are discarded as burn-in period.
No thinning is applied.

Note that the value for \code{n_mcmc_iterations} required for a good approximation of the posterior
distributions depends on the analysis model, the investigated scenarios, and the use case.
The default value might be a good compromise between run-time and approximation for
the estimation of decision probabilities, but
it should definitively be increased for the analysis of a single trial's outcome.

The analysis models will only be applied to the unique trial realizations across
all simulated scenarios.
The models can be applied in parallel by registering a parallel backend for the 'foreach'
framework, e.g. with \code{doFuture::registerDoFuture()} and \code{future::plan(future::multisession)}.
The parallelization is nested, so that the resources of a HPC environment can be used
efficiently.
For more on this topic, kindly see the respective vignette.
The tasks that are to be performed in parallel are chunked according to the number of workers
determined with \code{foreach::getDoParWorkers()}.

The JAGS code for the BHM \code{"exnex"} was taken from Neuenschwander et al. (2016).
The JAGS code for the BHM \code{"exnex_adj"} is based on the JAGS code for \code{"exnex"}.
The JAGS code for the BHM \code{"exnex_mix"} extends the \code{"exnex"} model by using
a mixture prior for the Nex component.
The JAGS code for the BHM \code{"exnex_adj_mix"} extends the \code{"exnex_adj"} model by using
a mixture prior for the Nex component.
}
\examples{
 trial_data <- createTrial(
   n_subjects   = c(10, 20, 30),
   n_responders = c(1, 2, 3))

 analysis_list <- performAnalyses(
   scenario_list      = trial_data,
   target_rates       = rep(0.5, 3),
   calc_differences   = matrix(c(3, 2, 1, 1), ncol = 2),
   n_mcmc_iterations  = 100)
}
\references{
Berry, Scott M., et al. "Bayesian hierarchical modeling of patient subpopulations:
efficient designs of phase II oncology clinical trials."
\emph{Clinical Trials} 10.5 (2013): 720-734.

Neuenschwander, Beat, et al. "Robust exchangeability designs
for early phase clinical trials with multiple strata."
\emph{Pharmaceutical statistics} 15.2 (2016): 123-134.

Plummer, Martyn. "JAGS: A program for analysis of Bayesian graphical models
using Gibbs sampling."
\emph{Proceedings of the 3rd international workshop on distributed statistical computing.}
Vol. 124. No. 125.10. 2003.
}
\seealso{
\code{\link[bhmbasket]{simulateScenarios}}
\code{\link[bhmbasket]{createTrial}}
\code{\link[bhmbasket]{getPriorParameters}}
}
\author{
Stephan Wojciekowski
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/DataFunctions.R
\name{simulateScenarios}
\alias{simulateScenarios}
\title{simulateScenarios}
\usage{
simulateScenarios(
  n_subjects_list,
  response_rates_list,
  scenario_numbers = seq_along(response_rates_list),
  n_trials = 10000
)
}
\arguments{
\item{n_subjects_list}{A list that contains for each scenario a vector for
the number of subjects per cohort.
A single vector can be provided if all scenarios should have the same number of subjects.}

\item{response_rates_list}{A list that contains for each scenario a vector for
the response rates per cohort.}

\item{scenario_numbers}{A vector of positive integers naming the scenarios,
Default: \code{seq_along(response_rates_list)}}

\item{n_trials}{An integer indicating the number of trial simulations per response rates,
Default: \code{10000}. If \code{n_trials} is present in \code{.GlobalEnv} and \code{missing(n_trials)},
the globally available value will be used.}
}
\value{
An object of class \code{scenario_list} with the scenario data for each specified scenario.
}
\description{
This function creates scenarios for the analysis with
\code{\link[bhmbasket]{performAnalyses}}.
}
\details{
The function simulates trials with binary outcome for each scenario.
Integer values for the response rates will be treated as observed outcomes.
}
\examples{
  n_subjects     <- c(10, 20, 30)

  rr_negative    <- rep(0.1, 3)
  rr_nugget      <- c(0.9, 0.1, 0.1)
  rr_positive    <- rep(0.9, 3)

  scenarios_list <- simulateScenarios(
    n_subjects_list     = list(n_subjects,
                               n_subjects,
                               n_subjects),
    response_rates_list = list(rr_negative,
                               rr_nugget,
                               rr_positive))
}
\seealso{
\code{\link[bhmbasket]{saveScenarios}}
\code{\link[bhmbasket]{createTrial}}
\code{\link[bhmbasket]{performAnalyses}}
}
\author{
Stephan Wojciekowski
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/PriorFunctions.R
\name{getPriorParameters}
\alias{getPriorParameters}
\title{getPriorParameters}
\usage{
getPriorParameters(
  method_names,
  target_rates,
  n_worth = 1,
  tau_scale = 1,
  w_j = 0.5
)
}
\arguments{
\item{method_names}{A vector of strings for the names of the methods to be used.
Available methods: \verb{c("berry", "exnex", "exnex_adj", "exnex_mix", "exnex_adj_mix", "pooled", "stratified", "stratified_mix)}}

\item{target_rates}{A vector of numerics in \verb{(0, 1)} for the
target rate of each cohort}

\item{n_worth}{An integer for the number of subjects the variability of the prior should reflect
response rate scale, Default: \code{1}}

\item{tau_scale}{A numeric for the scale parameter of the Half-normal distribution of \eqn{\tau}
in the methods \code{"berry"}, \code{"exnex"}, \verb{"exnex"_mix}, \code{"exnex_adj"} and \code{"exnex_adj_mix"}, Default: \code{1}}

\item{w_j}{A numeric in \verb{(0, 1)} for the weight of the Ex component in the methods \code{"exnex"}
and \code{"exnex_adj"}, Default: \code{0.5}}
}
\value{
A list with prior parameters of class \code{prior_parameters_list}
}
\description{
This function provides default prior parameters for the analysis methods
that can be used in \code{\link[bhmbasket]{performAnalyses}}.
}
\details{
Regarding the default prior parameters for \code{"berry"}, \code{"exnex"}, \code{"exnex_mix"}, \code{"exnex_adj"} and \code{"exnex_adj_mix"}:
\itemize{
\item \code{"berry"}: The mean of \eqn{\mu} is set to \code{0}.
Its variance is calculated as proposed in "Robust exchangeability designs for early
phase clinical trials with multiple strata" (Neuenschwander et al. (2016))
with regard to \code{n_worth}.
The scale parameter of \eqn{\tau} is set to \code{tau_scale}.
\item \code{"exnex"}: The weight of the Ex component is set to \code{w_j}.
For the Ex component:
The target rate that results in the greatest variance is determined.
The mean of \eqn{\mu} is set to that target rate.
The variance of \eqn{\mu} is calculated as proposed in "Robust exchangeability designs for early
phase clinical trials with multiple strata" (Neuenschwander et al. (2016))
with regard to \code{n_worth}.
The scale parameter of \eqn{\tau} is set to \code{tau_scale}.

For the Nex components:
The means of \eqn{\mu_j} are set to the respective target rates.
The variances of \eqn{\tau_j} are calculated as proposed in "Robust exchangeability designs for early
phase clinical trials with multiple strata" (Neuenschwander et al. (2016))
with regard to \code{n_worth}, see also \code{\link[bhmbasket]{getMuVar}}.
\item \code{"exnex_adj"}: The weight of the Ex component is set to \code{w_j}.
For the Ex component:
The target rate that results in the greatest variance is determined.
The mean of \eqn{\mu} is set to \code{0}.
The variance of \eqn{\mu} is calculated as proposed in "Robust exchangeability designs for early
phase clinical trials with multiple strata" (Neuenschwander et al. (2016))
with regard to \code{n_worth}, see also \code{\link[bhmbasket]{getMuVar}}.
The scale parameter of \eqn{\tau} is set to \code{tau_scale}.

For the Nex components:
The means of \eqn{\mu_j} are set to the \code{0}.
The variances of \eqn{\tau_j} are calculated as proposed in "Robust exchangeability designs for early
phase clinical trials with multiple strata" (Neuenschwander et al. (2016))
with regard to \code{n_worth}, see also \code{\link[bhmbasket]{getMuVar}}.
\item \code{"exnex_mix"}: Uses the same default Ex prior construction as \code{"exnex"}.
The Nex part default parameters are specified as a one-component mixture prior with
\code{w_nex = 1}, \code{mean_nex = matrix(logit(target_rates), nrow = 1)},
and \code{sd_nex = matrix(sqrt(getMuVar(target_rates, 0, n_worth)), nrow = 1)}.
This keeps the default mixture representation compatible with the \code{getPriorParameter()}'s input.
\item \code{"exnex_adj_mix"}: Uses the same default Ex prior construction as \code{"exnex_adj"}.
The Nex part is specified as a one-component mixture prior with
\code{w_nex = 1}, \code{mean_nex = matrix(logit(target_rates), nrow = 1)},
and \code{sd_nex = matrix(sqrt(getMuVar(target_rates, 0, n_worth)), nrow = 1)}.
The Ex component is centered as in \code{"exnex_adj"}.
\item \code{"pooled"}: The target rate that results in the greatest variance is determined.
The scale parameter \eqn{\alpha} is set to that target rate times \code{n_worth}.
The scale parameter \eqn{\beta} is set to 1 - that target rate times \code{n_worth}.
\item \code{"stratified"}:
The scale parameters \eqn{\alpha_j} are set to \code{target_rates * n_worth}.
The scale parameters \eqn{\beta_j} are set to \code{(1 - target_rates) * n_worth}.
\item \code{"stratified_mix"}:
A two-component beta mixture prior is created by default for each cohort.
The first component uses \code{a_j = target_rates * n_worth} and
\code{b_j = (1 - target_rates) * n_worth}.
The second component is a vague prior with \code{a_j = 1} and \code{b_j = 1}.
The default mixture weights are \code{c(0.8, 0.2)}.
}
}
\examples{
prior_parameters_list <- getPriorParameters(
  method_names = c("berry", "exnex", "exnex_mix", "exnex_adj",
                   "exnex_adj_mix", "pooled", "stratified", "stratified_mix"),
  target_rates = c(0.1, 0.2, 0.3))
}
\references{
Berry, Scott M., et al. "Bayesian hierarchical modeling of patient subpopulations:
efficient designs of phase II oncology clinical trials."
\emph{Clinical Trials} 10.5 (2013): 720-734.

Neuenschwander, Beat, et al. "Robust exchangeability designs
for early phase clinical trials with multiple strata."
\emph{Pharmaceutical statistics} 15.2 (2016): 123-134.
}
\seealso{
\code{\link[bhmbasket]{performAnalyses}}
\code{\link[bhmbasket]{setPriorParametersBerry}}
\code{\link[bhmbasket]{setPriorParametersExNex}}
\code{\link[bhmbasket]{setPriorParametersExNexAdj}}
\code{\link[bhmbasket]{setPriorParametersPooled}}
\code{\link[bhmbasket]{setPriorParametersStratified}}
\code{\link[bhmbasket]{setPriorParametersStratifiedMix}}
\code{\link[bhmbasket]{combinePriorParameters}}
\code{\link[bhmbasket]{getMuVar}}
}
\author{
Stephan Wojciekowski
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/DataFunctions.R
\name{saveScenarios}
\alias{saveScenarios}
\title{saveScenarios}
\usage{
saveScenarios(scenario_list, save_path = tempdir())
}
\arguments{
\item{scenario_list}{An object of class \code{scenario_list}, e.g. created with \code{simulateScenarios()}}

\item{save_path}{A string providing the path of the directory in which the scenario data
should be saved, Default: \code{\link[base]{tempfile}}}
}
\value{
A named list of length 2 with the scenario numbers and the \code{save_path}
}
\description{
Saves the scenario data in a newly created or existing directory
}
\examples{
  scenarios_list <- simulateScenarios(
    n_subjects_list     = list(c(10, 20, 30)),
    response_rates_list = list(rep(0.9, 3)),
    n_trials            = 10)

  save_info      <- saveScenarios(scenarios_list)
  scenarios_list <- loadScenarios(scenario_numbers = save_info$scenario_numbers,
                                  load_path        = save_info$path)
}
\seealso{
\code{\link[bhmbasket]{simulateScenarios}}
\code{\link[bhmbasket]{loadScenarios}}
\code{\link[base]{tempfile}}
}
\author{
Stephan Wojciekowski
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/OCFunctions.R
\name{getAverageNSubjects}
\alias{getAverageNSubjects}
\title{getAverageNSubjects}
\usage{
getAverageNSubjects(scenario_list)
}
\arguments{
\item{scenario_list}{An object of class \code{scenario_list},
as created with \code{\link[bhmbasket]{simulateScenarios}} or
\code{\link[bhmbasket]{continueRecruitment}}}
}
\value{
A named list of vectors for the average number of subjects in each scenario.
}
\description{
This function calculates the average number of subjects per scenario.
}
\details{
This function can be useful to assess decision rules with regard to the average number
of subjects across scenarios when performing interim analyses.
}
\examples{
interim_scenarios <- simulateScenarios(
  n_subjects_list     = list(c(10, 20, 30)),
  response_rates_list = list(rep(0.9, 3)),
  n_trials            = 10)

interim_analyses <- performAnalyses(
  scenario_list       = interim_scenarios,
  target_rates        = rep(0.5, 3),
  n_mcmc_iterations   = 100)

interim_gos <- getGoDecisions(
  analyses_list       = interim_analyses,
  cohort_names        = c("p_1", "p_2", "p_3"),
  evidence_levels     = c(0.5, 0.8, 0.5),
  boundary_rules      = quote(c(x[1] > 0.8, x[2] > 0.6, x[3] > 0.7)))

scenarios_list <- continueRecruitment(
  n_subjects_add_list = list(c(30, 20, 10)),
  decisions_list      = interim_gos,
  method_name         = "exnex_adj")

getAverageNSubjects(scenarios_list)

}
\seealso{
\code{\link[bhmbasket]{simulateScenarios}}
\code{\link[bhmbasket]{continueRecruitment}}
}
\author{
Stephan Wojciekowski
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/DataFunctions.R
\name{continueRecruitment}
\alias{continueRecruitment}
\title{continueRecruitment}
\usage{
continueRecruitment(n_subjects_add_list, decisions_list, method_name = NULL)
}
\arguments{
\item{n_subjects_add_list}{A list that contains for each scenario an integer vector for
the number of subjects per cohort to be additionally recruited.}

\item{decisions_list}{A list with decisions per scenario created with
\code{\link[bhmbasket]{getGoDecisions}}}

\item{method_name}{A string for the method name of the analysis the decisions are based on.
Can be \code{NULL} if only one method has been used for analysis, Default: \code{NULL}}
}
\value{
An object of class \code{scenario_list} with the scenario data for each specified scenario.
}
\description{
This function continues the recruitment of subjects for a set of scenarios
based on the Go / NoGo decisions in the simulated trial outcomes of said scenarios.
}
\details{
This function is intended to be used for analyses with the following work flow:\cr
\code{simulateScenarios()} -> \code{performAnalyses()} -> \code{getGoDecisions()}-> \cr
\code{continueRecruitment()} -> \code{performAnalyses()} -> \code{getGoDecisions()}-> \cr
\code{continueRecruitment()} -> ...

Note that \code{n_subjects_add_list} takes the additional number of subjects to be recruited,
not the overall number of subjects.
This way the work flow can be repeated as often as
required, which can be useful e.g. for interim analyses.
}
\examples{
interim_scenarios <- simulateScenarios(
  n_subjects_list     = list(c(10, 20, 30)),
  response_rates_list = list(rep(0.9, 3)),
  n_trials            = 10)

interim_analyses <- performAnalyses(
  scenario_list       = interim_scenarios,
  target_rates        = rep(0.5, 3),
  n_mcmc_iterations   = 100)

interim_gos <- getGoDecisions(
  analyses_list       = interim_analyses,
  cohort_names        = c("p_1", "p_2", "p_3"),
  evidence_levels     = c(0.5, 0.8, 0.5),
  boundary_rules      = quote(c(x[1] > 0.8, x[2] > 0.6, x[3] > 0.7)))

scenarios_list <- continueRecruitment(
  n_subjects_add_list = list(c(30, 20, 10)),
  decisions_list      = interim_gos,
  method_name         = "exnex_adj")
}
\seealso{
\code{\link[bhmbasket]{simulateScenarios}}
\code{\link[bhmbasket]{performAnalyses}}
\code{\link[bhmbasket]{getGoDecisions}}
}
\author{
Stephan Wojciekowski
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/DataFunctions.R
\name{createTrial}
\alias{createTrial}
\title{createTrial}
\usage{
createTrial(n_subjects, n_responders)
}
\arguments{
\item{n_subjects}{A vector of integers for the number of subjects in the trial outcome}

\item{n_responders}{A vector of integers for the number of responders in the trial outcome}
}
\value{
An object of class \code{scenario_list} with the scenario data for a single trial outcome.
}
\description{
This function creates an object of class \code{scenario_list}
for a single trial outcome, which can subsequently be analyzed with other functions of
\code{bhmbasket}, e.g. \code{\link[bhmbasket]{performAnalyses}}
}
\details{
This function is a wrapper for \code{\link[bhmbasket]{simulateScenarios}} with

\if{html}{\out{<div class="sourceCode">}}\preformatted{simulateScenarios(
 n_subjects_list     = list(n_subjects),
 response_rates_list = list(n_responders),
 n_trials            = 1)
}\if{html}{\out{</div>}}
}
\examples{
 trial_outcome <- createTrial(n_subjects   = c(10, 20, 30, 40),
                              n_responders = c( 1,  2,  3,  4))
}
\seealso{
\code{\link[bhmbasket]{simulateScenarios}}
\code{\link[bhmbasket]{performAnalyses}}
}
\author{
Stephan Wojciekowski
}
